/*
 * nasd_cheops_cl_init.c
 */
/*
 * initialization module for the CHEOPS client clerk
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */
#include <nasd/nasd_options.h>
#include <sys/errno.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>

#include <dce/dce_error.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_pdev.h>

#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_cheops_types.h>
#include <nasd/nasd_cheops_common.h>
#include <nasd/nasd_cheops_cache.h>
#include <nasd/nasd_cheops_mgr.h>
/*
 * XXX this should be fixed. It will not be correctly included
 * as a dependency on many platforms. Also, the relative path
 * thing makes it hard for alternatively-organized trees, and
 * just give up on including something like this in the kernel.
 */
#include "../cheops/shared/nasd_cheops_cl_drive.h"

#define _NASD_CHEOPS_MANAGER_RESOURCE_FILE "./cheops.mtab"

#define _NASD_CHEOPS_INIT_DEBUG 0

_nasd_cheops_mgr_handles_t _nasd_cheops_mgr_handles;
int _nasd_cheops_dr_handle_cache_size=16; /* jmb from 8 */
int _nasd_cheops_bs_handle_cache_size=60; /* jmb from 60 */
_nasd_cheops_cache_t *_nasd_cheops_bs_handle_cachep;
_nasd_cheops_cache_t *_nasd_cheops_dr_handle_cachep;
char manager_names[MAX_NUM_MANAGERS][_NASD_CHEOPS_MAX_FNAME_LEN];
char manager_ports[MAX_NUM_MANAGERS][_NASD_CHEOPS_MAX_FNAME_LEN];

/* prototypes */
extern void _nasd_cheops_doio_init();
int _nasd_cheops_dr_handle_read(_nasd_cheops_cache_key_t key, caddr_t bp);
int _nasd_cheops_dr_handle_writeback(_nasd_cheops_cache_ent_t *e, int bsize);
int _nasd_cheops_bs_handle_read(_nasd_cheops_cache_key_t key, caddr_t bp);
int _nasd_cheops_bs_handle_writeback(_nasd_cheops_cache_ent_t *e, int bsize);
int _nasd_cheops_read_mtab_file (char *fname, int *num_managers);

int
_nasd_cheops_cl_init(char *resource_arg)
{
	char name[64];
	char port[64];
	int  num_managers;
	int  i, rc;
	nasd_status_t nrc;
	_nasd_cheops_mgr_handle_t *hp;
	char *resource_file;

	if(resource_arg) {
	  resource_file = resource_arg;
	} else {
	  resource_file = getenv("NASD_CHEOPSMGR_RESOURCE_FILE");
	  if(!resource_file) {
	    resource_file = _NASD_CHEOPS_MANAGER_RESOURCE_FILE;
	  }
	}

	nrc = nasd_cl_p_init();
	if (nrc)
	  return(-1);

	/* read mtab file, discover and bind to a few storage managers */
	rc = _nasd_cheops_read_mtab_file(resource_file, &num_managers);
	if (rc)
	  return rc;
	
	_nasd_cheops_mgr_handles.num_managers = num_managers;
        NASD_Malloc(hp,  num_managers*sizeof(_nasd_cheops_mgr_handle_t), (_nasd_cheops_mgr_handle_t *));
	if (hp == NULL ) {
	  return ENOMEM;
	}
	else
	  _nasd_cheops_mgr_handles.hl = hp;

	for (i=0; i< num_managers; i++) {
	  _nasd_cheops_mgr_handles.hl[i].mgr_name = manager_names[i];
	  _nasd_cheops_mgr_handles.hl[i].mgr_port = manager_ports[i];
	  nrc = nasd_dce_direct_bind_to_server((unsigned_char_t *) "ncadg_ip_udp",
					      (unsigned_char_t *) manager_names[i],
					      (unsigned_char_t *) manager_ports[i],
					      &_nasd_cheops_mgr_handles.hl[i].h);
	  if(nrc)
	    return EIO;
	}

	/* create a cache for drive handles */
	rc = _nasd_cheops_cache_create(_nasd_cheops_dr_handle_cache_size, 
				      sizeof(_nasd_cheops_dr_handle_t),
				      _NASD_CHEOPS_CACHE_TYPE_DMAP, &_nasd_cheops_dr_handle_cachep);
	if(rc)
	  return rc;

	_nasd_cheops_dr_handle_cachep->handle_read_fault = _nasd_cheops_dr_handle_read;
	_nasd_cheops_dr_handle_cachep->writeback = _nasd_cheops_dr_handle_writeback;

	/* create a cache for object stripe maps */
	rc = _nasd_cheops_cache_create(_nasd_cheops_bs_handle_cache_size, 
				      sizeof(_nasd_cheops_bs_handle_t),
				      _NASD_CHEOPS_CACHE_TYPE_SMAP, &_nasd_cheops_bs_handle_cachep);
	if(rc)
	  return rc;

	_nasd_cheops_bs_handle_cachep->handle_read_fault = _nasd_cheops_bs_handle_read;
	_nasd_cheops_bs_handle_cachep->writeback = _nasd_cheops_bs_handle_writeback;

	 _nasd_cheops_doio_init();

	return 0;
}

int
_nasd_cheops_read_mtab_file (char *fname, int *num_managers)
{
        char line[1024];
	char foo[128];
	FILE *fd;
	int i;
	int j;
	int len;

	
	if (fname==NULL) {
         (cheops_Msg "%s (%d):_nasd_cheops_read_mtab_file received empty mtab filename\n", __FILE__, __LINE__);
	 return EINVAL;
	}
	
	if ( (fd = fopen(fname, "r")) == NULL){
	  (cheops_Msg "%s (%d): _nasd_cheops_read_mtab_file() unable to open file %s\n", __FILE__, __LINE__, fname);
	  return ENOENT;
	}

	i=0;
	while ( fgets(line, 1024, fd) != NULL){
	  if (line[0]=='#')
	    continue;
	  len = strlen(line);
	  line[len-1]='\0';
	  if(!strncmp(line,"manager",7)) {
	    /* NASD_Malloc(manager_names[i], len, (char *));
	    NASD_Malloc(manager_ports[i], len, (char *)); */
	    sscanf(line,"%s %s %s", foo, manager_names[i], manager_ports[i]);
	    i++;
	  }
	  if (i>= MAX_NUM_MANAGERS)
	    break;
	}
	*num_managers = i;
	return 0;
}

void
_nasd_cheops_cl_show_mgrs()
{
        int i;
	_nasd_cheops_mgr_handles_t *handles;

	handles = &_nasd_cheops_mgr_handles;
	(cheops_Msg "num managers: %d\n", handles->num_managers);
	for (i=0; i< handles->num_managers; i++)
	  (cheops_Msg "%s: %s\n", handles->hl[i].mgr_name, handles->hl[i].mgr_port);	

}

int
_nasd_cheops_dr_handle_writeback(_nasd_cheops_cache_ent_t *e, int bsize)
{
     return 0;
};

int
_nasd_cheops_dr_handle_read(_nasd_cheops_cache_key_t key,
			   caddr_t bp)
{
       int rc=0;
       int mgr_id;
       int fd=0;
       nasd_disk_ident_t dr_id;
       handle_t mh;
       nasd_drive_handle_t out_h;
       _nasd_cheops_dr_handle_t dh;
       _nasd_cheops_cl_dinfo_t  dinfo;
       nasd_cookie_t in_cookie;
       nasd_status_t nasd_status;
       error_status_t op_status;

       dr_id = key.record_id;
       mgr_id = 0;
       mh = _nasd_cheops_mgr_handles.hl[mgr_id].h;
       nasd_cheops_dr_lookup_cm(mh, in_cookie, dr_id,
			        &dinfo, &nasd_status, &op_status);

       if (op_status)
	 return op_status;
       else if  ( (nasd_status != NASD_SUCCESS) )
	 return nasd_status;

       rc = nasd_bind_to_drive((char *) dinfo.dr_name,
			       (char *) dinfo.dr_port,
			       NASD_BIND_DCE_DIRECT, NULL, 0, &out_h);
       if (rc) {
	 (cheops_Msg "CHEOPS error: failed to bind to drive %s at port %s\n", 
	  dinfo.dr_name, dinfo.dr_port);
	 return rc;
       }  else {
	 dh.h = out_h;
       }

       dh.fd = fd;
       dh.partnum = dinfo.partnum;
       bcopy(dinfo.dr_port, dh.dr_port, _NASD_CHEOPS_DR_NAMELEN);
       bcopy(dinfo.dr_name, dh.dr_name, _NASD_CHEOPS_DR_NAMELEN);
       bcopy((char *)&dh, bp, sizeof(_nasd_cheops_dr_handle_t));

       return 0;
}

int
_nasd_cheops_bs_handle_writeback(_nasd_cheops_cache_ent_t *e, int bsize)
{
      return 0;
};


int
_nasd_cheops_bs_handle_read(_nasd_cheops_cache_key_t key,
			   caddr_t bp)
{
       int mgr_id;
       handle_t h;
       nasd_cookie_t in_cookie;
       nasd_status_t nasd_status;
       error_status_t op_status;
       nasd_identifier_t bs_id;
       _nasd_cheops_bs_handle_t out_bs_handle;
       
       bs_id = key.record_id;

       mgr_id = 0;
       h = _nasd_cheops_mgr_handles.hl[mgr_id].h;

#if _NASD_CHEOPS_INIT_DEBUG > 0
       (cheops_Msg "clerk: issuing read due to cache miss (object=%" NASD_BIGd_FMT ")..\n", bs_id);
#endif

       nasd_cheops_bs_lookup_cm(h, in_cookie, bs_id,
			        &out_bs_handle, &nasd_status, &op_status);

       if  ( (nasd_status != NASD_SUCCESS) )
	 return nasd_status;
       else if (op_status)
	 return op_status;
     
       bcopy((char *)&out_bs_handle, bp, sizeof(_nasd_cheops_bs_handle_t));
       return 0;
}




