/*
 * nasd_cheops_cl_raid1.c
 */
/*
 * RAID1 access module
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */
#include <nasd/nasd_options.h>
#include <stdio.h>
#include <malloc.h>
#include <sys/errno.h>
#include <string.h>

#include <nasd/nasd_types.h>
#include <nasd/nasd_cheops_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_cheops_types.h>
#include <nasd/nasd_cheops_common.h>
#include <nasd/nasd_cheops_raidmap.h>
#include <nasd/nasd_cheops_ios.h>
#include "nasd_cheops_time.h"

#define _NASD_CHEOPS_RAID1_DEBUG 3

extern void 
_nasd_cheops_asm_freeall(nasd_asm_t *);
extern void 
_nasd_cheops_io_freeall(_nasd_cheops_aggr_io_t *);
extern void 
_nasd_cheops_init_io(NASD_pda_t *, _nasd_cheops_aggr_io_t *, _nasd_cheops_io_t *);
extern int 
_nasd_cheops_enq_io(_nasd_cheops_io_t *io);
extern void 
_nasd_cheops_print_asm(nasd_asm_t *);

int
_nasd_cheops_raid1_read(nasd_asm_t *asm, nasd_len_t *out_len)
{
        int rc =0, i;
	_nasd_cheops_aggr_io_t *pio; /* parent io */
	_nasd_cheops_io_t *prev_io, *new_io;
	NASD_stripe_access_t *sac;

	/* enqueue io items corresponding to each pda in the asm  */
	NASD_Malloc(pio, sizeof(_nasd_cheops_aggr_io_t), (_nasd_cheops_aggr_io_t *));
	if (pio == NULL)
	  return ENOMEM;

	rc =  nasd_mutex_init(&(pio->flag_mutex));
        if (rc) {
          (cheops_Msg "CHEOPS Error: can not initialize mutex\n");
          return rc; 
        }
	rc = nasd_cond_init(&(pio->ios_all_done));
        if (rc) {
          (cheops_Msg "CHEOPS Error: can not initialize condition variable\n");
          return rc;
        }
	pio->num_ios = asm->num_su_accessed;
	pio->num_done = 0;
	pio->in_datalen = asm->datalen;
	pio->first_io = NULL;
	pio->out_datalen = 0;
	pio->status = pio->nasd_status = 0;

#if _NASD_CHEOPS_RAID1_DEBUG > 0
	(cheops_Msg "raid1_read: issuing io with %d children\n", pio->num_ios);
	_nasd_cheops_print_asm(asm);
#endif
	prev_io= NULL;
	for (sac= asm->first_stripe_access; sac != NULL; sac = sac->next) {
	  if (!sac)
	    break;
	  for (i=0; i< _NASD_CHEOPS_MAX_STRIPE_SIZE; i++) {
	    if (sac->pda_list[i].buf != NULL ) {
	      NASD_Malloc(new_io, sizeof(_nasd_cheops_io_t), (_nasd_cheops_io_t *));
	      bzero((char *)new_io, sizeof(_nasd_cheops_io_t));
	      if (new_io == NULL) {
		return ENOMEM;
	      } else {
		  new_io->cnext = prev_io;	      
	      }
	      new_io->type = _NASD_CHEOPS_READ_IO;
	      prev_io = new_io;
	      _nasd_cheops_init_io(&sac->pda_list[i], pio, new_io);
	      _nasd_cheops_enq_io(new_io);  
	    }
	  }
	}
	
	/* wait until all ios complete */ 
	NASD_LOCK_MUTEX(pio->flag_mutex);
	while ( pio->num_done < pio->num_ios ) {
	  NASD_WAIT_COND(pio->ios_all_done, pio->flag_mutex);
	}
	NASD_UNLOCK_MUTEX(pio->flag_mutex);
	
	pio->first_io= new_io; /* child ios are chained in the reverse order they were enqueued */

	/* warning!: pio can not be touched until 
	   all ios complete, cuz otherwise we could into races with the doio threads */

	*out_len = pio->out_datalen;
	if (pio->status)
	  rc = pio->status;
	else
	  rc = pio->nasd_status;
	/* if everything is ok, free the pdas, and the io items */
	_nasd_cheops_asm_freeall(asm); 
	_nasd_cheops_io_freeall(pio); 

	return 0;
}

int
_nasd_cheops_raid1_write(nasd_asm_t *asm, nasd_len_t *out_len)
{
        int rc =0, i;
	_nasd_cheops_aggr_io_t *pio;                /* parent io */
	_nasd_cheops_io_t      *prev_io, *new_io;
	NASD_stripe_access_t   *sac;

	/* enqueue io items corresponding to each pda in the asm  */
	NASD_Malloc(pio, sizeof(_nasd_cheops_aggr_io_t), (_nasd_cheops_aggr_io_t *));
	if (pio == NULL)
	  return ENOMEM;

	rc =  nasd_mutex_init(&(pio->flag_mutex));
        if (rc) {
          (cheops_Msg "Error: can not initialize mutex\n");
          return rc; 
        }
	rc = nasd_cond_init(&(pio->ios_all_done));
        if (rc) {
          (cheops_Msg "Error: can not initialize condition variable\n");
          return rc;
        }
	pio->num_ios = asm->num_su_accessed;
	pio->num_done = 0;
	pio->in_datalen = asm->datalen;
	pio->first_io = NULL;
	pio->out_datalen = 0;
	pio->status = pio->nasd_status = 0;

#if _NASD_CHEOPS_RAID1_DEBUG > 2
	(cheops_Msg "raid1 write: sending pio with %d children\n", pio->num_ios);
	_nasd_cheops_print_asm(asm);
#endif
	prev_io = NULL;
	for (sac= asm->first_stripe_access; sac != NULL; sac = sac->next) {
	  if (!sac)
	    break;

	  for (i=0; i< _NASD_CHEOPS_MAX_STRIPE_SIZE; i++) {
	    if (sac->pda_list[i].buf != NULL) {
	      NASD_Malloc(new_io, sizeof(_nasd_cheops_io_t), (_nasd_cheops_io_t *));
	      bzero((char *)new_io, sizeof(_nasd_cheops_io_t));
	      if (new_io == NULL) {
		(cheops_Msg "malloc failed ...\n");
		return ENOMEM;
	      } else {
		  new_io->cnext = prev_io;	      
	      }
	      new_io->type = _NASD_CHEOPS_RAID1_WRITE_IO;
	      prev_io = new_io;
	      _nasd_cheops_init_io(&sac->pda_list[i], pio, new_io);
	      _nasd_cheops_enq_io(new_io);  
	    }
	  }

	}

	/* wait until all ios complete */
	NASD_LOCK_MUTEX(pio->flag_mutex);
	while ( pio->num_done < pio->num_ios ) {
	  NASD_WAIT_COND(pio->ios_all_done, pio->flag_mutex);
	}
	NASD_UNLOCK_MUTEX(pio->flag_mutex);

	pio->first_io= new_io; /* child ios are chained in the reverse order they were enqueued */

	/* warning!: pio can not be touched until 
	   all ios complete, cuz otherwise we could into races with the doio threads */

	*out_len = pio->out_datalen;	
	if (pio->status)
	  rc = pio->status;
	else
	  rc = pio->nasd_status;

	/* if everything is ok, free the pdas, and the io items */
	_nasd_cheops_asm_freeall(asm); 
	_nasd_cheops_io_freeall(pio); 

	return rc;
}


