/*
 * nasd_cheops_mgr_init.c
 */
/*
 * Initialization and all for the Cheops storage manager
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */
#include <nasd/nasd_options.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <sys/errno.h>
#include <stdio.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_cheops_types.h>
#include <nasd/nasd_cheops_cache.h>
#include <nasd/nasd_cheops_common.h>
#include <nasd/nasd_cheops_mgr_common.h>
#include <nasd/nasd_cheops_mgr_structs.h>
/*
 * XXX this should be fixed. It will not be correctly included
 * as a dependency on many platforms. Also, the relative path
 * thing makes it hard for alternatively-organized trees, and
 * just give up on including something like this in the kernel.
 */
#include "../cheops/shared/nasd_cheops_cl_drive.h"

/* extern proto */
extern int nasd_cheops_bsdir_cache_init();
extern int nasd_cheops_bsdir_cache_shutdown();
extern int nasd_cheops_sst_mgr_init();
extern int nasd_cheops_sst_mgr_shutdown();
extern int nasd_cheops_bs_qos_init();

/* local proto */
void nasd_cheops_mgr_io_batcher(nasd_threadarg_t ignored);
int nasd_cheops_mgr_init_drives();
int dmap_add_drive(nasd_cheops_dmap_t *dm,nasd_drive_handle_t h, char *dr_name, char *dr_port, int di, int *partnum);
void dmap_print(nasd_cheops_dmap_t *dm);
int nasd_cheops_read_exports_file (char *fname, int *num_drives);
int nasd_cheops_mgr_shutdown();

#define NASD_CHEOPS_DRIVE_RESOURCE_FILE "./cheops.dtab"
#define NASD_CHEOPS_DMAP_FILE "./cheops.dmap"
#define NASD_CHEOPS_PARTSIZE 1024*100 /* x 8K block size = 800 Mbyte partition */

/* nasty globals */
nasd_thread_t dr_thread; /* thread to prefetch */
int Nasd_dmap_fd;
nasd_cheops_dmap_t *dmap;
int lastDrive;
int NumDrives;
int Partition_id;
int  dr_id[MAX_NUM_DRIVES];
char dr_name[MAX_NUM_DRIVES][_NASD_CHEOPS_MAX_FNAME_LEN];
char dr_port[MAX_NUM_DRIVES][_NASD_CHEOPS_MAX_FNAME_LEN];

int 
nasd_cheops_mgr_init()
{
       int i, rc;

       rc = nasd_cl_p_init();
       if(rc) {
	 (cheops_Msg "client library initialization failed: %u (%s)\n",
	  rc, nasd_error_string(rc));
	 return rc;
       }
       
       rc = nasd_cheops_mgr_init_drives();
       if(rc) {
	 (cheops_Msg "cheops_mgr_init_drives() failed: %u (%s)\n",
	  rc, nasd_error_string(rc));
	 return rc;
       }
       
       rc = nasd_cheops_bs_qos_init();
       if(rc) {
	 (cheops_Msg "cheops_bs_qos_init() failed: %u (%s)\n",
	  rc, nasd_error_string(rc));
	 return rc;
       }
       
       rc = nasd_cheops_bsdir_cache_init();    
       if (rc) 
	 return rc;
       
       rc = nasd_cheops_sst_mgr_init();
       if (rc)
	 {	    
	   nasd_cheops_bsdir_cache_shutdown(); 
	   return rc;
	 }
       
       (cheops_Msg "initialization done...\n");
       (cheops_Msg "full dmap:\n");
        dmap_print(dmap);

    rc = nasd_thread_create(&dr_thread, nasd_cheops_mgr_io_batcher, (void *)((unsigned long)i));
    if (rc) {
	  (cheops_Msg "error spawning thread");
	}

       return 0; /* init success */

} 


/* Note: if MAX_NUM_DRIVES is increased to a large size, stack allocation should be
changed to mallocs */
int
nasd_cheops_mgr_init_drives()
{
	int  i, nb, rc, num_drives;
	int  partnum;
	nasd_cheops_dmap_t *dm;
	nasd_drive_handle_t hlist[MAX_NUM_DRIVES];
	nasd_drive_handle_t h;
	char *bp;

	/* read drive resources (export) */
	rc = nasd_cheops_read_exports_file(NASD_CHEOPS_DRIVE_RESOURCE_FILE, 
					   &num_drives);
	if (rc) {
	  (cheops_Msg "drive initization failed ..\n");
	  return _NASD_CHEOPS_ERR_INIT_FAILED;
	}
	
	/* bind to drives */
	for (i=0; i< num_drives; i++) {
	  rc = nasd_bind_to_drive(dr_name[i], dr_port[i],
	    NASD_BIND_DCE_DIRECT, NULL, 0, &h);
	  if(rc != NASD_SUCCESS) {
	    (cheops_Msg "could not bind to drive [%s:%s]: status=%u (%s)\n",
	     dr_name[i], dr_port[i], rc, nasd_error_string(rc));
	    return _NASD_CHEOPS_ERR_INIT_FAILED;
	  }
	  hlist[i] = h;
        }

	/* update drive map and write to disk*/
	NASD_Malloc(dm, sizeof(nasd_cheops_dmap_t), (nasd_cheops_dmap_t *) );
	bzero((char *)dm, sizeof(nasd_cheops_dmap_t));
	bp = (char *) dm;

	if ( (Nasd_dmap_fd = open(NASD_CHEOPS_DMAP_FILE, O_RDWR, 0666)) > 0 ) {

	     lseek(Nasd_dmap_fd, 0, SEEK_SET);
	     nb = read(Nasd_dmap_fd, bp, sizeof(nasd_cheops_dmap_t));
	     if (nb != sizeof(nasd_cheops_dmap_t)) {
	       (cheops_Msg "found old dmap file, could not read drive map!..exiting...\n");
	       nasd_cheops_mgr_shutdown();
	       exit(1);
	     } else
	       (cheops_Msg "found old dmap file, read drive map successfully!...\n");

	} else { 

	  if ( (Nasd_dmap_fd = open(NASD_CHEOPS_DMAP_FILE, O_RDWR|O_CREAT, 0666))==0 ) {
	    return _NASD_CHEOPS_ERR_CANT_CREATE_FILE;
	  }

	}

	/* add new drives to dmap (if not already there) */
	for (i=0; i< num_drives; i++) {
	  dmap_add_drive(dm, hlist[i], dr_name[i], dr_port[i], dr_id[i], &partnum);
	}
	
	/* write dmap to file */
	nb = write(Nasd_dmap_fd, (caddr_t )dm, sizeof(nasd_cheops_dmap_t));
	if (nb != sizeof(nasd_cheops_dmap_t)) {
	  (cheops_Msg "found old imap file, could not write drive map!\n");
	  nasd_cheops_mgr_shutdown();
	  exit(1);
	}

	dmap = dm;
	lastDrive = 0;
	NumDrives = num_drives;

	return 0;
}


void
nasd_cheops_mgr_io_batcher(
  nasd_threadarg_t  ignored)
{
     while (1) {      
       /* */
       sleep(1);
       pthread_testcancel();
     }
}

int 
nasd_cheops_mgr_shutdown()
{
      int rc;

      rc = pthread_cancel(dr_thread);
      if (rc) {
	(cheops_Msg "unclean shutdown...\n");
      } else
	(cheops_Msg "clean shutdown...\n");

      return 0;
}

int
nasd_cheops_read_exports_file (char *fname, int *num_drives)
{
        char line[1024];
        char foo[128];
        FILE *fd;
        int i;
        int j;
        int len;

	(cheops_Msg "nasd_cheops_read_exports_file %s\n",
	 fname);

        if (fname==NULL) {
         (cheops_Msg "%s (%d):nasd_cheops_read_exports_file received empty filename\n", __FILE__, __LINE__);
         return EINVAL;
        }

        if ( (fd = fopen(fname, "r")) == NULL){
          (cheops_Msg "%s (%d):nasd_cheops_read_exports_file() unable to open file %s\n", __FILE__, __LINE__, fname);
          return ENOENT;
        }

        i=0;
        while ( fgets(line, 1024, fd) != NULL){
          if (line[0]=='#')
            continue;
          len = strlen(line);
          line[len-1]='\0';
          if(!strncmp(line,"drive", 5)) {
	    /* NASD_Malloc(dr_name[i], len, (char *));            NASD_Malloc(dr_port[i], len, (char *));
	    */
            sscanf(line,"%s %d %s %s", foo, &(dr_id[i]), dr_name[i], dr_port[i]);
            i++;
          }
          if (i>= MAX_NUM_DRIVES)
            break;
        }
        *num_drives = i;
        return 0;
 
}

int
nasd_cheops_mgr_bind_to_drive(_nasd_cheops_dr_handle_t *dh)
{
        int rc;

        rc = nasd_bind_to_drive(dh->dr_name, dh->dr_port,
				 NASD_BIND_DCE_DIRECT, NULL, 0,
				 &dh->h);
	if (rc) {
	  (cheops_Msg "ERROR: cannot bind to server %s\n", dh->dr_name);
	  dh->h = NULL;
	  return rc;
	}

	return 0;
}


int
dmap_add_drive(nasd_cheops_dmap_t *dm, nasd_drive_handle_t h, char *dr_name, char *dr_port, int dr_id, int *partnum)
{
     int i;
     int rc;
     int pn;
     nasd_boolean_t found;
     nasd_cookie_t in_cookie;
     nasd_status_t nasd_status;
     error_status_t status;
     nasd_key_t in_key;

     i=0;
     found = NASD_FALSE;
     while ( (found==NASD_FALSE) && (dm->dinfo[i].dr_name[0] != '\0') ) {
       if ( (strcmp(dm->dinfo[i].dr_name, dr_name)==0) || (dm->dinfo[i].dr_id==dr_id) ) {
	 found = NASD_TRUE;
	 if ( ! ((strcmp(dm->dinfo[i].dr_name, dr_name)==0) && (dm->dinfo[i].dr_id==dr_id)) ) {
	   (cheops_Msg "drive %s has changed name/id....\n", dr_name);
	   (cheops_Msg "old id %d, new id %d \n",  dm->dinfo[i].dr_id,dr_id);
	   (cheops_Msg "old name %s, new name %s \n",  dm->dinfo[i].dr_name, dr_name);
	   exit(1);
	 }
       }
       if (found==NASD_TRUE) 
	 break;
       else
	 i++;
     }
     
     *partnum = dm->dinfo[i].partnum;
  
     if (found ==NASD_FALSE) {
       dm->dinfo[i].dr_id = dr_id;
       dm->dinfo[i].h = h;
       strcpy(dm->dinfo[i].dr_name, dr_name);
       strcpy(dm->dinfo[i].dr_port, dr_port);
       
       (cheops_Msg "Creating partition on drive %d: %s \n", i, dr_name);

       rc = _nasd_cheops_part_create_h_dr(h, Partition_id, in_key, NASD_CHEOPS_PARTSIZE, &nasd_status, &status);

       if (rc) {
	 (cheops_Msg "could not create partition! \n");
	 return rc; /* change to handle errors */
       } else
	 (cheops_Msg "success...\n");

       dm->dinfo[i].partnum = Partition_id;
       dm->dinfo[i].max_capacity = NASD_CHEOPS_PARTSIZE;
       *partnum = Partition_id;
     }    
  
     return 0;

}

void
dmap_print(nasd_cheops_dmap_t *dm) 
{
     int i=0;
     
     while ( dm->dinfo[i].dr_name[0] != '\0') {
       (cheops_Msg "drive name: %s, ", dm->dinfo[i].dr_name);
       (cheops_Msg "drive id: %d, ", dm->dinfo[i].dr_id);
       (cheops_Msg "drive part num: %d\n", dm->dinfo[i].partnum);
       i++;
     }
}

int
dmap_alloc_chunk(nasd_cheops_dmap_t *dm, int g, nasd_disk_ident_t *dr_id, int *dr_index)
{
     int i, su;

     i = lastDrive;
     for (su=0; su<g; su++) {
       dr_id[su] = dm->dinfo[i].dr_id;
       dr_index[su] = i;
       i = (i+1) % NumDrives;
     }

     /* lastDrive = (lastDrive+1)%NumDrives;  */
     return 0;

}






