/*
 * nasd_cheops_mgr_qos.c
 */
/*
 * Qos allocation
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_cheops_types.h>
#include <nasd/nasd_cheops_common.h>
#include <nasd/nasd_cheops_mgr_common.h>

/* map a Qos requirement structure onto a layout                       */
/* algorithms projects the requirements onto a set of classes, one for */
/* access size, one for throughput/bandwidth, one for reliability, and */
/* for access pattern (write ratio for now)                            */
/* The list of classes form a tuple which is used to index a table     */
/* storing the best layout for each tuple of requirement classes.      */

extern int NumDrives;
int Nasd_cheops_layouts[4][4][4];

#define CUTOFF_SMALL_IO        16
#define CUTOFF_VERY_LARGE_IO   512
#define CUTOFF_LOW_WRITES      0.2
#define CUTOFF_HIGH_WRITES     0.5
 
#define NASD_CHEOPS_STRIPE_WIDTH 8      /* the default stripe width for cheops objects */
#define NASD_CHEOPS_STRIPE_SIZE 128*1024  /* the default stripe unit size for cheops objects */

/* debugging */
#define NASD_CHEOPS_QOS_DEBUG 0

int 
nasd_cheops_bs_req_to_layout(_nasd_cheops_qos_req_t qr,
			     _nasd_cheops_bs_layout_t *lp)
{
        _nasd_cheops_ac_pattern_class_t    ac_pattern_class;
        _nasd_cheops_wratio_class_t        wratio_class;
	_nasd_cheops_reli_class_t          reli_class;

	
	/* classify the write ratio */
	if (qr.write_ratio < CUTOFF_LOW_WRITES)
	  wratio_class = _NASD_CHEOPS_QOS_MOSTLY_READ;
	else if (qr.write_ratio > CUTOFF_HIGH_WRITES)
	  wratio_class = _NASD_CHEOPS_QOS_MOSTLY_WRITE;
	else
	  wratio_class =  _NASD_CHEOPS_QOS_READWRITE;
	
	/* classify the bandwidth/throughput class */
	if ( (qr.parallelism==1) && (qr.typ_access_size<CUTOFF_SMALL_IO) )
	  ac_pattern_class = _NASD_CHEOPS_QOS_ONE_SMALL_IO;
	else if ( (qr.parallelism>1) && (qr.typ_access_size<CUTOFF_SMALL_IO) )
	  ac_pattern_class = _NASD_CHEOPS_QOS_MANY_SMALL_IOS;
	else if ( (qr.parallelism==1) )
	  ac_pattern_class = _NASD_CHEOPS_QOS_ONE_LARGE_IO;
	else
	  ac_pattern_class = _NASD_CHEOPS_QOS_MANY_LARGE_IOS;
	         
	reli_class = qr.reli_class;

	lp->raid_level = Nasd_cheops_layouts[ac_pattern_class][wratio_class][reli_class];
	lp->stripe_unit_size = NASD_CHEOPS_STRIPE_SIZE;
	if (lp->stripe_unit_size < qr.min_su_size)
	  lp->stripe_unit_size = qr.min_su_size;
	lp->num_data_col = NASD_CHEOPS_STRIPE_WIDTH; 
#if NASD_CHEOPS_QOS_DEBUG > 0	
	(cheops_Msg "inside bs_req to layout, su=%d, gsize=%d \n",
	 lp->stripe_unit_size, lp->num_data_col);
#endif

	if (NumDrives == 1)
	  lp->raid_level = RAID0;
	else if (NumDrives == 2) {
	  if (lp->raid_level == RAID5)
	    lp->raid_level = RAID1;
	} 

	switch(lp->raid_level) {
	case RAID0:
	  lp->num_col = lp->num_data_col;
	  break;
	case RAID1: 
	  lp->num_col = 2*lp->num_data_col;
	  if (lp->num_col > NumDrives) {
	    lp->num_col = (NumDrives/2) *2;
	    lp->num_data_col = NumDrives/2;
	  }
	  break;
	case RAID5:
	  lp->num_col = lp->num_data_col + 1;
	  if (lp->num_col > NumDrives) {
	    lp->num_col = NumDrives;
	    lp->num_data_col = NumDrives-1;
	  }
	  break;
	default:
	  NASD_ASSERT(0);
	}
       
	return 0;
};

int
nasd_cheops_bs_qos_init()
{

       /* object requires low reliability: 
	. map onto a RAID level 0
	*/

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                 [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                 [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_READWRITE]
	                 [_NASD_CHEOPS_QOS_LOW_RELI] = RAID0;

      /* object requires medium reliability: 
	 . map onto a RAID level 5 (if mostly read)
	 . map onto a RAID level 1 (if mostly write)
	 */

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                 [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                 [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_READWRITE]
	                 [_NASD_CHEOPS_QOS_MOD_RELI] = RAID5;

      /* object requires high reliability: 
	 . map onto a RAID level 1 
	 */

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_MOSTLY_READ]
	                 [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_MOSTLY_WRITE]
	                 [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_SMALL_IO]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_SMALL_IOS]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

     Nasd_cheops_layouts[_NASD_CHEOPS_QOS_ONE_LARGE_IO]
	                [_NASD_CHEOPS_QOS_READWRITE]
	                [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

      Nasd_cheops_layouts[_NASD_CHEOPS_QOS_MANY_LARGE_IOS]
	                 [_NASD_CHEOPS_QOS_READWRITE]
	                 [_NASD_CHEOPS_QOS_HIGH_RELI] = RAID1;

      return 0;
}



