/*
 * nasd_cheops_cache.c
 */
/*
 * Cache module for the CHEOPS (storage manager and client clerk)
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */
#include <nasd/nasd_options.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>

#include <sys/errno.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_cheops_cache.h>
#include <nasd/nasd_cheops_common.h>
#include <nasd/nasd_cheops_locks.h>

#define KEY_EQ(key1, key2) ( ( (key1).object_id == (key2).object_id ) && \
( (key1).record_id == (key2).record_id ) )

#define _NASD_CHEOPS_CACHE_DEBUG 0

/* prototypes (forward declarations) */
int _nasd_cheops_cache_compute_htab_size(int);
int _nasd_cheops_cache_hash_key(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key);
void _nasd_cheops_cache_print_key(_nasd_cheops_cache_key_t key);
int _nasd_cheops_cache_lookup_ent(_nasd_cheops_cache_t *cp, 
				 _nasd_cheops_cache_key_t key, caddr_t re, nasd_boolean_t miss_allocate);
int _nasd_cheops_cache_get_ent (_nasd_cheops_cache_t *cp, 
			       _nasd_cheops_cache_ent_t **ep);
int _nasd_cheops_cache_init_ent(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key, int flags,
			       _nasd_cheops_cache_ent_t *ep);

/* deallocate cache object */
int 
_nasd_cheops_cache_destroy(_nasd_cheops_cache_t *cp)
{
        NASD_Free(cp->buckets, cp->nbuckets*sizeof(_nasd_cheops_cache_ent_t));
	NASD_Free(cp->ents, cp->csize*sizeof(_nasd_cheops_cache_ent_t));
	NASD_Free(cp->bufs, cp->csize*cp->bsize );
	NASD_Free(cp, sizeof(_nasd_cheops_cache_t));
	return 0;

};

/* create cache of size csize with entries of size bsize    */
int 
_nasd_cheops_cache_create(int csize, int bsize, _nasd_cheops_cache_type_t ctype, _nasd_cheops_cache_t **cpp)
{
        _nasd_cheops_cache_t      *cp;
	int                      hsize;
	int                      i, rc;
	_nasd_cheops_cache_ent_t  *p, *q, *bu;
	caddr_t                  buf;
	_nasd_cheops_cache_key_t  zkey;
	
	zkey.object_id = 0;
	zkey.record_id = 0;

	/* verify arguments */
	if ((csize <= 1) || 
	    ( (ctype != _NASD_CHEOPS_CACHE_TYPE_SMAP) && (ctype != _NASD_CHEOPS_CACHE_TYPE_GMAP)
	      && (ctype != _NASD_CHEOPS_CACHE_TYPE_DMAP) && (ctype != _NASD_CHEOPS_CACHE_TYPE_IMAP)))
	  return(EINVAL);
	
	if (csize > _NASD_CHEOPS_MAX_CACHE_SIZE)
	  return (EDOM);
	
	/* allocate memory for cache data strutures */
	NASD_Malloc(cp, sizeof(_nasd_cheops_cache_t), (_nasd_cheops_cache_t *));
	if (cp == NULL) {
	  return(ENOMEM);
	}
	bzero((char *)cp, sizeof(_nasd_cheops_cache_t));
	*cpp = cp;
	cp->ctype = ctype;
	cp->victim = 0;
	
	cp->csize = csize;
	NASD_Malloc(cp->ents, csize*sizeof(_nasd_cheops_cache_ent_t), (_nasd_cheops_cache_ent_t *));
	if ( cp->ents == NULL) {
	  NASD_Free(cp, sizeof(_nasd_cheops_cache_t));
	  *cpp = NULL;
	  return (ENOMEM);
	}

	/* set up hash table */
	hsize = _nasd_cheops_cache_compute_htab_size(csize);
	cp->nbuckets = hsize;
	NASD_Malloc(cp->buckets, cp->nbuckets*sizeof(_nasd_cheops_cache_ent_t),
		    (_nasd_cheops_cache_ent_t *));
	if (cp->buckets == NULL) {
	  NASD_Free(cp, sizeof(_nasd_cheops_cache_t));
	  *cpp = NULL;
	  return ENOMEM;
	}
	bzero((char*)cp->buckets, cp->nbuckets*sizeof(_nasd_cheops_cache_ent_t));
	
	for(i=0;i<cp->nbuckets;i++) {
	  bu = &cp->buckets[i];
	  bu->bnext = bu->bprev = bu;
	}

	/* put all entries on freelist */
	for (i=0; i<csize; i++) {
	  p = &(cp->ents[i]);
	  if (i==0) {
	    cp->free_list = p;
	    q = p;
	  } else {
	    q->fnext = p;
	    p->fnext = NULL;
	    q = p;
	  }
	  
	  rc = nasd_mutex_init (&cp->ents[i].m);
	  if (rc) {
	    NASD_Free(cp->ents, cp->csize*sizeof(_nasd_cheops_cache_ent_t));
	    NASD_Free(cp, sizeof(_nasd_cheops_cache_t));
	    *cpp = NULL;
	    return (ENOMEM);
	  }
	  /* FIX -- should we bzero entries ? */
	  cp->ents[i].flags = _NASD_CHEOPS_CACHE_ENT_FREE;
	  cp->ents[i].hold_wait_refcnt = 0;
	  cp->ents[i].key = zkey;	  
	  cp->ents[i].data = NULL;
	}

	/* buffers */
	cp->bsize = bsize;
	NASD_Malloc( buf, csize*bsize, (caddr_t ));
	if (buf==NULL) {
	  NASD_Free(cp->buckets, cp->nbuckets*sizeof(_nasd_cheops_cache_ent_t));
	  NASD_Free(cp->ents, cp->csize*sizeof(_nasd_cheops_cache_ent_t));
	  NASD_Free(cp, sizeof(_nasd_cheops_cache_t));
	  *cpp = NULL;
	  return ENOMEM;
	}

	for (i=0;i<csize;i++)
	  cp->ents[i].data = buf + i*bsize;	
	cp->bufs = buf;

	if (rc) {
	  NASD_Free(cp->buckets, cp->nbuckets*sizeof(_nasd_cheops_cache_ent_t));
	  NASD_Free(cp->ents, cp->csize*sizeof(_nasd_cheops_cache_ent_t));
	  NASD_Free(cp->bufs, csize*bsize);
	  NASD_Free(cp, sizeof(_nasd_cheops_cache_t));	  
	  *cpp = NULL;
	  return ENOMEM;
	}

	/* init cache lock */
	_nasd_cheops_lockInit(&cp->lock);

	/* init statistics */
	cp->stats.misses = cp->stats.hits = cp->stats.updates = cp->stats.invalidates = 0;

	/* init lru and dirty */
	cp->lru.lnext = cp->lru.lprev = &cp->lru;
	cp->dirty.dnext = cp->dirty.dprev = &cp->dirty;

	return 0;

};

/* read an entry from the cache, copy it out into user buffer */
int
_nasd_cheops_cache_read(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key, caddr_t re)
{
       nasd_boolean_t found, inserted;
       _nasd_cheops_cache_ent_t      *e, *q;
       _nasd_cheops_cache_hash_ent_t *hp;
       int rc;
       int flags;
       nasd_boolean_t miss_allocate; /* allocate space and read in entry into the cache on miss */

#if _NASD_CHEOPS_CACHE_DEBUG > 0
       (cheops_Msg "\ncache: received read for object %d ->", key.record_id);
#endif

       /* try to look for the entry with a shared lock */
       miss_allocate = NASD_FALSE;
       _nasd_cheops_lockShared(&cp->lock);
       found = _nasd_cheops_cache_lookup_ent(cp, key, re, miss_allocate);
       _nasd_cheops_unlockShared(&cp->lock);
       if (found) {
#if _NASD_CHEOPS_CACHE_DEBUG > 0
	 (cheops_Msg "cache: found object %d\n", key.record_id);
#endif
	 _NASD_CHEOPS_CACHE_INC_STAT(cp, hits);
	 return 0;
       } else {
	 miss_allocate = NASD_TRUE;
	 _nasd_cheops_lockExclusive(&cp->lock);
	 inserted = _nasd_cheops_cache_lookup_ent(cp, key, re, miss_allocate);
#if _NASD_CHEOPS_CACHE_DEBUG > 0
	 (cheops_Msg "cache: brought in object %d\n", key.record_id);
#endif
	 _nasd_cheops_unlockExclusive(&cp->lock);
	 if (!inserted) {
	   return ENOMEM;
	 }
	 else
	   return 0;
       }	 	 

}

/* invalidate an entry */
int 
_nasd_cheops_cache_mark_invalid (_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key)
{
        nasd_boolean_t  found = NASD_FALSE;
	_nasd_cheops_cache_hash_ent_t *hp;
	_nasd_cheops_cache_ent_t      *e;
	_nasd_cheops_cache_ent_t      *enext, *eprev;

	_NASD_CHEOPS_CACHE_INC_STAT(cp, invalidates);

	return 0;

};

int
_nasd_cheops_cache_mark_dirty (_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key)
{

        nasd_boolean_t  found = NASD_FALSE;
	_nasd_cheops_cache_ent_t *bp;
	_nasd_cheops_cache_ent_t *e;
	_nasd_cheops_cache_ent_t *enext, *eprev;

	_NASD_CHEOPS_CACHE_INC_STAT(cp, updates);
	bp = &( cp->buckets[_nasd_cheops_cache_hash_key(cp, key)] );
	_nasd_cheops_lockExclusive( &cp->lock );
	for (e=bp->bnext; e!=bp; e=e->bnext) {
	     if (KEY_EQ(key, e->key)) {
	       found = NASD_TRUE; 
	       break;
	     }
	}

	if ( !found ) {
	     _nasd_cheops_unlockExclusive( &cp->lock);
	     return _NASD_CHEOPS_ERR_NOTFOUND;
	} else {
	     cp->writeback(e, cp->bsize); 	 
	     _nasd_cheops_unlockExclusive( &cp->lock);
	     return 0;
	     /* FIX -- put on dirty list and flush asynchronously later?
	     e->flags = e->flags | _NASD_CHEOPS_CACHE_ENT_DIRTY;
	     e->dnext = &cp->dirty;
	     e->dprev = cp->dirty.dprev;;
	     e->dnext->dprev = e;
	     e->dprev->dnext = e;
	     */
	     return 0;
	}

}

/* show statistics                                     */
void 
_nasd_cheops_cache_show_stats(_nasd_cheops_cache_t *cp)
{
        (cheops_Msg "Cache statistics: \n");
        (cheops_Msg "\t hits: %lu", cp->stats.hits);
        (cheops_Msg "\t misses: %lu", cp->stats.misses);
        (cheops_Msg "\t invalidates: %lu", cp->stats.invalidates);
        (cheops_Msg "\t fixes: %lu\n", cp->stats.fixes);

};

/* */
int
_nasd_cheops_cache_hold_ent(_nasd_cheops_cache_t *cp, 
			   _nasd_cheops_cache_key_t key, 
			   _nasd_cheops_cache_ent_t **ep)
{
        nasd_boolean_t found=NASD_FALSE;
	_nasd_cheops_cache_ent_t *e, *ne;
	_nasd_cheops_cache_ent_t *bp;
	int rc;
	int flags;
	
	_nasd_cheops_lockExclusive( &cp->lock);
	
	bp = &( cp->buckets[_nasd_cheops_cache_hash_key(cp, key)] );
	for (e=bp->bnext; e!=bp; e=e->bnext) {
	  if (KEY_EQ(key, e->key)) {
	    found = NASD_TRUE; 
	    break;
	  }
	}
	
	if (found) {
	  _NASD_CHEOPS_CACHE_ENT_LOCK(e);
	  /* quiesce, so that we do not get junk when we bcopy */
	  while(e->flags&_NASD_CHEOPS_CACHE_ENT_HOLD) {
	    e->hold_wait_refcnt++;
	   _NASD_CHEOPS_CACHE_ENT_WAIT(e);
	   e->hold_wait_refcnt--;
	  }
	  e->flags |= _NASD_CHEOPS_CACHE_ENT_HOLD;
	  _NASD_CHEOPS_CACHE_ENT_UNLOCK(e);
	  /* remove from LRU */
	  e->lnext->lprev = e->lprev;
	  e->lprev->lnext = e->lnext;
	  /* put at head of LRU */
	  e->lnext = &cp->lru;
	  e->lprev = cp->lru.lprev;
	  e->lnext->lprev = e;
	  e->lprev->lnext = e;
	  *ep = e;
	  _nasd_cheops_unlockExclusive( &cp->lock);
	  return 0;
       }
		
	/* entry was not found */
	rc = _nasd_cheops_cache_get_ent(cp, &ne);
	if (rc) {
	  _nasd_cheops_unlockExclusive( &cp->lock);
	  return rc;
	}
	flags = 0;
	_nasd_cheops_cache_init_ent(cp, key, flags, ne);
	
	ne->bnext = bp->bnext;
	ne->bprev = ne->bnext->bprev;
	ne->bnext->bprev = ne;
	ne->bprev->bnext = ne;
	
	ne->lnext = &cp->lru;
	ne->lprev = ne->lnext->lprev;
	ne->lnext->lprev = ne;
	ne->lprev->lnext = ne;
	*ep = ne;
       _nasd_cheops_unlockExclusive( &cp->lock);
       
       _NASD_CHEOPS_CACHE_ENT_LOCK(ne);
       while(ne->flags&_NASD_CHEOPS_CACHE_ENT_HOLD) {
	 ne->hold_wait_refcnt++;
	 _NASD_CHEOPS_CACHE_ENT_WAIT(ne);
	 ne->hold_wait_refcnt--;
       }

       ne->flags |= _NASD_CHEOPS_CACHE_ENT_HOLD;
       _NASD_CHEOPS_CACHE_ENT_UNLOCK(ne);       
       
       return 0;

}


/* unfix an entry, so it can be replaced */
int
_nasd_cheops_cache_release_ent(_nasd_cheops_cache_t *cp, 
			      _nasd_cheops_cache_ent_t *ep)
{
	/* decrement refcnt and broadcast condition */
        _NASD_CHEOPS_CACHE_ENT_LOCK(ep);
	ep->flags = ep->flags & ~_NASD_CHEOPS_CACHE_ENT_HOLD;
	if ( (ep->hold_wait_refcnt) ) /* if there is someone waiting to hold the entry */
	  _NASD_CHEOPS_CACHE_ENT_BCAST(ep);
	_NASD_CHEOPS_CACHE_ENT_UNLOCK(ep);	
	return 0;	
  
};


/*************************** Internal methods **********************************/

/* lookup an entry in the cache given a key, call with cp->lock held (shared or exclusive) */
/* returns 1 if entry is found or successfully brought in, 0 if not                        */

int 
_nasd_cheops_cache_lookup_ent (_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key, 
			      caddr_t re, nasd_boolean_t miss_allocate)
{
        nasd_boolean_t  found = NASD_FALSE;
	_nasd_cheops_cache_ent_t      *e;
	_nasd_cheops_cache_ent_t      *bp;
	_nasd_cheops_cache_ent_t      *ne;
	int rc, flags;

	bp = &( cp->buckets[_nasd_cheops_cache_hash_key(cp, key)] );
	for (e=bp->bnext; e!=bp; e=e->bnext) {
	  if (KEY_EQ(key, e->key)) {
	    found = NASD_TRUE; 
	    break;
	  }
	}

	if (found) {
	  _NASD_CHEOPS_CACHE_ENT_LOCK(e);
	  /* quiesce, so that we do not get junk when we bcopy */
	  while(e->flags&_NASD_CHEOPS_CACHE_ENT_HOLD) {
	    e->hold_wait_refcnt++;
	    _NASD_CHEOPS_CACHE_ENT_WAIT(e);
	    e->hold_wait_refcnt--;
	  }
	  bcopy(e->data, re, cp->bsize);
	  _NASD_CHEOPS_CACHE_ENT_UNLOCK(e);
	  /* remove from LRU */
	  e->lnext->lprev = e->lprev;
	  e->lprev->lnext = e->lnext;
	  /* put at head of LRU */
	  e->lnext = &cp->lru;
	  e->lprev = cp->lru.lprev;
	  e->lnext->lprev = e;
	  e->lprev->lnext = e;
	  return found;
	}

	/* entry not found */
	if (miss_allocate==NASD_FALSE)
	  return NASD_FALSE; /* is not in cache (not inserted) */
	else {
	  /* miss allocate -- assumes cache locked exclusively !! */
	  rc = _nasd_cheops_cache_get_ent(cp, &ne);
	  if (rc)
	    return rc;
	  flags = _NASD_CHEOPS_CACHE_ENT_LRU;
	  rc = _nasd_cheops_cache_init_ent(cp, key, flags, ne);
	  /* insert into hash chain */
	  ne->bnext = bp->bnext;
	  ne->bprev = ne->bnext->bprev;
	  ne->bnext->bprev = ne;
	  ne->bprev->bnext = ne;
	  /* insert at head of LRU */
	  ne->lnext = &cp->lru;
	  ne->lprev = ne->lnext->lprev;
	  ne->lnext->lprev = ne;
	  ne->lprev->lnext = ne;
	  bcopy(ne->data, re, cp->bsize);
	  return NASD_TRUE; /* entry has been successfully inserted */
	}

}

/* get an entry, called with cp->lock held in exclusive mode */
int
_nasd_cheops_cache_get_ent (_nasd_cheops_cache_t *cp, _nasd_cheops_cache_ent_t **ep)
{
	_nasd_cheops_cache_ent_t      *e;

	/* take an entry from the free list if one is there  */
	if (cp->free_list != NULL) {
	  e = cp->free_list;
	  cp->free_list = e->fnext;
	  e->fnext = NULL;
	  *ep = e;
	  return 0;
	}

	/* otherwise, eject an entry from the lru list */
	for(e=cp->lru.lnext;e!= &cp->lru;e=e->lnext) {
	  if (e->flags&_NASD_CHEOPS_CACHE_ENT_DIRTY) {
	    continue;
	  }
	  if (e->flags&_NASD_CHEOPS_CACHE_ENT_BUSY) {
	    continue;
	  }
	  if (e->flags&_NASD_CHEOPS_CACHE_ENT_HOLD) {
	    continue;
	  }
	  if (e->hold_wait_refcnt) {
	    continue;
	  }

	  /* found a block */
	  e->lprev->lnext = e->lnext;
	  e->lnext->lprev = e->lprev;
	  e->bnext->bprev = e->bprev;
	  e->bprev->bnext = e->bnext;
	  *ep = e;
	  return 0;
	}

	return ENOMEM;

}


int
_nasd_cheops_cache_init_ent(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key, int flags, _nasd_cheops_cache_ent_t *ep)
{
        int rc;

	rc = (cp->handle_read_fault)(key, ep->data);
	ep->key = key;
	ep->flags = flags; 
	return rc;
}


int
_nasd_cheops_cache_compute_htab_size(int csize)
{
         int n, i;

	 n = csize / 10;
	 for (i=1; i<=n; i++)  
	   i = 2*i;	 
	 i--;
	 return ( (i==0)? 1: i);

};

int
_nasd_cheops_cache_hash_key(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key)
{
        int part1, part2;
	int bitmask = 0xffaffa;
	int hash;
	int prime1 = 41;

	part1 = key.object_id;
	part2 = key.record_id;       
	/* hash = (part1*prime1) + part2&bitmask); */
	hash = part2;
	hash = hash % cp->nbuckets;
	return hash;

};

void
_nasd_cheops_cache_print_key(_nasd_cheops_cache_key_t key)
{

	(cheops_Msg "(%d,%d)", key.object_id, key.record_id);
  
}





