/*
 * nasd_cheops_locks.c
 */
/*
 * lock module: implements multi-reader single-writer locks
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 20 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */
#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <sys/errno.h>
#include <nasd/nasd_cheops_locks.h>

int 
_nasd_cheops_lockInit( _nasd_cheops_rwlock_t *l )
{
        int rc;

        /* rc = pthread_mutex_init( &(l->m), pthread_mutexattr_default);  FIXED -- using nasd macros */

        rc = nasd_mutex_init( &(l->m) );
	if (rc) {
	  return(ENOMEM);
	}
	/* rc = pthread_cond_init ( &(l->canRead), pthread_condattr_default); FIXED -- using nasd macros */

	rc = nasd_cond_init ( &(l->canRead) );
	if (rc) {
	  return(ENOMEM);
	}

	/* rc = pthread_cond_init ( &(l->canWrite),  pthread_condattr_default);  FIXED -- using nasd macros */

	rc = nasd_cond_init ( &(l->canWrite) );
	if (rc) {
	  return(ENOMEM);
	}
	
        l->nActive = 0;     /* number of active readers, -1 if active writer*/
	l->nPendingReaders = 0;
	l->nPendingWriters = 0;

	return 0;
}

int 
_nasd_cheops_lockShared( _nasd_cheops_rwlock_t *l )
{

        NASD_LOCK_MUTEX( l->m );

	l->nPendingReaders++;
	if (l->nPendingWriters >0)
	   NASD_WAIT_COND( l->canRead, l->m );
        while ( l->nActive < 0)
	   NASD_WAIT_COND ( l->canRead, l->m);
	l->nActive++;
	l->nPendingReaders--;
	
	NASD_UNLOCK_MUTEX ( l->m );
        return 0;
}

int 
_nasd_cheops_unlockShared( _nasd_cheops_rwlock_t *l)
{
        NASD_LOCK_MUTEX( l->m );

	l->nActive--;
	if ( l->nActive == 0) {			/* last reader to leave */
           NASD_UNLOCK_MUTEX( l->m );
	   NASD_SIGNAL_COND ( l->canWrite);
        } else
           NASD_UNLOCK_MUTEX( l->m );

	return 0;
}

int 
_nasd_cheops_lockExclusive( _nasd_cheops_rwlock_t *l )
{
        NASD_LOCK_MUTEX( l->m );
	l->nPendingWriters++;
	while ( l->nActive )
	   NASD_WAIT_COND(l->canWrite, l->m);
	l->nPendingWriters--;
	l->nActive = -1;
	NASD_UNLOCK_MUTEX( l->m );	

	return 0;
}

int 
_nasd_cheops_unlockExclusive( _nasd_cheops_rwlock_t *l )
{
	nasd_boolean_t wakeReaders;

	NASD_LOCK_MUTEX( l->m );
	l->nActive = 0;
	wakeReaders = (l->nPendingReaders != 0 );
	NASD_UNLOCK_MUTEX( l->m);
	if (wakeReaders) {
	   NASD_BROADCAST_COND(l->canRead);
	}
	else {
	   NASD_SIGNAL_COND(l->canWrite);
	}
	return 0;
}

int 
_nasd_cheops_downgrade( _nasd_cheops_rwlock_t *l )
{
	nasd_boolean_t wakeReaders;

	NASD_LOCK_MUTEX ( l->m );
	NASD_ASSERT(l->nActive == -1); /* to downgrade, need to be writer */
	l->nActive = 1;
	wakeReaders = (l->nPendingReaders != 0);
	NASD_UNLOCK_MUTEX( l->m);
	if ( wakeReaders)
	  NASD_BROADCAST_COND(l->canRead );
	return 0;
}

int 
_nasd_cheops_upgrade( _nasd_cheops_rwlock_t *l )
{
	NASD_LOCK_MUTEX( l->m );
	if ( l->nActive == 1) {
	   l->nActive == -1;
	} else {              /* there are other active readers */
	   l->nPendingWriters++;
	   l->nActive--;
	   while (l->nActive)
	      NASD_WAIT_COND(l->canWrite, l->m);
	   l->nPendingWriters--;
	   l->nActive= -1;
        }
	NASD_UNLOCK_MUTEX ( l->m );
	return 0;
	
}



