/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Howard Gobioff 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/***********************************************************************
 *
 * nasd_nonce_mgmt.c
 *
 * Handle checking if a nonce is fresh or resued
 ***********************************************************************/
#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_nonce_mgmt.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_security_time.h>

#define PRINT_ERROR

#ifdef DEBUG_ALL
#define PRINT_ERROR
#define DEBUG_INIT_NONCE
#define DEBUG_CHECK_NONCE
#endif
nasd_nonce_el_t * nonce_table[NASD_NONCE_BUCKETS];
NASD_DECLARE_MUTEX(nonce_table_lock[NASD_NONCE_BUCKETS]);

nasd_freelist_t *nasd_cl_nonce_freelist = NULL;
#define NASD_MAX_FREE_NONCE NASD_NONCE_BUCKETS*3
#define NASD_NONCE_INC      4
#define NASD_NONCE_INITIAL  NASD_NONCE_BUCKETS

#define GETNONCE(_n_) \
  NASD_FREELIST_GET(nasd_cl_nonce_freelist,_n_,next,(nasd_nonce_el_t *))
#define FREENONCE(_n_) \
  NASD_FREELIST_FREE(nasd_cl_nonce_freelist,_n_,next)

void
nasd_cl_shutdown_nonce_freelist(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_cl_nonce_freelist,next,(nasd_nonce_el_t *));
}

nasd_status_t
nasd_init_nonce_mgmt(
  nasd_shutdown_list_t  *sl)
{
#if NASD_SECURE_RPCS_ENABLE > 0
  int i;
  nasd_status_t rc=NASD_SUCCESS;
  nasd_nonce_el_t * curr;
  nasd_nonce_el_t * next;

#ifdef DEBUG_INIT_NONCE
  printf("DEBUG(nasd_init_nonce): entered\n");
#endif /* DEBUG_INIT_NONCE */

  for(i=0;i<NASD_NONCE_BUCKETS;i++) {
    nonce_table[i]=NULL;
    rc=nasd_mutex_init(&(nonce_table_lock[i]));
    if (rc) {
      return(rc);
    }
    rc = nasd_shutdown_proc(sl, nasd_shutdown_mutex,
                            &nonce_table_lock[i]);
    if (rc) {
      nasd_shutdown_mutex(&nonce_table_lock[i]);
      return(rc);
    }
  }

  NASD_FREELIST_CREATE(nasd_cl_nonce_freelist, NASD_MAX_FREE_NONCE,
    NASD_NONCE_INC, sizeof(nasd_nonce_el_t));
  if (nasd_cl_nonce_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(sl, nasd_cl_shutdown_nonce_freelist, NULL);
  if (rc) {
    nasd_cl_shutdown_nonce_freelist(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_cl_nonce_freelist, NASD_NONCE_INITIAL,
    next, (nasd_nonce_el_t *));

#ifdef DEBUG_INIT_NONCE
  printf("DEBUG(nasd_init_nonce): exited with nonce_count=%d\n",nonce_count);
#endif /* DEBUG_INIT_NONCE */

  rc = nasd_shutdown_proc(sl, nasd_cl_shutdown_nonce_table, NULL);
  if (rc) {
    nasd_cl_shutdown_nonce_table(NULL);
    return(rc);
  }

#endif /* NASD_SECURE_RPCS_ENABLE >0 */

  return(NASD_SUCCESS);
}

void
nasd_cl_shutdown_nonce_table(
  void  *ignored)
{
  nasd_nonce_el_t *r, *next;
  int i;

  for(i=0;i<NASD_NONCE_BUCKETS;i++) {
    for(r=nonce_table[i];r;r=next) {
      next = r->next;
      FREENONCE(r);
    }
  }
}

nasd_status_t nasd_check_nonce(nasd_capability_args_t *cap_args,
                               nasd_timespec_t *nonce){
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_timespec_t now;
  nasd_uint64 now_l,then_l,diff;
  nasd_nonce_el_t * curr,*new_el,*tmp;
  int  bucket;
  int check_first=NASD_TRUE;
  nasd_status_t rc;
#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce): entered %" NASD_64u_FMT " for drive %d\n",*(nasd_uint64*)nonce,cap_args->drive_name);
#endif /* DEBUG_CHECK_NONCE */

  nasd_security_get_drive_time(cap_args->drive_name,&now);

  

  now_l=(nasd_uint64)now.ts_nsec+(nasd_uint64)now.ts_sec*NASD_NSEC_PER_SEC;
  then_l=(nasd_uint64)nonce->ts_nsec+(nasd_uint64)nonce->ts_sec*NASD_NSEC_PER_SEC;

#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce): then was  %" NASD_64u_FMT "\n",then_l);
#endif /* DEBUG_CHECK_NONCE */

  diff=now_l-then_l;
  if (NASD_ABS(diff) > NASD_ACCEPTABLE_SKEW) {

#ifdef PRINT_ERROR
    printf("DEBUG(nasd_check_nonce) too far from now FAILED - NASD_BAD_NONCE \n");
    printf("DEBUG(nasd_check_nonce) now= %ld:%ld \n",now.ts_sec,now.ts_nsec);
    printf("DEBUG(nasd_check_nonce) then=%ld:%ld \n",nonce->ts_sec,nonce->ts_nsec);
#endif /* PRINT_ERROR */
    return(NASD_BAD_NONCE);
  }

  /* We only expire  capabilities */
  /* This check is here so we only goto the clock once -hgobioff */
  if (((cap_args->type == NASD_RED_CAPABILITY) || (cap_args->type == NASD_BLACK_CAPABILITY) )&& (now.ts_sec > cap_args->expiration_seconds)) 
    {
#ifdef PRINT_ERROR
    printf("DEBUG(nasd_check_nonce) expired capability \n");
#endif /* PRINT_ERROR */
      return(NASD_EXPIRED_CAPABILITY);
    }



  /* Neither too old nor too young so we look in the table */

  GETNONCE(new_el);
  if (new_el == NULL) {
    printf("ERROR: %s:%d out of memory\n", __FILE__, __LINE__);
    return(NASD_NO_MEM);
  }

/* We have a nasd_nonce_el_t */

  new_el->time=then_l;
  new_el->ni=cap_args->ni;
  new_el->expiration=cap_args->expiration_seconds;
  new_el->drive_name=cap_args->drive_name;


  bucket=NONCE_HASH(new_el);
  bucket=NASD_ABS(bucket%NASD_NONCE_BUCKETS);

#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) looking for %ld using bucket %d  (<%d)\n",new_el->time,bucket,NASD_NONCE_BUCKETS);
#endif /* DEBUG_CHECK_NONCE */
 
  NASD_LOCK_MUTEX(nonce_table_lock[bucket]);

#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) grabbed lock for bucket %d\n",bucket);
#endif /* DEBUG_CHECK_NONCE */
  
  /* Check the first element in the bucket */
  while (check_first) {
#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) in check_first loop \n");
#endif /* DEBUG_CHECK_NONCE */

    /* Case 1 - Empty bucket */
    if (nonce_table[bucket] == NULL ) {
      
#ifdef DEBUG_CHECK_NONCE
      printf("DEBUG(nasd_check_nonce) 1  inserting in an empty bucket \n");
#endif /* DEBUG_CHECK_NONCE */
      nonce_table[bucket]=new_el;
      new_el->next=NULL;

      NASD_UNLOCK_MUTEX(nonce_table_lock[bucket]);
      return(NASD_SUCCESS);
    } else if (nonce_table[bucket]->drive_name == new_el->drive_name ) {
      /* Case 2 - First element should be freed */ 
#ifdef DEBUG_CHECK_NONCE
	printf("DEBUG(nasd_check_nonce) first element is for the correct drive\n");
#endif /* DEBUG_CHECK_NONCE */
      if (NASD_ABS(now_l-nonce_table[bucket]->time) > NASD_ACCEPTABLE_SKEW) {
	/* We can free it */

#ifdef DEBUG_CHECK_NONCE
	printf("DEBUG(nasd_check_nonce) 2 **freeing the first element\n");
#endif /* DEBUG_CHECK_NONCE */
	tmp=nonce_table[bucket];
	nonce_table[bucket]=nonce_table[bucket]->next;
	FREENONCE(tmp);
      }  else if ((nonce_table[bucket]->time == new_el->time) &&
		  (nonce_table[bucket]->ni == new_el->ni) &&
		  (nonce_table[bucket]->expiration == new_el->expiration))
	{/* Case 3 - First element is a match */
	  
#ifdef DEBUG_CHECK_NONCE
	  printf("DEBUG(nasd_check_nonce) 3 found a match in the bucket FAILED - NASD_BAD_NONCE\n");
#endif /* DEBUG_CHECK_NONCE */
         
	  FREENONCE(new_el);
	  NASD_UNLOCK_MUTEX(nonce_table_lock[bucket]);
	  return(NASD_BAD_NONCE);
	} else {
#ifdef DEBUG_CHECK_NONCE
	  printf("DEBUG(nasd_check_nonce) first element is uninteresting\n");
#endif /* DEBUG_CHECK_NONCE */
	  check_first=NASD_FALSE;
	}
    } else {
      check_first=NASD_FALSE;
    }
  }
  
#ifdef DEBUG_CHECK_NONCE
    printf("DEBUG(nasd_check_nonce) checking the chain\n");
#endif /* DEBUG_CHECK_NONCE */
  curr=nonce_table[bucket];
  NASD_ASSERT(curr!=NULL);


  while(curr->next!=NULL) {
   if (curr->next->drive_name == new_el->drive_name ) {
     if (NASD_ABS(now_l-curr->next->time) > NASD_ACCEPTABLE_SKEW) {
       /* We can free it */
      
#ifdef DEBUG_CHECK_NONCE
       printf("DEBUG(nasd_check_nonce) 4 **freeing an element in the chain\n");
#endif /* DEBUG_CHECK_NONCE */
       tmp=curr->next;
       curr->next=curr->next->next;
       FREENONCE(tmp);
     } else if ((curr->next->time == new_el->time) &&
		(curr->next->ni == new_el->ni) &&
		(curr->next->expiration == new_el->expiration))
       {
	
#ifdef PRINT_ERROR
	 printf("DEBUG(nasd_check_nonce) 5 found a match in the chain FAILED=NASD_BAD_NONCE\n");
#endif /* PRINT_ERROR */
	 FREENONCE(new_el);
	 NASD_UNLOCK_MUTEX(nonce_table_lock[bucket]);
	 return(NASD_BAD_NONCE);
       }
   }
   curr=curr->next;
  }
  

  
  /* We insert it at the end which means it will take longer
     to search for but we will walk the list and cleanup as we search */
  new_el->next=NULL;
  curr->next=new_el;
#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) finished successfully\n");
#endif /* DEBUG_CHECK_NONCE */
  

  NASD_UNLOCK_MUTEX(nonce_table_lock[bucket]);
#endif /* NASD_SECURE_RPCS_ENABLE >0*/
  
  return(NASD_SUCCESS);
    
}







