/*
 * nasd_security_rpc.c
 *
 * Arg encryptiong and result decryption for NASD RPCs.
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Howard Gobioff
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */
 
#include <nasd/nasd_options.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_net_convert.h>
#ifdef _KERNEL
/* XXX - including this file causes a kernel compile to blow up */
#else /* _KERNEL */
#include <nasd/nasd_convert_types.h>
#endif /* _KERNEL */
#include <nasd/nasd_security.h>
#include <nasd/nasd_security_time.h>
#include <nasd/nasd_keymgmt.h>
#include <nasd/nasd_nonce_mgmt.h>
#include <cryptolib/libcrypt.h>
#include <sys/time.h>
#include <sys/resource.h>
#ifndef _KERNEL
#include <string.h>
#endif /* _KERNEL */

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif


/*
 * Normal version for everyday RPCs
 */


/* This code should be connected to a protocol where the drive and
file manager agree on the time. */

int nasd_disable_reply_crypto_check=FALSE;

nasd_status_t nasd_security_digest_args(nasd_args_t *in_args,
                                        nasd_key_t in_key,
					nasd_byte_t *out_args_otw,
					nasd_args_len_t out_args_len,
                                        nasd_digest_nonce_otw_t in_digest_otw,
                                        nasd_security_context_t *contextp)
{
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_args_t  *otw_args=(nasd_args_t *)out_args_otw;
  HMAC_SHA1_CTX HMAC_context;
  nasd_byte md_out[SHA_DIGEST_LENGTH];
  nasd_digest_nonce_t *in_digest=(nasd_digest_nonce_t*)in_digest_otw;
  nasd_timespec_t now;
  nasd_key_t integrity_key;

#if DEBUG_DIGEST_ARGS > 0
  int cnt;
#endif /* DEBUG_DIGEST_ARGS */

#if DEBUG_DIGEST_ARGS > 0
  printf("DEBUG(nasd_security_digest_args): entered with security=%d \n",in_args->in_security_param.actual_protection );
#endif /* DEBUG_DIGEST_ARGS */

#if DEBUG_DIGEST_ARGS > 0
  printf("DEBUG(nasd_security_digest_args): "
         "key=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n",
         ((nasd_uint64*)in_key)[0],((nasd_uint64*)in_key)[1] );
#endif /* DEBUG_DIGEST_ARGS */
  
  if (in_args->in_security_param.actual_protection == NASD_NO_PROTECTION) {
    /* xxx Superflous */
    bzero((char *)in_digest, sizeof(nasd_digest_nonce_t));
  }  else {
    /* Add Nonce */
    /* Get the Nonce */
    
    nasd_security_get_drive_time(in_args->in_capability_args.drive_name,&now);
    nasd_timespec_t_to_net(&now,(nasd_byte *)&in_digest->nonce);
#if DEBUG_DIGEST_ARGS > 0
    printf("DEBUG(nasd_security_digest_args): now.ts_sec %ld        now.ts_nsec %ld\n",
           now.ts_sec,now.ts_nsec);
#endif /* DEBUG_DIGEST_ARGS */

        
    
    if ((in_args->in_security_param.actual_protection & ( NASD_INTEGRITY_DATA|NASD_INTEGRITY_ARGS)))
      
      nasd_security_generate_integrity_key(in_key, integrity_key);
      HMAC_SHA1_Init(&HMAC_context,(char *)integrity_key,sizeof(nasd_key_t));
    if ((in_args->in_security_param.actual_protection  & NASD_INTEGRITY_ARGS) == NASD_INTEGRITY_ARGS) {
#if DEBUG_DIGEST_ARGS > 0
  printf("DEBUG(nasd_security_digest_args): digesting args \n");
#endif /* DEBUG_DIGEST_ARGS */
      HMAC_SHA1_Update(&HMAC_context,(unsigned char *)out_args_otw+sizeof(nasd_args_t),out_args_len-(sizeof(nasd_args_t)+(sizeof(nasd_digest_nonce_t))));
      HMAC_SHA1_Update(&HMAC_context,(unsigned char *)&in_digest->nonce,sizeof(nasd_timespec_t));
      
      
    }
    if ((in_args->in_security_param.actual_protection & ( NASD_INTEGRITY_DATA|NASD_INTEGRITY_ARGS))) {
      
      
      
      if (contextp != NULL){

        bcopy((char *)&HMAC_context, (char *)&contextp->HMAC_context, sizeof(HMAC_SHA1_CTX));   

        contextp->started=1;

        contextp->protection=in_args->in_security_param.actual_protection;
      }

      /* Stick a digest of the args into the request */
      /* Data digests will be cumulative and follow later */
      HMAC_SHA1_Final((unsigned char *)in_digest->digest,&HMAC_context);   

#if DEBUG_DIGEST_ARGS > 0
      printf("DEBUG(nasd_security_digest_args):  args digest=");
      for(cnt=0;cnt<NASD_DIGEST_SIZE;cnt++)
	printf("%x",((nasd_byte*)in_digest->digest)[cnt]);
      printf("]\n");
#endif /* DEBUG_DIGEST_ARGS */
    } else {
      /* I zero fill when we don't use digests. This probably
	 isn't necessary */
      bzero((char *)in_digest->digest,NASD_DIGEST_SIZE);
    }
  }

#if DEBUG_DIGEST_ARGS > 0
  printf("DEBUG(nasd_security_digest_args): successfull return \n");
#endif /* DEBUG_DIGEST_ARGS */

  return(NASD_SUCCESS); 
#else /* NASD_SECURE_RPCS_ENABLE > 0 */

  return(NASD_SUCCESS); /* cheap security - nice and fast */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

}

nasd_status_t nasd_security_verify_res(nasd_args_t *in_args,
                                       nasd_key_t in_key,
                                       nasd_byte_t *in_res_otw,
                                       nasd_args_len_t in_res_len,
                                       nasd_status_t *result,
                                       nasd_digest_nonce_otw_t in_digest,
                                       nasd_security_context_t *contextp){
#if NASD_SECURE_RPCS_ENABLE > 0

HMAC_SHA1_CTX HMAC_context, *HMAC_contextp;
nasd_byte md_out[SHA_DIGEST_LENGTH]; 
nasd_timespec_t now,then;
nasd_uint64 now_l,then_l,diff;
nasd_digest_nonce_t *otw_digest= (nasd_digest_nonce_t *)in_digest;
nasd_status_t otw_stat;
nasd_key_t tmp_key, integrity_key;
nasd_res_t *in_res=(nasd_res_t*)in_res_otw;
nasd_status_t rc;
int i,bcount;
nasd_digest_t tdigest;


#if DEBUG_VERIFY_RES > 0
  printf("DEBUG(nasd_security_verify_res): enter \n");
#endif /* DEBUG_VERIFY_RES */
  if ( in_args->in_security_param.actual_protection  != NASD_NO_PROTECTION) {
#if DEBUG_VERIFY_RES > 0
    printf("DEBUG(nasd_security_verfiy_res): "
           "key=[%08" NASD_64x_FMT " %08" NASD_64x_FMT "]\n",
           ((nasd_uint64*)in_key)[0],((nasd_uint64*)in_key)[1]);
#endif /* DEBUG_VERIFY_RES */
    /* Check the nonce */ 
    
    nasd_timespec_t_from_net((nasd_byte *)&otw_digest->nonce,&then);
    rc=nasd_check_nonce(&(in_args->in_capability_args),&then);
    if (rc) {
      *result=rc;
      return rc;
    }

    /* Decrypt BEFORE we check the DIGEST */
    
    /* If we have a context, use that */

    if ( in_args->in_security_param.actual_protection & (NASD_INTEGRITY_DATA|NASD_INTEGRITY_ARGS)    ) 
      if (contextp != NULL) {
        HMAC_contextp=&contextp->HMAC_context;
      } else {
        HMAC_contextp=&HMAC_context;
        if (!nasd_disable_reply_crypto_check){
          nasd_security_generate_integrity_key(in_key, integrity_key);
          HMAC_SHA1_Init(HMAC_contextp,(char *)integrity_key,sizeof(nasd_key_t));
        }
      } 

#if (NASD_USE_HMAC_ONLY == 0)    
    if (( in_args->in_security_param.actual_protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) {


      /* At this point, most of the data should have been digest in
         the pipe functions if it was needed. However, we may have 
         an unfinshed unkeyed digest that we must Finalize and then
         add to the keyed digest */
      if (contextp != NULL) {
        if (contextp->curr_off % NASD_OD_BLOCK_SIZE) {
          /* Leftover data in an unkeyed digest (a trailing partial datablock) */
          if (!nasd_disable_reply_crypto_check) {
            SHA1_Final(FALSE, tdigest, &contextp->SHA1_ctx);
          
#if DEBUG_VERIFY_RES > 0
            printf("verify_res (COMPUTED from running digest): "
                   "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
                   ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]   );
#endif /* DEBUB_VERIFY_RES */
            HMAC_SHA1_Update(&contextp->HMAC_context,tdigest, NASD_DIGEST_SIZE);    
          }
        }
      }
    }
    
#endif /* (NASD_USE_HMAC_ONLY == 0) */
    if (( in_args->in_security_param.actual_protection & NASD_INTEGRITY_ARGS) == NASD_INTEGRITY_ARGS) {
      if (!nasd_disable_reply_crypto_check){
        HMAC_SHA1_Update(HMAC_contextp,(unsigned char*)in_res_otw,in_res_len-sizeof(nasd_digest_nonce_t));
        HMAC_SHA1_Update(HMAC_contextp,(unsigned char*)&otw_digest->nonce,sizeof(nasd_timespec_t));
      }
    }
    
    if ( in_args->in_security_param.actual_protection & (NASD_INTEGRITY_DATA|NASD_INTEGRITY_ARGS)    ) {
      if (!nasd_disable_reply_crypto_check) {
        HMAC_SHA1_Final(md_out,HMAC_contextp);
      }
      if ((!nasd_disable_reply_crypto_check) && bcmp((char *)md_out, (char *)otw_digest->digest, NASD_MIN(SHA_DIGEST_LENGTH, NASD_DIGEST_SIZE)) ) {
	*result=NASD_BAD_DIGEST_REPLY;
#if PRINT_ERROR > 0 
	printf("DEBUG(nasd_security_verify_res): FAILED NASD_BAD_DIGEST_REPLY \n");
#endif /* PRINT_ERROR */
        
#if DEBUG_VERIFY_RES > 0 
	printf("DEBUG(nasd_security_verify_res): "
               "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n",
               ((nasd_uint64*)md_out)[0], ((nasd_uint64*)md_out)[1]);
	printf("DEBUG(nasd_security_verify_res): "
               "expected digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n",
               ((nasd_uint64*)otw_digest->digest)[0],
               ((nasd_uint64*)otw_digest->digest)[1]);
	printf("DEBUG(nasd_security_verify_res): error NASD_BAD_DIGEST_REPLY\n");
#endif /* DEBUG_VERIFY_RES */
	return(NASD_BAD_DIGEST_REPLY);       
      }
    }
  }
#if DEBUG_VERIFY_RES > 0
  printf("DEBUG(nasd_security_verify_res): "
         "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n",
         ((nasd_uint64*)md_out)[0], ((nasd_uint64*)md_out)[1]);
  printf("DEBUG(nasd_security_verify_res): exit \n");
#endif /* DEBUG_VERIFY_RES */
  return (NASD_SUCCESS);
  
#else /* NASD_SECURE_RPCS_ENABLE > 0 */
  return(NASD_SUCCESS); /* see, nice and fast */
#endif /*  NASD_SECURE_RPCS_ENABLE > 0  */
}


/* Updates the context on a read call -

   performs the digest and then encryption. The
   caller will then ship off the data to a client */

void nasd_security_context_update_receive(nasd_security_context_t *sec_ctx,
                                          nasd_byte_t *data, nasd_len_t len,
                                          nasd_offset_t curr_file_off)
{
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_digest_t tdigest;

#if DEBUG_CONTEXT_UPDATE > 0
  printf("DEBUG(nasd_security_context_update_receive): entered \n");
#endif /* DEBUG_CONTEXT_UPDATE */

 NASD_ASSERT(sec_ctx->started != 0);

 if ((data!=NULL) && (len>0) ) {
   if (!nasd_disable_reply_crypto_check) { 
     if ((sec_ctx->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
     }
     
     /* Should I be worring about locking here ? */
     if ((sec_ctx->protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) {
    
       
#if (NASD_USE_HMAC_ONLY == 0)   
    
 
       SHA1_Update(FALSE, &sec_ctx->SHA1_ctx, data, len);
       if ((( curr_file_off + len) % NASD_OD_BLOCK_SIZE) == 0) {
         SHA1_Final(FALSE, tdigest, &sec_ctx->SHA1_ctx);
         SHA1_Init(FALSE, &sec_ctx->SHA1_ctx);
#if DEBUG_UPDATE_RECEIVE > 0
         /* This should be ok in kernel because it should be allocated aligned
            on stack */
         printf("update_receive : "
                "SHA1 digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
                ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]   );
#endif /* DEBUB_UPDATE_RECEIVE */
         HMAC_SHA1_Update(&sec_ctx->HMAC_context, tdigest, sizeof(nasd_digest_t));
       }
#else
   HMAC_SHA1_Update( &sec_ctx->HMAC_context, data, len);
       


#endif /* (NASD_USE_HMAC_ONLY == 0)    */
     }
   }
   sec_ctx->curr_off+= len;
 }
 
#if DEBUG_CONTEXT_UPDATE > 0
  printf("DEBUG(nasd_security_context_update_receive): exited \n");
#endif /* DEBUG_CONTEXT_UPDATE */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

}



/* We set up the digest to partial digest to hold the 
   message digest context and the key for our running
   digest and encryption */
void nasd_security_reply_init(nasd_args_t *in_args, 
                              nasd_key_t in_key, 
                              nasd_byte_t *icv,
                              nasd_security_context_t *sec_ctx)

{
#if NASD_SECURE_RPCS_ENABLE > 0
  unsigned char enc_key[3][8];
  unsigned char *keys[3];
  nasd_key_t integrity_key;
  nasd_key_t priv_key;

#ifdef DEBUG_REPLY_INIT
  printf("DEBUG(nasd_security_reply_init): entered \n");
  printf("DEBUG(nasd_security_reply_init): "
         "in_key=%08" NASD_64x_FMT " %08" NASD_64x_FMT "\n",
         ((nasd_uint64*)in_key)[0], ((nasd_uint64*)in_key)[1]);
#endif /* DEBUG_REPLY_INIT */

  
  sec_ctx->protection=in_args->in_security_param.actual_protection;
  sec_ctx->started=1;
  if ((sec_ctx->protection & (NASD_INTEGRITY_DATA|NASD_INTEGRITY_ARGS) ) !=0 ) {

    /* We are digesting something */
    SHA1_Init(FALSE, &sec_ctx->SHA1_ctx);
    nasd_security_generate_integrity_key(in_key, integrity_key);
    HMAC_SHA1_Init(&sec_ctx->HMAC_context,(char *)integrity_key,sizeof(nasd_key_t));
  }

 if ((in_args->in_security_param.actual_protection & (NASD_PRIVACY_DATA | NASD_PRIVACY_ARGS)) > 0) {
    NASD_ASSERT( sizeof(nasd_key_t) > 8 );

    nasd_security_generate_privacy_key(in_key, priv_key);
    bcopy(&priv_key[0], enc_key[0], 8);
    bcopy(&priv_key[sizeof(nasd_key_t)-8], enc_key[1], 8);
    keys[0]=enc_key[0];
    keys[1]=enc_key[1];
    keys[2]=enc_key[0];
 }
#ifdef DEBUG_REPLY_INIT
  printf("DEBUG(nasd_security_reply_init): exited \n");
#endif /* DEBUG_REPLY_INIT */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}
