/*
 * drive_sys.c
 *
 * In-kernel NASD system support
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd_server.h>
#include <nasd/nasd_options.h>

#include <sys/secdefines.h>
#if SEC_BASE > 0
#include <sys/security.h>
#include <sys/audit.h>
#include <sys/secpolicy.h>
#endif /* SEC_BASE */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <sys/kernel.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/buf.h>
#include <sys/proc.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/conf.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/specdev.h>
#include <sys/flock.h>
#include <ufs/quota.h>
#include <ufs/inode.h>
#include <ufs/fs.h>
#include <ufs/ufsmount.h>
#include <ufs/fs_proto.h>
#include <ufs/ufs_stats.h>
#include <sys/syslimits.h>
#include <kern/assert.h>
#include <kern/parallel.h>
#include <mach/mach_types.h>
#include <vm/vm_page.h>
#include <vm/vm_vppage.h>
#include <sys/vfs_ubc.h>
#include <vm/vm_mmap.h>
#include <vm/vm_vp.h>
#include <vm/vm_debug.h>
#include <kern/ipc_globals.h>
#include <sys/malloc.h>
#include <dce/dce_error.h>
#include <dce/ker/pthread_exc.h>

#include <nasd/nasd_types.h>
#include <nasd/nasd_threadstuff.h>
#include <rpc/rpc.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_sys.h>
#include <nasd/nasd_security.h>
#include <dce/rpc.h>
#include <dce/stubbase.h>

#include <mach/kern_return.h>

#define PRINT_ERROR 1


#define DOSYSRETURN(_val_) { \
  *retval = _val_; \
  return(_val_); \
}

struct nasd_svinfo *nasd_sva = NULL;
struct nasd_nfs_svinfo *nasd_nfs_sva = NULL;

extern kern_return_t vm_map_copy_overwrite(vm_map_t dest_map,
  vm_offset_t dest_addr, vm_map_copy_t copy, boolean_t interruptible);

extern vm_offset_t vm_alloc_kva(vm_size_t size);

int nasd_wire_err_verbose = 1;
int nasd_wire_tries = 5;
unsigned long nasd_wire_retries = 0;

NASD_DECLARE_ONCE(nasd_dux_kinit_once)

void
nasd_dux_kinit()
{
  nasd_status_t rc;

  rc = nasd_srv_init();
  if (rc) {
    dprintf("NASD: unable to init srv subsys\n");
  }
}

int
nasd_srv(p, args, retval)
  struct proc  *p;
  void         *args;
  long         *retval;
{
  struct args {
    int    cmd;
    void  *arg;
  } *uap = (struct args *)args;
  struct nasd_nfs_mount *nfsmounts;
  struct nasd_nfs_config nfscfg;
  struct nasd_serv nasdsrv;
  nasd_options_t opts;
  nasd_status_t rc;
  int error;

  nasd_once(&nasd_dux_kinit_once, nasd_dux_kinit);

  PRIV_SUSER(u.u_cred, &u.u_acflag, SEC_REMOTE, 0, EPERM, error);
  if (error) {
    DOSYSRETURN(error);
  }

  switch(uap->cmd) {

    case NASD_GET_OPTIONS:
      bzero((char *)&opts, sizeof(opts));
#if NASD_SECURE_RPCS_ENABLE > 0
      opts.opts1 |= NASD_OPTS1_SECURITY;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
      opts.opts1 |= NASD_OPTS1_NFS_CLIENT;
      opts.opts1 |= NASD_OPTS1_NFS_FM;
      opts.opts1 |= NASD_OPTS1_DRIVE;
      error = copyout((caddr_t)&opts, (caddr_t)uap->arg, sizeof(opts));
      DOSYSRETURN(error);
      break;

    case NASD_PSRV_GO:
      NASD_D_LOCK();

      if (nasd_sva) {
        NASD_D_UNLOCK();
        DOSYSRETURN(EADDRINUSE);
      }
      error = copyin((caddr_t)uap->arg, (caddr_t)&nasdsrv, sizeof(nasdsrv));
      if (error) {
        NASD_D_UNLOCK();
        DOSYSRETURN(error);
      }
      rc = nasd_threads_init();
      if (rc) {
        NASD_D_UNLOCK();
        DOSYSRETURN(EIO);
      }
      rc = nasd_mem_init();
      if (rc) {
        nasd_threads_shutdown();
        NASD_D_UNLOCK();
        DOSYSRETURN(EIO);
      }
#ifdef malloc
#define dce_still_sucks_goats malloc
#undef malloc
#endif /* malloc */
      NASD_Malloc(nasd_sva, sizeof(struct nasd_svinfo),
        (struct nasd_svinfo *));
#ifdef dce_still_sucks_goats
#define malloc dce_still_sucks_goats
#undef dce_still_sucks_goats
#endif /* dce_still_sucks_goats */
      if (nasd_sva == NULL) {
        NASD_D_UNLOCK();
        nasd_mem_shutdown();
        nasd_threads_shutdown();
        DOSYSRETURN(ENOMEM);
      }
      bcopy((char *)&nasdsrv, (char *)&nasd_sva->srv,
        sizeof(struct nasd_svinfo));
      error = nasd_psrv_go(nasd_sva); /* consumes D_LOCK */
      /* nasd_mem_shutdown() done by nasd_psrv_go() */
      /* D_LOCK released by nasd_psrv_go() */
      DOSYSRETURN(error);
      break;
  
    case NASD_SRV_PANIC:
      NASD_PANIC();
      break;

    case NASD_NFS_SRV_GO:
      NASD_D_LOCK();

      if (nasd_nfs_sva) {
        NASD_D_UNLOCK();
        DOSYSRETURN(EADDRINUSE);
      }
      error = copyin((caddr_t)uap->arg, (caddr_t)&nfscfg, sizeof(nfscfg));
      if (error) {
        NASD_D_UNLOCK();
        DOSYSRETURN(error);
      }
      if (nfscfg.mount_array_len <= 0) {
        NASD_D_UNLOCK();
        DOSYSRETURN(EINVAL);
      }
      rc = nasd_threads_init();
      if (rc) {
        NASD_D_UNLOCK();
        DOSYSRETURN(EIO);
      }
      rc = nasd_mem_init();
      if (rc) {
        nasd_threads_shutdown();
        NASD_D_UNLOCK();
        DOSYSRETURN(EIO);
      }
#ifdef malloc
#define dce_still_sucks_goats malloc
#undef malloc
#endif /* malloc */
      NASD_Malloc(nasd_nfs_sva, sizeof(struct nasd_nfs_svinfo),
        (struct nasd_nfs_svinfo *));
#ifdef dce_still_sucks_goats
#define malloc dce_still_sucks_goats
#undef dce_still_sucks_goats
#endif /* dce_still_sucks_goats */
      if (nasd_nfs_sva == NULL) {
       NASD_D_UNLOCK();
       nasd_mem_shutdown();
       nasd_threads_shutdown();
       DOSYSRETURN(ENOMEM);
      }
#ifdef malloc
#define dce_still_sucks_goats malloc
#undef malloc
#endif /* malloc */
      NASD_Malloc(nfsmounts,
        sizeof(struct nasd_nfs_mount)*nfscfg.mount_array_len,
        (struct nasd_nfs_mount *));
#ifdef dce_still_sucks_goats
#define malloc dce_still_sucks_goats
#undef dce_still_sucks_goats
#endif /* dce_still_sucks_goats */
      if (nfsmounts == NULL) {
#ifdef free
#define dce_sucks_goats free
#undef free
#endif /* free */
        NASD_Free(nasd_nfs_sva, sizeof(struct nasd_nfs_svinfo));
#ifdef dce_sucks_goats
#define free dce_sucks_goats
#undef dce_sucks_goats
#endif /* dce_sucks_goats */
        nasd_nfs_sva = NULL;
        NASD_D_UNLOCK();
        nasd_mem_shutdown();
        nasd_threads_shutdown();
        DOSYSRETURN(ENOMEM);
      }
      bcopy((char *)&nfscfg, (char *)&nasd_nfs_sva->config,
        sizeof(struct nasd_nfs_config));
      error = copyin((caddr_t)nfscfg.mount_array, (caddr_t)nfsmounts,
        sizeof(struct nasd_nfs_mount)*nfscfg.mount_array_len);
      if (error) {
#ifdef free
#define dce_sucks_goats free
#undef free
#endif /* free */
        NASD_Free(nfsmounts,
          sizeof(struct nasd_nfs_mount)*nfscfg.mount_array_len);
        NASD_Free(nasd_nfs_sva, sizeof(struct nasd_nfs_svinfo));
#ifdef dce_sucks_goats
#define free dce_sucks_goats
#undef dce_sucks_goats
#endif /* dce_sucks_goats */
        nasd_nfs_sva = NULL;
        NASD_D_UNLOCK();
        nasd_mem_shutdown();
        nasd_threads_shutdown();
        DOSYSRETURN(error);
      }
      nasd_nfs_sva->config.mount_array = nfsmounts;
      error = nasd_nfs_srv_go(nasd_nfs_sva); /* consumes D_LOCK */
      /*
       * nasd_nfs_srv_go() consumes D_LOCK, frees the mount array
       * (nfsmounts) and the svinfo structure (nasd_nfs_sva).
       */
      nasd_mem_shutdown();
      nasd_threads_shutdown();
#if NASD_MEM_COUNT_ALLOC > 0
      if (nasd_mem_allocated) {
        printf(
          "NFS WARNING: %ld bytes of memory still outstanding, something leaked core.\n",
          nasd_mem_allocated);
      }
      else {
        printf("NFS: %ld bytes of memory still outstanding.\n", nasd_mem_allocated);
      }
#endif /* NASD_MEM_COUNT_ALLOC > 0 */
      DOSYSRETURN(error);
      break;

    default:
      NASD_D_UNLOCK();
      DOSYSRETURN(EINVAL);
  }

  DOSYSRETURN(0);
}

#if 0
#define NASD_WIRE_WRITE_MAP kernel_copy_map
#else
#define NASD_WIRE_WRITE_MAP ipc_kernel_map
#endif

int
nasd_wire_buf(
  void   *buf,
  int     len,
  void  **iobufp,
  int     dir)
{
  void *m;
  int rc;

#if 0
  if (dir == NASD_WIRE_DIR_WRITE) {
    vm_map_copy_t vmcpy;
    kern_return_t ret;
    vm_offset_t kaddr;
    vm_map_t umap;
    task_t utask;

    utask = current_task();
    umap = utask->map;

    vm_map_reference(umap);

    vmcpy = NULL;
    ret = vm_map_copyin(umap, buf, len, FALSE, &vmcpy);
    vm_map_deallocate(umap); /* must do no matter success of above */
    if (ret != KERN_SUCCESS) {
      if (nasd_wire_err_verbose)
        dprintf("failed vm_map_copyin len=%d ret=%d\n", len, ret);
      return(EIO);
    }

    kaddr = 0;
    ret = vm_map_copyout(NASD_WIRE_WRITE_MAP, &kaddr, vmcpy);
    if (ret != KERN_SUCCESS) {
      if (nasd_wire_err_verbose)
        dprintf("failed vm_map_copyout len=%d vmcpy=0x%lx ret=%d\n", len, vmcpy, ret);
      vm_map_copy_discard(vmcpy);
      return(EIO);
    }

    *iobufp = (void *)kaddr;
    return(0);
  }
  else {
#endif
    vm_offset_t ustart, uend, ud;
    vm_offset_t kaddr, k1;
    vm_map_copy_t vmcpy;
    kern_return_t ret;
    vm_map_t umap;
    task_t utask;
    int ulen, try;

    utask = current_task();
    umap = utask->map;

    ustart = trunc_page((vm_offset_t)buf);
    uend = round_page(((vm_offset_t)buf) + len);
    ulen = uend - ustart;
    ud = ((vm_offset_t)buf) - ustart;

    vm_map_reference(umap);

    try = 0;

retry_wire:
    ret = vm_map_pageable(umap, ustart, uend, VM_PROT_READ|VM_PROT_WRITE);
    if (ret) {
      if (nasd_wire_err_verbose) {
        extern boolean_t vm_page_free_wanted;
        extern int ubc_pages, vm_page_free_reserved;
#if NTIP == 0
        extern int ubc_lru_page_count;
#endif /* NTIP == 0 */
        extern int vm_page_free_count, vm_page_zeroed_count, vm_managed_pages;
        extern int vm_page_kluster, vm_page_bad_count, vm_page_free_target;
        extern int vm_page_inactive_count;

        dprintf("failed vm_map_pageable ustart=0x%lx uend=0x%lx ret=%d\n", ustart, uend, ret);

        dprintf("ubc_pages=%d\n", ubc_pages);
        dprintf("nasd_wire_retries=%lu\n", nasd_wire_retries);
#if (NTIP==0)
        dprintf("ubc_lru_page_count=%d\n", ubc_lru_page_count);
#endif /* NTIP==0 */
        dprintf("vm_page_free_target=%d\n", vm_page_free_target);
        dprintf("vm_page_free_reserved=%d\n", vm_page_free_reserved);
        dprintf("vm_page_inactive_count=%d\n", vm_page_inactive_count);
        dprintf("vm_page_free_wanted=%d\n", vm_page_free_wanted);
        dprintf("vm_page_free_count=%d\n", vm_page_free_count);
        dprintf("vm_page_zeroed_count=%d\n", vm_page_zeroed_count);
        dprintf("vm_managed_pages=%d\n", vm_managed_pages);
        dprintf("vm_page_kluster=%d\n", vm_page_kluster);
        dprintf("vm_page_bad_count=%d\n", vm_page_bad_count);
      }
      try++;
      if ((try < nasd_wire_tries) && (ret == KERN_RESOURCE_SHORTAGE)) {
        vm_wait();
        nasd_wire_retries++;
        goto retry_wire;
      }
      vm_map_deallocate(umap);
      return(EIO);
    }

    ret = vm_dup_va(umap, ustart, ulen, &kaddr, VM_PROT_READ|VM_PROT_WRITE);
    if (ret) {
      if (nasd_wire_err_verbose) {
        dprintf("failed vm_dup_va umap=0x%lx ustart=0x%lx ulen=0x%lx ret=%d\n",
          umap, ustart, ulen, ret);
      }
      ret = vm_map_pageable(umap, ustart, uend, VM_PROT_NONE);
      if (ret) {
        dprintf("nasd_pdev.c line %d failed unwiring 0x%lx..0x%lx ret=%d\n", ustart, uend, ret);
      }
      return(EIO);
    }

    vm_map_deallocate(umap);

    *iobufp = (void *)(kaddr+ud);
    return(0);
#if 0
  }
#endif
}

int
nasd_unwire_buf(
  void  *buf,
  int    in_len,
  void  *iobuf,
  int    dir)
{
#if 0
  if (dir == NASD_WIRE_DIR_WRITE) {
    vm_deallocate(NASD_WIRE_WRITE_MAP, iobuf, in_len);
    return(0);
  }
  else {
#endif
    vm_offset_t ustart, uend, ud;
    vm_offset_t start, end;
    vm_map_copy_t vmcpy;
    kern_return_t ret;
    vm_map_t umap;
    int len, ulen;
    task_t utask;

    utask = current_task();
    umap = utask->map;

    vm_map_reference(umap);

    ustart = trunc_page((vm_offset_t)buf);
    uend = round_page(((vm_offset_t)buf) + in_len);
    ulen = uend - ustart;
    ud = ((vm_offset_t)buf) - ustart;

    start = trunc_page((vm_offset_t)iobuf);
    end = round_page(((vm_offset_t)iobuf)+in_len);
    len = end - start;

    NASD_ASSERT(len == ulen);
    ret = vm_free_dup_va(start, ulen);
    if (ret) {
      dprintf("nasd_pdev.c line %d ret=%d from vm_free_dup_va 0x%lx len=%d\n",
        __LINE__, ret, start, ulen);
    }

    ret = vm_map_pageable(umap, ustart, uend, VM_PROT_NONE);
    if (ret) {
      dprintf("nasd_pdev.c line %d failed unwiring 0x%lx..0x%lx\n",
        __LINE__, ustart, uend);
    }

    vm_map_deallocate(umap);

    return(0);
#if 0
  }
#endif
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
