/*
 * nasd_pthread_support.c
 *
 * pthread support routines
 */
/*
 * Copyright (c) 1995,1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Authors: Jim Zelenka, Marc Unangst
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_general.h>

pthread_mutexattr_t nasd_pthread_mutex_attr;
int nasd_pthread_mutex_attr_good = 0;

nasd_status_t
nasd_sys_threads_init()
{
  /* nothing to do here */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threads_init_once()
{
  int rc;

  nasd_pthread_mutex_attr_good = 0;

#ifdef PTHREAD_EXC
  pthread_mutexattr_create(&nasd_pthread_mutex_attr);
  pthread_mutexattr_setkind_np(&nasd_pthread_mutex_attr, MUTEX_FAST_NP);
#else /* PTHREAD_EXC */
  rc = pthread_mutexattr_create(&nasd_pthread_mutex_attr);
  if (rc)
    return(NASD_FAIL);
  rc = pthread_mutexattr_setkind_np(&nasd_pthread_mutex_attr, MUTEX_FAST_NP);
  if (rc)
    return(NASD_FAIL);
#endif /* PTHREAD_EXC */
  nasd_pthread_mutex_attr_good = 1;

  return(NASD_SUCCESS);
}

void
nasd_sys_threads_shutdown()
{
  /* nothing to do here */
}

nasd_status_t
nasd_sys_mutex_init(
  pthread_mutex_t  *m)
{
  int rc;

#ifdef PTHREAD_EXC
  if (nasd_pthread_mutex_attr_good == 0)
    return(NASD_FAIL);
  pthread_mutex_init(m, nasd_pthread_mutex_attr);
#else /* PTHREAD_EXC */
#if defined(DEC_OSF) || defined(LINUX)
  if (nasd_pthread_mutex_attr_good == 0)
    return(NASD_FAIL);
  rc = pthread_mutex_init(m, nasd_pthread_mutex_attr);
  if (rc)
    return(NASD_FAIL);
#endif /* DEC_OSF || LINUX */
#ifdef AIX
	rc = pthread_mutex_init(m, &pthread_mutexattr_default);
  if (rc)
    return(NASD_FAIL);
#endif /* AIX */
#endif /* PTHREAD_EXC */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_mutex_destroy(
  pthread_mutex_t  *m)
{
  int rc;

#ifdef PTHREAD_EXC
  pthread_mutex_destroy(m);
#else
  rc = pthread_mutex_destroy(m);
  if (rc)
    return(NASD_FAIL);
#endif /* PTHREAD_EXC */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_cond_init(
  pthread_cond_t  *c)
{
  int rc;

#ifdef PTHREAD_EXC
  pthread_cond_init(c, pthread_condattr_default);
#else /* PTHREAD_EXC */
#if defined(DEC_OSF) || defined(LINUX)
  rc = pthread_cond_init(c, pthread_condattr_default);
#endif /* DEC_OSF || LINUX */
#ifdef AIX
  rc = pthread_cond_init(c, &pthread_condattr_default);
#endif /* AIX */
  if (rc)
    return(NASD_FAIL);
#endif /* PTHREAD_EXC */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_cond_destroy(
  pthread_cond_t  *c)
{
  int rc;

#ifdef PTHREAD_EXC
  pthread_cond_destroy(c);
#else /* PTHREAD_EXC */
  rc = pthread_cond_destroy(c);
  if (rc)
    return(NASD_FAIL);
#endif /* PTHREAD_EXC */
  return(NASD_SUCCESS);
}

int
_nasd_once(
  pthread_once_t         *once_block,
  pthread_initroutine_t   init_routine,
  char                   *file,
  int                     line)
{
  int rc = NASD_SUCCESS;

#ifdef PTHREAD_EXC
  pthread_once(once_block, init_routine);
#else /* PTHREAD_EXC */
  rc = pthread_once(once_block, init_routine);
#endif /* PTHREAD_EXC */
  return(rc);
}

nasd_status_t
nasd_sys_threadattr_create(
  nasd_sys_threadattr_t  *attrp)
{
  int ret;

  ret = pthread_attr_create(attrp);
  if (ret)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threadattr_destroy(
  nasd_sys_threadattr_t  *attrp)
{
  int ret;

  ret = pthread_attr_delete(attrp);
  if (ret)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

#ifdef LINUX

/*
 * Don't do this on Linux, because Linux pthreads have a dynamic
 * stack that grows on demand, and thus setstacksize/getstacksize
 * are not supported.
 */

nasd_status_t
nasd_sys_threadattr_setstacksize(
  nasd_sys_threadattr_t  *attrp,
  int                     size)
{
  return(NASD_OP_NOT_SUPPORTED);
}

nasd_status_t
nasd_sys_threadattr_getstacksize(
  nasd_sys_threadattr_t  *attrp,
  int                    *sizep)
{
  return(NASD_OP_NOT_SUPPORTED);
}

#else /* LINUX */

nasd_status_t
nasd_sys_threadattr_setstacksize(
  nasd_sys_threadattr_t  *attrp,
  int                     size)
{
  int ret;

  ret = pthread_attr_setstacksize(attrp, (long)size);
  if (ret)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threadattr_getstacksize(
  nasd_sys_threadattr_t  *attrp,
  int                    *sizep)
{
  long size;

  size = pthread_attr_getstacksize(*attrp);
  if (size < 0)
    return(NASD_FAIL);
  *sizep = (int)size;
  return(NASD_SUCCESS);
}

#endif /* LINUX */

nasd_status_t
nasd_sys_thread_create_attr(
  nasd_sys_thread_t      *threadp,
  nasd_sys_threadattr_t  *attrp,
  nasd_threadfunc_t       func,
  nasd_threadarg_t        arg)
{
  int ret;

  ret = pthread_create(threadp, *attrp, (pthread_startroutine_t)func, arg);
  if (ret)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_thread_create(
  nasd_sys_thread_t  *threadp,
  nasd_threadfunc_t   func,
  nasd_threadarg_t    arg)
{
  int ret;

  ret = pthread_create(threadp, pthread_attr_default,
    (pthread_startroutine_t)func, arg);
  if (ret)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
