/*
 * nasd_common.c
 *
 * Shared "general" functionality for NASD entities
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <sys/errno.h>
#include <dce/dce_error.h>
#include <dce/rpc.h>
#include <sys/time.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_general.h>

#ifdef _KERNEL
#include <sys/kernel.h>
#define NULL 0L
#else /* _KERNEL */
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#endif /* _KERNEL */

/* external utility routines */
extern char* nasd_error_string(nasd_status_t error);

#ifdef KERNEL
char nasd_panicbuf[NASD_PANICBUF_LEN];
#endif /* KERNEL */


nasd_status_t
nasd_dce_direct_bind_to_server(
  unsigned_char_t       *protseq,        /* "ncadg_ip_udp" */
  unsigned_char_t       *host,           /* server hostname */
  unsigned_char_t       *endpoint,       /* port number as a string */ 
  rpc_binding_handle_t  *binding_handle) /* OUT: binding handle */
{
    unsigned_char_t             *string_binding;
    unsigned_char_t             *netoptions = 0;
    unsigned32                  status;
    int i, retries = 5;

    /*
     * compose a string binding from the arguments
     */
    rpc_string_binding_compose (
				NULL,                               /* no object uuid */
				(unsigned_char_t *) protseq,        /* protocol sequence */
				(unsigned_char_t *) host,        /* host address */
				(unsigned_char_t *) endpoint,        /* transport endpoint */
				(unsigned_char_t *) netoptions,        /* network options */
				&string_binding,
				&status);

  if (status != rpc_s_ok)  {
    return(NASD_FAIL);
  }

  /*
   * convert the string binding to a binding handle
   */
  for(i=0;i<retries;i++) {
    rpc_binding_from_string_binding (string_binding, binding_handle, &status);
    if (status == rpc_s_ok) {
      break;
    }
  }

  rpc_string_free(&string_binding, &status); /* XXX - no cleanup on failure above */

  if (status != rpc_s_ok) {
    return(NASD_FAIL);
  }

  return(NASD_SUCCESS);
}

nasd_status_t
nasd_dce_direct_unbind_server(rpc_binding_handle_t *binding_handle)
{
  unsigned32 status;

  rpc_binding_free(binding_handle, &status);

  if (status != rpc_s_ok)
    return(NASD_FAIL);

  return(NASD_SUCCESS);
}



#define DOTIME(_tv_) _nasd_timestr((long)((_tv_).ts_sec)), (_tv_).ts_sec, (_tv_).ts_nsec

char *
_nasd_timestr(
  time_t  t)
{
  static char str[128];
#ifdef KERNEL
  sprintf(str, "%lu", (unsigned long)t);
#else /* KERNEL */
  strcpy(str, ctime(&t));
  if (str[strlen(str)-1] == '\n') {
    str[strlen(str)-1] = '\0';
  }
#endif /* KERNEL */
  return(str);
}

char *
_nasd_timestr_r(
  time_t   t,
  char    *str)
{
#ifdef KERNEL
  sprintf(str, "%lu", (unsigned long)t);
#else /* KERNEL */
  strcpy(str, ctime(&t));
  if (str[strlen(str)-1] == '\n') {
    str[strlen(str)-1] = '\0';
  }
#endif /* KERNEL */
  return(str);
}

char *
nasd_timestr_r(
  nasd_timespec_t   ts,
  char             *str)
{
#ifdef KERNEL
  sprintf(str, "%d:%09d", ts.ts_sec, ts.ts_nsec);
#else /* KERNEL */
  time_t t;

  t = ts.ts_sec;
  strcpy(str, ctime(&t));
  if (str[strlen(str)-1] == '\n') {
    str[strlen(str)-1] = '\0';
  }
#endif /* KERNEL */
  return(str);
}

/*
 * Print the contents of a nasd attribute
 */
void
nasd_printstat(
  nasd_attribute_t  *attr)
{
  printf("  block_preallocation     %6" NASD_64u_FMT "\n", attr->block_preallocation);
  printf("  blocks_used             %6" NASD_64u_FMT "\n", attr->blocks_used);
  printf("  block_size              %6d\n", attr->block_size);
  printf("  object_len              %6" NASD_64u_FMT "\n", attr->object_len);
  printf("  attr_modify_time        %s (%d:%09d)\n", DOTIME(attr->attr_modify_time));
  printf("  object_modify_time      %s (%d:%09d)\n", DOTIME(attr->object_modify_time));
  printf("  object_create_time      %s (%d:%09d)\n", DOTIME(attr->object_create_time));
  printf("  fs_attr_modify_time     %s (%d:%09d)\n", DOTIME(attr->fs_attr_modify_time));
  printf("  fs_object_modify_time   %s (%d:%09d)\n", DOTIME(attr->fs_object_modify_time));
  printf("  av                      %6u\n", (unsigned)attr->av);
}


nasd_status_t
nasd_str_to_nasd_id(
  char               *str,
  nasd_identifier_t  *nidp)
{
  nasd_identifier_t n;
  nasd_uint64 v, j;
  int base, i, f;
  char c;

  *nidp = n = 0;

  if ((str[0] == '0') && (str[1] == 'x')) {
    base = 16;
    f = 2;
  }
  else {
    base = 10;
    f = 0;
  }

  for(j=1,i=strlen(&str[f])-1;i>=0;i--,j*=base) {
    c = str[f+i];
    v = (-1);
    if ((c >= '0') && (c <= '9')) {
      v = c - '0';
    }
    else if (base == 16) {
      if ((c >= 'a') && (c <= 'f'))
        v = (c - 'a') + 10;
      else if ((c >= 'A') && (c <= 'F'))
        v = (c - 'A') + 10;
    }
    if (v < 0)
      return(NASD_BAD_IDENTIFIER);
    n += v * j;
  }

  *nidp = n;

  return(NASD_SUCCESS);
}

#ifndef KERNEL
nasd_status_t
nasd_raw_disk_len(
  int           fd,
  nasd_uint64  *lenp)
{
  off_t max, min, pos, val;
  int i, rc;

  max = (off_t) ~((off_t)0x80<<((sizeof(max)-1)*CHAR_BIT));
  min = 0;
  val = min + ((max - min) >> 1);

  do {
    pos = lseek(fd, val, SEEK_SET);
    if (pos < 0) {
      max = min + ((max - min) >> 1);
    }
    else {
      rc = read(fd, &i, 1);
      if (rc == 0 ||
          ((rc < 0) && (errno == ENOSPC)))
        max = val;
      else
        min = val;
    }
    val = min + ((max - min) >> 1);
  } while (max > min + 1);
  *lenp = (nasd_uint64)(val+1);
  return(NASD_SUCCESS);
}
#endif /* !KERNEL */
