/*
 * nasd_error.c
 *
 * NASD error table
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_nfs_types.h>

#define KASE(_v_,_s_) case _v_: return(_s_);

typedef struct NASD_ErrorEnt_s NASD_ErrorEnt_t;

#define PUNT_STR "Unknown NASD error and internal error reporting mechanism failure"

struct NASD_ErrorEnt_s {
  nasd_status_t     status;
  char              str[48];
  NASD_ErrorEnt_t  *next;
};

NASD_DECLARE_ONCE(nasd_err_init_once)
NASD_DECLARE_STATIC_MUTEX(err_mutex)

static NASD_ErrorEnt_t *err_list;

static void
nasd_errsys_init()
{
  int rc;

  rc = nasd_mutex_init(&err_mutex);
  if (rc) {
    NASD_PANIC();
  }
  err_list = NULL;
}

char *
nasd_error_string(err)
  nasd_status_t  err;
{
  NASD_ErrorEnt_t *r;
  char *str;
  
  switch(err) {
    KASE(NASD_SUCCESS,"Success")
    KASE(NASD_NO_MEM,"Out of memory (or other resource)")
    KASE(NASD_FAIL,"Internal operation failure")
    KASE(NASD_NO_MORE_OBJECTS,"No more objects may be created in partition");
    KASE(NASD_AUTHCHECK_FAILED,"Authorization check failed")
    KASE(NASD_BAD_IDENTIFIER,"Bad NASD id")
    KASE(NASD_BAD_ATTR_SET,"Bad attribute fieldmask value")
    KASE(NASD_OP_NOT_SUPPORTED,"Operation not supported")
    KASE(NASD_BAD_TIMESPEC,"Bad timespec")
    KASE(NASD_NO_FREE_IDS,"No free NASD ids")
    KASE(NASD_GUARD_CHECK_FAILED,"Guard check failed")
    KASE(NASD_BAD_IO_LIST_LEN,"Bad I/O list length")
    KASE(NASD_BAD_IO_LIST,"Bad I/O list")
    KASE(NASD_BAD_OFFSET,"Bad offset")
    KASE(NASD_BAD_LEN,"Bad length")
    KASE(NASD_RPC_TRAP,"RPC exception")
    KASE(NASD_BAD_FIELDMASK,"Bad fieldmask")
    KASE(NASD_BAD_NONCE,"Invalid Nonce")
    KASE(NASD_INSUFFICIENT_PERM,"Insufficient permissions for operation")
    KASE(NASD_BAD_KEYTYPE,"Key type not valid for operation")
    KASE(NASD_WEAK_COOKIE,"Cookie does not have required protection for drive minimums")
    KASE(NASD_WRONG_NI,"Cookie NI and Request NI do no match")
    KASE(NASD_WRONG_DRIVE,"Cookie drive_name does not match expected value")
    KASE(NASD_TIMEOUT,"Timed out")
    KASE(NASD_NO_SPACE,"No space left on disk or partition")
    KASE(NASD_EXPIRED_CAPABILITY,"Capability expired")
    KASE(NASD_BAD_PARTITION,"Bad or invalid partition identifier")
    KASE(NASD_NOT_ON_CONTROL,"Operation not allowed on control object")
    KASE(NASD_KEYINIT_FAILED,"Unable to initialize cryptographic keys")
    KASE(NASD_BAD_DIGEST_REQ,"Digest on the REQUEST did not have the expected value")
    KASE(NASD_BAD_DIGEST_REPLY,"Digest on the REPLY did not have the expected value")
    KASE(NASD_OFFLINE,"Unit off-line")
    KASE(NASD_REMOTE_TRAP,"Remote exception trapped")
    KASE(NASD_DRIVE_INITIALIZED,"Drive was already initialized")
    KASE(NASD_DRIVE_UNINITIALIZED,"Drive has not been initialized")
    KASE(NASD_CAP_PROTECTION,"Request used protection that did not meet capability specified minimum")
    KASE(NASD_PART_PROTECTION,"Request used protection that did not meet partition specified minimum")
    KASE(NASD_SECURITY_MISMATCH,"Fields in the security_param and capability_args did not match when they were expected to.")
    KASE(NASD_IMMUTABLE_KEY,"The master key is immutable and may not be changed")
    KASE(NASD_WRONG_PARTITION,"The capability is not valid for the requested partition")
    KASE(NASD_REQUEST_OUTSIDE_RANGE,"The bytes addressed in the request are outside the range permitted by the capability")

    /*
     * API
     */
    KASE(NASD_CTRL_ID_CHECK_FAILED,"Request control id didn't match reply")
    KASE(NASD_BAD_MEM_LIST,"Bad memory list")
    KASE(NASD_MEM_LIST_ERR,"Error walking memory list")
    KASE(NASD_REMOTE_OVERRUN,"Remote agent overran local buffer")
    KASE(NASD_BAD_HANDLE,"Request issued with bad drive handle")
    KASE(NASD_BAD_HANDLE_TYPE,"Request issued with bad or incompatible handle type")
    KASE(NASD_BAD_BINDING_PARAM,"Bad binding parameter")
    KASE(NASD_BAD_BINDING_PARAM_LEN,"Bad binding parameter length")
    KASE(NASD_BINDING_OP_NOSUPP,"Operation not supported by binding type")
    KASE(NASD_RPC_LOCAL_FAIL,"Local RPC failure")
    KASE(NASD_SYS_NOT_INITIALIZED,"Subsystem not initialized")

    /*
     * Drive-internal
     */
    KASE(NASD_BAD_LAYOUT_TYPE,"Bad layout type (drive-internal error)")
    KASE(NASD_BAD_BLKNO,"Bad block number (drive-internal error)")
    KASE(NASD_BAD_BLKCNT,"Bad block count (drive-internal error)")
    KASE(NASD_EJECTING,"Ejecting block (drive-internal status)")
    KASE(NASD_BLOCK_NOT_CACHED,"Block not cached (drive-internal status)")
    KASE(NASD_BLOCK_PARTIAL,"Partial block complete (drive-internal status)")
    KASE(NASD_OBJ_REMOVING,"Removing object (drive-internal status)")
    KASE(NASD_GOOD_DIGEST," XXX !!! DOCUMENT THIS ERROR CODE !!! ")
    KASE(NASD_BAD_DIGEST," XXX !!! DOCUMENT THIS ERROR CODE !!! ")
    KASE(NASD_IOSYS_FULL,"I/O subsystem is full (drive-internal error)")
    KASE(NASD_BAD_IOQUEUE_TYPE,"Bad I/O queue type (drive-internal error)")
    KASE(NASD_BAD_IOQUEUE_LEN,"Bad I/O queue length (drive-internal error)")

    /*
     * NFS errors
     */
    KASE(NASD_NFS_BAD_MARKER_VERF,"NFS: in_markerverf bad or stale")
    KASE(NASD_NFS_STALE_ID,"NFS: stale or bad id")
    KASE(NASD_NFS_BAD_NAME,"NFS: bad name")
    KASE(NASD_NFS_ALREADY_EXISTS,"NFS: name already exists")
    KASE(NASD_NFS_NOT_DIR,"NFS: not a directory")
    KASE(NASD_NFS_IS_DIR,"NFS: is a directory")
    KASE(NASD_NFS_DIR_NOT_EMPTY,"NFS: directory not empty")

    default:
      /*
       * Return an error string identifying the error
       * number. Thread-safe. Safe for children, pets.
       */
      nasd_once(&nasd_err_init_once, nasd_errsys_init);
      str = NULL;
      NASD_LOCK_MUTEX(err_mutex);
      for(r=err_list;r;r=r->next) {
        if (r->status == err) {
          str = r->str;
          break;
        }
      }
      if (str == NULL) {
        NASD_Malloc(r,sizeof(NASD_ErrorEnt_t),(NASD_ErrorEnt_t *));
      }
      if (r == NULL) {
        str = PUNT_STR;
      }
      else {
        r->status = err;
        sprintf(r->str, "Unknown NASD error %d", (int)err);
        r->next = err_list;
        err_list = r;
        str = r->str;
      }
      NASD_UNLOCK_MUTEX(err_mutex);
      return(str);
  }
}
