/*
 * nasd_check.c
 *
 * Fault, sanity checking for NASD drive.
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_cache.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_timer.h>
#include <nasd/nasd_security_dr.h>
#include <nasd/nasd_layout.h>

/*
 * Note that this does not take partition locks. To be protected
 * against modification, the caller should hold a partition lock.
 */
nasd_status_t
nasd_od_check_startup()
{
  nasd_status_t rc, check_rc;

  check_rc = NASD_SUCCESS;

  rc = nasd_od_check_npt();
  if (rc) {
    printf("DRIVE-CHECK: failed checking NPT error 0x%x (%s)\n",
      rc, nasd_error_string(rc));
    check_rc = NASD_FAIL;
  }

  if (check_rc == NASD_SUCCESS) {
    printf("DRIVE-CHECK: all sanity checks passed\n");
  }

  return(check_rc);
}

/*
 * Note that this does not take partition locks. To be protected
 * against modification, the caller should hold a partition lock.
 */
nasd_status_t
nasd_od_check_npt()
{
  nasd_blkcnt_t npt_sz, npt2_sz;
  nasd_odc_ent_t *npte, *npte2;
  nasd_status_t rc, check_rc;
  nasd_blkno_t nbb, nb, nb2;
  int npti;

  check_rc = NASD_SUCCESS;

  npt_sz = nasd_odc_state->disk.npt_ext.last
    - nasd_odc_state->disk.npt_ext.first + 1;
  npt2_sz = nasd_odc_state->disk.npt2_ext.last
    - nasd_odc_state->disk.npt2_ext.first + 1;

  if ((npt_sz != npt2_sz) || (npt_sz != nasd_odc_state->npt_sz)) {
    printf("DRIVE-CHECK: failed NPT size checks\n");
    printf("DRIVE-CHECK: npt  = %u..%u size %u\n",
      nasd_odc_state->disk.npt_ext.first, nasd_odc_state->disk.npt_ext.last,
      npt_sz);
    printf("DRIVE-CHECK: npt2 = %u..%u size %u\n",
      nasd_odc_state->disk.npt2_ext.first, nasd_odc_state->disk.npt2_ext.last,
      npt2_sz);
    printf("DRIVE-CHECK: npt_sz = %u\n", npt_sz);
    check_rc = NASD_FAIL;
  }

  for(nbb=0;nbb<nasd_odc_state->npt_sz;nbb++) {
    nb = nasd_odc_state->disk.npt_ext.first + nbb;
    nb2 = nasd_odc_state->disk.npt2_ext.first + nbb;

    rc = nasd_odc_block_get(NULL, nb,
      NASD_ODC_L_FORCE|NASD_ODC_L_BLOCK|NASD_ODC_L_LOAD,
      &npte, NASD_ODC_T_NPT1, NULL);
    if (rc) {
      printf("DRIVE-CHECK: nbb=%u nb=%u could not get npte\n",
        nbb, nb);
      check_rc = NASD_FAIL;
      continue;
    }

    rc = nasd_odc_block_get(NULL, nb2,
      NASD_ODC_L_FORCE|NASD_ODC_L_BLOCK|NASD_ODC_L_LOAD,
      &npte2, NASD_ODC_T_NPT2, NULL);
    if (rc) {
      printf("DRIVE-CHECK: nbb=%u nb2=%u could not get npte2\n",
        nbb, nb);
      nasd_odc_block_release(npte);
      check_rc = NASD_FAIL;
      continue;
    }

    NASD_ODC_RLOCK_BLOCK_DATA(npte);
    NASD_ODC_RLOCK_BLOCK_DATA(npte2);

    NASD_ODC_LOCK_BLOCK(npte);
    nasd_odc_wait_not_busy_invalid(npte);
    NASD_ODC_UNLOCK_BLOCK(npte);

    NASD_ODC_LOCK_BLOCK(npte2);
    nasd_odc_wait_not_busy_invalid(npte2);
    NASD_ODC_UNLOCK_BLOCK(npte2);

    for(npti=0;npti<NASD_OD_NODES_PER_NPT_BLOCK;npti++) {
      if (npte->data.pte[npti].blkno != npte2->data.pte[npti].blkno) {
        printf("DRIVE-CHECK: NPT blocks do not match\n");
        printf("DRIVE-CHECK: npti=%d nb=%u nb2=%u nbb=%u\n",
          npti, nb, nb2, nbb);
        printf("DRIVE-CHECK: npte->data.pte[npti].blkno = %u\n",
          npte->data.pte[npti].blkno);
        printf("DRIVE-CHECK: npte2->data.pte[npti].blkno = %u\n",
          npte2->data.pte[npti].blkno);
        check_rc = NASD_FAIL;
      }
    }

    NASD_ODC_RUNLOCK_BLOCK_DATA(npte2);
    NASD_ODC_RUNLOCK_BLOCK_DATA(npte);

    nasd_odc_block_release(npte);
    nasd_odc_block_release(npte2);
  }

  return(check_rc);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
