/*
 * nasd_nonce_mgmt_dr.c
 *
 * Handle checking if a nonce is fresh or resued
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Howard Gobioff 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/***********************************************************************
 *
 * nasd_nonce_mgmt.c
 *
 * Handle checking if a nonce is fresh or resued
 ***********************************************************************/

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_cache.h>
#include <nasd/nasd_nonce_mgmt_dr.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_security_time.h>
#include <nasd/nasd_freelist.h>

#define PRINT_ERROR

#ifdef DEBUG_ALL
#define PRINT_ERROR
#define DEBUG_INIT_NONCE
#define DEBUG_CHECK_NONCE
#endif

nasd_nonce_el_t * nasd_nonce_table_dr[NASD_NONCE_BUCKETS];
NASD_DECLARE_MUTEX(nasd_nonce_table_dr_lock[NASD_NONCE_BUCKETS]);

nasd_freelist_t *nasd_nonce_freelist;
#define NASD_MAX_FREE_NONCE 1024
#define NASD_NONCE_INC        32
#define NASD_NONCE_INITIAL   512

#define NASD_NONCE_GET(_noncep_) \
  NASD_FREELIST_GET(nasd_nonce_freelist,_noncep_,next,(nasd_nonce_el_t *))

#define NASD_NONCE_FREE(_noncep_) \
  NASD_FREELIST_FREE(nasd_nonce_freelist,_noncep_,next)

void
nasd_shutdown_nonce_table(
  void  *ignored)
{
  nasd_nonce_el_t *r, *n;
  int i;

  for(i=0;i<NASD_NONCE_BUCKETS;i++) {
    for(r=nasd_nonce_table_dr[i];r;r=n) {
      n = r->next;
      NASD_NONCE_FREE(r);
    }
  }
}

void
nasd_shutdown_free_nonce(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_nonce_freelist,next,(nasd_nonce_el_t *));
}

nasd_status_t nasd_init_nonce_mgmt_dr(){
#if NASD_SECURE_RPCS_ENABLE > 0
  int i;
  nasd_status_t rc=NASD_SUCCESS;
  nasd_nonce_el_t * curr;
  nasd_nonce_el_t * next;

#ifdef DEBUG_INIT_NONCE
  printf("DEBUG(nasd_init_nonce): entered\n");
#endif /* DEBUG_INIT_NONCE */

    NASD_FREELIST_CREATE(nasd_nonce_freelist, NASD_MAX_FREE_NONCE,
      NASD_NONCE_INC, sizeof(nasd_nonce_el_t));
    if (nasd_nonce_freelist == NULL)
      return(NASD_NO_MEM);
    NASD_FREELIST_PRIME(nasd_nonce_freelist, NASD_NONCE_INITIAL,next,
      (nasd_nonce_el_t *));
    rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_free_nonce,
      NULL);
    if (rc) {
      nasd_shutdown_free_nonce(NULL);
      return(rc);
    }

    for(i=0;i<NASD_NONCE_BUCKETS;i++) {
      nasd_nonce_table_dr[i]=NULL;
      rc=nasd_mutex_init(&(nasd_nonce_table_dr_lock[i]));
      if (rc) {
        printf("ERROR: %s:%d nasd_mutex_init() failed, rc=%d\n", __FILE__, __LINE__, rc);
        return(rc);
      }
      rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_mutex,
                              &nasd_nonce_table_dr_lock[i]);
      if (rc) {
        printf("ERROR: %s:%d cannot establish mutex autodestroy, rc=%d\n", __FILE__, __LINE__, rc);
        nasd_mutex_destroy(&nasd_nonce_table_dr_lock[i]);
        return(rc);
      }
    }

    rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_nonce_table,
      NULL);
    if (rc) {
      nasd_shutdown_nonce_table(NULL);
      return(rc);
    }

#ifdef DEBUG_INIT_NONCE
  printf("DEBUG(nasd_init_nonce): done\n");
#endif /* DEBUG_INIT_NONCE */

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  return(NASD_SUCCESS);
}

nasd_status_t nasd_check_nonce_dr(nasd_capability_args_t *cap_args,
                               nasd_timespec_t *nonce){
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_timespec_t now;
  nasd_uint64 now_l,then_l,diff;
  nasd_nonce_el_t * curr,*new_el,*tmp;
  int  bucket;
  int check_first=1;
  nasd_status_t rc;
#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce): entered %" NASD_64s_FMT " for drive %d\n",*(nasd_uint64*)nonce,cap_args->drive_name);
#endif

  nasd_security_get_drive_time(cap_args->drive_name,&now);

  

  now_l=(nasd_uint64)now.ts_nsec+(nasd_uint64)now.ts_sec*NASD_NSEC_PER_SEC;
  then_l=(nasd_uint64)nonce->ts_nsec+(nasd_uint64)nonce->ts_sec*NASD_NSEC_PER_SEC;

#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce): then was  %" NASD_64s_FMT "\n",then_l);
#endif

  diff=now_l-then_l;
  if (NASD_ABS(diff) > NASD_ACCEPTABLE_SKEW) {

#ifdef PRINT_ERROR
    printf("DEBUG(nasd_check_nonce) too far from now FAILED - NASD_BAD_NONCE \n");
    printf("DEBUG(nasd_check_nonce) now= %ld:%ld \n",now.ts_sec,now.ts_nsec);
    printf("DEBUG(nasd_check_nonce) then=%ld:%ld \n",nonce->ts_sec,nonce->ts_nsec);

#endif
    return(NASD_BAD_NONCE);
  }

  /* We only expire  capabilities */
  /* This check is here so we only goto the clock once -hgobioff */
  if (((cap_args->type == NASD_RED_CAPABILITY) || (cap_args->type == NASD_BLACK_CAPABILITY) )&& (now.ts_sec > cap_args->expiration_seconds)) 
    {
#ifdef PRINT_ERROR
    printf("DEBUG(nasd_check_nonce) expired capability \n");
#endif
      return(NASD_EXPIRED_CAPABILITY);
    }

  NASD_NONCE_GET(new_el);

/* We have a nasd_nonce_el_t */

  new_el->time=then_l;
  new_el->ni=cap_args->ni;
  new_el->expiration=cap_args->expiration_seconds;
  new_el->drive_name=cap_args->drive_name;


  bucket=NONCE_HASH(new_el);
  bucket=NASD_ABS(bucket%NASD_NONCE_BUCKETS);

#ifdef DEBUG_CHECK_NONCE
    printf("DEBUG(nasd_check_nonce) looking for %" NASD_64s_FMT " using bucket %d  (<%d)\n",new_el->time,bucket,NASD_NONCE_BUCKETS);
#endif  
 
  NASD_LOCK_MUTEX(nasd_nonce_table_dr_lock[bucket]);

#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) grabbed lock for bucket %d\n",bucket);
#endif
  
  /* Check the first element in the bucket */
  while (check_first) {
#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) in check_first loop \n");
#endif

    /* Case 1 - Empty bucket */
    if (nasd_nonce_table_dr[bucket] == NULL ) {
      
#ifdef DEBUG_CHECK_NONCE
      printf("DEBUG(nasd_check_nonce) 1  inserting in an empty bucket \n");
#endif
      nasd_nonce_table_dr[bucket]=new_el;
      new_el->next=NULL;

      NASD_UNLOCK_MUTEX(nasd_nonce_table_dr_lock[bucket]);
      return(NASD_SUCCESS);
    } else if (nasd_nonce_table_dr[bucket]->drive_name == new_el->drive_name ) {
      /* Case 2 - First element should be freed */ 
#ifdef DEBUG_CHECK_NONCE
	printf("DEBUG(nasd_check_nonce) first element is for the correct drive\n");
#endif
      if (NASD_ABS(now_l-nasd_nonce_table_dr[bucket]->time) > NASD_ACCEPTABLE_SKEW) {
	/* We can free it */

#ifdef DEBUG_CHECK_NONCE
	printf("DEBUG(nasd_check_nonce) 2 **freeing the first element\n");
#endif
	tmp=nasd_nonce_table_dr[bucket];
	nasd_nonce_table_dr[bucket]=nasd_nonce_table_dr[bucket]->next;
	NASD_NONCE_FREE(tmp);
      }  else if ((nasd_nonce_table_dr[bucket]->time == new_el->time) &&
		  (nasd_nonce_table_dr[bucket]->ni == new_el->ni) &&
		  (nasd_nonce_table_dr[bucket]->expiration == new_el->expiration))
	{/* Case 3 - First element is a match */
	  
#ifdef DEBUG_CHECK_NONCE
	  printf("DEBUG(nasd_check_nonce) 3 found a match in the bucket FAILED - NASD_BAD_NONCE\n");
         
#endif
	  NASD_NONCE_FREE(new_el);
	  NASD_UNLOCK_MUTEX(nasd_nonce_table_dr_lock[bucket]);
	  return(NASD_BAD_NONCE);
	} else {
#ifdef DEBUG_CHECK_NONCE
	  printf("DEBUG(nasd_check_nonce) first element is uninteresting\n");
#endif
	  check_first=0;
	}
    } else {
      check_first=0;
    }
  }
  
#ifdef DEBUG_CHECK_NONCE
    printf("DEBUG(nasd_check_nonce) checking the chain\n");
#endif  
  curr=nasd_nonce_table_dr[bucket];
  NASD_ASSERT(curr!=NULL);


  while(curr->next!=NULL) {
   if (curr->next->drive_name == new_el->drive_name ) {
     if (NASD_ABS(now_l-curr->next->time) > NASD_ACCEPTABLE_SKEW) {
       /* We can free it */
      
#ifdef DEBUG_CHECK_NONCE
       printf("DEBUG(nasd_check_nonce) 4 **freeing an element in the chain\n");
#endif
       tmp=curr->next;
       curr->next=curr->next->next;
       NASD_NONCE_FREE(tmp);
     } else if ((curr->next->time == new_el->time) &&
		(curr->next->ni == new_el->ni) &&
		(curr->next->expiration == new_el->expiration))
       {
	
#ifdef PRINT_ERROR
	 printf("DEBUG(nasd_check_nonce) 5 found a match in the chain FAILED=NASD_BAD_NONCE\n");
#endif
	 NASD_NONCE_FREE(new_el);
	 NASD_UNLOCK_MUTEX(nasd_nonce_table_dr_lock[bucket]);
	 return(NASD_BAD_NONCE);
       }
   }
   curr=curr->next;
  }
  

  
  /* We insert it at the end which means it will take longer
     to search for but we will walk the list and cleanup as we search */
  new_el->next=NULL;
  curr->next=new_el;
#ifdef DEBUG_CHECK_NONCE
  printf("DEBUG(nasd_check_nonce) finished successfully\n");
#endif
  

  NASD_UNLOCK_MUTEX(nasd_nonce_table_dr_lock[bucket]);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  
  return(NASD_SUCCESS);
    
}
