/*
 * nasd_od_drive.c
 *
 * User-level RPC front-end for on-disk filesystem.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#undef  NASD_PTHREAD_EXC
#define NASD_PTHREAD_EXC 1
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_cache.h>
#include <nasd/nasd_common.h>
#ifndef KERNEL
#include <nasd/nasd_mq.h>
#endif /* !KERNEL */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <dce/dce_error.h>
#include <nasd/nasd_pdrive.h>

#include <dce/dce_error.h>
#ifndef LINUX
#include <rpc/rpc.h>
#endif /* !LINUX */
#include <dce/stubbase.h>

#include <nasd/nasd_drive_defaults.h>

char *progname;
char *devname;

extern char *optarg;
extern int   optind;

static int stopped = 0;

extern int nasd_od_mq_ignore_eexist;
extern int nasd_od_rusage_at_shutdown;

int service_threads = NASD_DEF_SERVICE_THREADS;

int io_outstanding = NASD_DEF_IO_OUTSTANDING;

int def_cachesize;

extern nasd_pdrive_v0_1_epv_t nasd_pdrive_epv;

extern unsigned_char_t *nasd_od_protseq_str;

nasd_layout_type_t layout = NASD_DEF_LAYOUT;

nasd_ioqueue_type_t ioqueue = NASD_DEF_IOQUEUE;

nasd_od_config_t config;

#if LINUX
/*
 * Under Linux, the SIGINT signal is delivered to *all* of the
 * threads, not just one (as under DU).  Because it is impossible to
 * determine which thread will hit this code first, it is possible
 * for the thread that runs the shutdown code to be a thread that
 * also has cleanup work to do before it exits.  If that happens,
 * the drive will deadlock.  Therefore, we create a "supervisor"
 * thread to do the shutdown stuff, and allow the thread that ran
 * the signal handler to return. When signals come in, we simply
 * ensure that this thread awakens.
 */
nasd_thread_id_t nasd_od_master_shutdown_thread_id;
nasd_threadgroup_t nasd_od_master_shutdown_thread_group;
int nasd_od_linux_shutdown_signalled = 0;
nasd_thread_t nasd_od_shutdown_worker;
#endif /* LINUX */

void handle_sigint(int);
static void handle_sigint2(nasd_threadarg_t ignored);

unsigned long max_len = 0;

#ifdef LINUX

static void
nasd_od_linux_shutdown_worker(
  nasd_threadarg_t  ignored)
{
  struct timeval tv;
  int ret;

  nasd_od_master_shutdown_thread_id = nasd_thread_self();
  NASD_THREADGROUP_RUNNING(&nasd_od_master_shutdown_thread_group);

  while(!nasd_od_linux_shutdown_signalled) {
    tv.tv_sec = 7200;
    tv.tv_usec = 0;
    ret = select(1, NULL, NULL, NULL, &tv);
    if (nasd_od_linux_shutdown_signalled)
      break;
    if (ret) {
      fprintf(stderr, "DRIVE WARNING: linux shutdown worker thread got
        ret %d errno %d\n", ret, errno);
      fflush(stderr);
    }
  }

  handle_sigint2(NULL);

  /* I don't think we should ever get back here. */
}

void
handle_linux_sigint(
  int  sig)
{
  nasd_od_linux_shutdown_signalled = 1;
}

#else /* LINUX */

void
handle_sigint(
  int  sig)
{
  /* Under DU, only one thread receives the SIGINT (the "master"
     thread), so it's safe to run the shutdown handler code
     ourselves. */
  stopped = 1;
  handle_sigint2(NULL);
}

#endif /* LINUX */

static void
handle_sigint2(
  nasd_threadarg_t  ignored)
{
  nasd_thread_id_t self;
  nasd_status_t rc;

  self = nasd_thread_self();

  fprintf(stderr, "INTERRUPTED (thread id %" NASD_THREAD_ID_FMT ")\n", self);
  fflush(stderr);

  nasd_stop_rpc();

#if MJU_DEBUG
  fprintf(stderr, "shutdown[%lu]: stop_rpc() complete\n", self);
  fflush(stderr);
#endif /* MJU_DEBUG */

  rc = nasd_basic_shutdown();
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) from nasd_basic_shutdown()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  fprintf(stderr, "Completed shutdown.\n");
  fflush(stderr);
#if NASD_MEM_COUNT_ALLOC > 0
  if (nasd_mem_allocated) {
    fprintf(stderr,
            "WARNING: %ld bytes of memory still outstanding, something leaked core.\n",
            nasd_mem_allocated);
    fflush(stderr);
  }
  else {
    printf("%ld bytes of memory still outstanding.\n", nasd_mem_allocated);
  }
#endif /* NASD_MEM_COUNT_ALLOC > 0 */
  fflush(stdout);

  exit(0);
}

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] devname\n", progname);
  fprintf(stderr, "options:\n");
  fprintf(stderr, "  -c  cache_size [default %d]\n", def_cachesize);
  fprintf(stderr, "  -f  format\n");
  fprintf(stderr, "  -i  I/Os outstanding [default %d]\n",
    NASD_DEF_IO_OUTSTANDING);
  fprintf(stderr, "  -l  layout (default '%c')\n", NASD_DEF_LAYOUT);
  fprintf(stderr, "  -m  maximum disk length to use\n");
  fprintf(stderr, "  -q  disk queue policy (default '%c')\n",
    NASD_DEF_IOQUEUE);
  fprintf(stderr, "  -r  show rusage stats at shutdown\n");
  fprintf(stderr, "  -T  use TCP\n");
  fprintf(stderr, "  -t  service threads [default %d]\n",
    NASD_DEF_SERVICE_THREADS);
  fprintf(stderr, "  -x  ignore extant message queue\n");
  fflush(stderr);
  exit(1);
}

int
main(argc, argv)
  int     argc;
  char  **argv;
{
  unsigned long num_sect;
  error_status_t status;
  nasd_uint64 dlen, len;
  nasd_status_t rc;
  int fd, ret, val;
  struct stat s;
  dev_t dev;
  char c;

  progname = argv[0];
  def_cachesize = nasd_odc_size;

  nasd_od_setup_udp();

  while ((c = getopt(argc, argv, "c:fi:l:m:q:rTt:x")) != EOF) {
    switch(c) {
      case 'c':
        if (sscanf(optarg, "%d", &val) != 1)
          usage();
        if (val < 1)
          usage();
        nasd_odc_size = val;
        break;
      case 'f':
        nasd_odc_force_format = 1;
        break;
      case 'i':
        if (sscanf(optarg, "%d", &io_outstanding) != 1)
          usage();
        break;
      case 'l':
        if (strlen(optarg) != 1) {
          usage();
        }
        layout = optarg[0];
        break;
      case 'm':
        if (sscanf(optarg, "%lu", &max_len) != 1)
          usage();
        if (max_len == 0)
          usage();
        break;
      case 'q':
        if (strlen(optarg) != 1) {
          usage();
        }
        ioqueue = optarg[0];
        break;
      case 'r':
        nasd_od_rusage_at_shutdown = 1;
        break;
      case 'T':
        nasd_od_setup_tcp();
        break;
      case 't':
        if (sscanf(optarg, "%d", &service_threads) != 1)
          usage();
        break;
      case 'x':
        if(nasd_od_mq_ignore_eexist)
          usage();
        nasd_od_mq_ignore_eexist = 1;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }
  if (optind >= argc)
    usage();
  devname = argv[optind];
  optind++;
  if (optind < argc)
    usage();

  if (service_threads < 1)
    usage();

  if (io_outstanding < 1)
    usage();

  fd = open(devname, O_RDWR);
  if (fd < 0) {
    perror(devname);
    fprintf(stderr, "ERROR: cannot open %s\n", devname);
    fflush(stderr);
    exit(1);
  }

  rc = nasd_raw_disk_len(fd, &dlen);
  if (rc) {
    fprintf(stderr, "ERROR: cannot get raw disk length, error 0x%x (%s)\n",
      rc, nasd_error_string(rc));
    exit(1);
  }

  ret = fstat(fd, &s);
  if (ret) {
    perror("fstat");
    fprintf(stderr, "ERROR: could not stat %s\n", devname);
    fflush(stderr);
    exit(1);
  }
  dev = s.st_rdev;

  close(fd);

  config.layout_type = layout;
  config.ioqueue_type = ioqueue;
  config.ios_outstanding = io_outstanding;

  if (max_len) {
    len = NASD_MIN(max_len, dlen);
    num_sect = (unsigned long)len;
  }
  else {
    num_sect = (unsigned long)dlen;
  }
  num_sect >>= 9;

  printf("Initializing %s, %lu sectors\n", devname, num_sect);
#if NASD_SECURE_RPCS_ENABLE > 0
   printf("\nDrive compiled *WITH* security\n\n");
#else /* NASD_SECURE_RPCS_ENABLE > 0 */
   printf("\nDrive compiled WITHOUT security\n\n");
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  fflush(stdout);

  rc = nasd_basic_init();
  if (rc) {
    fprintf(stderr, "ERROR: got %s from nasd_basic_init()\n",
      nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_od_io_init(devname, &config);
  if (rc) {
    fprintf(stderr, "ERROR: got %s from nasd_od_io_init()\n",
      nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  printf("Setting up...\n");
  fflush(stdout);

  rc = nasd_setup_disk(num_sect, dev, &config);
  if (rc) {
    fprintf(stderr, "ERROR: got %s from nasd_setup_disk()\n",
      nasd_error_string(rc));
    fflush(stderr);
    goto abort_drive;
  }

#ifdef LINUX
  rc = nasd_init_threadgroup(&nasd_od_master_shutdown_thread_group);
  if (rc) {
    fprintf(stderr,
      "ERROR: got 0x%x (%s) initializing nasd_od_master_shutdown_thread_group\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    goto abort_drive;
  }

  rc = nasd_thread_create(&nasd_od_shutdown_worker,
    nasd_od_linux_shutdown_worker, NULL);
  if (rc) {
    fprintf(stderr,
      "ERROR: got 0x%x (%s) creating shutdown worker thread\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    goto abort_drive;
  }
  NASD_THREADGROUP_WAIT_START(&nasd_od_master_shutdown_thread_group);
#endif /* LINUX */

  rc = nasd_od_io_go();
  if (rc) {
    fprintf(stderr, "ERROR: got %s from nasd_od_io_go()\n",
      nasd_error_string(rc));
    fflush(stderr);
    goto abort_drive;
  }

  printf("Setup RPC subsystem\n");

  rc = nasd_startup_rpc();
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) from nasd_startup_rpc()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

#ifndef KERNEL
  printf("Setup message queue subsystem\n");

  rc = nasd_startup_msgq();
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) from nasd_startup_msgq()\n",
            rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }
#endif /* !KERNEL */

#ifndef LINUX
  signal(SIGUSR1, handle_sigint);
  signal(SIGINT, handle_sigint);
#endif /* LINUX */

#ifdef LINUX
  /* Linux's pthreads library uses SIGUSR1 and SIGUSR2, so we can't. */
  signal(SIGPWR, handle_linux_sigint);
  signal(SIGINT, handle_linux_sigint);
#endif /* LINUX */

  printf("Running with pid %d.\n", getpid());
  fflush(stdout);

  rpc_server_listen(service_threads, &status);

  printf("Listener aborted! status=%lx\n", (unsigned long)status);
  fflush(stdout);

abort_drive:
  if (!stopped) {
    nasd_stop_rpc();

    rc = nasd_basic_shutdown();
    if (rc) {
      fprintf(stderr, "ERROR: got %s (%lx) from nasd_basic_shutdown()\n",
        nasd_error_string(rc), (u_long)rc);
      fflush(stderr);
      exit(1);
    }
  }

  exit(2);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
