/*
 * nasd_od_ops.c
 *
 * RPC handlers for embedded NASD
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#if defined(DEC_OSF) || (defined(OSF) && defined(MACH))
#include <sys/secdefines.h>
#if SEC_BASE > 0
#include <sys/security.h>
#include <sys/audit.h>
#include <sys/secpolicy.h>
#endif /* SEC_BASE */
#endif /* DEC_OSF || (OSF && MACH) */
#include <sys/errno.h>

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#undef  NASD_PTHREAD_EXC
#define NASD_PTHREAD_EXC 1
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_sys.h>
#include <nasd/nasd_cache.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_keymgmt.h>

#include <nasd/nasd_security_dr.h>
#include <nasd/nasd_security_time.h>

#include <nasd/nasd_timer.h>
#include <nasd/nasd_trace_dr.h>

#include <nasd/nasd_timeout.h>

#include <dce/dce_error.h>

#if defined(DEC_OSF) && defined(KERNEL)
#include <dce/ker/pthread_exc.h>
#endif /* DEC_OSF && KERNEL */

#ifndef LINUX
#include <rpc/rpc.h>
#endif /* !LINUX */

#include <dce/stubbase.h>

#if defined(DEC_OSF) && defined(KERNEL)
/* I don't really want to use DCE's malloc and free */
#undef malloc
#undef free
#endif /* DEC_OSF && KERNEL */

#define NASD_OD_RPC_NASD_RET out_res.nasd_status
#include <nasd/nasd_od_rpc.h>

nasd_uint64 nasd_od_threads_in_rpc = 0;
nasd_uint64 nasd_od_shutdown_timeouts = 0;
extern nasd_uint64 nasd_od_shutdown_timeouts_max;
extern nasd_timespec_t nasd_od_shutdown_timeout_interval;
int nasd_od_wait_shutdown = 0;
nasd_timeout_handle_t nasd_od_shutdown_timeout_handle;

NASD_DECLARE_COND(nasd_od_shutdown_cond)
NASD_DECLARE_MUTEX(nasd_od_rpc_cnt_lock)
NASD_DECLARE_MUTEX(nasd_od_rpc_qdepth_lock)

#if NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0
NASD_DECLARE_MUTEX(nasd_dce_pipe_lock)
#endif /* NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0 */

nasd_drive_name_t my_name = 0;

int nasd_pdrive_protseq_installed = 0;
int nasd_pdrive_active = 0;

int nasd_pdrive_if_registered = 0;

int nasd_od_rpc_qdepth = 0;
nasd_timer_t nasd_od_rpc_qdepth_timer;

int nasd_od_rpc_cnt_lock_initted = 0;

extern char nasd_odc_zeroblk[NASD_OD_BASIC_BLOCKSIZE];
extern char nasd_odc_bitbucket[NASD_OD_BASIC_BLOCKSIZE];
extern nasd_freelist_t *nasd_ebuff_freelist;
nasd_ctrl_drive_opstats_t nasd_drive_opstats;

nasd_drive_opholder_t nasd_drive_fail_opholder;

unsigned_char_t *nasd_od_protseq_str = (unsigned_char_t *)"ncadg_ip_udp";

/* Don't use __FILE__ here to avoid getting kernel-mangled names */
#define DOBEGIN(_opname_) NASD_OD_RPC_DOBEGIN(_opname_,"nasd_od_ops.c")
#define DORETURN(_opname_) NASD_OD_RPC_DORETURN(_opname_,"nasd_od_ops.c")

#define NASD_OD_DCE_EXCEPTION_CAUGHT() { \
}

/*
 * No-op
 */
void
nasd_sp_null_dr(
  handle_t         client_handle, 
  nasd_status_t   *nasd_status,
  error_status_t  *status)
{
  nasd_od_rpc_basic_t out_res;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(null);
  nasd_dt_noop_args(&trace_seq);

  *nasd_status = out_res.nasd_status = NASD_SUCCESS;

  nasd_dt_noop_res(trace_seq);
  DORETURN(null);
}

/*
 * Sync
 */
void
nasd_sp_sync_dr(
  handle_t         client_handle, 
  nasd_status_t   *nasd_status,
  error_status_t  *status)
{
  nasd_od_rpc_basic_t out_res;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(sync);
  nasd_dt_sync_args(&trace_seq);

  TRY {
    out_res.nasd_status = nasd_odc_flush_dirty(1);
    *nasd_status = out_res.nasd_status;
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("sync caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
      
  nasd_dt_sync_res(trace_seq);
  DORETURN(sync);
}

/*
 * Create partition
 */
void
nasd_sp_part_creat_otw_dr(
  handle_t                          client_handle, 
  nasd_p_part_creat_dr_args_otw_t   in_args_otw,
  nasd_p_part_creat_dr_res_otw_t    out_res_otw,
  error_status_t                   *status)
{
  nasd_p_part_creat_dr_args_t in_args;
  nasd_p_part_creat_dr_res_t out_res;
  nasd_args_len_t in_args_len = sizeof(nasd_p_part_creat_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_part_creat_dr_res_otw_t);

  nasd_identifier_t nid;
  nasd_blkcnt_t blkcnt;
  nasd_status_t rc;
  int partnum;

  nasd_key_t op_key;
  nasd_uint16 min_protection;
  nasd_status_t verification;
  nasd_iv_t icv;
  nasd_trace_seq_t trace_seq; 

  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(part_creat);

  nasd_dt_part_creat_args(in_args_otw, &trace_seq);
  TRY {
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_part_creat_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_part_creat_dr_args_t, in_args_otw, icv);
    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    } else if (in_args.in_security_param.type != NASD_DRIVE_KEY ) {
      out_res.nasd_status=NASD_BAD_KEYTYPE;
    } else if ((in_args.in_security_param.actual_protection & NASD_SECURE_OPS)
               != NASD_SECURE_OPS) {
      out_res.nasd_status=NASD_WEAK_COOKIE;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {
      
  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

    nasd_p_part_creat_dr_unmarshall_args(in_args_otw, &in_args);
#if NASD_SECURE_RPCS_ENABLE > 0
    if ((verification= nasd_security_verify_args((nasd_args_t*)&in_args,
                                               (nasd_byte_t*)in_args_otw,
                                               in_args_len,
                                               NULL,
                                               0,
                                               (nasd_byte_t * )&(((nasd_p_part_creat_dr_args_t*)in_args_otw)->in_digest),
                                               op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else {

#if HBG_DEBUG
printf("nasd_sp_part_creat_otw_dr: partnum = %d protection = %d\n",in_args.in_partnum,PART(in_args.in_partnum).min_protection);
#endif
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    partnum = (int)in_args.in_partnum;
    blkcnt = (nasd_blkcnt_t)in_args.in_blkcnt;
    rc = nasd_od_create_partition(partnum, blkcnt,
                                in_args.in_min_protection,
                                in_args.in_partition_key,
                                in_args.in_red_key,
                                in_args.in_black_key,
                                &nid);

    out_res.nasd_status = rc;
  
#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
   
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("part_creat caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  nasd_p_part_creat_dr_marshall_res(&out_res, out_res_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  nasd_security_digest_res((nasd_args_t*)&in_args, op_key, (nasd_byte_t*)out_res_otw, out_res_len, NULL, 0, NULL, (nasd_byte_t *)&(((nasd_p_part_creat_dr_res_t*)out_res_otw)->out_digest));
  
 SETUP_REPLY_IV(icv);


  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_part_creat_res(out_res_otw, trace_seq);
  DORETURN(part_creat);

}

/*
 * Create object
 */
void
nasd_sp_create_otw_dr(
  handle_t                      client_handle, 
  nasd_p_create_dr_args_otw_t   in_args_otw,
  nasd_p_create_dr_res_otw_t    out_res_otw,
  error_status_t               *status)
{
  nasd_drive_opholder_t *cr_opholder;
  nasd_od_rpc_basic_t out_res;
  nasd_status_t rc;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(create);
  nasd_dt_create_args(in_args_otw, &trace_seq);

  NASD_DRIVE_GET_OPHOLDER(cr_opholder);
  if (cr_opholder == NULL)
    goto done_create;

  TRY {
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_create_dr_unmarshall_args(in_args_otw, &cr_opholder->opholder_create.in_args);
    SETUP_IV(nasd_p_create_dr_args_t, in_args_otw, cr_opholder->opholder_create.icv);

    if (!DISK.initialized) {
      out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(cr_opholder->opholder_create.in_args.in_security_param.partnum)) {
      out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(cr_opholder->opholder_create.in_args.in_security_param.partnum))){
      out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status =
          nasd_security_get_request_key((nasd_args_t*)&cr_opholder->opholder_create.in_args,
            (nasd_byte_t*)in_args_otw, 
            sizeof(nasd_p_create_dr_args_otw_t),
            cr_opholder->opholder_create.op_key);
      if (cr_opholder->opholder_create.out_res.nasd_status == NASD_SUCCESS) {
      
  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    nasd_p_create_dr_unmarshall_args(in_args_otw, &cr_opholder->opholder_create.in_args);

#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
    
    
    if  ((cr_opholder->opholder_create.verification = nasd_security_verify_args((nasd_args_t*)&cr_opholder->opholder_create.in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        sizeof(nasd_p_create_dr_args_otw_t),
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_create_dr_args_t*)in_args_otw)->in_digest),
                                                        cr_opholder->opholder_create.op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status = cr_opholder->opholder_create.verification; 
  } else  if  ( (cr_opholder->opholder_create.in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (cr_opholder->opholder_create.in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {
    /* No create capabilities */
    out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status=NASD_BAD_KEYTYPE;
  } else if ((cr_opholder->opholder_create.in_args.in_security_param.actual_protection & PART(cr_opholder->opholder_create.in_args.in_partnum).min_protection)  != PART(cr_opholder->opholder_create.in_args.in_partnum).min_protection) {
    out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status=NASD_PART_PROTECTION;
  }
  
  /* If we have no errors - do real work */     
  if (cr_opholder->opholder_create.out_res.nasd_status==NASD_SUCCESS) {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    
    rc = nasd_obj_create(
      (int)cr_opholder->opholder_create.in_args.in_partnum,
      &cr_opholder->opholder_create.in_args.in_attribute,
      cr_opholder->opholder_create.in_args.in_fieldmask,
      &cr_opholder->opholder_create.out_res.out_identifier,
      &cr_opholder->opholder_create.out_res.out_attribute, 0);
    cr_opholder->opholder_create.out_res.nasd_status = rc;
    out_res.nasd_status = rc;
#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = cr_opholder->opholder_create.out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("create caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

done_create:
  if (cr_opholder) {
    nasd_p_create_dr_marshall_res(&cr_opholder->opholder_create.out_res,
      out_res_otw);
  }
  else {
    nasd_p_create_dr_marshall_res(
      &nasd_drive_fail_opholder.opholder_create.out_res,
      out_res_otw);
  }

#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&cr_opholder->opholder_create.in_args,cr_opholder->opholder_create.op_key,(nasd_byte_t*)out_res_otw,sizeof(nasd_p_create_dr_res_otw_t),NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_create_dr_res_t*)out_res_otw)->out_digest));

     SETUP_REPLY_IV(cr_opholder->opholder_create.icv);


  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  if (cr_opholder) {
    NASD_DRIVE_FREE_OPHOLDER(cr_opholder);
  }

  nasd_dt_create_res(out_res_otw, trace_seq);
  DORETURN(create);
}

/*
 * Get object attributes
 */
void
nasd_sp_getattr_otw_dr(
  handle_t                       client_handle, 
  nasd_p_getattr_dr_args_otw_t   in_args_otw,
  nasd_p_getattr_dr_res_otw_t    out_res_otw,
  error_status_t                *status)
{
  nasd_p_getattr_dr_args_t in_args;
  nasd_p_getattr_dr_res_t out_res;
  nasd_status_t rc;
  nasd_args_len_t in_args_len = sizeof(nasd_p_getattr_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_getattr_dr_res_otw_t);
  nasd_status_t verification;
  nasd_key_t op_key;
  nasd_iv_t icv;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(getattr);
  nasd_dt_getattr_args(in_args_otw, &trace_seq);
  TRY {
  
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_getattr_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_getattr_dr_args_t, in_args_otw, icv); 

    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if  (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
        out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                         
                                                        (nasd_byte_t*)in_args_otw, 
                                                          in_args_len,   
                                                          op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {
         
  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    nasd_p_getattr_dr_unmarshall_args(in_args_otw, &in_args);


#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
  
  if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        in_args_len,
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_getattr_dr_args_t*)in_args_otw)->in_digest),
                                                        op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {


   /* Checks for a capability only */
  
    if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_GETATTR)!=
        NASD_ACCESS_RIGHTS_GETATTR) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    } else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
      out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    }
  } 
  if ((out_res.nasd_status == NASD_SUCCESS) && ((in_args.in_security_param.actual_protection& PART(in_args.in_partnum).min_protection)!= PART(in_args.in_partnum).min_protection)) {
    out_res.nasd_status=NASD_PART_PROTECTION;
  }
  
  /* If we have no errors - do real work */     
  if (out_res.nasd_status==NASD_SUCCESS) {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

   rc = nasd_obj_getattr(in_args.in_partnum, in_args.in_identifier,
      &out_res.out_attribute);
    out_res.nasd_status = rc;
#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("getattr caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY



 
 nasd_p_getattr_dr_marshall_res(&out_res, out_res_otw);


#if NASD_SECURE_RPCS_ENABLE > 0

  nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_getattr_dr_res_t*)out_res_otw)->out_digest));
    SETUP_REPLY_IV(icv);

  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_getattr_res(out_res_otw, trace_seq);
  DORETURN(getattr);
}

/*
 * Read single byterange in object - internal
 * function (does read and read2, which are only
 * different for stats-gathering).
 */
void
nasd_sp_read_simple_otw_dr_internal(
  handle_t                       client_handle, 
  nasd_p_smpl_op_dr_args_otw_t   in_args_otw,
  nasd_byte_pipe_t               byte_pipe,
  nasd_p_fastread_dr_res_otw_t   out_res_otw,
  error_status_t                *status,
  int                            is_read2,
  nasd_od_rpc_basic_t           *out_res_basic_p)
{
  nasd_p_smpl_op_dr_args_t in_args;
  nasd_p_fastread_dr_res_t out_res;
  nasd_len_t out_len, len, l;
  nasd_status_t rc, rc2;
  nasd_ebuff_t *my_ebuff;
  nasd_byte_t  *sendp;
  nasd_args_len_t in_args_len = sizeof(nasd_p_smpl_op_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_fastread_dr_res_otw_t);
  nasd_status_t verification;
  nasd_security_context_t security_context;
  nasd_key_t op_key;
  nasd_iv_t icv;
  nasd_procpipe_t procpipe;
  int is_special; /* Partinfo or drive info */


  TRY {
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_smpl_op_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_smpl_op_dr_args_t, in_args_otw, icv);

/* XXX THIS IS WRONG
 * This should use nasd_od_decompose_control() !!!
 */
    is_special=( (in_args.in_identifier ==  (nasd_identifier_t)NASD_CTRL_PART_INFO) ||  (in_args.in_identifier ==  (nasd_identifier_t)NASD_CTRL_DRIVE_INFO));

      if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
      printf("A");
    } else if ((NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))
) && (!is_special )) {
      printf("B");
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      if (is_special ) {
        out_res.nasd_status=NASD_SUCCESS;
      } else {
        out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      }

      if (out_res.nasd_status == NASD_SUCCESS) {
        if (!is_special ) 

      
  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

        nasd_p_smpl_op_dr_unmarshall_args(in_args_otw, &in_args);

#if NASD_SECURE_RPCS_ENABLE > 0
        if (!is_special ) {
        /* Assume everything is good until proven otherwise */
    if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                   (nasd_byte_t*)in_args_otw,
                                                   in_args_len,
                                                   NULL,
                                                   0,
                                                   (nasd_byte_t * )&(((nasd_p_smpl_op_dr_args_t*)in_args_otw)->in_digest),
                                                   op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {

    /* Checks for a capability only */
    
    if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_READ)!=
        NASD_ACCESS_RIGHTS_READ) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    } else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
      out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    } else if (in_args.in_capability_args.region_start > in_args.in_offset) {
      out_res.nasd_status=NASD_REQUEST_OUTSIDE_RANGE;
    } else if (in_args.in_capability_args.region_end < (in_args.in_offset+in_args.in_len)) {
      out_res.nasd_status=NASD_REQUEST_OUTSIDE_RANGE;
    } else if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection) != PART(in_args.in_partnum).min_protection) {
      out_res.nasd_status=NASD_PART_PROTECTION;
    }
  }
        }
   if (out_res.nasd_status==NASD_SUCCESS) {
    
     bzero(&security_context,sizeof(nasd_security_context_t));
    SETUP_REPLY_IV(icv); 
     if (is_special)
       security_context.started=NASD_TRUE;
     else 
       nasd_security_reply_init_dr((nasd_args_t *)&in_args, op_key, icv, &security_context,
                                   in_args.in_offset);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */ 
    out_len = 0;
    procpipe.state = &byte_pipe;
    procpipe.push = nasd_od_dce_push;
    procpipe.pull = NULL;
    rc = nasd_obj_read_simple(in_args.in_partnum, in_args.in_identifier,
                              in_args.in_offset, in_args.in_len, NULL,
                              is_read2, &procpipe, &out_len,
                              &security_context);
    out_res.nasd_status = rc;
    out_res.out_datalen = out_len;
    NASD_ASSERT(in_args.in_len >= out_len);
    if (rc == NASD_RPC_TRAP) {
      /*
       * This most likely means we failed shipping data on the pipe inside
       * nasd_obj_read_simple(). Trying to fix things up for the client
       * at this point will most likely crash us. Screw the client.
       */
      goto hose_the_client_not_us;
    }
    len = in_args.in_len - out_len;
    /*
     * Must send correct number of bytes to avoid angering DCE gods. Grrn.
     */
    sendp=(nasd_byte_t *)nasd_odc_zeroblk;
#if NASD_SECURE_RPCS_ENABLE > 0
    if ( (len) && (security_context.protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      NASD_FREELIST_GET( nasd_ebuff_freelist, my_ebuff, payload.next,
                       (nasd_ebuff_t *));
      sendp=my_ebuff->payload.data;
    } else
      my_ebuff=NULL;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

    while(len) {
      l = NASD_MIN(len, NASD_OD_BASIC_BLOCKSIZE);
#if NASD_SECURE_RPCS_ENABLE > 0
      if ((security_context.protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      bzero(my_ebuff->payload.data, l);      
   }
      nasd_security_context_update_send_dr(&security_context, sendp , l, NULL, NASD_TRUE);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
      rc2 = nasd_od_dce_push(&byte_pipe, sendp, l);
      if (rc2) {
        out_res.nasd_status = NASD_RPC_TRAP;
        break;
      }
      len -= l;
    }
    /* terminate DCE pipe */
    if (out_res.nasd_status != NASD_RPC_TRAP) {
      nasd_od_dce_push(&byte_pipe, (nasd_byte_t *)nasd_odc_zeroblk, 0);
    }
#if NASD_SECURE_RPCS_ENABLE > 0
    if ( my_ebuff != NULL) 
      NASD_FREELIST_FREE( nasd_ebuff_freelist, my_ebuff, payload.next);
   }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("read_simple caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

hose_the_client_not_us:

  nasd_p_fastread_dr_marshall_res(&out_res, out_res_otw);
  out_res_basic_p->nasd_status = out_res.nasd_status;

#if NASD_SECURE_RPCS_ENABLE > 0
  if (!is_special ) {
/* If we do not succeed, we don't care about the partial digest */
  if (out_res.nasd_status==NASD_SUCCESS) {
    nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,&security_context,(nasd_byte_t *)&(((nasd_p_fastread_dr_res_t*)out_res_otw)->out_digest));
    


    
  } else {
    nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_fastread_dr_res_t*)out_res_otw)->out_digest));
    


  }
  }
  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

/*
 * Read single byterange in object
 */
void
nasd_sp_read_simple_otw_dr(
  handle_t                       client_handle, 
  nasd_p_smpl_op_dr_args_otw_t   in_args_otw,
  nasd_byte_pipe_t               byte_pipe,
  nasd_p_fastread_dr_res_otw_t   out_res_otw,
  error_status_t                *status)
{
  nasd_od_rpc_basic_t out_res;
  nasd_trace_seq_t trace_seq;
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(read_simple);
  nasd_dt_read_simple_args(in_args_otw, &trace_seq);
  nasd_sp_read_simple_otw_dr_internal(client_handle,
    in_args_otw, byte_pipe, out_res_otw, status, 0, &out_res);
  nasd_dt_read_simple_res(out_res_otw, trace_seq);
  DORETURN(read_simple);
}

/*
 * Read single byterange in object
 */
void
nasd_sp_read2_simple_otw_dr(
  handle_t                       client_handle, 
  nasd_p_smpl_op_dr_args_otw_t   in_args_otw,
  nasd_byte_pipe_t               byte_pipe,
  nasd_p_fastread_dr_res_otw_t   out_res_otw,
  error_status_t                *status)
{
  nasd_od_rpc_basic_t out_res;
  nasd_trace_seq_t trace_seq;
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(read2_simple);
  nasd_dt_read2_simple_args(in_args_otw, &trace_seq);
  nasd_sp_read_simple_otw_dr_internal(client_handle,
    in_args_otw, byte_pipe, out_res_otw, status, 1,
    &out_res);
  nasd_dt_read2_simple_res(out_res_otw, trace_seq);
  DORETURN(read2_simple);
}

/*
 * Throttled read single byterange in object
 */
void
nasd_sp_tread_simple_otw_dr(
  handle_t                        client_handle, 
  nasd_p_thrtl_op_dr_args_otw_t   in_args_otw,
  nasd_byte_pipe_t                byte_pipe,
  nasd_p_fastread_dr_res_otw_t    out_res_otw,
  error_status_t                 *status)
{
  nasd_p_thrtl_op_dr_args_t in_args;
  nasd_p_fastread_dr_res_t out_res;
  nasd_len_t out_len, len, l;
  nasd_status_t rc, rc2;
  nasd_byte_t *my_ebuf, *sendp;
  nasd_args_len_t in_args_len = sizeof(nasd_p_thrtl_op_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_fastread_dr_res_otw_t);
  nasd_status_t verification;
  nasd_security_context_t security_context;
  nasd_iv_t icv;
  nasd_key_t op_key;
  nasd_procpipe_t procpipe;
  nasd_trace_seq_t trace_seq;
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(tread_simple);
  nasd_dt_tread_simple_args(in_args_otw, &trace_seq);

  TRY {
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_thrtl_op_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_thrtl_op_dr_args_t, in_args_otw, icv);

    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {
      
  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

    nasd_p_thrtl_op_dr_unmarshall_args(in_args_otw, &in_args);



#if NASD_SECURE_RPCS_ENABLE > 0
 if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        in_args_len,
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_thrtl_op_dr_args_t*)in_args_otw)->in_digest),
                                                        op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {
   /* Checks for a capability only */
   
   if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_READ)!=
               NASD_ACCESS_RIGHTS_READ) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    } else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
      out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    } else if (in_args.in_capability_args.region_start > in_args.in_offset) {
      out_res.nasd_status=NASD_REQUEST_OUTSIDE_RANGE;
    } else if (in_args.in_capability_args.region_end < (in_args.in_offset+in_args.in_len)) {
      out_res.nasd_status=NASD_REQUEST_OUTSIDE_RANGE;
    } else if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection) != PART(in_args.in_partnum).min_protection) {
      out_res.nasd_status=NASD_PART_PROTECTION;
    }
  }
      
      
      if (out_res.nasd_status==NASD_SUCCESS) {
    
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

   SETUP_REPLY_IV(icv); 
    bzero((char *)&security_context, sizeof(nasd_security_context_t));
    nasd_security_reply_init_dr((nasd_args_t *)&in_args, op_key, icv, 
                             &security_context, in_args.in_offset);
    out_len = 0;
    procpipe.state = &byte_pipe;
    procpipe.push = nasd_od_dce_push;
    procpipe.pull = NULL;
    rc = nasd_obj_read_simple(in_args.in_partnum, in_args.in_identifier,
      in_args.in_offset, in_args.in_len, &in_args.in_bms_targ, 0, &procpipe,
      &out_len, &security_context);
    out_res.nasd_status = rc;
    out_res.out_datalen = out_len;
    NASD_ASSERT(in_args.in_len >= out_len);
    len = in_args.in_len - out_len;
    /*
     * Must send correct number of bytes to avoid angering DCE gods. Grrn.
     */
 sendp=(nasd_byte_t *)nasd_odc_zeroblk;
#if NASD_SECURE_RPCS_ENABLE > 0
    if ( (len) && (security_context.protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      NASD_Malloc(my_ebuf, NASD_OD_BASIC_BLOCKSIZE, (nasd_byte_t *));
      sendp=my_ebuf;
    } else
      my_ebuf=NULL;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

    while(len) {
      l = NASD_MIN(len, NASD_OD_BASIC_BLOCKSIZE);
#if NASD_SECURE_RPCS_ENABLE > 0
      if ((security_context.protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      bzero(my_ebuf, l);      
   }
      nasd_security_context_update_send_dr(&security_context, sendp , l, NULL, NASD_TRUE);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
      rc2 = nasd_od_dce_push(&byte_pipe, sendp, l);
      if (rc2) {
        out_res.nasd_status = NASD_RPC_TRAP;
        break;
      }

      len -= l;
    }
    /* terminate DCE pipe */
    if (out_res.nasd_status != NASD_RPC_TRAP) {
      nasd_od_dce_push(&byte_pipe, (nasd_byte_t *)nasd_odc_zeroblk, 0);
    }
#if NASD_SECURE_RPCS_ENABLE > 0
    if ( my_ebuf != NULL) 
      NASD_Free(my_ebuf, NASD_OD_BASIC_BLOCKSIZE);
      }
    }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("tread_simple caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  nasd_p_fastread_dr_marshall_res(&out_res, out_res_otw);
#if NASD_SECURE_RPCS_ENABLE > 0
  /* If we do not succeed, we don't care about the partial digest */
  if (out_res.nasd_status==NASD_SUCCESS) {
    nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,&security_context,(nasd_byte_t *)&(((nasd_p_fastread_dr_res_t*)out_res_otw)->out_digest));
   SETUP_REPLY_IV(icv);


  }  else {
    nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_fastread_dr_res_t*)out_res_otw)->out_digest)); 
   SETUP_REPLY_IV(icv);
  }
  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_tread_simple_res(out_res_otw, trace_seq);
  DORETURN(tread_simple);
}

/*
 * Write single byterange in object
 */
void
nasd_sp_write_simple_otw_dr(
  handle_t                        client_handle, 
  nasd_p_smpl_op_dr_args_otw_t    in_args_otw,
  nasd_byte_pipe_t                byte_pipe,
  nasd_p_fastwrite_dr_res_otw_t   out_res_otw,
  error_status_t                 *status)
{
  nasd_p_smpl_op_dr_args_t in_args;
  nasd_p_fastwrite_dr_res_t out_res;
  nasd_len_t out_len, len, l;
  idl_ulong_int count;
  nasd_status_t rc, rc2;
  nasd_len_t tcount;
  nasd_security_context_t security_context;
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_args_len_t in_args_len = sizeof(nasd_p_smpl_op_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_fastwrite_dr_res_otw_t);
  nasd_status_t verification;
  nasd_iv_t icv;
  nasd_key_t op_key;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_procpipe_t procpipe;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(write_simple);
  nasd_dt_write_simple_args(in_args_otw, &trace_seq);

  TRY {

#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_smpl_op_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_smpl_op_dr_args_t, in_args_otw, icv);

    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      
      if (out_res.nasd_status == NASD_SUCCESS) {


#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

    nasd_p_smpl_op_dr_unmarshall_args(in_args_otw, &in_args);



#if NASD_SECURE_RPCS_ENABLE > 0
    if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                   (nasd_byte_t*)in_args_otw,
                                                   in_args_len,
                                                   NULL,
                                                   0,
                                                   (nasd_byte_t * )&(((nasd_p_smpl_op_dr_args_t*)in_args_otw)->in_digest),
                                                   op_key,
                                                   &security_context)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {
   /* Checks for a capability only */
  
   if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_WRITE)!=
               NASD_ACCESS_RIGHTS_WRITE) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    } else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
      out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    }
  }
  if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection)!= PART(in_args.in_partnum).min_protection) {
    out_res.nasd_status=NASD_PART_PROTECTION;
  }
  /* If we have no errors - do real work */     
  if (out_res.nasd_status==NASD_SUCCESS) {
    security_context.curr_off=in_args.in_offset;
    security_context.start_off=in_args.in_offset;
#if (NASD_USE_HMAC_ONLY == 0)
   SHA1_Init(NASD_FALSE, &security_context.SHA1_ctx);
#endif /* (NASD_USE_HMAC_ONLY == 0) */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

    out_len = 0;
    procpipe.state = &byte_pipe;
    procpipe.push = NULL;
    procpipe.pull = nasd_od_dce_pull;
    rc = nasd_obj_write_simple(in_args.in_partnum, in_args.in_identifier,
      in_args.in_offset, in_args.in_len, &procpipe,
      &security_context, &out_len);
    out_res.nasd_status = rc;
    out_res.out_datalen = out_len;
    NASD_ASSERT(in_args.in_len >= out_len);
    /*
     * Must recv correct number of bytes to avoid angering DCE gods. Grrn.
     */
    len = in_args.in_len - out_len;
    while(len) {
      l = NASD_MIN(len, NASD_OD_BASIC_BLOCKSIZE);
      nasd_od_dce_pull(&byte_pipe, (nasd_byte_t *)nasd_odc_bitbucket, l, &tcount);

      len -= tcount;      
#if NASD_SECURE_RPCS_ENABLE > 0
      rc=nasd_security_context_update_receive_dr(&security_context,
                                                 (nasd_byte_t *)nasd_odc_bitbucket,
                                               tcount,
                                               &procpipe,
                                                 len,
                                                 NULL,
                                                 NULL);

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

      if (tcount == 0)
        break;
    }
    /* terminate DCE pipe */
    rc2 = nasd_od_dce_pull(&byte_pipe, (nasd_byte_t *)nasd_odc_bitbucket, 0, &tcount);
    if (rc2) {
      if (rc == NASD_SUCCESS)
        rc = NASD_RPC_TRAP;
    }
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  else {
    /*
     * We failed some kind of check at the beginning of the RPC.
     * Be sure to terminate the pipe.
     */
    nasd_od_dce_drainpull(&byte_pipe);
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("write_simple caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  nasd_p_fastwrite_dr_marshall_res(&out_res, out_res_otw);
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len, NULL, 0, NULL, (nasd_byte_t *)&(((nasd_p_fastwrite_dr_res_t*)out_res_otw)->out_digest));
   SETUP_REPLY_IV(icv);

  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_write_simple_res(out_res_otw, trace_seq);
  DORETURN(write_simple);
}

/*
 * Set object attributes
 */
void
nasd_sp_setattr_otw_dr(
  handle_t                       client_handle, 
  nasd_p_setattr_dr_args_otw_t   in_args_otw,
  nasd_p_setattr_dr_res_otw_t      out_res_otw,
  error_status_t                *status)
{
  nasd_p_setattr_dr_args_t in_args;
  nasd_p_setattr_dr_res_t out_res;
  nasd_status_t rc;
  nasd_args_len_t in_args_len = sizeof(nasd_p_setattr_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_setattr_dr_res_otw_t);
  nasd_status_t verification;
  nasd_key_t op_key;
  nasd_iv_t icv;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL  

  DOBEGIN(setattr);
  nasd_dt_setattr_args(in_args_otw, &trace_seq);
  TRY {
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_setattr_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_setattr_dr_args_t, in_args_otw, icv);
 
    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

        nasd_p_setattr_dr_unmarshall_args(in_args_otw, &in_args);

#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
        if ((verification =
             nasd_security_verify_args((nasd_args_t*)&in_args,
                                       (nasd_byte_t*)in_args_otw,
                                       in_args_len,
                                       NULL,
                                       0,
                                       (nasd_byte_t * )&(((nasd_p_setattr_dr_args_t*)in_args_otw)->in_digest),
                                       op_key,
                                       NULL)) != NASD_SUCCESS) {
          out_res.nasd_status = verification; 
        } else if ((in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
                   (in_args.in_security_param.type == NASD_RED_CAPABILITY)) {

          /* Checks for a capability only */
          if ((in_args.in_capability_args.rights &
               NASD_ACCESS_RIGHTS_SETATTR) != NASD_ACCESS_RIGHTS_SETATTR) {
            out_res.nasd_status=NASD_INSUFFICIENT_PERM;
          } else if (in_args.in_capability_args.partnum !=
                     in_args.in_partnum) {
            out_res.nasd_status=NASD_WRONG_PARTITION;
          } else if (in_args.in_capability_args.min_protection &
                     in_args.in_security_param.actual_protection !=
                     in_args.in_capability_args.min_protection) {
            out_res.nasd_status=NASD_CAP_PROTECTION;
          } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
            out_res.nasd_status=NASD_WRONG_NI;
          } else if (in_args.in_capability_args.drive_name != my_name) {
            out_res.nasd_status=NASD_WRONG_DRIVE;
          }
        }
        if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection)!= PART(in_args.in_partnum).min_protection) {
          out_res.nasd_status=NASD_PART_PROTECTION;
        }
        /* If we have no errors - do real work */     
        if (out_res.nasd_status==NASD_SUCCESS) {

          /* If we are changeing the AV, we want to wipe out cached
             capabilities */
          if ((in_args.in_fieldmask & NASD_ATTR_AV) == NASD_ATTR_AV) 
            nasd_security_cache_invalidate(in_args.in_partnum,in_args.in_identifier);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

          rc = nasd_obj_setattr(in_args.in_partnum, in_args.in_identifier,
                                &in_args.in_attribute, in_args.in_fieldmask,
                                &out_res.out_attribute);
          out_res.nasd_status = rc;

#if NASD_SECURE_RPCS_ENABLE > 0
        }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  } CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("setattr caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
  nasd_p_setattr_dr_marshall_res(&out_res, out_res_otw);

#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&in_args,
                           op_key,
                           (nasd_byte_t*)out_res_otw,
                           out_res_len,
                           NULL,
                           0,
                           NULL,
                           (nasd_byte_t *)&(((nasd_p_setattr_dr_res_t*)out_res_otw)->out_digest));
 
  SETUP_REPLY_IV(icv);


  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_setattr_res(out_res_otw, trace_seq);
  DORETURN(setattr);
}

void
nasd_sp_flush_obj_otw_dr(
  handle_t                         client_handle, 
  nasd_p_flush_obj_dr_args_otw_t   in_args_otw,
  nasd_p_flush_obj_dr_res_otw_t    out_res_otw,
  error_status_t                  *status)
{
  nasd_p_flush_obj_dr_args_t in_args;
  nasd_p_flush_obj_dr_res_t out_res;
  nasd_status_t rc;
  nasd_args_len_t in_args_len = sizeof(nasd_p_flush_obj_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_flush_obj_dr_res_otw_t);
  nasd_status_t verification;
  nasd_key_t op_key;
  nasd_iv_t icv;
  nasd_trace_seq_t trace_seq;
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(flush_obj);
  nasd_dt_flush_args(in_args_otw, &trace_seq);
  TRY {

#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_flush_obj_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_flush_obj_dr_args_t, in_args_otw, icv);
  
    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {
      
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_p_flush_obj_dr_unmarshall_args(in_args_otw, &in_args);

#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
 if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        in_args_len,
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_flush_obj_dr_args_t*)in_args_otw)->in_digest),
                                                        op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {


   /* Checks for a capability only */
  
   if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_FLUSH)!=
               NASD_ACCESS_RIGHTS_FLUSH) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    }  else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
      out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    }
  }
  if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection ) != PART(in_args.in_partnum).min_protection) {
    out_res.nasd_status=NASD_PART_PROTECTION;
  }
  
  /* If we have no errors - do real work */     
  if (out_res.nasd_status==NASD_SUCCESS) {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  rc = nasd_obj_flush(in_args.in_partnum, in_args.in_identifier);
  out_res.nasd_status = rc;

#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("flush caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
  nasd_p_flush_obj_dr_marshall_res(&out_res, out_res_otw);


#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_flush_obj_dr_res_t*)out_res_otw)->out_digest));
 
   SETUP_REPLY_IV(icv);


  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_flush_res(out_res_otw, trace_seq);
  DORETURN(flush_obj);
}


void
nasd_sp_remove_otw_dr(
  handle_t                      client_handle, 
  nasd_p_remove_dr_args_otw_t   in_args_otw,
  nasd_p_remove_dr_res_otw_t    out_res_otw,
  error_status_t               *status)
{
  nasd_p_remove_dr_args_t in_args;
  nasd_p_remove_dr_res_t out_res;
  nasd_status_t rc;
  nasd_args_len_t in_args_len = sizeof(nasd_p_remove_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_remove_dr_res_otw_t);
  nasd_status_t verification;
  nasd_key_t op_key;
  nasd_iv_t icv;
  nasd_trace_seq_t trace_seq;
  static my_local_count=0;
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(remove);
 
  nasd_dt_remove_args(in_args_otw, &trace_seq);

  TRY{
#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_remove_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_remove_dr_args_t, in_args_otw, icv);
 
   if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {

        
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_p_remove_dr_unmarshall_args(in_args_otw, &in_args);


#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
 
  if ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        in_args_len,
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_remove_dr_args_t*)in_args_otw)->in_digest),
                                                        op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {


   /* Checks for a capability only */
  
   if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_REMOVE)!=
               NASD_ACCESS_RIGHTS_REMOVE) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    }  else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
       out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    }
  }
  if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection )!= PART(in_args.in_partnum).min_protection) {
    out_res.nasd_status=NASD_PART_PROTECTION;
  }

  /* If we have no errors - do real work */     
  if (out_res.nasd_status==NASD_SUCCESS) {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  rc = nasd_obj_remove(in_args.in_partnum, in_args.in_identifier);
  out_res.nasd_status = rc;

#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("obj_remove caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
  nasd_p_remove_dr_marshall_res(&out_res, out_res_otw);




#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_remove_dr_res_t*)out_res_otw)->out_digest));

   SETUP_REPLY_IV(icv);

  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_remove_res(out_res_otw, trace_seq);
  DORETURN(remove);
}



void
nasd_sp_initialize_otw_dr(
  handle_t                          client_handle, 
  nasd_p_initialize_dr_args_otw_t   in_args_otw,
  nasd_p_initialize_dr_res_otw_t    out_res_otw,
  error_status_t                   *status)
{
  nasd_p_initialize_dr_args_t in_args;
  nasd_p_initialize_dr_res_t out_res;
  nasd_status_t rc;

  NASD_OD_RPC_OP_STATS_DECL
    
  DOBEGIN(initialize);
  TRY {
  nasd_p_initialize_dr_unmarshall_args(in_args_otw, &in_args);
  
  rc = nasd_security_initialize_drive(in_args.in_master_key,in_args.in_drive_key);

  out_res.nasd_status = rc;
 }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("initialize caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
  nasd_p_initialize_dr_marshall_res(&out_res, out_res_otw);
  DORETURN(initialize);
}



void
nasd_sp_change_key_otw_dr(
  handle_t                          client_handle, 
  nasd_p_change_key_dr_args_otw_t   in_args_otw,
  nasd_p_change_key_dr_res_otw_t    out_res_otw,
  error_status_t                   *status)
{
  nasd_p_change_key_dr_args_t in_args;
  nasd_p_change_key_dr_res_t out_res;
  nasd_status_t rc;
  nasd_args_len_t in_args_len = sizeof(nasd_p_change_key_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_change_key_dr_res_otw_t);
  nasd_status_t verification;
  nasd_key_t op_key;
  nasd_iv_t icv;
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(change_key);
  TRY {

#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_change_key_dr_unmarshall_args(in_args_otw, &in_args);   
    SETUP_IV(nasd_p_change_key_dr_args_t, in_args_otw, icv);
     
   if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  nasd_p_change_key_dr_unmarshall_args(in_args_otw, &in_args);

#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
  if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        in_args_len,
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_change_key_dr_args_t*)in_args_otw)->in_digest),
                                                        op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  }

 /* If we have no errors - do real work */     
  if (out_res.nasd_status==NASD_SUCCESS) {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  rc = nasd_security_change_key_drive(in_args.in_type,in_args.in_new_key,in_args.in_partnum);
  out_res.nasd_status = rc;

#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
 }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("change_key caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
  nasd_p_change_key_dr_marshall_res(&out_res, out_res_otw);


#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_change_key_dr_res_t*)out_res_otw)->out_digest));
  
  SETUP_REPLY_IV(icv);

  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  DORETURN(change_key);
}


void
nasd_sp_eject_obj_otw_dr(
  handle_t                         client_handle, 
  nasd_p_eject_obj_dr_args_otw_t   in_args_otw,
  nasd_p_eject_obj_dr_res_otw_t    out_res_otw,
  error_status_t                  *status)
{
  nasd_p_eject_obj_dr_args_t in_args;
  nasd_p_eject_obj_dr_res_t out_res;
  nasd_status_t rc;
  nasd_args_len_t in_args_len = sizeof(nasd_p_eject_obj_dr_args_otw_t);
  nasd_args_len_t out_res_len = sizeof(nasd_p_eject_obj_dr_res_otw_t);
  nasd_status_t verification;
  nasd_key_t op_key;
  nasd_iv_t icv;
  nasd_trace_seq_t trace_seq; 
  NASD_OD_RPC_OP_STATS_DECL

  DOBEGIN(eject_obj);
  nasd_dt_eject_args(in_args_otw, &trace_seq);
  TRY {

#if NASD_SECURE_RPCS_ENABLE > 0
    /* All we really want unmarshalled is the capability args and
       security param */
    nasd_p_eject_obj_dr_unmarshall_args(in_args_otw, &in_args);
    SETUP_IV(nasd_p_eject_obj_dr_args_t, in_args_otw, icv);
  
    if (!DISK.initialized) {
      out_res.nasd_status=NASD_DRIVE_UNINITIALIZED;
    }  else if (!NASD_OD_PARTNUM_VALID(in_args.in_security_param.partnum)) {
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else if (NASD_OD_INVALID_PART(&PART(in_args.in_security_param.partnum))){
      out_res.nasd_status=NASD_BAD_PARTITION;
    } else {
      out_res.nasd_status=nasd_security_get_request_key((nasd_args_t*)&in_args,                                              (nasd_byte_t*)in_args_otw, 
                                              in_args_len,   
                                              op_key);
      if (out_res.nasd_status == NASD_SUCCESS) {

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_p_eject_obj_dr_unmarshall_args(in_args_otw, &in_args);

#if NASD_SECURE_RPCS_ENABLE > 0
  /* Assume everything is good until proven otherwise */
 if  ((verification = nasd_security_verify_args((nasd_args_t*)&in_args,
                                                        (nasd_byte_t*)in_args_otw,
                                                        in_args_len,
                                                        NULL,
                                                        0,
                                                        (nasd_byte_t * )&(((nasd_p_eject_obj_dr_args_t*)in_args_otw)->in_digest),
                                                        op_key, NULL)) != NASD_SUCCESS) {
    out_res.nasd_status = verification; 
  } else  if  ( (in_args.in_security_param.type == NASD_BLACK_CAPABILITY) ||
               (in_args.in_security_param.type == NASD_RED_CAPABILITY) ) {


   /* Checks for a capability only */
  
   if ((in_args.in_capability_args.rights & NASD_ACCESS_RIGHTS_EJECT)!=
               NASD_ACCESS_RIGHTS_EJECT) {
      out_res.nasd_status=NASD_INSUFFICIENT_PERM;
    } else if (in_args.in_capability_args.partnum != in_args.in_partnum) {
      out_res.nasd_status=NASD_WRONG_PARTITION;
    }  else if (in_args.in_capability_args.min_protection & in_args.in_security_param.actual_protection != in_args.in_capability_args.min_protection) {
      out_res.nasd_status=NASD_CAP_PROTECTION;
    } else if (in_args.in_capability_args.ni != in_args.in_identifier) {
      out_res.nasd_status=NASD_WRONG_NI;
    } else if (in_args.in_capability_args.drive_name != my_name) {
      out_res.nasd_status=NASD_WRONG_DRIVE;
    }
  }
  if ((in_args.in_security_param.actual_protection & PART(in_args.in_partnum).min_protection ) != PART(in_args.in_partnum).min_protection) {
    out_res.nasd_status=NASD_PART_PROTECTION;
  }
  
  /* If we have no errors - do real work */     
  if (out_res.nasd_status==NASD_SUCCESS) {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  rc = nasd_obj_eject(in_args.in_partnum, in_args.in_identifier);
  out_res.nasd_status = rc;

#if NASD_SECURE_RPCS_ENABLE > 0
  }
      }
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  }
  CATCH_ALL {
    NASD_OD_DCE_EXCEPTION_CAUGHT();
    out_res.nasd_status = NASD_REMOTE_TRAP;
#if NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("eject caught DCE exception\n");
#endif /* NASD_OD_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY
  nasd_p_eject_obj_dr_marshall_res(&out_res, out_res_otw);


#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_digest_res((nasd_args_t*)&in_args,op_key,(nasd_byte_t*)out_res_otw,out_res_len,NULL,0,NULL,(nasd_byte_t *)&(((nasd_p_eject_obj_dr_res_t*)out_res_otw)->out_digest));
 
   SETUP_REPLY_IV(icv);


  *status=0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  nasd_dt_eject_res(out_res_otw, trace_seq);
  DORETURN(eject_obj);
}


/*
 * !!! Do not hand-modify! Generate this stanza with "make mepv"
 */
static nasd_pdrive_v0_1_epv_t nasd_pdrive_epv = {
  nasd_sp_null_dr
  ,nasd_sp_sync_dr
  ,nasd_sp_part_creat_otw_dr
  ,nasd_sp_create_otw_dr
  ,nasd_sp_getattr_otw_dr
  ,nasd_sp_read_simple_otw_dr
  ,nasd_sp_read2_simple_otw_dr
  ,nasd_sp_tread_simple_otw_dr
  ,nasd_sp_write_simple_otw_dr
  ,nasd_sp_setattr_otw_dr
  ,nasd_sp_flush_obj_otw_dr
  ,nasd_sp_remove_otw_dr
  ,nasd_sp_initialize_otw_dr
  ,nasd_sp_change_key_otw_dr
  ,nasd_sp_eject_obj_otw_dr
};


void
nasd_od_setup_tcp()
{
  rpc_if_rep_t *rep;

  nasd_od_protseq_str = (unsigned_char_t *)"ncacn_ip_tcp";
  rep = (rpc_if_rep_t *)nasd_pdrive_v0_1_s_ifspec;
  rep->endpoint_vector.endpoint_vector_elt[0].rpc_protseq = nasd_od_protseq_str;
}

void
nasd_od_setup_udp()
{
  rpc_if_rep_t *rep;

  nasd_od_protseq_str = (unsigned_char_t *)"ncadg_ip_udp";
  rep = (rpc_if_rep_t *)nasd_pdrive_v0_1_s_ifspec;
  rep->endpoint_vector.endpoint_vector_elt[0].rpc_protseq = nasd_od_protseq_str;
}

/*
 * We have our own shutdown routine so we can keep
 * track of whether or not this is already
 * initialized.
 */
void
nasd_od_rpc_shutdown_rpc_cnt_lock(
  void  *ignored)
{
  nasd_shutdown_mutex(&nasd_od_rpc_cnt_lock);
  nasd_od_rpc_cnt_lock_initted = 0;
}

nasd_status_t
nasd_od_rpc_init()
{
  nasd_status_t rc;

  nasd_od_rpc_qdepth = 0;

  bzero((char *)&nasd_drive_fail_opholder, sizeof(nasd_drive_fail_opholder));
  nasd_drive_fail_opholder.opholder_create.out_res.nasd_status = NASD_NO_MEM;

  if (nasd_od_threads_in_rpc) {
    /* already/still running */
    return(NASD_FAIL);
  }

  bzero((char *)&nasd_drive_opstats, sizeof(nasd_drive_opstats));
  nasd_drive_opstats.ctrl_id = NASD_CTRL_DRIVE_INFO;

  nasd_od_wait_shutdown = 0;
  nasd_od_threads_in_rpc = 0;
  nasd_od_shutdown_timeouts = 0;

  if (nasd_od_rpc_cnt_lock_initted) {
    printf("DRIVE ERROR: nasd_od_rpc_cnt_lock_initted unexpectedly "
           "has value %d\n", nasd_od_rpc_cnt_lock);
    return(NASD_FAIL);
  }

  rc = nasd_mutex_init(&nasd_od_rpc_cnt_lock);
  if (rc)
    return(rc);
  nasd_od_rpc_cnt_lock_initted = 1;
  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_od_rpc_shutdown_rpc_cnt_lock,
    NULL);
  if (rc) {
    nasd_od_rpc_shutdown_rpc_cnt_lock(NULL);
    return(rc);
  }

  rc = nasd_cond_init(&nasd_od_shutdown_cond);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_cond,
    &nasd_od_shutdown_cond);
  if (rc) {
    nasd_cond_destroy(&nasd_od_shutdown_cond);
    return(rc);
  }

  rc = nasd_mutex_init(&nasd_od_rpc_qdepth_lock);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_mutex,
    &nasd_od_rpc_qdepth_lock);
  if (rc) {
    nasd_mutex_destroy(&nasd_od_rpc_qdepth_lock);
    return(rc);
  }

#if NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0
  rc = nasd_mutex_init(&nasd_dce_pipe_lock);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_mutex,
    &nasd_dce_pipe_lock);
  if (rc) {
    nasd_mutex_destroy(&nasd_dce_pipe_lock);
    return(rc);
  }
#endif /* NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0 */

  return(NASD_SUCCESS);
}

void
nasd_od_begin_rpc(
  nasd_opstat_t  *opstat)
{
  NASD_ATOMIC_INC64(&nasd_od_threads_in_rpc);
}

void
nasd_od_end_rpc(
  nasd_opstat_t  *opstat)
{
  NASD_ATOMIC_DEC64(&nasd_od_threads_in_rpc);
  if (nasd_od_wait_shutdown && (nasd_od_threads_in_rpc == 0)) {
    NASD_BROADCAST_COND(nasd_od_shutdown_cond);
  }
}

void
nasd_od_timeout_shutdown_rpc(
  nasd_timeout_handle_t   tmh,
  void                   *arg1,
  void                   *arg2)
{
  nasd_timespec_t ts;
  char str[128];

#if (NASD_OD_RPC_DEBUG_RPC_SHUTDOWN > 0) || (NASD_DRIVE_SHUTDOWN_DEBUG > 0)
  nasd_gettime(&ts);
  printf("DRIVE: shutdown timeout at %s, nasd_od_threads_in_rpc=%"
    NASD_64u_FMT "\n", nasd_timestr_r(ts, str), nasd_od_threads_in_rpc);
#endif /* NASD_OD_RPC_DEBUG_RPC_SHUTDOWN > 0 || NASD_DRIVE_SHUTDOWN_DEBUG > 0 */
  if (nasd_od_threads_in_rpc == 0) {
    NASD_BROADCAST_COND(nasd_od_shutdown_cond);
  }
  nasd_od_shutdown_timeouts++;
}

void
nasd_od_wait_shutdown_rpc()
{
  nasd_status_t rc;

  nasd_od_wait_shutdown = 1;
  rc = NASD_FAIL;

  if (nasd_od_rpc_cnt_lock_initted == 0)
    return;

  NASD_LOCK_MUTEX(nasd_od_rpc_cnt_lock);
  while(nasd_od_threads_in_rpc) {
    if (rc) {
      rc = nasd_timeout_add(&nasd_od_shutdown_timeout_handle,
        nasd_od_timeout_shutdown_rpc, NULL, NULL,
        nasd_od_shutdown_timeout_interval,
        nasd_od_shutdown_timeout_interval,
        NASD_TIMEOUT_F_PERIODIC);
    }
    if (rc) {
      printf("DRIVE: got 0x%x (%s) adding timeout for shutdown-quiesce\n",
        rc, nasd_error_string(rc));
    }
    NASD_WAIT_COND(nasd_od_shutdown_cond,nasd_od_rpc_cnt_lock);
  }
  NASD_UNLOCK_MUTEX(nasd_od_rpc_cnt_lock);
}

void
nasd_stop_rpc()
{
  rpc_if_handle_t nasd_pdrive_ifspec;
  dce_error_string_t err_str;
  error_status_t status, ps;

  nasd_pdrive_active = 0;

  nasd_pdrive_ifspec = nasd_pdrive_v0_1_s_ifspec;

#if 0
  /*
   * This just doesn't work. Hurray DCE.
   */
  rpc_mgmt_stop_server_listening(NULL, &status);
  if (status) {
    dce_error_inq_text(status, err_str, &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("DRIVE: cannot stop server listener status %d (%s)\n",
      status, err_str);
  }
#endif

  if (nasd_pdrive_if_registered) {
    rpc_server_unregister_if(nasd_pdrive_ifspec, NULL, &status);
    if (status) {
      dce_error_inq_text(status, err_str, (int *) &ps);
      printf("DRIVE: cannot unregister server_if status %d (%s)\n",
        status, err_str);
    }
    nasd_pdrive_if_registered = 0;
  }
}

void
nasd_dump_opstat(
  nasd_opstat_t  *os,
  char           *name)
{
  unsigned long s, ns;

  s = os->op_nsecs / 1000000000UL;
  ns = os->op_nsecs % 1000000000UL;
  printf("Operation: %s\n", name);
  printf("  time       %lu:%09lu\n", s, ns);
  printf("  num_ops    %5lu\n", os->num_ops);
  printf("  in_flight  %5lu\n", os->in_flight);
  printf("  invalid    %5ld\n", os->invalid);
}

#define DUMP_OPSTAT(_opname_) \
  nasd_dump_opstat(&nasd_drive_opstats._opname_,NASD_STRING(_opname_))

void
nasd_shutdown_rpc()
{
#if NASD_OD_RPC_DEBUG_RPC_SHUTDOWN > 0
  printf("nasd_shutdown_rpc(): nasd_od_threads_in_rpc=%" NASD_64u_FMT "\n",
    nasd_od_threads_in_rpc);
#endif /* NASD_OD_RPC_DEBUG_RPC_SHUTDOWN > 0 */

  nasd_od_wait_shutdown_rpc();

#if NASD_OD_RPC_DUMP_OP_STATS > 0
  DUMP_OPSTAT(null);
  DUMP_OPSTAT(sync);
  DUMP_OPSTAT(part_creat);
  DUMP_OPSTAT(create);
  DUMP_OPSTAT(getattr);
  DUMP_OPSTAT(read_simple);
  DUMP_OPSTAT(tread_simple);
  DUMP_OPSTAT(write_simple);
  DUMP_OPSTAT(setattr);
  DUMP_OPSTAT(flush_obj);
  DUMP_OPSTAT(remove);
  DUMP_OPSTAT(initialize);
  DUMP_OPSTAT(change_key);
#endif /* NASD_OD_RPC_DUMP_OP_STATS > 0 */

#if NASD_OD_RPC_DEBUG_RPC_SHUTDOWN > 0
  printf("done nasd_shutdown_rpc()\n");
#endif /* NASD_OD_RPC_DEBUG_RPC_SHUTDOWN > 0 */
}

nasd_status_t
nasd_startup_rpc()
{
  rpc_binding_vector_p_t binding_vector;
  rpc_if_handle_t nasd_pdrive_ifspec;
  rpc_if_rep_p_t nasd_pdrive_ifrep;
  dce_error_string_t err_str;
  error_status_t status, ps;
  nasd_status_t rc;

  nasd_pdrive_active = 1;

  nasd_pdrive_ifrep = (rpc_if_rep_p_t)nasd_pdrive_v0_1_s_ifspec;
  nasd_pdrive_ifrep->manager_epv = (rpc_mgr_epv_t)&nasd_pdrive_epv;
  nasd_pdrive_ifspec = nasd_pdrive_v0_1_s_ifspec;

  if (nasd_pdrive_protseq_installed == 0) {
    rpc_server_use_protseq_if(nasd_od_protseq_str,
      rpc_c_protseq_max_reqs_default, nasd_pdrive_ifspec, &status);
    if (status != rpc_s_ok) {
      dce_error_inq_text(status, err_str, (int *) &ps);
      if (ps)
        strcpy(err_str, "unknown");
      printf("ERROR: cannot register protocol sequence status=0x%x (%s)\n",
        status, err_str);
      return(NASD_FAIL);
    }
    nasd_pdrive_protseq_installed = 1;
  }

  if (nasd_pdrive_if_registered) {
    printf("ERROR: nasd_pdrive_if_registered already has value %d"
           " in nasd_startup_rpc()\n",
           nasd_pdrive_if_registered);
    return(NASD_FAIL);
  }

  rpc_server_register_if(nasd_pdrive_ifspec, NULL, NULL, &status);
  if (status != rpc_s_ok) {
    dce_error_inq_text(status, err_str, (int *) &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("ERROR: cannot register interface status=0x%x (%s)\n",
      status, err_str);
    return(NASD_FAIL);
  }
  nasd_pdrive_if_registered = 1;

  rpc_server_inq_bindings(&binding_vector, &status);
  if (status != rpc_s_ok) {
    dce_error_inq_text(status, err_str, (int *) &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("ERROR: cannot retrieve binding information status=0x%x (%s)\n",
      status, err_str);
    return(NASD_FAIL);
  }

  return(NASD_SUCCESS);
}

/*
 * DCE pipe stuff
 */

nasd_status_t
nasd_od_dce_push(
  void        *real_pipe,
  void        *buf,
  nasd_len_t   in_len)
{
  nasd_byte_pipe_t *byte_pipe;
  nasd_status_t rc;

#define CATCHPUSHPIPE(_exc_) \
 CATCH(_exc_) { \
   printf("DCE ex (%s) performing byte_pipe.push(0x%lx, 0x%lx, %d)\n", \
     NASD_STRING(_exc_), \
     (unsigned long)byte_pipe->state, (unsigned long)buf, \
     in_len); \
   rc = NASD_RPC_TRAP; \
 }

  byte_pipe = (nasd_byte_pipe_t *)real_pipe;

#if NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0
  NASD_LOCK_MUTEX(nasd_dce_pipe_lock);
#endif /* NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0 */

  TRY {
    byte_pipe->push(byte_pipe->state, buf, in_len);
    rc = NASD_SUCCESS;
  }
  CATCHPUSHPIPE(rpc_x_ss_char_trans_open_fail)
  CATCHPUSHPIPE(rpc_x_ss_char_trans_short_file)
  CATCHPUSHPIPE(rpc_x_ss_context_damaged)
  CATCHPUSHPIPE(rpc_x_ss_context_mismatch)
  CATCHPUSHPIPE(rpc_x_ss_in_null_context)
  CATCHPUSHPIPE(rpc_x_ss_pipe_closed)
  CATCHPUSHPIPE(rpc_x_ss_pipe_comm_error)
  CATCHPUSHPIPE(rpc_x_ss_pipe_discipline_error)
  CATCHPUSHPIPE(rpc_x_ss_pipe_empty)
  CATCHPUSHPIPE(rpc_x_ss_pipe_memory)
  CATCHPUSHPIPE(rpc_x_ss_pipe_order)
  CATCHPUSHPIPE(rpc_x_ss_remote_comm_failure)
  CATCHPUSHPIPE(rpc_x_ss_remote_no_memory)
  CATCHPUSHPIPE(rpc_x_ss_bad_buffer)
  CATCHPUSHPIPE(rpc_x_ss_bad_es_action)
  CATCHPUSHPIPE(rpc_x_ss_wrong_es_version)
  CATCHPUSHPIPE(rpc_x_stub_protocol_error)
  CATCHPUSHPIPE(rpc_x_unknown_stub_rtl_if_vers)
  CATCH_ALL {
    printf("DCE ex (other) performing byte_pipe.push(0x%lx, 0x%lx, %d)\n",
      (unsigned long)byte_pipe->state, (unsigned long)buf,
      in_len);
    rc = NASD_RPC_TRAP;
  } ENDTRY

#if NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0
  NASD_UNLOCK_MUTEX(nasd_dce_pipe_lock);
#endif /* NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0 */

  return(rc);
}

nasd_status_t
nasd_od_dce_pull(
  void        *real_pipe,
  void        *buf,
  nasd_len_t   in_len,
  nasd_len_t  *out_lenp)
{
  nasd_byte_pipe_t *byte_pipe;
  idl_ulong_int count;
  nasd_status_t rc;

  byte_pipe = (nasd_byte_pipe_t *)real_pipe;

#define CATCHPULLPIPE(_exc_) \
 CATCH(_exc_) { \
   printf("DCE ex (%s) performing byte_pipe.pull(0x%lx, 0x%lx, %d, 0x%lx)\n", \
     NASD_STRING(_exc_), \
     (unsigned long)byte_pipe->state, (unsigned long)buf, \
     in_len, (unsigned long)&count); \
   *out_lenp = 0; \
   rc = NASD_RPC_TRAP; \
 }

#if NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0
  NASD_LOCK_MUTEX(nasd_dce_pipe_lock);
#endif /* NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0 */

  TRY {
    byte_pipe->pull(byte_pipe->state, (nasd_byte_t *)buf, in_len, &count);
    *out_lenp = count;
    rc = NASD_SUCCESS;
  }
  CATCHPULLPIPE(rpc_x_ss_char_trans_open_fail)
  CATCHPULLPIPE(rpc_x_ss_char_trans_short_file)
  CATCHPULLPIPE(rpc_x_ss_context_damaged)
  CATCHPULLPIPE(rpc_x_ss_context_mismatch)
  CATCHPULLPIPE(rpc_x_ss_in_null_context)
  CATCHPULLPIPE(rpc_x_ss_pipe_closed)
  CATCHPULLPIPE(rpc_x_ss_pipe_comm_error)
  CATCHPULLPIPE(rpc_x_ss_pipe_discipline_error)
  CATCHPULLPIPE(rpc_x_ss_pipe_empty)
  CATCHPULLPIPE(rpc_x_ss_pipe_memory)
  CATCHPULLPIPE(rpc_x_ss_pipe_order)
  CATCHPULLPIPE(rpc_x_ss_remote_comm_failure)
  CATCHPULLPIPE(rpc_x_ss_remote_no_memory)
  CATCHPULLPIPE(rpc_x_ss_bad_buffer)
  CATCHPULLPIPE(rpc_x_ss_bad_es_action)
  CATCHPULLPIPE(rpc_x_ss_wrong_es_version)
  CATCHPULLPIPE(rpc_x_stub_protocol_error)
  CATCHPULLPIPE(rpc_x_unknown_stub_rtl_if_vers)
  CATCH_ALL {
    printf("DCE ex (other) performing byte_pipe.pull(0x%lx, 0x%lx, %d, 0x%lx)\n",
      (unsigned long)byte_pipe->state, (unsigned long)buf,
      in_len, (unsigned long)&count);
    *out_lenp = 0;
    rc = NASD_RPC_TRAP;
  } ENDTRY

#if NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0
  NASD_UNLOCK_MUTEX(nasd_dce_pipe_lock);
#endif /* NASD_DRIVE_DISABLE_CONCURRENT_DCE_PIPES > 0 */

  return(rc);
}

nasd_status_t
nasd_od_dce_drainpull(
  void        *real_pipe)
{
  nasd_len_t tcount;
  nasd_status_t rc;

  do {
    rc = nasd_od_dce_pull(real_pipe, (nasd_byte_t *)nasd_odc_bitbucket,
      NASD_OD_BASIC_BLOCKSIZE, &tcount);
  } while ((rc == NASD_SUCCESS) && tcount);

  return(rc);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
