/*
 * nasd_cheops_cache.h
 */
/*
 * Cache module for the Cheops (clerk and storage manager)
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#ifndef _NASD_CHEOPS_CACHE_H_
#define _NASD_CHEOPS_CACHE_H_

#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_cheops_locks.h>

#define _NASD_CHEOPS_MAX_CACHE_SIZE         100

typedef struct _nasd_cheops_cache_s           _nasd_cheops_cache_t;
typedef struct _nasd_cheops_cache_ent_s       _nasd_cheops_cache_ent_t;
typedef struct _nasd_cheops_cache_hash_ent_s  _nasd_cheops_cache_hash_ent_t;
typedef struct _nasd_cheops_cache_stats_s     _nasd_cheops_cache_stats_t;
typedef int                          _nasd_cheops_cache_err_t;

/* not clear whether this is going to be ever needed */
typedef enum {_NASD_CHEOPS_CACHE_TYPE_SMAP, /* stripe map (for cheops objects) */
	      _NASD_CHEOPS_CACHE_TYPE_GMAP, 
	      _NASD_CHEOPS_CACHE_TYPE_DMAP, 
	      _NASD_CHEOPS_CACHE_TYPE_IMAP} _nasd_cheops_cache_type_t;

/* cache ent flags */
#define _NASD_CHEOPS_CACHE_ENT_BUSY       (1<<0) /* io in progress */
#define _NASD_CHEOPS_CACHE_ENT_DIRTY      (1<<1) /* dirty */
#define _NASD_CHEOPS_CACHE_ENT_FREE       (1<<2) /* free */
#define _NASD_CHEOPS_CACHE_ENT_LRU        (1<<3) /* lru */
#define _NASD_CHEOPS_CACHE_ENT_HOLD       (1<<4) /* held */

typedef struct _nasd_cheops_cache_key_s {
  int  object_id;
  int  record_id;
 } _nasd_cheops_cache_key_t;

struct _nasd_cheops_cache_stats_s {
  unsigned long  hits;
  unsigned long  misses;
  unsigned long  updates;
  unsigned long  fixes;
  unsigned long  invalidates;
};

struct _nasd_cheops_cache_ent_s {
  NASD_DECLARE_MUTEX(m)
  NASD_DECLARE_COND(c)
  int                          flags;
  int                          hold_wait_refcnt; /* the number of threads waiting to lock this entry */
  caddr_t                      data;  
  _nasd_cheops_cache_key_t      key;     /* key depends on what kind of entries are being cached */
  _nasd_cheops_cache_ent_t      *bnext;  /* hash chain links */
  _nasd_cheops_cache_ent_t      *bprev;
  _nasd_cheops_cache_ent_t      *lnext;     /* LRU links */
  _nasd_cheops_cache_ent_t      *lprev;
  _nasd_cheops_cache_ent_t      *dnext;     /* dirty links */
  _nasd_cheops_cache_ent_t      *dprev;
  _nasd_cheops_cache_ent_t      *fnext;  /* next in the free list */
};

struct _nasd_cheops_cache_s {
  NASD_DECLARE_MUTEX(m)
  _nasd_cheops_rwlock_t          lock;
  int                            csize;      /* size of cache in number of entries */
  int                            bsize;      /* size of data buffer pointed by each entry */
  int                            nbuckets;      /* size of of hash table */
  int                            victim;
  _nasd_cheops_cache_type_t       ctype;
  _nasd_cheops_cache_ent_t        *buckets;   /* hash buckets*/
  _nasd_cheops_cache_ent_t        *free_list; /* free list */
  _nasd_cheops_cache_ent_t        *ents;      /* entry list */
  _nasd_cheops_cache_ent_t        lru;
  _nasd_cheops_cache_ent_t        dirty;
  caddr_t                        bufs;
  _nasd_cheops_cache_stats_t      stats;
  int                            (*handle_read_fault)(_nasd_cheops_cache_key_t, caddr_t);
  int                            (*writeback)(_nasd_cheops_cache_ent_t *, int);
};

/* courtesy of jimz, greatest hacker of all times */
#define _NASD_CHEOPS_CACHE_INC_STAT(_cp_,_stat_) 
/* (nasd_atomic_inc64(&((_cp_)->stats._stat_))) */
#define _NASD_CHEOPS_CACHE_LOCK(_cp_)  NASD_LOCK_MUTEX((_cp_)->m)
#define _NASD_CHEOPS_CACHE_UNLOCK(_cp_) NASD_UNLOCK_MUTEX((_cp_)->m)
#define _NASD_CHEOPS_CACHE_ENT_LOCK(_e_)   NASD_LOCK_MUTEX((_e_)->m)
#define _NASD_CHEOPS_CACHE_ENT_UNLOCK(_e_) NASD_UNLOCK_MUTEX((_e_)->m)
#define _NASD_CHEOPS_CACHE_ENT_WAIT(_e_)   NASD_WAIT_COND((_e_)->c,(_e_)->m)
#define _NASD_CHEOPS_CACHE_ENT_SIGNAL(_e_) NASD_SIGNAL_COND((_e_)->c)
#define _NASD_CHEOPS_CACHE_ENT_BCAST(_e_)  NASD_BROADCAST_COND((_e_)->c)

/* Prototypes */
int _nasd_cheops_cache_destroy(_nasd_cheops_cache_t *cp);
int _nasd_cheops_cache_create(int csize, int bsize, _nasd_cheops_cache_type_t ctype, _nasd_cheops_cache_t **cpp);
int _nasd_cheops_cache_read(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key, caddr_t re);
void _nasd_cheops_cache_show_stats(_nasd_cheops_cache_t *cp);
int _nasd_cheops_cache_mark_invalid(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key);
int _nasd_cheops_cache_mark_dirty(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key);
int _nasd_cheops_cache_hold_ent(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_key_t key, _nasd_cheops_cache_ent_t **ep);
int _nasd_cheops_cache_release_ent(_nasd_cheops_cache_t *cp, _nasd_cheops_cache_ent_t *ep);

#endif /* !_NASD_CHEOPS_CACHE_H_ */





