/*
 * nasd_nfs_rpc.h
 *
 * Assist macros and definitions for NASD NFS server
 * RPC subsystem
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */


#ifndef _NASD_NFS_RPC_H_
#define _NASD_NFS_RPC_H_

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_timer.h>
#include <nasd/nasd_nfs_internal.h>

#define NASD_NFS_RPC_KEEP_OP_STATS         1
#define NASD_NFS_RPC_DUMP_OP_STATS         0
#define NASD_NFS_RPC_OP_QUEUE_TIMES        0

#define NASD_NFS_RPC_DEBUG_RPC_START_STOP  0
#define NASD_NFS_RPC_DEBUG_RPC_RETURNS     0
#define NASD_NFS_RPC_DEBUG_RPC_ERR_RETURNS 0
#define NASD_NFS_RPC_DEBUG_RPC_SHUTDOWN    1
#define NASD_NFS_RPC_DEBUG_DCE_EXCEPTION   1

extern nasd_uint64 nasd_nfs_threads_in_rpc;
extern int nasd_nfs_wait_shutdown;
NASD_DECLARE_EXTERN_COND(nasd_nfs_shutdown_cond)
NASD_DECLARE_EXTERN_MUTEX(nasd_nfs_rpc_cnt_lock)
NASD_DECLARE_EXTERN_MUTEX(nasd_nfs_rpc_qdepth_lock)

extern int nasd_nfs_rpc_qdepth;
extern nasd_timer_t nasd_nfs_rpc_qdepth_timer;
extern nasd_nfs_rpc_opdepths_t nasd_nfs_rpc_opdepths;

extern int nasd_nfs_protseq_installed;
extern int nasd_nfs_active;

extern nasd_nfs_opstats_t nasd_nfs_opstats;
extern nasd_nfs_cachestats_t nasd_nfs_cache_stats;

#if NASD_NFS_RPC_KEEP_OP_STATS > 0

#define NASD_NFS_RPC_OP_STATS_DECL \
  nasd_timer_t _op_timer; \
  nasd_uint64 _el;

#define NASD_NFS_RPC_OP_BEGIN(_opname_) { \
  NASD_ATOMIC_INC64(&nasd_nfs_opstats._opname_.in_flight); \
  NASD_TM_START(&_op_timer); \
}

#define NASD_NFS_RPC_OP_END(_opname_) { \
  NASD_TM_STOP(&_op_timer); \
  NASD_TM_ELAPSED_NSEC(&_op_timer,&_el); \
  NASD_ATOMIC_INC64(&nasd_nfs_opstats._opname_.invalid); \
  NASD_ATOMIC_INC64(&nasd_nfs_opstats._opname_.num_ops); \
  NASD_ATOMIC_DEC64(&nasd_nfs_opstats._opname_.in_flight); \
  NASD_ATOMIC_ADD64(&nasd_nfs_opstats._opname_.op_nsecs,_el); \
  NASD_ATOMIC_DEC64(&nasd_nfs_opstats._opname_.invalid); \
}

#else /* NASD_NFS_RPC_KEEP_OP_STATS > 0 */

#define NASD_NFS_RPC_OP_BEGIN(_opname_)
#define NASD_NFS_RPC_OP_END(_opname_)
#define NASD_NFS_RPC_OP_STATS_DECL \
  nasd_uint64 _el=0;

#endif /* NASD_NFS_RPC_KEEP_OP_STATS > 0 */

#if NASD_NFS_RPC_OP_QUEUE_TIMES > 0

#define NASD_NFS_RPC_QBEGIN() { \
  nasd_timespec_t _ts; \
  int _qd; \
  NASD_LOCK_MUTEX(nasd_nfs_rpc_qdepth_lock); \
  if (nasd_nfs_rpc_qdepth) { \
    NASD_TM_STOP(&nasd_nfs_rpc_qdepth_timer); \
    NASD_TM_ELAPSED_TS(&nasd_nfs_rpc_qdepth_timer, &_ts); \
  } \
  NASD_TM_START(&nasd_nfs_rpc_qdepth_timer); \
  if (nasd_nfs_rpc_qdepth) { \
    _qd = NASD_MIN(nasd_nfs_rpc_qdepth, NASD_NFS_RPC_OPDEPTHS); \
    NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_rpc_opdepths.opdepths[_qd-1], &_ts); \
  } \
  nasd_nfs_rpc_qdepth++; \
  NASD_UNLOCK_MUTEX(nasd_nfs_rpc_qdepth_lock); \
}

#define NASD_NFS_RPC_QEND() { \
  nasd_timespec_t _ts; \
  int _qd; \
  NASD_LOCK_MUTEX(nasd_nfs_rpc_qdepth_lock); \
  NASD_ASSERT(nasd_nfs_rpc_qdepth > 0); \
  NASD_TM_STOP(&nasd_nfs_rpc_qdepth_timer); \
  NASD_TM_ELAPSED_TS(&nasd_nfs_rpc_qdepth_timer, &_ts); \
  nasd_nfs_rpc_qdepth--; \
  if (nasd_nfs_rpc_qdepth) { \
    NASD_TM_START(&nasd_nfs_rpc_qdepth_timer); \
  } \
  _qd = NASD_MIN(nasd_nfs_rpc_qdepth, NASD_NFS_RPC_OPDEPTHS); \
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_rpc_opdepths.opdepths[_qd], &_ts); \
  NASD_UNLOCK_MUTEX(nasd_nfs_rpc_qdepth_lock); \
}

#else /* NASD_NFS_RPC_OP_QUEUE_TIMES > 0 */

#define NASD_NFS_RPC_QBEGIN() { /* noop */ }
#define NASD_NFS_RPC_QEND()   { /* noop */ }

#endif /* NASD_NFS_RPC_OP_QUEUE_TIMES > 0 */

#if NASD_NFS_RPC_DEBUG_RPC_START_STOP > 0
#define NASD_NFS_RPC_SS_DBG_PRINTF(_a_) printf _a_
#else /* NASD_NFS_RPC_DEBUG_RPC_START_STOP > 0 */
#define NASD_NFS_RPC_SS_DBG_PRINTF(_a_)
#endif /* NASD_NFS_RPC_DEBUG_RPC_START_STOP > 0 */

#if NASD_NFS_RPC_DEBUG_RPC_RETURNS > 0
#define NASD_NFS_RPC_RET_DBG_PRINTF(_a_) printf _a_
#else /* NASD_NFS_RPC_DEBUG_RPC_RETURNS > 0 */
#if NASD_NFS_RPC_DEBUG_RPC_ERR_RETURNS > 0
#define NASD_NFS_RPC_RET_DBG_PRINTF(_a_) if (NASD_NFS_RPC_NASD_RET) { printf _a_; }
#else /* NASD_NFS_RPC_DEBUG_RPC_ERR_RETURNS > 0 */
#define NASD_NFS_RPC_RET_DBG_PRINTF(_a_)
#endif /* NASD_NFS_RPC_DEBUG_RPC_ERR_RETURNS > 0 */
#endif /* NASD_NFS_RPC_DEBUG_RPC_RETURNS > 0 */


#define NASD_NFS_RPC_DOBEGIN(_opname_,_file_) { \
  if (nasd_nfs_active == 0) { \
    /* \
     * XXX should never happen, but DCE RPC \
     * is not as good at "shutdown" as we \
     * want. Ideally, this would set *op_status \
     * equal to 0, and a useful error code in \
     * the nasd status field, but that would \
     * involve marshalling a reply here, which \
     * is more than we want to do. \
     */ \
    *op_status = 999; \
    return; \
  } \
  NASD_NFS_RPC_QBEGIN(); \
  nasd_nfs_begin_rpc(&nasd_nfs_opstats._opname_); \
  NASD_NFS_RPC_OP_BEGIN(_opname_); \
  NASD_NFS_RPC_SS_DBG_PRINTF(("Begin %s (file %s line %d) thread %" \
    NASD_THREAD_ID_FMT "\n", \
    NASD_STRING(_opname_), _file_, __LINE__, nasd_thread_self())); \
}

#define NASD_NFS_RPC_DORETURN(_opname_,_file_) { \
  NASD_NFS_RPC_SS_DBG_PRINTF(("End %s (file %s line %d)\n", \
    NASD_STRING(_opname_), _file_, __LINE__)); \
  NASD_NFS_RPC_RET_DBG_PRINTF(("%s (file %s line %d) returning %d (%s)\n", \
    NASD_STRING(_opname_), _file_, __LINE__, NASD_NFS_RPC_NASD_RET, \
    nasd_error_string(NASD_NFS_RPC_NASD_RET))); \
  *op_status = 0; \
  NASD_NFS_RPC_OP_END(_opname_); \
  nasd_nfs_end_rpc(&nasd_nfs_opstats._opname_); \
  NASD_NFS_RPC_QEND(); \
  return; \
}

typedef struct nasd_nfs_rpc_basic_s {
  nasd_status_t  nasd_status;
} nasd_nfs_rpc_basic_t;

extern void
nasd_dump_nfs3_attributes(
  nasd_nfs3_attributes_t  *nfs3attrp);

extern void
nasd_nfs_real_mount(
  nasd_credential_t       in_credential,
  nasd_string_t           dirpath[],
  nasd_list_len_t         in_listlen,
  nasd_net_address_t      out_drivelist[],
  nasd_list_len_t        *out_listlen,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_fsstat(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsstat_t          *out_stat,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_fsinfo(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsinfo_t          *out_info,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_lookup(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  char                   *in_dirpath,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_post_attribute_t  *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_readdir(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_marker_t           in_marker,    
  nasd_markerverf_t       in_markerverf,
  nasd_len_t              in_count,
  nasd_marker_t          *out_marker,
  nasd_markerverf_t      *out_markerverf,
  nasd_len_t             *out_count,
  nasd_boolean_t         *out_eof,
  nasd_dirent_t           out_entries[],
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_access(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_access_t           in_access,
  nasd_access_t          *out_access,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_setattr(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_guard_t            in_guard,
  nasd_attribute_t       *out_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_create(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_createmask_t       in_createmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_symlink(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_string_t           in_symlink[],
  nasd_attribute_t        in_attribute,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_remove(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_mkdir(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_rmdir(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_newcookie(
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_cookie_t          *out_cookie,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_rename(
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_from_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_from_path[],
  nasd_nfs_identifier_t   in_to_directory,
  nasd_string_t           in_to_path[],
  nasd_post_attribute_t  *post_from_attribute,
  nasd_post_attribute_t  *post_to_attribute,
  nasd_status_t          *nasd_status);

extern void
nasd_nfs_real_getstats(
  nasd_nfs_opstats_t       *opstats,
  nasd_nfs_cachestats_t    *cachestats,
  nasd_nfs_rpc_opdepths_t  *opdepths,
  nasd_status_t            *nasd_status);

#endif /* !_NASD_NFS_RPC_H_ */
