/*
 * nasd_nfs_client_switch.c
 *
 * RPC wrappers, dispatch to correct binding type
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_nfs_client.h>

#include <dce/dce_error.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_mem.h>

#include <nasd/nasd_pdev.h>
#include <sys/errno.h>
#include <sys/ioctl.h>

#ifndef KERNEL
#include <fcntl.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#endif /* !KERNEL */

#include <nasd/nasd_nfs_rpc.h>

#if defined(DEC_OSF) && defined(KERNEL)
#include <mach/vm_param.h>
#include <nasd/dux/nasd_dux_opt.h>
#endif /* DEC_OSF && KERNEL */

NASD_DECLARE_ONCE(nasd_nfscli_init_once)
NASD_DECLARE_MUTEX(nasd_nfscli_use_counter_lock)
int nasd_nfscli_use_counter = 0;
nasd_shutdown_list_t *nasd_nfscli_shutdown_list = NULL;

/* Don't use __FILE__ here to avoid getting kernel-mangled names */
#define DOBEGIN(_opname_) NASD_NFS_RPC_DOBEGIN(_opname_,"nasd_nfs_client_switch.c")
#define DORETURN(_opname_) NASD_NFS_RPC_DORETURN(_opname_,"nasd_nfs_client_switch.c")

#define LOCK_COUNTER_LOCK()   NASD_LOCK_MUTEX(nasd_nfscli_use_counter_lock)
#define UNLOCK_COUNTER_LOCK() NASD_UNLOCK_MUTEX(nasd_nfscli_use_counter_lock)

/*
 * nasd_nfscli_real_init
 *
 * Called when the system comes into use after not being
 * used (start-of-day call to nasd_nfscli_init(), or first
 * such call after the last call to nasd_nfscli_shutdown()
 * which actually deactivated the system).
 */
nasd_status_t
nasd_nfscli_real_init()
{
  nasd_status_t rc;

  /* already did threads_init() */

  rc = nasd_mem_init();
  if (rc) {
    nasd_threads_shutdown();
    return(rc);
  }

  rc = nasd_shutdown_sys_init();
  if (rc) {
    nasd_mem_shutdown();
    nasd_threads_shutdown();
    return(rc);
  }

  rc = nasd_shutdown_list_init(&nasd_nfscli_shutdown_list);
  if (rc) {
    nasd_shutdown_cleanup();
    nasd_mem_shutdown();
    nasd_threads_shutdown();
    return(rc);
  }

  rc = nasd_nfscli_dce_init(nasd_nfscli_shutdown_list);
  if (rc)
    goto bad;

  rc = nasd_cl_p_init();
  if (rc)
    goto bad;
  rc = nasd_shutdown_proc(nasd_nfscli_shutdown_list,
    nasd_nfscli_shutdown_drivecli, NULL);
  if (rc) {
    nasd_nfscli_shutdown_drivecli(NULL);
    goto bad;
  }

  return(NASD_SUCCESS);

bad:
  rc = nasd_shutdown_list_shutdown(nasd_nfscli_shutdown_list);
  if (rc) {
    NASD_PANIC();
  }
  nasd_shutdown_cleanup();
  nasd_mem_shutdown();
  nasd_threads_shutdown();
  return(rc);
}

/*
 * nasd_nfscli_real_shutdown
 *
 * Called when last user of the nfscli system calls nasd_nfscli_shutdown().
 * Clean up and deallocate resources.
 */
void
nasd_nfscli_real_shutdown()
{
  nasd_status_t rc;

  rc = nasd_shutdown_list_shutdown(nasd_nfscli_shutdown_list);
  if (rc) {
    NASD_PANIC();
  }
  nasd_nfscli_shutdown_list = NULL;
  nasd_shutdown_cleanup();
  nasd_mem_shutdown();
  nasd_threads_shutdown();

}

/*
 * nasd_nfscli_sys_init
 *
 * Executed exactly once, the first time nasd_nfscli_init() is
 * called. Initialize counter tracking number of times system
 * is initted.
 */
void
nasd_nfscli_sys_init()
{
  nasd_status_t rc;

  rc = nasd_mutex_init(&nasd_nfscli_use_counter_lock);
  if (rc) {
    NASD_PANIC();
  }
  nasd_nfscli_use_counter = 0;
}

/*
 * nasd_nfscli_init
 *
 * Keep a counter of the number of times we're initted and
 * shutdown. When the last shutdown arrives, really deallocate.
 * This lets multiple subsystems use us without knowing about
 * one another.
 */
nasd_status_t
nasd_nfscli_init()
{
  nasd_status_t rc;

  rc = nasd_threads_init();
  if (rc)
    return(rc);
  nasd_once(&nasd_nfscli_init_once, nasd_nfscli_sys_init);
  LOCK_COUNTER_LOCK();
  nasd_nfscli_use_counter++;
  if (nasd_nfscli_use_counter == 1) {
    rc = nasd_nfscli_real_init();
    if (rc) {
      nasd_nfscli_use_counter = 0;
    }
  }
  else {
    /* get rid of extra threads init above */
    nasd_threads_shutdown();
    rc = NASD_SUCCESS;
  }
  UNLOCK_COUNTER_LOCK();

  return(rc);
}

/*
 * nasd_nfscli_shutdown
 *
 * Previous caller of nasd_nfscli_init() not using the nfscli
 * subsystem any more. Deallocate and cleanup iff necessary.
 */
void
nasd_nfscli_shutdown()
{
  NASD_ASSERT(nasd_nfscli_use_counter != 0);

  LOCK_COUNTER_LOCK();
  nasd_nfscli_use_counter--;
  if (nasd_nfscli_use_counter == 0) {
    nasd_nfscli_real_shutdown();
  }
  UNLOCK_COUNTER_LOCK();
}

void
nasd_nfscli_shutdown_drivecli(
  void  *ignored)
{
  nasd_cl_p_shutdown();
}

nasd_status_t
nasd_bind_to_nfs_server(
  char               *server_name,
  char               *portnum,
  int                 binding_type,
  void               *binding_param,
  int                 binding_param_len,
  nasd_nfs_handle_t  *handlep)
{
  nasd_nfs_handle_t handle;

  if (nasd_nfscli_use_counter == 0)
    return(NASD_SYS_NOT_INITIALIZED);

#ifndef KERNEL
{
  extern char *getenv();
  char *env;

  /* set this to something like "/dev/nasdkp0" */
  env = getenv("NASD_JIMZ_MONKEY_COLOCATE");
  /*
   * If we've already got a nasd_nfs_param_dce_kpdev_t
   * argument, we use it anyway.
   */
  if (env) {
    if (binding_type != NASD_BIND_COLOCATE) {
      if (binding_type != NASD_BIND_DCE_KPDEV) {
        static nasd_nfs_param_dce_kpdev_t kpdev_args;
        strcpy(kpdev_args.devname, env);
        binding_param = &kpdev_args;
        binding_param_len = sizeof(kpdev_args);
      }
      binding_type = NASD_BIND_COLOCATE;
    }
  }
}
#endif /* !KERNEL */

  *handlep = NULL;
  if ((binding_type == NASD_BIND_DCE_DIRECT)
#ifndef KERNEL
    || (binding_type == NASD_BIND_DCE_DIRECT_TCP)
#endif /* !KERNEL */
    )
  {
    nasd_status_t rc;

    rc = NASD_FAIL;
    NASD_Malloc(handle, sizeof(struct nasd_drive_handle_s), (nasd_nfs_handle_t));
    if (handle == NULL)
      return(NASD_NO_MEM);
    bzero((char *)handle, sizeof(struct nasd_drive_handle_s));
    handle->type = NASD_BIND_DCE_DIRECT;
    handle->real_type = binding_type;
    if (binding_type == NASD_BIND_DCE_DIRECT) {
      rc = nasd_dce_direct_bind_to_server((unsigned_char_t *)"ncadg_ip_udp",
        (unsigned_char_t *)server_name, (unsigned_char_t *)portnum,
        &handle->h);
    }
#ifndef KERNEL
    else if (binding_type == NASD_BIND_DCE_DIRECT_TCP) {
      rc = nasd_dce_direct_bind_to_server((unsigned_char_t *)"ncacn_ip_tcp",
        (unsigned_char_t *)server_name, (unsigned_char_t *)portnum,
        &handle->h);
    }
#endif /* !KERNEL */
    if (rc) {
      NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
      return(rc);
    }
    *handlep = handle;
    LOCK_COUNTER_LOCK();
    nasd_nfscli_use_counter++;
    UNLOCK_COUNTER_LOCK();
    return(NASD_SUCCESS);
  }
  else if (binding_type == NASD_BIND_COLOCATE) {
#ifndef KERNEL
    nasd_nfs_param_dce_kpdev_t *param;
    NASDPD_bind_t kpbind;
    int ret, fd;

    if (binding_param_len != sizeof(nasd_nfs_param_dce_kpdev_t))
      return(NASD_BAD_BINDING_PARAM_LEN);
    param = (nasd_nfs_param_dce_kpdev_t *)binding_param;
    fd = open(param->devname, O_RDWR);
    if (fd < 0) {
      return(NASD_FAIL);
    }
#endif /* !KERNEL */

    NASD_Malloc(handle, sizeof(struct nasd_drive_handle_s), (nasd_nfs_handle_t));
    if (handle == NULL)
      return(NASD_NO_MEM);
    bzero((char *)handle, sizeof(struct nasd_drive_handle_s));
#ifdef KERNEL
    handle->type = binding_type;
    handle->real_type = binding_type;
#else /* KERNEL */
    handle->type = NASD_BIND_DCE_KPDEV;
    handle->real_type = binding_type;
    handle->fd = fd;
    strcpy(kpbind.drive_name, "colocated drive");
    strcpy(kpbind.drive_port, portnum);
    kpbind.binding_type = binding_type;
    kpbind.binding_param = NULL;
    kpbind.binding_param_len = 0;
    ioctl(handle->fd, NASDPD_PUNBIND);
    ret = ioctl(handle->fd, NASDPD_PBIND, &kpbind);
    if (ret) {
      NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
      return(NASD_FAIL);
    }
#endif /* KERNEL */
    *handlep = handle;
    LOCK_COUNTER_LOCK();
    nasd_nfscli_use_counter++;
    UNLOCK_COUNTER_LOCK();
    return(NASD_SUCCESS);
  }

  return(NASD_BAD_HANDLE_TYPE);
}

nasd_status_t
nasd_unbind_nfs_server(
  nasd_nfs_handle_t  *handlep)
{
  nasd_nfs_handle_t handle;

  handle = *handlep;
  if (handle == NULL) {
    return(NASD_BAD_HANDLE);
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    int rc;

    rc = nasd_dce_direct_unbind_server(&handle->h);
    if (rc)
      return(NASD_FAIL);
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_nfscli_shutdown();
    return(NASD_SUCCESS);
  }
#ifndef KERNEL
  else if ((handle->type == NASD_BIND_DCE_KPDEV)
    || (handle->type == NASD_BIND_COLOCATE))
  {
    int rc;

    rc = ioctl(handle->fd, NASDPD_PUNBIND);
    rc = close(handle->fd);
    if (rc)
      return(NASD_FAIL);
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_nfscli_shutdown();
    return(NASD_SUCCESS);
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_nfscli_shutdown();
    return(NASD_SUCCESS);
  }
#endif /* KERNEL */
  return(NASD_BAD_HANDLE_TYPE);
}

/*
 * Operation wrappers
 */

void
nasd_nfscli_null(
  nasd_nfs_handle_t     handle,
  nasd_status_t        *nasd_status,
  nasd_rpc_status_t    *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_null(handle->h, nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_nfs_rpc_basic_t out_res;
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(null);
    *nasd_status = out_res.nasd_status = NASD_SUCCESS;
    DORETURN(null);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_mount(
  nasd_nfs_handle_t       handle,
  nasd_credential_t       in_credential,
  nasd_string_t           dirpath[],
  nasd_list_len_t         in_listlen,
  nasd_net_address_t      out_drivelist[],
  nasd_list_len_t        *out_listlen,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_mount(handle->h, in_credential, dirpath,
      in_listlen, out_drivelist, out_listlen, out_cookie,
      out_identifier, nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_nfs_rpc_basic_t out_res;
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(mount);
    nasd_nfs_real_mount(
      in_credential, dirpath,
      in_listlen, out_drivelist, out_listlen, out_cookie,
      out_identifier, nasd_status);
    out_res.nasd_status = *nasd_status;
    DORETURN(mount);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_fsstat(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsstat_t          *out_stat,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_fsstat(handle->h, in_cookie, in_identifier,
      in_credential, out_stat, post_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(fsstat);

    nasd_nfs_real_fsstat(in_cookie, in_identifier, in_credential,
      out_stat, post_attribute, nasd_status);

    DORETURN(fsstat);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_fsinfo(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsinfo_t          *out_info,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_fsinfo(handle->h, in_cookie, in_identifier,
      in_credential, out_info, post_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(fsinfo);

    nasd_nfs_real_fsinfo(in_cookie, in_identifier, in_credential,
      out_info, post_attribute, nasd_status);

    DORETURN(fsinfo);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_lookup_otw_provided(
  nasd_nfs_handle_t        handle,
  nasd_nfs_lookup_args_t  *args,
  nasd_nfs_lookup_res_t   *res,
  nasd_byte_t             *args_otw_p,
  nasd_byte_t             *res_otw_p,
  nasd_rpc_status_t       *op_status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_lookup_otw_provided(handle->h,
      args, res, args_otw_p, res_otw_p, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(lookup_otw);

    nasd_nfs_real_lookup(args->in_cookie, args->in_identifier,
      args->in_credential, (char *)args->in_dirpath,
      &res->out_cookie, &res->out_identifier,
      &res->out_attribute, &res->post_attribute,
      &res->nasd_status);

    DORETURN(lookup_otw);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_lookup(
  nasd_nfs_handle_t        handle,
  nasd_nfs_lookup_args_t  *args,
  nasd_nfs_lookup_res_t   *res,
  nasd_rpc_status_t       *op_status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_lookup(handle->h, args, res, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(lookup_otw);

    nasd_nfs_real_lookup(args->in_cookie, args->in_identifier,
      args->in_credential, (char *)args->in_dirpath,
      &res->out_cookie, &res->out_identifier,
      &res->out_attribute, &res->post_attribute,
      &res->nasd_status);

    DORETURN(lookup_otw);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_readdir(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_marker_t           in_marker,    
  nasd_markerverf_t       in_markerverf,
  nasd_len_t              in_count,
  nasd_marker_t          *out_marker,
  nasd_markerverf_t      *out_markerverf,
  nasd_len_t             *out_count,
  nasd_boolean_t         *out_eof,
  nasd_dirent_t           out_entries[],
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_readdir(handle->h, in_cookie, in_identifier,
      in_credential, in_marker, in_markerverf, in_count,
      out_marker, out_markerverf, out_count, out_eof,
      out_entries, post_attribute, nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(readdir);

    nasd_nfs_real_readdir(in_cookie, in_identifier, in_credential,
      in_marker, in_markerverf, in_count, out_marker, out_markerverf,
      out_count, out_eof, out_entries, post_attribute, nasd_status);

    DORETURN(readdir);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_access(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_access_t           in_access,
  nasd_access_t          *out_access,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_access(handle->h, in_cookie, in_identifier,
      in_credential, in_access, out_access, post_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(access);

    nasd_nfs_real_access(in_cookie, in_identifier, in_credential,
      in_access, out_access, post_attribute, nasd_status);

    DORETURN(access);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_setattr(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_guard_t            in_guard,
  nasd_attribute_t       *out_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_setattr(handle->h, in_cookie, in_identifier,
      in_credential, in_attribute, in_fieldmask,
      in_guard, out_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(setattr);

    nasd_nfs_real_setattr(in_cookie, in_identifier,
      in_credential, in_attribute, in_fieldmask, in_guard,
      out_attribute, nasd_status);

    DORETURN(setattr);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_create(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_createmask_t       in_createmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_create(handle->h, in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, in_createmask,
      out_cookie, out_identifier, out_attribute, post_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(create);

    nasd_nfs_real_create(in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, in_createmask,
      out_cookie, out_identifier, out_attribute, post_attribute,
      nasd_status);

    DORETURN(create);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_symlink(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_string_t           in_symlink[],
  nasd_attribute_t        in_attribute,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_symlink(handle->h, in_cookie, in_directory, in_credential,
      in_dirpath, in_symlink, in_attribute, out_cookie, out_identifier,
      out_attribute, post_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(symlink);

    nasd_nfs_real_symlink(in_cookie, in_directory, in_credential,
      in_dirpath, in_symlink, in_attribute, out_cookie, out_identifier,
      out_attribute, post_attribute, nasd_status);

    DORETURN(symlink);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_remove(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_remove(handle->h, in_cookie, in_directory, in_credential,
      in_dirpath, nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(remove);

    nasd_nfs_real_remove(in_cookie, in_directory, in_credential,
      in_dirpath, nasd_status);

    DORETURN(remove);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_mkdir(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_mkdir(handle->h, in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, out_cookie,
      out_identifier, out_attribute, post_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(mkdir);

    nasd_nfs_real_mkdir(in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, out_cookie, out_identifier,
      out_attribute, post_attribute, nasd_status);

    DORETURN(mkdir);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_rmdir(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_rmdir(handle->h, in_cookie, in_directory,
      in_credential, in_dirpath,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(rmdir);

    nasd_nfs_real_rmdir(in_cookie, in_directory, in_credential,
      in_dirpath, nasd_status);

    DORETURN(rmdir);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_newcookie(
  nasd_nfs_handle_t       handle,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_cookie_t          *out_cookie,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_newcookie(handle->h, in_identifier, in_credential,
      out_cookie, nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(newcookie);

    nasd_nfs_real_newcookie(in_identifier, in_credential,
      out_cookie, nasd_status);
  
    DORETURN(newcookie);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_rename(
  nasd_nfs_handle_t       handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_from_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_from_path[],
  nasd_nfs_identifier_t   in_to_directory,
  nasd_string_t           in_to_path[],
  nasd_post_attribute_t  *post_from_attribute,
  nasd_post_attribute_t  *post_to_attribute,
  nasd_status_t          *nasd_status,
  nasd_rpc_status_t      *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_rename(handle->h, in_cookie, in_from_directory,
      in_credential, in_from_path, in_to_directory, in_to_path,
      post_from_attribute, post_to_attribute,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(rename);

    nasd_nfs_real_rename(in_cookie, in_from_directory, in_credential,
      in_from_path, in_to_directory, in_to_path, post_from_attribute,
      post_to_attribute, nasd_status);

    DORETURN(rename);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

void
nasd_nfscli_getstats(
  nasd_nfs_handle_t         handle,
  nasd_nfs_opstats_t       *opstats,
  nasd_nfs_cachestats_t    *cachestats,
  nasd_nfs_rpc_opdepths_t  *opdepths,
  nasd_status_t            *nasd_status,
  nasd_rpc_status_t        *op_status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *op_status = 0;
    return;
  }
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_nfsdce_getstats(handle->h, opstats, cachestats, opdepths,
      nasd_status, op_status);
    return;
  }
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_NFS_RPC_OP_STATS_DECL

    DOBEGIN(getstats);

    nasd_nfs_real_getstats(opstats, cachestats, opdepths, nasd_status);

    DORETURN(getstats);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *op_status = 0;
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
