/*
 * nfs_dir.c
 *
 * Basic directory management for NASD NFS.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_nfs_internal.h>
#include <nasd/nasd_timer.h>

/*
 * Directory structure:
 * Each directory file is composed of "pages"
 * Each page is broken into "slots"
 * The first slot in a page is the page header - remaining
 * slots are directory entries.
 * Pages are 8192 bytes, slots are 64 bytes, so there are
 * 128 slots, or 1 header slot and 127 data slots.
 *
 * Header slot:
 * 16 bytes bitmap of slot use
 *   (lowest bit always on - it's the header slot itself)
 *  1 byte free slot count
 * 47 bytes spare
 *
 * Data slot:
 *  8 bytes flags
 *  8 bytes nasdid
 * 48 bytes name (null-terminated, so 39 bytes actually available)
 *
 * Data flags format:
 * bits 0..4: "extended" - name > 39 bytes, so next N slots are
 *  extended dance remix name, where N is the value stored in 0..4
 * bits 5..6: type
 *   00 - bogus
 *   01 - dir
 *   10 - link
 *   11 - data
 * bits 7..63: spare
 */

int nasd_nfs_max_cached_dirs  = 768; /* max dirs to cache */
int nasd_nfs_max_cached_pages = 768; /* max dir pages in cache */

NASD_DECLARE_COND(nasd_nfs_dir_clean_cond)

nasd_nfs_core_dir_t *nasd_nfs_dirty_dir_head = NULL;
nasd_nfs_core_dir_t *nasd_nfs_dirty_dir_tail = NULL;

#define SIGNAL_DIRC() NASD_SIGNAL_COND(nasd_nfs_dir_clean_cond)

nasd_nfs_core_dirpage_t *nasd_nfs_free_dirpages = NULL;
int nasd_nfs_num_free_dirpages = 0;

NASD_DECLARE_MUTEX(nasd_nfs_dircache_lock)

nasd_nfs_core_dir_t *nasd_nfs_free_dirs = NULL;
nasd_nfs_core_dir_t nasd_nfs_clean_dirs;

nasd_nfs_core_dir_t nasd_nfs_obj_buckets[NASD_NFS_OBJHASH_BUCKETS];

#define DIRB_UNHASH(_cd_) { \
  (_cd_)->hprev->hnext = (_cd_)->hnext; \
  (_cd_)->hnext->hprev = (_cd_)->hprev; \
  (_cd_)->hprev = (_cd_)->hnext = NULL; \
}

/*
 * Called to dealloc resources associated with a core dir
 */
void
nasd_nfs_coredir_disassoc(
  nasd_nfs_core_dir_t  *cd)
{
  nasd_nfs_dir_hash_check(0);
  NASD_ASSERT(cd->ndirty == 0);
  NASD_ASSERT(cd->busy == 0);
  NASD_ASSERT(cd->refcnt == 0);
  if (cd->lprev) {
    cd->lprev->lnext = cd->lnext;
    cd->lnext->lprev = cd->lprev;
  }
  else {
    NASD_ASSERT(cd->lnext == NULL);
  }
  DIRB_UNHASH(cd);
  /* now get rid of entries associated with this dir */
  NASD_NFS_LOCK_DIR(cd,1);
  nasd_nfs_name_cleandir(cd);
  NASD_NFS_UNLOCK_DIR(cd);
  if (cd->first_page) {
    cd->last_page->next = nasd_nfs_free_dirpages;
    nasd_nfs_free_dirpages = cd->first_page;
    nasd_nfs_num_free_dirpages += cd->npages;
    cd->first_page = cd->last_page = NULL;
    cd->npages = 0;
  }
  else {
    NASD_ASSERT(cd->last_page == NULL);
    NASD_ASSERT(cd->npages == 0);
  }
  cd->lprev = cd->lnext = NULL;
  cd->dprev = cd->dnext = NULL;
  if (cd->attr)
    nasd_nfs_release_attr(cd->attr);
  cd->attr = NULL;
  cd->refcnt = 0;
  cd->lnext = nasd_nfs_free_dirs;
  nasd_nfs_free_dirs = cd;
  nasd_nfs_dir_hash_check(0);
}

/*
 * Called at shutdown time- cleans up LRU (which is all
 * clean directories).
 */
void
nasd_nfs_dir_cleanup()
{
  nasd_nfs_core_dir_t *cd, *next;

  for(cd=nasd_nfs_clean_dirs.lnext;cd!=&nasd_nfs_clean_dirs;cd=next) {
    next = cd->lnext;
    nasd_nfs_coredir_disassoc(cd);
  }
}

/*
 * call with dircache LOCKED
 */
void
nasd_nfs_dir_hash_ins(
  nasd_nfs_core_dir_t  *cd)
{
  nasd_nfs_core_dir_t *bucket;
  int h;

  nasd_nfs_dir_hash_check(0);
  h = nasd_nfs_obj_hash(cd->nid);
  bucket = &nasd_nfs_obj_buckets[h];
  cd->hnext = bucket->hnext;
  cd->hprev = bucket;
  cd->hprev->hnext = cd;
  cd->hnext->hprev = cd;
  nasd_nfs_dir_hash_check(0);
}

/*
 * DEBUGGING
 */
void
_nasd_nfs_dir_hash_check(
  int  verbose)
{
  static nasd_int64 calls = 0;
  nasd_nfs_core_dir_t *bucket, *r;
  int i, size, bad, say_size, h;

  say_size = verbose;

  NASD_ATOMIC_INC64(&calls);
  if (calls%50000 == 0) {
    say_size = 1;
    printf("\n\nnasd_nfs_dir_hash_check calls %" NASD_64u_FMT "\n", calls);
  }

  bad = 0;
  for(i=0;i<NASD_NFS_OBJHASH_BUCKETS;i++) {
    bucket = &nasd_nfs_obj_buckets[i];
    size = 0;
    if (verbose)
      printf("bucket 0x%lx\n", (unsigned long)bucket);
    for(r=bucket->hnext;r!=bucket;r=r->hnext) {
      size++;
      h = nasd_nfs_obj_hash(r->nid);
      if (h != i) {
        printf("\nERROR r=0x%lx r->nid=0x%" NASD_ID_FMT " i=%d h=%d\n",
          (unsigned long)r, r->nid, i, h);
        bad = 1;
        break;
      }
      if (r->visited) {
        printf("[REPEAT 0x%lx (0x%lx 0x%lx 0x%lx)]\n",
          (unsigned long)r, (unsigned long)r->hprev,
          (unsigned long)r, (unsigned long)r->hnext);
        bad = 1;
        break;
      }
      r->visited = 1;
      if (verbose)
        printf("[0x%lx 0x%lx 0x%lx] ", (unsigned long)r->hprev,
          (unsigned long)r, (unsigned long)r->hnext);
      if (r->hprev->hnext != r) {
        printf("\nERROR r->hprev->hnext=0x%lx r=0x%lx\n",
          (unsigned long)r->hprev->hnext, (unsigned long)r);
        bad = 1;
        break;
      }
      if (r->hnext->hprev != r) {
        printf("\nERROR r->hnext->hprev=0x%lx r=0x%lx\n",
          (unsigned long)r->hnext->hprev, (unsigned long)r);
        bad = 1;
        break;
      }
    }
    if (verbose)
      printf("\n");
    for(r=bucket->hnext;r!=bucket;r=r->hnext) {
      if (r->visited == 0)
        break;
      r->visited = 0;
    }
    if (bad)
      break;
    if (say_size) {
      printf("bucket 0x%lx (i=%d) has size %d\n",
        (unsigned long)bucket, i, size);
    }
  }
}

/*
 * call with dircache LOCKED
 * returns with ref on dir
 */
nasd_status_t
nasd_nfs_dir_hash_lookup(
  nasd_nfs_identifier_t    nfs_id,
  int                      force_load,
  nasd_nfs_core_dir_t    **out_dirp,
  nasd_nfs_attr_t        **out_attrp)
{
  nasd_nfs_core_dir_t *bucket, *r, *cd, *parent;
  nasd_nfs3_attributes_t *nfs3attr;
  nasd_nfs_identifier_t p_nfs_id;
  nasd_nfs_dirent_t *de, *found;
  nasd_nfs_attr_t *dh_attr;
  nasd_identifier_t p_nid;
  nasd_nfs_drive_t *dr;
  nasd_status_t rc;
  int h, lru_take;

  NASD_NFS_CSINC(dir_lookup);

  nasd_nfs_dir_hash_check(0);

  *out_dirp = NULL;
  *out_attrp = NULL;
  h = nasd_nfs_obj_hash(nfs_id.nasd_identifier);
  bucket = &nasd_nfs_obj_buckets[h];
  for(r=bucket->hnext;r!=bucket;r=r->hnext) {
    if ((r->nid == nfs_id.nasd_identifier)
      && (r->di == nfs_id.disk_identifier))
    {
      if (r->deleted) {
        /* Oh, yeah, it's "deleted." */
        nasd_nfs_dir_hash_check(0);
        return(NASD_BAD_IDENTIFIER);
      }
      NASD_NFS_LOCK_DIR(r,1);
      r->refcnt++;
      if ((r->refcnt == 1) && (r->isroot == 0)) {
        /* remove from clean LRU */
        r->lprev->lnext = r->lnext;
        r->lnext->lprev = r->lprev;
        lru_take = 1;
      }
      while(r->busy) {
        NASD_WAIT_COND(r->cond,r->lock);
      }
      NASD_NFS_UNLOCK_DIR(r);
      if (r->deleted) {
        /* Oh, yeah, it's "deleted." */
        nasd_nfs_dir_release(r);
        nasd_nfs_dir_hash_check(0);
        return(NASD_BAD_IDENTIFIER);
      }
      *out_dirp = r;
      nasd_nfs_dir_hash_check(0);
      return(NASD_SUCCESS);
    }
  }
  if (*out_dirp || (force_load == 0)) {
    nasd_nfs_dir_hash_check(0);
    return(NASD_SUCCESS);
  }
  /*
   * Go retrieve directory from disk
   */
  NASD_NFS_CSINC(dir_force);
  rc = nasd_nfs_get_drive(nfs_id, &dr);
  if (rc) {
    nasd_nfs_dir_hash_check(0);
    return(rc);
  }
  NASD_ASSERT(dr != NULL);
  cd = nasd_nfs_dir_get_coredir(1);
  NASD_ASSERT(cd != NULL);
  cd->drive = dr;
  cd->nid = nfs_id.nasd_identifier;
  cd->di = dr->di;
  /* @@@ init cd->cookie here */
  NASD_ASSERT(cd->refcnt == 1);
  cd->busy = 1;
  cd->ndirty = 0;
  cd->attr = nasd_nfs_get_attr();
  if (cd->attr == NULL) {
    NASD_PANIC();
  }
  cd->isroot = 0;

  nasd_nfs_dir_hash_ins(cd);

  /*
   * It is safe to unlock and relock here because we've left the
   * directory cache in a consistent state. If the object is
   * deleted or moved in the meantime, we'll discover that
   * below when we recheck. The directory we're reading in will
   * now appear correctly in the cache, but will be marked
   * busy, which will prevent other lookups from completing
   * until the read is done.
   */
  NASD_NFS_UNLOCK_DIRCACHE();

  NASD_NFS_CSINC(dirc_load_attr);
  rc = nasd_nfs_dir_load_attr(cd);
  if (rc != NASD_SUCCESS) {
    cd->busy = 0;
    NASD_BROADCAST_COND(cd->cond);
    cd->deleted = 1;
    nasd_nfs_dir_release(cd);
    nasd_nfs_dir_hash_check(0);
    return(rc);
  }
  nfs3attr =
    (nasd_nfs3_attributes_t *)NASD_NFS_ATTR_ATTR(cd->attr).fs_specific;
  if (nfs3attr->type != NF3DIR) {
    /*
     * Not a directory- we return NASD_NFS_NOT_DIR,
     * and give the caller the attribute, in case it's
     * useful to them (for example, ACCESS will still
     * want to check permission bits). The caller is
     * responsible for calling nasd_nfs_release_attr()
     * on this attribute.
     */
    NASD_NFS_CSINC(dirc_not_dir);
    NASD_NFS_LOCK_DIRCACHE();
    cd->attr->refcnt++;
    *out_attrp = cd->attr;
    NASD_NFS_UNLOCK_DIRCACHE();
    cd->busy = 0;
    NASD_BROADCAST_COND(cd->cond);
    cd->deleted = 1;
    nasd_nfs_dir_release(cd);
    nasd_nfs_dir_hash_check(0);
    return(NASD_NFS_NOT_DIR);
  }

  NASD_NFS_CSINC(dirc_load_dir);
  rc = nasd_nfs_load_dir(cd);
  if (rc != NASD_SUCCESS) {
    /*
     * Someone else may have tried to grab in the
     * meantime, get rid of them.
     */
    cd->busy = 0;
    NASD_BROADCAST_COND(cd->cond);
    cd->deleted = 1;
    nasd_nfs_dir_release(cd);
    nasd_nfs_dir_hash_check(0);
    return(rc);
  }

  NASD_NFS_LOCK_DIRCACHE();

  nasd_nfs_dir_hash_check(0);

  /*
   * Now extract the parent ent, and see if it's
   * in-core. If so, link up with it.
   *
   * First, find the parent dir.
   */
  p_nid = cd->first_page->page->data_slots[1].nasdid;
  p_nfs_id.nasd_identifier = p_nid;
  p_nfs_id.disk_identifier = dr->di;
  rc = nasd_nfs_dir_hash_lookup(p_nfs_id, 0, &parent, &dh_attr);
  if (rc) {
    /* this should never happen */
    NASD_PANIC();
  }
  if (parent) {
    /* found ../, now get our entry */
    NASD_NFS_LOCK_DIR(parent,1);
    h = nasd_nfs_obj_hash(cd->nid);
    found = NULL;
    for(de=parent->idbuckets[h];de;de=de->ihnext) {
      if (de->data->nasdid == cd->nid) {
        found = de;
        break;
      }
    }
    if (found) {
      found->dir = cd;
      found->attr = cd->attr;
      cd->attr->refcnt++;
    }
    else {
      if (cd->deleted) {
        /* deleted while loading- how ignominous! */
        cd->busy = 0;
        NASD_BROADCAST_COND(cd->cond);
        nasd_nfs_dir_release(cd);
        nasd_nfs_dir_hash_check(0);
        return(NASD_BAD_IDENTIFIER);
      }
      else {
        /* perhaps we're moving - declare "staleness" and bail */
        cd->busy = 0;
        NASD_BROADCAST_COND(cd->cond);
        nasd_nfs_dir_release(cd);
        nasd_nfs_dir_hash_check(0);
        return(NASD_NFS_STALE_ID);
      }
    }
    NASD_NFS_UNLOCK_DIR(parent);
    nasd_nfs_dir_release(parent);
  }

  nasd_nfs_dir_hash_check(0);

  cd->busy = 0;
  NASD_BROADCAST_COND(cd->cond);

  *out_dirp = cd;
  return(NASD_SUCCESS);
}

/*
 * call with dircache LOCKED
 * call with dir LOCKED
 */
void
nasd_nfs_dir_dirty_page(
  nasd_nfs_core_dir_t      *dir,
  nasd_nfs_core_dirpage_t  *cdp)
{
  NASD_NFS_CSINC(dirc_dirty);
  if (cdp->dirty)
    return;
  NASD_NFS_CSINC(dirc_new_dirty);
  cdp->dirty = 1;
  dir->ndirty++;
}

/*
 * call with dircache LOCKED
 * call with dir LOCKED
 */
nasd_status_t
nasd_nfs_dir_add_entry(
  nasd_nfs_core_dir_t   *dir,
  char                  *name,
  nasd_identifier_t      nasdid,
  nasd_uint64            flags,
  nasd_nfs_dirent_t    **dep)
{
  nasd_nfs_core_dirpage_t *cdp, *found;
  int nentries, len, i, slot, j;
  nasd_nfs_dirheader_t *dh;
  nasd_nfs_dirent_t *de;

  /*
   * compute how many contig entries we need
   * to hold this name
   */
  len = strlen(name);
  if (len > NASD_NFS_MAX_NAME_LEN)
    return(NASD_NFS_BAD_NAME);
  if (len < 1)
    return(NASD_NFS_BAD_NAME);
  len++; /* hold the NULL */
  nentries = 1;
  if (len > NASD_NFS_DIRDATA_EMBEDDED_NAME_SIZE) {
    len -= NASD_NFS_DIRDATA_EMBEDDED_NAME_SIZE;
    nentries += (len+NASD_NFS_DIRDATA_SIZE-1)/NASD_NFS_DIRDATA_SIZE;
  }

  found = NULL;
  slot = (-1);
  for(cdp=dir->first_page;cdp;cdp=cdp->next) {
    dh = &cdp->page->header;
    if (dh->freecnt < nentries)
      continue;
    /* page has enough entries- are they contiguous? */
    for(i=0;i<127;i++) {
      if (!NASD_NFS_DH_SLOTFULL(dh,i)) {
        /* slot i is empty, do we have enough contig? */
        for(j=1;((j<nentries)&&((i+j)<127));j++) {
          if (NASD_NFS_DH_SLOTFULL(dh,i+j)) {
            break;
          }
        }
        if (j < nentries)
          continue;
        /* found enough contig */
        slot = i;
        found = cdp;
        break;
      }
    }
    if (found)
      break;
  }
  if (found) {
    NASD_ASSERT(slot >= 0);
    NASD_ASSERT(slot < 127);
    cdp = found;
  }
  else {
    /* no pages with enough available slots, add one */
    nasd_nfs_dir_mk_freepages(1);
    cdp = nasd_nfs_free_dirpages;
    nasd_nfs_free_dirpages = cdp->next;
    nasd_nfs_num_free_dirpages--;
    /* cdp is our new page */
    dir->last_page->next = cdp;
    cdp->next = NULL;
    cdp->prev = dir->last_page;
    cdp->offset = dir->last_page->offset + sizeof(nasd_nfs_dirpage_t);
    dir->last_page = cdp;
    dir->npages++;
    bzero((char *)cdp->page, sizeof(nasd_nfs_dirpage_t));
    cdp->page->header.freecnt = 127;
    slot = 0;
  }

  /*
   * Add the entry to the page
   */
  nasd_nfs_dir_dirty_page(dir, cdp);
  for(i=slot;i<slot+nentries;i++) {
    NASD_ASSERT((cdp->page->header.usemap[i/8]&(1<<(i%8))) == 0);
    cdp->page->header.usemap[i/8] |= 1<<(i%8);
  }
  cdp->page->header.freecnt -= nentries;
  cdp->page->data_slots[slot].flags = flags|(nentries-1);
  cdp->page->data_slots[slot].nasdid = nasdid;
  cdp->page->data_slots[slot].name_len = len - 1; /* take null back off */
  bcopy(name, cdp->page->data_slots[slot].name, len);

  de = nasd_nfs_name_add(dir, cdp, &cdp->page->data_slots[slot]);
  if (de == NULL) {
    NASD_PANIC();
  }

  *dep = de;

  return(NASD_SUCCESS);
}

/*
 * call with dircache LOCKED
 */
void
nasd_nfs_dir_release(
  nasd_nfs_core_dir_t  *dir)
{
  int ndirty, cleaned;
  nasd_status_t rc;

  cleaned = 0;
  NASD_ASSERT(dir->refcnt > 0);
  dir->refcnt--;
  if ((dir->ndirty) && (dir->refcnt == 0)) {
    nasd_gettime(&dir->dirty_time);
    /*
     * XXX
     *
     * For now, nothing goes into the dirty list.
     * At some point, these should be queued for
     * async writes.
     */
    dir->busy = 1;
    ndirty = dir->ndirty;
    rc = nasd_nfs_dir_write_dirty(dir);
    if (rc != NASD_SUCCESS) {
      printf("NFS: got rc=0x%lx (%s) writing dirty dir pages\n",
        (unsigned long)rc, nasd_error_string(rc));
      NASD_PANIC();
    }
    dir->busy = 0;
    NASD_BROADCAST_COND(dir->cond);
    cleaned = 1;
  }
  if ((dir->refcnt == 0) && (dir->isroot == 0)) {
    dir->lnext = nasd_nfs_clean_dirs.lnext;
    dir->lprev = &nasd_nfs_clean_dirs;
    dir->lprev->lnext = dir;
    dir->lnext->lprev = dir;
    if (cleaned) {
      SIGNAL_DIRC();
    }
  }
  NASD_NFS_LOCK_DIR(dir,1);
  if ((dir->refcnt == 1) && (dir->deleted)) {
    NASD_SIGNAL_COND(dir->cond);
    if (cleaned) {
      SIGNAL_DIRC();
    }
  }
  if ((dir->refcnt == 0) && (dir->deleted)) {
    NASD_NFS_UNLOCK_DIR(dir);
    nasd_nfs_coredir_disassoc(dir);
    if (cleaned)
      SIGNAL_DIRC();
    return;
  }
  NASD_NFS_UNLOCK_DIR(dir);
}

/*
 * call with dircache LOCKED
 */
void
nasd_nfs_dir_release_post_attribute(
  nasd_nfs_core_dir_t    *cd,
  nasd_nfs_identifier_t   in_identifier,
  nasd_post_attribute_t  *post_attribute)
{
  nasd_nfs_dir_release(cd);
  if (post_attribute)
    nasd_nfs_post_attribute(cd, in_identifier, post_attribute);
}

void
nasd_nfs_dir_writer_shutdown(
  void  *ignored)
{
  nasd_status_t rc;

  /*
   * Mark all the directories clean
   */
  rc = nasd_nfs_mark_dirs_clean();
  if (rc) {
    NASD_PANIC();
  }
}

nasd_status_t
nasd_nfs_dir_init()
{
  nasd_nfs_core_dirpage_t *dirpages, *cdp;
  nasd_nfs_core_dir_t *dirs, *cd;
  nasd_nfs_dirpage_t *pages, *p;
  nasd_status_t rc;
  int i, size;

  for(i=0;i<NASD_NFS_OBJHASH_BUCKETS;i++) {
    cd = &nasd_nfs_obj_buckets[i];
    bzero((char *)cd, sizeof(nasd_nfs_core_dir_t));
    cd->hnext = cd->hprev = cd;
  }

  nasd_nfs_free_dirpages = NULL;
  nasd_nfs_free_dirs = NULL;
  nasd_nfs_num_free_dirpages = 0;
  nasd_nfs_dirty_dir_head = NULL;
  nasd_nfs_dirty_dir_tail = NULL;

  bzero((char *)&nasd_nfs_clean_dirs, sizeof(nasd_nfs_clean_dirs));
  nasd_nfs_clean_dirs.lnext = &nasd_nfs_clean_dirs;
  nasd_nfs_clean_dirs.lprev = &nasd_nfs_clean_dirs;

  rc = nasd_mutex_init(&nasd_nfs_dircache_lock);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_mutex,
    &nasd_nfs_dircache_lock);
  if (rc) {
    nasd_mutex_destroy(&nasd_nfs_dircache_lock);
    return(rc);
  }

  rc = nasd_cond_init(&nasd_nfs_dir_clean_cond);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_cond,
    &nasd_nfs_dir_clean_cond);
  if (rc) {
    nasd_cond_destroy(&nasd_nfs_dir_clean_cond);
    return(rc);
  }

  size = sizeof(nasd_nfs_core_dirpage_t)*nasd_nfs_max_cached_pages;
  NASD_Malloc(dirpages, size, (nasd_nfs_core_dirpage_t *));
  if (dirpages == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_mem(nasd_nfs_shutdown, dirpages, size);
  if (rc) {
    NASD_Free(dirpages, size);
    return(NASD_NO_MEM);
  }

  size = sizeof(nasd_nfs_core_dir_t)*nasd_nfs_max_cached_dirs;
  NASD_Malloc(dirs, size, (nasd_nfs_core_dir_t *));
  if (dirs == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_mem(nasd_nfs_shutdown, dirs, size);
  if (rc) {
    NASD_Free(dirs, size);
    return(NASD_NO_MEM);
  }

  size = sizeof(nasd_nfs_dirpage_t)*nasd_nfs_max_cached_pages;
  NASD_Malloc(pages, size, (nasd_nfs_dirpage_t *));
  if (pages == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_mem(nasd_nfs_shutdown, pages, size);
  if (rc) {
    NASD_Free(pages, size);
    return(NASD_NO_MEM);
  }

  for(i=0;i<nasd_nfs_max_cached_pages;i++) {
    cdp = &dirpages[i];
    p = &pages[i];
    cdp->page = p;
    cdp->prev = NULL;
    cdp->dirty = 0;
    cdp->offset = 0;
    cdp->next = nasd_nfs_free_dirpages;
    nasd_nfs_free_dirpages = cdp;
    nasd_nfs_num_free_dirpages++;
  }

  for(i=0;i<nasd_nfs_max_cached_dirs;i++) {
    cd = &dirs[i];
    rc = nasd_mutex_init(&cd->lock);
    if (rc) {
      return(rc);
    }
    rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_mutex,
      &cd->lock);
    if (rc) {
      nasd_mutex_destroy(&cd->lock);
      return(rc);
    }
    rc = nasd_cond_init(&cd->cond);
    if (rc) {
      return(rc);
    }
    rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_cond,
      &cd->cond);
    if (rc) {
      nasd_cond_destroy(&cd->cond);
      return(rc);
    }
    cd->refcnt = 0;
    cd->first_page = NULL;
    cd->last_page = NULL;
    cd->lprev = NULL;
    cd->lnext = nasd_nfs_free_dirs;
    nasd_nfs_free_dirs = cd;
  }

  return(NASD_SUCCESS);
}

void
nasd_nfs_dealloc_coredir(
  nasd_nfs_core_dir_t  *cd,
  int                   dircache_lock_held)
{
  if (dircache_lock_held == 0) {
    NASD_NFS_LOCK_DIRCACHE();
  }

  nasd_nfs_dir_hash_check(0);

  cd->last_page->next = nasd_nfs_free_dirpages;
  nasd_nfs_free_dirpages = cd->first_page;
  nasd_nfs_num_free_dirpages += cd->npages;
  cd->npages = 0;
  cd->first_page = cd->last_page = NULL;

  nasd_nfs_name_cleandir(cd);

  if (cd->attr) {
    nasd_nfs_release_attr(cd->attr);
    cd->attr = NULL;
  }

  cd->lnext = nasd_nfs_free_dirs;
  nasd_nfs_free_dirs = cd;

  nasd_nfs_dir_hash_check(0);

  if (dircache_lock_held == 0) {
    NASD_NFS_UNLOCK_DIRCACHE();
  }
  SIGNAL_DIRC();
}

void
nasd_nfs_ditch_coredir(
  nasd_nfs_core_dir_t  *cd,
  int                   dircache_lock_held)
{
  if (dircache_lock_held == 0) {
    NASD_NFS_LOCK_DIRCACHE();
  }
  cd->nid = NASDID_NULL;
  cd->lnext = nasd_nfs_free_dirs;
  nasd_nfs_free_dirs = cd;
  if (dircache_lock_held == 0) {
    NASD_NFS_UNLOCK_DIRCACHE();
  }
  SIGNAL_DIRC();
}

nasd_nfs_core_dir_t *
nasd_nfs_dir_get_coredir(
  int  dircache_lock_held)
{
  nasd_nfs_core_dir_t *cd;

  if (dircache_lock_held == 0) {
    NASD_NFS_LOCK_DIRCACHE();
  }
  NASD_NFS_CSINC(dir_get);
reget_dir:
  NASD_NFS_CSINC(dir_reget);
  cd = nasd_nfs_free_dirs;
  if (cd) {
    nasd_nfs_free_dirs = cd->lnext;
    goto got_dir;
  }
  /*
   * No free dir, must pull one off lru
   */
  cd = nasd_nfs_clean_dirs.lprev;
  if (cd == &nasd_nfs_clean_dirs) {
    /* no clean dirs - wait for one */
    NASD_WAIT_COND(nasd_nfs_dir_clean_cond, nasd_nfs_dircache_lock);
    goto reget_dir;
  }
  NASD_NFS_CSINC(dir_replace);
  NASD_ASSERT(cd->ndirty == 0);
  NASD_ASSERT(cd->busy == 0);
  cd->lprev->lnext = cd->lnext;
  cd->lnext->lprev = cd->lprev;
  DIRB_UNHASH(cd);
  /*
   * Now get rid of entries associated with this dir
   */
  NASD_NFS_LOCK_DIR(cd,1);
  nasd_nfs_name_cleandir(cd);
  NASD_NFS_UNLOCK_DIR(cd);
  cd->last_page->next = nasd_nfs_free_dirpages;
  nasd_nfs_free_dirpages = cd->first_page;
  nasd_nfs_num_free_dirpages += cd->npages;

  if (cd->attr) {
    nasd_nfs_release_attr(cd->attr);
  }

got_dir:
  cd->npages = 0;
  cd->busy = 0;
  cd->first_page = cd->last_page = NULL;
  cd->lprev = cd->lnext = NULL;
  cd->hprev = cd->hnext = NULL;
  cd->dprev = cd->dnext = NULL;
  cd->attr = NULL;
  cd->refcnt = 1;
  cd->deleted = 0;
  cd->nentries = 0;
  cd->isroot = 0;
  bzero((char *)cd->buckets, sizeof(cd->buckets));
  bzero((char *)cd->idbuckets, sizeof(cd->idbuckets));
  nasd_nfs_dir_hash_check(0);
  if (dircache_lock_held == 0) {
    NASD_NFS_UNLOCK_DIRCACHE();
  }
  return(cd);
}

/*
 * call with dircache LOCKED
 */
void
nasd_nfs_dir_mk_freepages(
  int  npages)
{
  nasd_nfs_core_dir_t *cd_tmp;

  while (nasd_nfs_num_free_dirpages < npages) {
    NASD_NFS_CSINC(dirc_mkfree_loop);
    /*
     * Keep pulling dirs off the LRU until we have enough pages
     */
    cd_tmp = nasd_nfs_clean_dirs.lprev;
    while (cd_tmp == &nasd_nfs_clean_dirs) {
      NASD_NFS_CSINC(dirc_mkfree_wait);
      NASD_WAIT_COND(nasd_nfs_dir_clean_cond, nasd_nfs_dircache_lock);
      cd_tmp = nasd_nfs_clean_dirs.lprev;
    }
    nasd_nfs_coredir_disassoc(cd_tmp);
    SIGNAL_DIRC();
  }
}

/*
 * call with dircache UNLOCKED
 * call with dir UNLOCKED but BUSY
 */
nasd_status_t
nasd_nfs_load_dir(
  nasd_nfs_core_dir_t  *cd)
{
  int npages, i, pgbyte, pgbit, pgmask, exl, typ, h;
  nasd_mem_list_t memlist[NASD_NFS_MEMLIST_LEN];
  nasd_nfs_core_dir_t *r, *found;
  nasd_nfs_core_dirpage_t *cdp;
  nasd_error_string_t err_str;
  nasd_status_t nasd_status;
  nasd_rpc_status_t status;
  nasd_offset_t off, loff;
  nasd_len_t out_len, len;
  nasd_nfs_dirdata_t *dd;
  nasd_nfs_dirent_t *de;
  nasd_timespec_t ts;
  nasd_timer_t tm;

  NASD_ASSERT((cd->attr->attr_res.out_attribute.object_len
    % sizeof(nasd_nfs_dirpage_t)) == 0);
  npages = cd->attr->attr_res.out_attribute.object_len
    / sizeof(nasd_nfs_dirpage_t);

  NASD_NFS_CSINC(dir_load);

  NASD_NFS_LOCK_DIRCACHE();

  nasd_nfs_dir_hash_check(0);

  nasd_nfs_dir_mk_freepages(npages);

  cd->first_page = cd->last_page = nasd_nfs_free_dirpages;
  nasd_nfs_free_dirpages = cd->last_page->next;
  nasd_nfs_num_free_dirpages--;
  cd->npages = npages;
  cd->first_page->prev = NULL;
  cd->last_page->next = NULL;
  cd->last_page->offset = 0;
  for(i=1;i<npages;i++) {
    cd->last_page->next = nasd_nfs_free_dirpages;
    nasd_nfs_free_dirpages = nasd_nfs_free_dirpages->next;
    nasd_nfs_num_free_dirpages--;
    cd->last_page->next->prev = cd->last_page;
    cd->last_page = cd->last_page->next;
    cd->last_page->next = NULL;
    cd->last_page->offset = cd->last_page->prev->offset
      + sizeof(nasd_nfs_dirpage_t);
  }

  NASD_NFS_UNLOCK_DIRCACHE();

  /*
   * Now we have enough pages, load the bits
   */
  loff = off = 0;
  len = 0;
  for(i=0,cdp=cd->first_page;cdp;cdp=cdp->next) {
    memlist[i].addr = (void *)cdp->page;
    memlist[i].len = sizeof(nasd_nfs_dirpage_t);
    memlist[i].stride = 0;
    memlist[i].nelem = 1;
    memlist[i].next = &memlist[i+1];
    loff += sizeof(nasd_nfs_dirpage_t);
    len += sizeof(nasd_nfs_dirpage_t);
    i++;
    if (i >= NASD_NFS_MEMLIST_LEN) {
      i--;
      memlist[i].next = NULL;
      NASD_TM_START(&tm);
      nasd_cl_p_range_read(cd->drive->handle, cd->cookie.key,cd->cookie.capability_args,
        cd->drive->partnum, cd->nid, memlist, off, len,
        &out_len, &nasd_status, &status);
      NASD_TM_STOP(&tm);
      NASD_TM_ELAPSED_TS(&tm, &ts);
      NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
      if (nasd_status || status) {
        printf(
          "NFS: loading directory drive %s partnum %d nasd_status 0x%x (%s) status 0x%x (%s)\n",
          cd->drive->drname, cd->drive->partnum, nasd_status,
          nasd_error_string(nasd_status), status,
          nasd_cl_error_string(cd->drive->handle, status, err_str));
        NASD_PANIC();
      }
      NASD_ASSERT(out_len == len);
      off = loff;
      len = 0;
      i = 0;
    }
  }
  if (i) {
    i--;
    memlist[i].next = NULL;
    NASD_TM_START(&tm);
    nasd_cl_p_range_read(cd->drive->handle, cd->cookie.key, cd->cookie.capability_args,
      cd->drive->partnum, cd->nid, memlist, off, len,
      &out_len, &nasd_status, &status);
    NASD_TM_STOP(&tm);
    NASD_TM_ELAPSED_TS(&tm, &ts);
    NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
    if (nasd_status || status) {
      printf(
        "NFS: loading directory drive %s partnum %d nasd_status 0x%x (%s) status 0x%x (%s)\n",
        cd->drive->drname, cd->drive->partnum, nasd_status,
        nasd_error_string(nasd_status), status,
        nasd_cl_error_string(cd->drive->handle, status, err_str));
      NASD_PANIC();
    }
    NASD_ASSERT(out_len == len);
  }

  NASD_NFS_LOCK_DIRCACHE();

  /*
   * Now we have the bits. Parse them.
   */
  for(cdp=cd->first_page;cdp;cdp=cdp->next) {
    for(i=0;i<127;i++) {
      pgbyte = i/8;
      pgbit = i%8;
      pgmask = (1<<pgbit);
      if (cdp->page->header.usemap[pgbyte]&pgmask) {
        /* slot contains a valid entry */
        dd = &cdp->page->data_slots[i];
        exl = NASD_NFS_DIRD_EXLEN(dd);
        typ = NASD_NFS_DIRD_TYPEOF(dd);
        de = nasd_nfs_name_add(cd, cdp, dd);
        NASD_ASSERT(de != NULL);
        if (typ == NASD_NFS_DIRD_TYPE_DIR) {
          /* wire up to subdir, if present */
          h = nasd_nfs_obj_hash(dd->nasdid);
          found = NULL;
          for(r=nasd_nfs_obj_buckets[h].hnext;r!=&nasd_nfs_obj_buckets[h];
            r=r->hnext)
          {
            if ((r->nid == dd->nasdid) && (r->di == cd->di)) {
              found = r;
              break;
            }
          }
          if (found) {
            /* subdir is in core */
            de->dir = found;
            if (found->attr) {
              de->attr = found->attr;
              de->attr->refcnt++;
            }
          }
        }
        /* skip extra name slots */
        i += exl;
      }
    }
  }

  NASD_NFS_UNLOCK_DIRCACHE();

  return(NASD_SUCCESS);
}

void
nasd_nfs_post_attribute(
  nasd_nfs_core_dir_t    *cd,
  nasd_nfs_identifier_t   in_identifier,
  nasd_post_attribute_t  *post_attribute)
{
  NASD_NFS_CSINC(post_attribute);
#if 0
  post_attribute->valid = NASD_ATTRIBUTE_INVALID;
  NASD_NFS_CSINC(post_attribute_invalid);
#else
  if (cd->attr) {
    post_attribute->valid = NASD_ATTRIBUTE_VALID;
    post_attribute->attribute = cd->attr->attr_res.out_attribute;
    NASD_NFS_CSINC(post_attribute_valid);
  }
  else {
    post_attribute->valid = NASD_ATTRIBUTE_INVALID;
    NASD_NFS_CSINC(post_attribute_invalid);
  }
#endif
}

nasd_status_t
nasd_nfs_lookup_post_attribute(
  nasd_nfs_identifier_t   in_identifier,
  nasd_post_attribute_t  *post_attribute)
{
  /*
   * XXX lookup + set this while we're here
   */
  NASD_NFS_CSINC(lookup_post_attribute);
  post_attribute->valid = NASD_ATTRIBUTE_INVALID;
  return(NASD_SUCCESS);
}

/* call with dir LOCKED */
nasd_status_t
nasd_nfs_dir_markerverf(
  nasd_nfs_core_dir_t  *dir,
  nasd_markerverf_t    *mvp)
{
  nasd_markerverf_t m;
  nasd_status_t rc;

  m = 0;
  if (dir->attr == NULL) {
    rc = nasd_nfs_dir_load_attr(dir);
    if (rc)
      return(rc);
  }
  /* XXX eventually, this should be the disk-maintained stamp, not fs */
  m = NASD_NFS_ATTR_ATTR(dir->attr).object_modify_time.ts_sec;
  m <<= 32;
  m |= ((nasd_markerverf_t)NASD_NFS_ATTR_ATTR(dir->attr).object_modify_time.ts_nsec)&0xffffffff;
  *mvp = m;
  return(NASD_SUCCESS);
}

/*
 * call with dircache LOCKED
 * call with dir LOCKED
 *
 * deallocates de
 */
nasd_status_t
nasd_nfs_dir_remove_entry(
  nasd_nfs_core_dir_t  *dir,
  nasd_nfs_dirent_t    *de)
{
  nasd_nfs_core_dirpage_t *cdp;
  int slot, nentries, i, st;
  nasd_status_t rc;
  unsigned long dd;

  nentries = NASD_NFS_DIRD_EXLEN(de->data) + 1;
  cdp = de->page;
  /*
   * I guess I'm really sorry about the following
   * couple lines of code. --jimz
   */
  dd = (unsigned long)de->data;
  dd -= (unsigned long)&cdp->page->data_slots[0];
  dd /= sizeof(nasd_nfs_dirdata_t);
  slot = dd;
  NASD_ASSERT(&cdp->page->data_slots[slot] == de->data);
  st = slot+nentries;

  /*
   * Remove the name from the in-core hash
   */
  rc = nasd_nfs_name_remove(dir, de, 0);
  if (rc) {
    NASD_PANIC();
  }

  /*
   * Remove the entry from the on-disk structure
   * (just turn off the valid bits for the entry
   * slots it occupies)
   */
  nasd_nfs_dir_dirty_page(dir, cdp);
  for(i=slot;i<st;i++) {
    NASD_ASSERT(cdp->page->header.usemap[i/8]&(1<<(i%8)));
    cdp->page->header.usemap[i/8] &= ~(1<<(i%8));
  }
  cdp->page->header.freecnt += nentries;

  return(NASD_SUCCESS);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
