/*
 * nasd_nfs_drive.c
 *
 * Interfaces to NASD drive for NASD NFS server.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_control.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_nfs_internal.h>
#include <nasd/nasd_timer.h>

#include <sys/errno.h>

NASD_DECLARE_MUTEX(nasd_nfs_mount_lock)
#define LOCK_MOUNT()   NASD_LOCK_MUTEX(nasd_nfs_mount_lock)
#define UNLOCK_MOUNT() NASD_UNLOCK_MUTEX(nasd_nfs_mount_lock)

nasd_nfs_drive_t *nasd_nfs_mountbuckets[NASD_NFS_NAMEHASH_BUCKETS];
nasd_nfs_drive_t *nasd_nfs_dimountbuckets[NASD_NFS_DIHASH_BUCKETS];

#define NASD_NFS_DISKIDENT_BASE_FIRST 1
nasd_disk_ident_t nasd_nfs_diskident_base = NASD_NFS_DISKIDENT_BASE_FIRST;
nasd_cookie_t mcookie; /* Current cookie we use for our accesses */

nasd_status_t
nasd_nfs_drsys_init()
{
  nasd_status_t rc;

  bzero((char *)nasd_nfs_mountbuckets, sizeof(nasd_nfs_mountbuckets));
  bzero((char *)nasd_nfs_dimountbuckets, sizeof(nasd_nfs_dimountbuckets));
  nasd_nfs_diskident_base = NASD_NFS_DISKIDENT_BASE_FIRST;

  rc = nasd_mutex_init(&nasd_nfs_mount_lock);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_mutex,
    &nasd_nfs_mount_lock);
  if (rc) {
    nasd_mutex_destroy(&nasd_nfs_mount_lock);
    return(rc);
  }

  return(NASD_SUCCESS);
}

/*
 * Called when shutting down to re-mark directories as "clean"
 */
nasd_status_t
nasd_nfs_mark_dirs_clean()
{
  return(NASD_SUCCESS);
}

/*
 * Clean up one mounted drive
 */
void
nasd_nfs_shutdown_drive(
  nasd_nfs_drive_t *dr)
{
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_timespec_t ts;
  nasd_status_t rc;
  nasd_timer_t tm;

  nasd_nfs_name_cleandir(dr->root);
  nasd_nfs_release_attr(dr->root->attr);
  dr->root->attr = NULL;
  nasd_nfs_ditch_coredir(dr->root, 0);
  NASD_TM_START(&tm);
  nasd_cl_p_sync_dr(dr->handle, &rc, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (rc || status) {
    printf("ERROR: status=0x%x (%s) nasd_status=0x%x (%s) on sync of %s\n",
      status, nasd_cl_error_string(dr->handle, status, err_str),
      rc, nasd_error_string(rc), dr->drname);
  }
  nasd_unbind_drive(&dr->handle);
  NASD_Free(dr, sizeof(nasd_nfs_drive_t));
}

/*
 * Cleanup mount table state
 */
void
nasd_nfs_shutdown_drives()
{
  nasd_nfs_drive_t *dr, *next;
  int i;

  for(i=0;i<NASD_NFS_NAMEHASH_BUCKETS;i++) {
    for(dr=nasd_nfs_mountbuckets[i];dr;dr=next) {
      next = dr->next;
      /* cleanup drive dr */
      nasd_nfs_shutdown_drive(dr);
    }
    nasd_nfs_mountbuckets[i] = NULL;
  }
}

/*
 * Note that this is not protected against simultaneous mount
 * adds. If mounts are dynamically added, the dynamic adder
 * should just serialize all such requests.
 */
nasd_status_t
nasd_nfs_add_mount(
  char         *path,
  char         *drive,
  char         *drive_bind_name,
  int           portnum,
  int           partnum,
  nasd_uint32   net_addr)
{
  nasd_nfs3_attributes_t *nfs3attr;
  nasd_ctrl_drive_info_t *drinfo;
  nasd_ctrl_part_info_t *ptinfo;
  nasd_p_getattr_dr_args_t args;
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_nfs_core_dir_t *cd;
  nasd_drive_handle_t dh;
  nasd_nfs_dirent_t *de;
  nasd_cookie_t cookie;
  nasd_nfs_drive_t *dr;
  nasd_timespec_t ts;
  nasd_status_t rc;
  nasd_timer_t tm;
  int h, l, ret;

  printf("Add mount %s from %s partnum %d\n", path, drive, partnum);

  rc = nasd_nfs_do_drive_bind(drive_bind_name, &dh);
  if (rc) {
    printf("NFS: cannot bind to %s\n", drive);
    return(rc);
  }
  NASD_TM_START(&tm);
  nasd_cl_p_null_dr(dh, &rc, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status || rc) {
    printf("NFS: cannot talk to %s status=0x%x (%s) nasd_status=0x%x (%s)\n",
      drive, status, nasd_cl_error_string(dh, status, err_str),
      rc, nasd_error_string(rc));
    nasd_unbind_drive(&dh);
    return(NASD_OFFLINE);
  }
  NASD_Malloc(dr,sizeof(nasd_nfs_drive_t),(nasd_nfs_drive_t *));
  if (dr == NULL) {
    nasd_unbind_drive(&dh);
    return(NASD_NO_MEM);
  }
  dr->handle = dh;

  LOCK_MOUNT();
  dr->di = nasd_nfs_diskident_base;
  nasd_nfs_diskident_base++;
  UNLOCK_MOUNT();

  NASD_Malloc(ptinfo, sizeof(nasd_ctrl_part_info_t),
    (nasd_ctrl_part_info_t *));
  if (ptinfo == NULL) {
    printf("NFS: could not allocate info page to mount %s\n", drive);
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(NASD_NO_MEM);
  }
  drinfo = (nasd_ctrl_drive_info_t *)ptinfo;
  /* @@@ init cookie here */
  NASD_TM_START(&tm);
  rc = nasd_cl_p_ctrl_get_drive_info(dh, cookie.key, cookie.capability_args,
    drinfo);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (rc) {
    printf("NFS: nasd_cl_p_get_drive_info got 0x%x (%s)\n",
      rc, nasd_error_string(rc));
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(rc);
  }
  dr->blocksize = drinfo->blocksize;
  /* @@@ init cookie here */
  NASD_TM_START(&tm);
  rc = nasd_cl_p_ctrl_get_part_info(dh, cookie.key, cookie.capability_args,
    partnum, ptinfo);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (rc) {
    printf("NFS: nasd_cl_p_get_part_info got 0x%x (%s)\n",
      rc, nasd_error_string(rc));
    if ((rc == NASD_CTRL_ID_CHECK_FAILED) && (ptinfo->ctrl_id == 0)) {
      printf(
        "NFS: got bad control id 0, suggesting partition %d does not exist on %s\n",
        partnum, drive);
      return(NASD_FAIL);
    }
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(rc);
  }

  cd = nasd_nfs_dir_get_coredir(0);
  NASD_ASSERT(cd != NULL);

  dr->root = cd;
  dr->portnum = portnum;
  dr->net_addr = net_addr;
  dr->partnum = partnum;
  l = NASD_MIN(NASD_NFS_DRIVENAME_LEN-1, strlen(drive));
  bcopy(drive, dr->drname, l);
  dr->drname[l] = '\0';
  l = NASD_MIN(NASD_NFS_DRPATH_LEN-1, strlen(path));
  bcopy(path, dr->pname, l);
  dr->pname[l] = '\0';

  cd->drive = dr;
  cd->nid = ptinfo->first_obj;
  cd->di = dr->di;
  /* @@@ init cd->cookie here */
  NASD_ASSERT(cd->refcnt == 1);
  cd->busy = 1;
  cd->ndirty = 0;
  cd->attr = nasd_nfs_get_attr();
  NASD_ASSERT(cd->attr != NULL);
  cd->isroot = 1;

  NASD_Free(ptinfo, sizeof(nasd_ctrl_part_info_t));

  args.in_partnum = partnum;
  args.in_identifier = cd->nid;
  args.in_capability_args=cookie.capability_args;
  NASD_TM_START(&tm);
  nasd_cl_p_getattr_dr(dh, cookie.key,&args, &cd->attr->attr_res, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status || cd->attr->attr_res.nasd_status) {
    printf(
      "NFS: cannot get attribute root %s nid=0x%" NASD_ID_FMT " status=0x%x (%s) nasd_status=0x%x (%s)\n",
      drive, (unsigned long)cd->nid,
      status, nasd_cl_error_string(dh, status, err_str),
      cd->attr->attr_res.nasd_status,
      nasd_error_string(cd->attr->attr_res.nasd_status));
    nasd_nfs_release_attr(cd->attr);
    cd->attr = NULL;
    nasd_nfs_ditch_coredir(cd, 0);
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(rc);
  }

  if (cd->attr->attr_res.out_attribute.object_len%sizeof(nasd_nfs_dirpage_t)) {
    printf("NFS: root object drive %s partition %d has bad length %lu\n",
      drive, partnum, (u_long)cd->attr->attr_res.out_attribute.object_len);
    nasd_nfs_release_attr(cd->attr);
    cd->attr = NULL;
    nasd_nfs_ditch_coredir(cd, 0);
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(NASD_FAIL);
  }

  nfs3attr =
    (nasd_nfs3_attributes_t *)NASD_NFS_ATTR_ATTR(cd->attr).fs_specific;
  if (nfs3attr->clean != NASD_NFS3_CLEAN) {
    printf("NFS: filesystem on %s partition %d is not clean\n",
      drive, partnum);
    nasd_nfs_release_attr(cd->attr);
    cd->attr = NULL;
    nasd_nfs_ditch_coredir(cd, 0);
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(NASD_FAIL);
  }
  if (nfs3attr->type != NF3DIR) {
    printf("NFS: filesystem on %s partition %d is not dir-rooted!\n",
      drive, partnum);
    nasd_nfs_release_attr(cd->attr);
    cd->attr = NULL;
    nasd_nfs_ditch_coredir(cd, 0);
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(NASD_FAIL);
  }

  NASD_NFS_LOCK_DIRCACHE();
  NASD_NFS_LOCK_DIR(cd,1);

  nasd_nfs_dir_hash_ins(cd);

  NASD_NFS_UNLOCK_DIR(cd);
  NASD_NFS_UNLOCK_DIRCACHE();

  rc = nasd_nfs_load_dir(cd);
  if (rc != NASD_SUCCESS) {
    printf("NFS: got 0x%x (%s) loading %s@%s (%s)\n", rc,
      nasd_error_string(rc), path, drive, drive_bind_name);
    cd->hnext->hprev = cd->hprev;
    cd->hprev->hnext = cd->hnext;
    nasd_nfs_name_cleandir(cd);
    nasd_nfs_release_attr(cd->attr);
    cd->attr = NULL;
    nasd_nfs_ditch_coredir(cd, 0);
    nasd_unbind_drive(&dh);
    NASD_Free(dr, sizeof(nasd_nfs_drive_t));
    return(NASD_FAIL);
  }

  /*
   * XXX mark directory dirty here
   */

  NASD_NFS_LOCK_DIRCACHE();
  NASD_NFS_LOCK_DIR(cd,1);

  cd->busy = 0;
  NASD_BROADCAST_COND(cd->cond);

  NASD_NFS_UNLOCK_DIR(cd);
  nasd_nfs_dir_release(cd);
  NASD_NFS_UNLOCK_DIRCACHE();

  LOCK_MOUNT();

  h = nasd_nfs_name_hash(dr->pname);
  dr->next = nasd_nfs_mountbuckets[h];
  nasd_nfs_mountbuckets[h] = dr;

  h = NASD_NFS_DIHASH(dr->di);
  dr->dinext = nasd_nfs_dimountbuckets[h];
  nasd_nfs_dimountbuckets[h] = dr;

  UNLOCK_MOUNT();

  return(NASD_SUCCESS);
}

nasd_status_t
nasd_nfs_mount_lookup(
  char                   *dirpath,
  nasd_list_len_t         in_listlen,
  nasd_net_address_t      out_drivelist[],
  nasd_list_len_t        *out_listlen,
  nasd_nfs_identifier_t  *out_identifier)
{
  nasd_nfs_drive_t *r, *dr;
  nasd_status_t rc;
  int h;

  if (in_listlen < 1) {
    return(NASD_BAD_LEN);
  }

  rc = NASD_SUCCESS;
  LOCK_MOUNT();

  h = nasd_nfs_name_hash(dirpath);
  for(dr=NULL,r=nasd_nfs_mountbuckets[h];r;r=r->next) {
    if (!strcmp(dirpath, r->pname)) {
      dr = r;
      break;
    }
  }
  if (dr == NULL) {
    rc = NASD_NFS_BAD_NAME;
    goto done;
  }
  out_drivelist[0].disk_identifier = dr->di;
  out_drivelist[0].network_address = dr->net_addr;
  out_drivelist[0].port_number = dr->portnum;
  out_drivelist[0].partnum = dr->partnum;
  *out_listlen = 1;
  out_identifier->disk_identifier = dr->di;
  out_identifier->nasd_identifier = dr->root->nid;
  out_identifier->partnum=dr->partnum;
  out_identifier->pad1 = 0;


done:
  UNLOCK_MOUNT();
  return(rc);
}

#define NASD_NFS_DIR_WRITER_FLS() { \
  nasd_timespec_t _ts; \
  nasd_timer_t _tm; \
  NASD_NFS_CSINC(dir_writer_fls); \
  memlist[i-1].next = NULL; \
  NASD_TM_START(&_tm); \
  nasd_cl_p_range_write(dir->drive->handle, dir->cookie.key, dir->cookie.capability_args, \
    dir->drive->partnum, dir->nid, memlist, off, len, \
    &out_len, &nasd_status, &status); \
  NASD_TM_STOP(&_tm); \
  NASD_TM_ELAPSED_TS(&_tm, &_ts); \
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &_ts); \
  if (nasd_status || status) { \
    printf( \
      "NFS: writing directory drive %s partnum %d nasd_status 0x%x (%s) status 0x%x (%s)\n", \
      dir->drive->drname, dir->drive->partnum, nasd_status, \
      nasd_error_string(nasd_status), status, \
      nasd_cl_error_string(dir->drive->handle, status, err_str)); \
    NASD_PANIC(); \
  } \
  if (out_len != len) { \
    printf( \
      "NFS: writing directory drive %s partnum %d, got partial %d/%d\n", \
      dir->drive->drname, dir->drive->partnum, (int)out_len, (int)len); \
    NASD_PANIC(); \
  } \
  if (dir->attr) { \
    nasd_status = nasd_nfs_dir_load_attr(dir); \
    if (nasd_status) { \
      printf("NFS: failed attribute readback\n"); \
    } \
  } \
  i = 0; \
  len = 0; \
  next_off = 0; \
  off = 0; \
}

/*
 * call with dir already marked BUSY
 */
nasd_status_t
nasd_nfs_dir_write_dirty(
  nasd_nfs_core_dir_t  *dir)
{
  nasd_mem_list_t memlist[NASD_NFS_MEMLIST_LEN];
  nasd_nfs_core_dirpage_t *cdp;
  nasd_offset_t next_off, off;
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_status_t nasd_status;
  nasd_len_t len, out_len;
  int i;

  i = 0;
  len = 0;
  next_off = 0;
  off = 0;
  for(cdp=dir->first_page;cdp;cdp=cdp->next) {
    if (cdp->dirty == 0)
      continue;
    if ((i != 0) && (cdp->offset != next_off)) {
      NASD_NFS_DIR_WRITER_FLS();
    }
    if (i == 0)
      off = cdp->offset;
    NASD_NFS_CSINC(dir_writer_pages);
    memlist[i].addr = (void *)cdp->page;
    memlist[i].len = sizeof(nasd_nfs_dirpage_t);
    memlist[i].stride = 0;
    memlist[i].nelem = 1;
    memlist[i].next = &memlist[i+1];
    next_off += sizeof(nasd_nfs_dirpage_t);
    len += sizeof(nasd_nfs_dirpage_t);
    i++;
    if (i >= NASD_NFS_MEMLIST_LEN) {
      NASD_NFS_DIR_WRITER_FLS();
    }
    cdp->dirty = 0;
  }
  if (i) {
    NASD_NFS_DIR_WRITER_FLS();
  }
  dir->ndirty = 0;

  return(NASD_SUCCESS);
}

/* call with dir LOCKED */
nasd_status_t
nasd_nfs_dir_load_attr(
  nasd_nfs_core_dir_t  *dir)
{
  nasd_p_getattr_dr_args_t args;
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_drive_handle_t dh;
  nasd_timespec_t ts;
  nasd_status_t rc;
  nasd_timer_t tm;

  NASD_NFS_CSINC(dir_load_attr);
  if (dir->attr == NULL) {
    dir->attr = nasd_nfs_get_attr();
  }
  dh = dir->drive->handle;
  args.in_partnum = dir->drive->partnum;
  args.in_identifier = dir->nid;
  NASD_TM_START(&tm);
  nasd_cl_p_getattr_dr(dh, mcookie.key,&args, &dir->attr->attr_res, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status || dir->attr->attr_res.nasd_status) {
    printf(
      "NFS: cannot get attribute drive %s nid=0x%" NASD_ID_FMT " status=0x%x (%s) nasd_status=0x%x (%s)\n",
      dir->drive->drname, (unsigned long)dir->nid, status,
      nasd_cl_error_string(dh, status, err_str),
      dir->attr->attr_res.nasd_status,
      nasd_error_string(dir->attr->attr_res.nasd_status));
    nasd_nfs_release_attr(dir->attr);
    dir->attr = NULL;
    return(NASD_FAIL);
  }
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_nfs_do_getattr(
  nasd_nfs_drive_t         *dr,
  nasd_identifier_t         nasdid,
  nasd_p_getattr_dr_res_t  *ga_res)
{
  nasd_p_getattr_dr_args_t args;
  nasd_rpc_status_t status;
  nasd_timespec_t ts;
  nasd_timer_t tm;

  NASD_NFS_CSINC(do_getattr);
  args.in_partnum = dr->partnum;
  args.in_identifier = nasdid;
  NASD_TM_START(&tm);
  nasd_cl_p_getattr_dr(dr->handle, mcookie.key,&args, ga_res, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status)
    return(NASD_FAIL);
  return(ga_res->nasd_status);
}

nasd_status_t
nasd_nfs_do_setattr(
  nasd_nfs_drive_t          *dr,
  nasd_p_setattr_dr_args_t  *args,
  nasd_p_setattr_dr_res_t   *res,
  nasd_byte_t               *args_otw,
  nasd_byte_t               *res_otw)
{
  nasd_rpc_status_t status;
  nasd_timespec_t ts;
  nasd_timer_t tm;

  NASD_NFS_CSINC(do_setattr);

  /* @@@ compute cookie */
  bzero((char *)&args->in_capability_args, sizeof(nasd_capability_args_t));

  NASD_TM_START(&tm);
  nasd_cl_p_setattr_dr__otw_provided(dr->handle, mcookie.key,
    args, res, args_otw, res_otw, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status)
    return(NASD_FAIL);
  return(res->nasd_status);
}

nasd_status_t
nasd_nfs_get_drive(
  nasd_nfs_identifier_t    in_identifier,
  nasd_nfs_drive_t       **drp)
{
  nasd_nfs_drive_t *r;
  nasd_status_t rc;
  int h;

  rc = NASD_BAD_IDENTIFIER;
  *drp = NULL;
  LOCK_MOUNT();

  h = NASD_NFS_DIHASH(in_identifier.disk_identifier);
  for(r=nasd_nfs_dimountbuckets[h];r;r=r->dinext) {
    if (r->di == in_identifier.disk_identifier) {
      *drp = r;
      rc = NASD_SUCCESS;
      break;
    }
  }

  UNLOCK_MOUNT();
  return(rc);
}

nasd_status_t
nasd_nfs_do_create(
  nasd_nfs_drive_t         *dr,
  nasd_p_create_dr_args_t  *cr_args,
  nasd_p_create_dr_res_t   *cr_res,
  nasd_byte_t              *cr_args_otw,
  nasd_byte_t              *cr_res_otw)
{
  nasd_rpc_status_t status;
  nasd_timespec_t ts;
  nasd_timer_t tm;

  /* @@@ cons up cookie for create or cache in drive */
  bzero((char *)&cr_args->in_capability_args, sizeof(nasd_capability_args_t));

  cr_args->in_partnum = dr->partnum;

  NASD_TM_START(&tm);
  nasd_cl_p_create_dr__otw_provided(dr->handle, mcookie.key,
    cr_args, cr_res, cr_args_otw, cr_res_otw, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status)
    return(NASD_FAIL);
  if (cr_res->nasd_status)
    return(cr_res->nasd_status);

  return(NASD_SUCCESS);
}

nasd_status_t
nasd_nfs_do_remove(
  nasd_nfs_drive_t   *dr,
  nasd_identifier_t   nasdid)
{
  nasd_p_remove_dr_args_t rm_args;
  nasd_p_remove_dr_res_t rm_res;
  nasd_rpc_status_t status;
  nasd_timespec_t ts;
  nasd_timer_t tm;

  /* @@@ cons up or cache cookie in dr */
  bzero((char *)&rm_args.in_capability_args, sizeof(nasd_capability_args_t));
  rm_args.in_identifier = nasdid;
  rm_args.in_partnum = dr->partnum;

  NASD_TM_START(&tm);
  nasd_cl_p_remove_dr(dr->handle,mcookie.key, &rm_args, &rm_res, &status);
  NASD_TM_STOP(&tm);
  NASD_TM_ELAPSED_TS(&tm, &ts);
  NASD_ATOMIC_TIMESPEC_ADD(&nasd_nfs_cache_stats.drive_time, &ts);
  if (status)
    return(NASD_FAIL);

  return(rm_res.nasd_status);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
