/*
 * nfs_name.c
 *
 * Name management for NASD NFS.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_nfs_internal.h>

nasd_freelist_t *nasd_nfs_dirent_freelist = NULL;
#define NASD_MAX_FREE_DIRENT 1536
#define NASD_DIRENT_INC        64
#define NASD_DIRENT_INITIAL   768

nasd_freelist_t *nasd_nfs_attr_freelist = NULL;
#define NASD_MAX_FREE_ATTR 2048
#define NASD_ATTR_INC       256
#define NASD_ATTR_INITIAL   512

#define GET_DIRENT(_d_) \
  NASD_FREELIST_GET(nasd_nfs_dirent_freelist,_d_,hnext,(nasd_nfs_dirent_t *))

/*
 * Low 5 bits of each char
 */

int
nasd_nfs_name_hash(
  char  *str)
{
  int l, i;
  long n;

  n = 0;
  l = strlen(str);
  if (l > 10)
    l = 10;
  for(i=0;i<l;i++) {
    n |= str[i]&0x1f;
    n <<= 5;
  }
  n %= NASD_NFS_NAMEHASH_BUCKETS;
  if(n < 0)
    n = -n;
  return((int)n);
}

void
nasd_nfs_name_shutdown_freelist(
  void  *arg)
{
  NASD_FREELIST_DESTROY(nasd_nfs_dirent_freelist,hnext,(nasd_nfs_dirent_t *));
}

void
nasd_nfs_attr_shutdown_freelist(
  void  *arg)
{
  NASD_FREELIST_DESTROY(nasd_nfs_attr_freelist,next,(nasd_nfs_attr_t *));
}

nasd_status_t
nasd_nfs_name_init()
{
  nasd_status_t rc;

  NASD_FREELIST_CREATE(nasd_nfs_dirent_freelist, NASD_MAX_FREE_DIRENT,
    NASD_DIRENT_INC, sizeof(nasd_nfs_dirent_t));
  if (nasd_nfs_dirent_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_name_shutdown_freelist,
    NULL);
  if (rc) {
    nasd_nfs_name_shutdown_freelist(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_nfs_dirent_freelist, NASD_DIRENT_INITIAL,hnext,
    (nasd_nfs_dirent_t *));

  NASD_FREELIST_CREATE(nasd_nfs_attr_freelist, NASD_MAX_FREE_ATTR,
    NASD_ATTR_INC, sizeof(nasd_nfs_attr_t));
  if (nasd_nfs_attr_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_attr_shutdown_freelist,
    NULL);
  if (rc) {
    nasd_nfs_attr_shutdown_freelist(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_nfs_attr_freelist, NASD_ATTR_INITIAL,next,
    (nasd_nfs_attr_t *));

  return(NASD_SUCCESS);
}

/*
 * call with dir LOCKED, dircache LOCKED
 * discard all namecache entries for dir
 */
void
nasd_nfs_name_cleandir(
  nasd_nfs_core_dir_t  *dir)
{
  nasd_nfs_dirent_t *de;
  int i;

  NASD_FREELIST_DO_LOCK(nasd_nfs_dirent_freelist);
  for(i=0;i<NASD_NFS_NAMEHASH_BUCKETS;i++) {
    while(dir->buckets[i]) {
      de = dir->buckets[i];
      dir->buckets[i] = de->hnext;
      if (de->attr) {
        nasd_nfs_release_attr(de->attr);
        de->attr = NULL;
      }
      NASD_FREELIST_FREE_NOLOCKING(nasd_nfs_dirent_freelist,de,hnext);
    }
  }
  /* already freed entries above */
#if 0
  /*
   * This is actually extraneous, since nasd_nfs_dir_get_coredir()
   * will zero the buckets. Callers of this function should be aware
   * that the resulting state of dir is inconsistent, however.
   */
  bzero((char *)dir->idbuckets, sizeof(dir->idbuckets));
#endif
  NASD_FREELIST_DO_UNLOCK(nasd_nfs_dirent_freelist);
}

/* call with dir LOCKED */
nasd_status_t
nasd_nfs_name_hash_ins(
  nasd_nfs_core_dir_t  *dir,
  nasd_nfs_dirent_t    *de)
{
  int h;

  h = nasd_nfs_name_hash(de->name);
  de->hprev = NULL;
  de->hnext = dir->buckets[h];
  if (de->hnext)
    de->hnext->hprev = de;
  dir->buckets[h] = de;

  h = nasd_nfs_obj_hash(de->data->nasdid);
  de->ihprev = NULL;
  de->ihnext = dir->idbuckets[h];
  if (de->ihnext)
    de->ihnext->ihprev = de;
  dir->idbuckets[h] = de;

  return(NASD_SUCCESS);
}

/* call with dir LOCKED */
nasd_nfs_dirent_t *
nasd_nfs_name_add(
  nasd_nfs_core_dir_t      *dir,
  nasd_nfs_core_dirpage_t  *page,
  nasd_nfs_dirdata_t       *data)
{
  nasd_nfs_dirent_t *de;
  int h;

  GET_DIRENT(de);
  if (de == NULL)
    return(NULL);
  de->name = data->name;
  de->page = page;
  de->data = data;
  de->attr = NULL;
  de->dir = NULL;

  h = nasd_nfs_name_hash(de->name);
  de->hprev = NULL;
  de->hnext = dir->buckets[h];
  if (de->hnext)
    de->hnext->hprev = de;
  dir->buckets[h] = de;

  h = nasd_nfs_obj_hash(de->data->nasdid);
  de->ihprev = NULL;
  de->ihnext = dir->idbuckets[h];
  if (de->ihnext)
    de->ihnext->ihprev = de;
  dir->idbuckets[h] = de;

  dir->nentries++;

  return(de);
}

/* call with dir LOCKED */
nasd_nfs_dirent_t *
nasd_nfs_name_lookup(
  nasd_nfs_core_dir_t  *dir,
  char                 *name)
{
  nasd_nfs_dirent_t *r;
  int h, l;

  h = nasd_nfs_name_hash(name);
  l = strlen(name);
  for(r=dir->buckets[h];r;r=r->hnext) {
    if (r->data->name_len != l)
      continue;
    if (!bcmp(r->name, name, l))
      return(r);
  }
  return(NULL);
}

/*
 * call with dir LOCKED
 *
 * deallocates de
 */
nasd_status_t
nasd_nfs_name_remove(
  nasd_nfs_core_dir_t  *dir,
  nasd_nfs_dirent_t    *de,
  int                   unhash_only)
{
  int h;

  h = nasd_nfs_name_hash(de->name);
  if (de->hprev) {
    de->hprev->hnext = de->hnext;
  }
  else {
    NASD_ASSERT(dir->buckets[h] == de);
    dir->buckets[h] = de->hnext;
    if (dir->buckets[h]) {
      dir->buckets[h]->hprev = NULL;
    }
  }
  if (de->hnext) {
    de->hnext->hprev = de->hprev;
  }
  de->hnext = NULL;
  de->hprev = NULL;

  h = nasd_nfs_obj_hash(de->data->nasdid);
  if (de->ihprev) {
    de->ihprev->ihnext = de->ihnext;
  }
  else {
    NASD_ASSERT(dir->idbuckets[h] == de);
    dir->idbuckets[h] = de->ihnext;
    if (dir->idbuckets[h]) {
      dir->idbuckets[h]->ihprev = NULL;
    }
  }
  if (de->ihnext) {
    de->ihnext->ihprev = de->ihprev;
  }
  de->ihnext = NULL;
  de->ihprev = NULL;

  if (unhash_only == 0) {
    if (de->attr) {
      nasd_nfs_release_attr(de->attr);
      de->attr = NULL;
    }
    NASD_FREELIST_FREE(nasd_nfs_dirent_freelist,de,hnext);
    dir->nentries--;
  }

  return(NASD_SUCCESS);
}

nasd_nfs_attr_t *
nasd_nfs_get_attr()
{
  nasd_nfs_attr_t *attr;

  NASD_FREELIST_GET(nasd_nfs_attr_freelist,attr,next,(nasd_nfs_attr_t *));
  attr->refcnt = 1;
  attr->next = NULL;
  return(attr);
}

/*
 * call with dircache lock HELD
 */
void
nasd_nfs_release_attr(
  nasd_nfs_attr_t  *attr)
{
  attr->refcnt--;
  if (attr->refcnt == 0) {
    NASD_FREELIST_FREE(nasd_nfs_attr_freelist,attr,next);
  }
}

void
nasd_nfs_debug_name_cache(
  nasd_nfs_core_dir_t  *dir)
{
  nasd_nfs_dirent_t *rr, *ss, *r1;
  unsigned long slot;
  int h;

  for(h=0;h<NASD_NFS_NAMEHASH_BUCKETS;h++) {
    for(ss=NULL,rr=dir->buckets[h];rr;ss=rr,rr=rr->hnext) {
      if (ss) {
        NASD_ASSERT(ss->hnext == rr);
        if (rr->hprev != ss) {
          printf("caught rr->hprev != ss\n");
          for(r1=dir->buckets[h];r1;r1=r1->hnext) {
            slot = (unsigned long)r1->data;
            slot -= (unsigned long)&r1->page->page->data_slots[0];
            slot /= sizeof(nasd_nfs_dirdata_t);
            printf("[ 0x%16lx 0x%16lx 0x%16lx %s slot %03lu slotused=%d]\n",
              (unsigned long)r1->hprev, (unsigned long)r1,
              (unsigned long)r1->hnext, r1->name, slot,
              (r1->page->page->header.usemap[slot/8]&(1<<(slot%8)))?1:0);
          }
        }
        NASD_ASSERT(rr->hprev == ss);
      }
      else {
        NASD_ASSERT(dir->buckets[h] == rr);
        NASD_ASSERT(rr->hprev == NULL);
      }
    }
  }
}

nasd_status_t
nasd_nfs_valid_name(
  char  *name)
{
  int i, len;

  len = strlen(name);
  if (len == 0)
    return(NASD_NFS_BAD_NAME);
  if (len > NASD_NFS_MAX_NAME_LEN)
    return(NASD_NFS_BAD_NAME);

  for(i=0;i<len;i++) {
    if (name[i] == '/')
      return(NASD_NFS_BAD_NAME);
    if (name[i]&0x80)
      return(NASD_NFS_BAD_NAME);
    if (name[i] == 127)
      return(NASD_NFS_BAD_NAME);
    if (name[i] < ' ')
      return(NASD_NFS_BAD_NAME);
  }

  return(NASD_SUCCESS);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
