/*
 * nasd_nfs_format.c
 *
 * Format a NASD partition for NFS.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_control.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_nfs_internal.h>

char *progname;

extern char *optarg;
extern int   optind;

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername partnum\n", progname);
  fprintf(stderr, "  No options supported.\n");
  fprintf(stderr, "  We have eliminated your need to make decisions.\n");
  fprintf(stderr, "  We have eliminated your need for choice.\n");
  fprintf(stderr, "  Have a nice day. (You have no choice.)\n");
  fflush(stderr);
  exit(1);
}

int
main(
  int     argc,
  char  **argv)
{
  nasd_nfs3_attributes_t *nfs3attr;
  nasd_p_smpl_op_dr_args_t wr_args;
  nasd_p_fastwrite_dr_res_t wr_res;
  nasd_p_getattr_dr_args_t ga_args;
  nasd_p_setattr_dr_args_t sa_args;
  nasd_p_getattr_dr_res_t ga_res;
  nasd_p_setattr_dr_res_t sa_res;
  nasd_ctrl_part_info_t ptinfo;
  nasd_nfs_dirpage_t dirpage;
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_attribute_t *attr;
  nasd_drive_handle_t h;
  nasd_identifier_t nid;
  nasd_cookie_t cookie;
  char c, *server_name;
  nasd_len_t out_len;
  nasd_status_t rc;
  int partnum;

  progname = argv[0];
  partnum = (-1);

  while ((c = getopt(argc, argv, "p:")) != EOF) {
    switch(c) {
      case 'p':
        if (sscanf(optarg, "%d", &partnum) != 1)
          usage();
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }
  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;
  if (optind >= argc)
    usage();
  if (sscanf(argv[optind], "%d", &partnum) != 1) {
    fprintf(stderr, "\"%s\" is not a valid partition number\n", argv[optind]);
    usage();
  }
  optind++;
  if (optind < argc) {
    fprintf(stderr, "Extra arguments detected on command line\n");
    usage();
  }

  if (partnum < 0) {
    fprintf(stderr, "%s: bad partition number %d\n", progname, partnum);
    fflush(stderr);
    exit(1);
  }

  rc = nasd_cl_p_init();
  if (rc) {
    fprintf(stderr,
      "ERROR: cannot initialize client library, error 0x%x (%s)\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_bind_to_drive(server_name, NASD_PDRIVE_PORT, NASD_BIND_DCE_DIRECT,
    NULL, 0, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s error 0x%x (%s)\n",
      server_name, rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  /* @@@ init cookie */
  bzero((char *)&cookie, sizeof(cookie));
  rc = nasd_cl_p_ctrl_get_part_info(h, cookie.key,cookie.capability_args, partnum, &ptinfo);
  if ((rc == NASD_CTRL_ID_CHECK_FAILED) && (ptinfo.ctrl_id == 0)) {
    /* got back a page of zeroes - no such partition */
    fprintf(stderr, "ERROR: partition %d does not exist\n", partnum);
    fflush(stderr);
    exit(1);
  }
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) getting partition info\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  if (ptinfo.num_obj != 1) {
    fprintf(stderr, "ERROR: partition contains multiple objects. Reformatting\n");
    fprintf(stderr, "  is not supported.\n");
    fflush(stderr);
    exit(1);
  }

  nid = ptinfo.first_obj;

  /* @@@ init cookie */
  bzero((char *)&cookie, sizeof(cookie));

  printf("Partition has %" NASD_64u_FMT " blocks, root is 0x%"
          NASD_ID_FMT "\n",
          ptinfo.part_size, nid);

  ga_args.in_partnum = partnum;
  ga_args.in_identifier = nid;

  nasd_cl_p_getattr_dr(h,cookie.key, &ga_args, &ga_res, &status);
  if (ga_res.nasd_status || status) {
    fprintf(stderr,
      "ERROR: got nasd_status 0x%x (%s) status 0x%x (%s) getting root obj status\n",
      ga_res.nasd_status, nasd_error_string(ga_res.nasd_status),
      status, nasd_cl_error_string(h, status, err_str));
    fflush(stderr);
    exit(1);
  }

  attr = &ga_res.out_attribute;
  if (attr->object_len) {
    fprintf(stderr, "Partition is not virgin: root object length %lu\n",
      (unsigned long)attr->object_len);
    fflush(stderr);
    exit(1);
  }

  if (bcmp((char *)&attr->object_create_time,
    (char *)&attr->object_modify_time, sizeof(nasd_timespec_t)))
  {
    fprintf(stderr, "Partition is not virgin: modify time != create time\n");
    fflush(stderr);
    exit(1);
  }
  if (bcmp((char *)&attr->object_create_time,
    (char *)&attr->attr_modify_time, sizeof(nasd_timespec_t)))
  {
    fprintf(stderr,
      "Partition is not virgin: attr modify time != create time\n");
    fflush(stderr);
    exit(1);
  }

  bzero((char *)&sa_args, sizeof(sa_args));
  sa_args.in_identifier = nid;
  sa_args.in_fieldmask = NASD_ATTR_FS_SPECIFIC;
  sa_args.in_partnum = partnum;
  sa_args.in_guard = 0;
  bcopy((char *)attr, (char *)&sa_args.in_attribute, sizeof(nasd_attribute_t));
  attr = &sa_args.in_attribute;
  nfs3attr = (nasd_nfs3_attributes_t *)attr->fs_specific;
  nfs3attr->type = NF3DIR;
  nfs3attr->mode = 0755;
  nfs3attr->nlink = 1;
  nfs3attr->uid = 0;
  nfs3attr->gid = 0;
  nfs3attr->clean = NASD_NFS3_CLEAN;

  nasd_cl_p_setattr_dr(h,cookie.key, &sa_args, &sa_res, &status);
  if (sa_res.nasd_status || status) {
    fprintf(stderr,
      "ERROR: got nasd_status 0x%x (%s) status 0x%x (%s) setting root obj status\n",
      sa_res.nasd_status, nasd_error_string(sa_res.nasd_status),
      status, nasd_cl_error_string(h, status, err_str));
    fflush(stderr);
    exit(1);
  }

  /* directory contains only "." and ".." entry */
  bzero((char *)&dirpage, sizeof(dirpage));

  dirpage.header.usemap[0] = 3;
  dirpage.header.freecnt = 125;

  dirpage.data_slots[0].flags = NASD_NFS_DIRD_TYPE_DIR<<NASD_NFS_DIRD_SHIFT;
  dirpage.data_slots[0].nasdid = nid;
  dirpage.data_slots[0].name_len = 1;
  dirpage.data_slots[0].name[0] = '.';
  dirpage.data_slots[0].name[1] = '\0';

  dirpage.data_slots[1].flags = NASD_NFS_DIRD_TYPE_DIR<<NASD_NFS_DIRD_SHIFT;
  dirpage.data_slots[1].nasdid = nid;
  dirpage.data_slots[1].name_len = 2;
  dirpage.data_slots[1].name[0] = '.';
  dirpage.data_slots[1].name[1] = '.';
  dirpage.data_slots[1].name[2] = '\0';

  wr_args.in_partnum = partnum;
  wr_args.in_identifier = nid;
  wr_args.in_offset = 0;
  wr_args.in_len = sizeof(dirpage);
  wr_args.in_capability_args = cookie.capability_args;
  wr_args.in_security_param.type = cookie.capability_args.type;
  wr_args.in_security_param.partnum = partnum;
  wr_args.in_security_param.actual_protection = cookie.capability_args.min_protection;
  nasd_cl_p_write_simple_dr(h, cookie.key, &wr_args, &dirpage,
    &wr_res, &status);
  out_len = wr_res.out_datalen;
  rc = wr_res.nasd_status;
  if (rc || status) {
    fprintf(stderr,
      "ERROR: got nasd_status 0x%x (%s) status 0x%x (%s) writing root dir\n",
      rc, nasd_error_string(rc), status,
      nasd_cl_error_string(h, status, err_str));
    fflush(stderr);
    exit(1);
  }
  if (out_len != sizeof(dirpage)) {
    fprintf(stderr, "ERROR: write %d/%d bytes for root directory\n",
      (int)out_len, (int)sizeof(dirpage));
    fflush(stderr);
    exit(1);
  }

  nasd_cl_p_sync_dr(h, &rc, &status);
  if (rc || status) {
    fprintf(stderr,
      "ERROR: got nasd_status 0x%x (%s) status 0x%x (%s) forcing drive writeback\n",
      rc, nasd_error_string(rc), status,
      nasd_cl_error_string(h, status, err_str));
    fflush(stderr);
    exit(1);
  }

  printf("Format complete.\n");

  nasd_cl_p_shutdown();

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
