/*
 * nasd_nfs_getstats.c
 *
 * Retrieve and print NFS server stats
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_nfs_client.h>
#include <stdio.h>

char *progname;

extern char *optarg;
extern int   optind;

nasd_nfs_opstats_t       opstats;
nasd_nfs_cachestats_t    cachestats;
nasd_nfs_rpc_opdepths_t  opdepths;

int print_cachestats = 0;
int print_opstats = 1;
int print_rpc_queuedepths = 0;

#define DUMP_OPSTAT(_opname_) \
  nasd_nfs_dump_opstat(&opstats._opname_,NASD_STRING(_opname_))

#define NASD_NFS_CSDUMP(_stat_) { \
  printf("  %24s %5lu\n", NASD_STRING(_stat_), cachestats._stat_); \
}

#define NASD_NFS_CSDUMP_TS(_stat_) { \
  printf("  %24s %5d:%09d\n", NASD_STRING(_stat_), \
    cachestats._stat_.ts_sec, cachestats._stat_.ts_nsec); \
}

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername\n", progname);
  fprintf(stderr, "  -C exclude cache stats\n");
  fprintf(stderr, "  -c include cache stats\n");
  fprintf(stderr, "  -O exclude opstats\n");
  fprintf(stderr, "  -o include opstats\n");
  fprintf(stderr, "  -Q exclude rpc queue depths\n");
  fprintf(stderr, "  -q include rpc queue depths\n");
  fflush(stderr);
  exit(1);
}

void
nasd_nfs_dump_opstat(
  nasd_opstat_t  *os,
  char           *name)
{
  unsigned long s, ns, rs, rns;
  double d;

  s = os->op_nsecs / 1000000000UL;
  ns = os->op_nsecs % 1000000000UL;
  if (os->num_ops) {
    d = (double)ns;
    d /= 1000000000.0;
    d += (double)s;
    d /= (double)os->num_ops;
    rs = (unsigned long)d;
    d -= (double)rs;
    d *= 1000000000.0;
    rns = (double)d;
  }
  else {
    d = 0.0;
    rs = 0;
    rns = 0;
  }
  printf("Operation: %s\n", name);
  printf("  time       %lu:%09lu\n", s, ns);
  printf("  num_ops    %5lu\n", os->num_ops);
  printf("  in_flight  %5lu\n", os->in_flight);
  printf("  invalid    %5ld\n", os->invalid);
  printf("  avg time   %5ld.%09ld\n", rs, rns);
}

int
main(
  int     argc,
  char  **argv)
{
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_nfs_handle_t h;
  char *server_name;
  nasd_status_t rc;
  int ret, i;
  char c;

  progname = argv[0];

  while ((c = getopt(argc, argv, "cCoOqQ")) != EOF) {
    switch(c) {
      case 'c':
        print_cachestats = 1;
        break;
      case 'C':
        print_cachestats = 0;
        break;
      case 'o':
        print_opstats = 1;
        break;
      case 'O':
        print_opstats = 0;
        break;
      case 'q':
        print_rpc_queuedepths = 1;
        break;
      case 'Q':
        print_rpc_queuedepths = 0;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }

  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;
  if (optind < argc) {
    fprintf(stderr, "Extra arguments detected on command line\n");
    usage();
  }

  rc = nasd_nfscli_init();
  if (rc) {
    fprintf(stderr, "ERROR: failed nasd_nfscli_init() rc=0x%x (%s)\n",
      rc, nasd_error_string(rc));
    exit(1);
  }

  rc = nasd_bind_to_nfs_server(server_name, NASD_NFS_SERVER_PORT,
    NASD_BIND_DCE_DIRECT, NULL, 0, &h);
  if (rc) {
    fprintf(stderr, "ERROR: unable to bind to server %s, rc=0x%x (%s)\n",
      server_name, rc, nasd_error_string(rc));
    exit(1);
  }

  nasd_nfscli_getstats(h, &opstats, &cachestats, &opdepths,
    &rc, &status);

  if (rc || status) {
    fprintf(stderr,
      "ERROR: nasd_nfscli_getstatsreturned nasd_status 0x%x (%s) status 0x%x (%s)\n",
      rc, nasd_error_string(rc),
      status, nasd_cl_error_string(h, status, err_str));
    exit(1);
  }

  if (print_opstats) {
    DUMP_OPSTAT(null);
    DUMP_OPSTAT(mount);
    DUMP_OPSTAT(fsstat);
    DUMP_OPSTAT(fsinfo);
    DUMP_OPSTAT(lookup);
    DUMP_OPSTAT(readdir);
    DUMP_OPSTAT(access);
    DUMP_OPSTAT(setattr);
    DUMP_OPSTAT(create);
    DUMP_OPSTAT(remove);
    DUMP_OPSTAT(mkdir);
    DUMP_OPSTAT(rmdir);
    DUMP_OPSTAT(newcookie);
    DUMP_OPSTAT(rename);
    DUMP_OPSTAT(symlink);
    DUMP_OPSTAT(getstats);
    DUMP_OPSTAT(lookup_otw);
  }

  if (print_rpc_queuedepths) {
    for(i=0;i<NASD_NFS_RPC_OPDEPTHS-1;i++) {
      printf("  rpc_queue_depth[%03d]     %4d:%09d\n", i+1,
        opdepths.opdepths[i].ts_sec, 
        opdepths.opdepths[i].ts_nsec);
    }
    printf("  rpc_queue_depth[%02d+]     %4d:%09d\n", i+1,
      opdepths.opdepths[i].ts_sec, 
      opdepths.opdepths[i].ts_nsec);
  }

  if (print_cachestats) {
    NASD_NFS_CSDUMP(dir_lookup);
    NASD_NFS_CSDUMP(dir_force);
    NASD_NFS_CSDUMP(dirc_load_attr);
    NASD_NFS_CSDUMP(dirc_not_dir);
    NASD_NFS_CSDUMP(dirc_load_dir);
    NASD_NFS_CSDUMP(dirc_dirty);
    NASD_NFS_CSDUMP(dirc_new_dirty);
    NASD_NFS_CSDUMP(dirc_mkfree_loop);
    NASD_NFS_CSDUMP(dirc_mkfree_wait);
    NASD_NFS_CSDUMP(dir_load);
    NASD_NFS_CSDUMP(post_attribute);
    NASD_NFS_CSDUMP(post_attribute_valid);
    NASD_NFS_CSDUMP(post_attribute_invalid);
    NASD_NFS_CSDUMP(lookup_post_attribute);
    NASD_NFS_CSDUMP(dir_writer_fls);
    NASD_NFS_CSDUMP(dir_writer_pages);
    NASD_NFS_CSDUMP(dir_load_attr);
    NASD_NFS_CSDUMP(do_getattr);
    NASD_NFS_CSDUMP(do_setattr);
    NASD_NFS_CSDUMP(access_lookup_fail);
    NASD_NFS_CSDUMP(access_not_dir);
    NASD_NFS_CSDUMP(access_fault_attr);
    NASD_NFS_CSDUMP(access_get_attr);
    NASD_NFS_CSDUMP(lookup_get_attr);
    NASD_NFS_CSDUMP(setattr_get_attr);
    NASD_NFS_CSDUMP_TS(drive_time);
  }

  rc = nasd_unbind_nfs_server(&h);
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) unbinding NFS server\n",
      rc, nasd_error_string(rc));
    exit(1);
  }

  nasd_nfscli_shutdown();

  exit(0);
}
