/*
 * pcnulls.c
 *
 * send noops to a NASD drive
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_timer.h>

char *progname;

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername niters\n", progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -M use message queues\n");
  fprintf(stderr, "  -T use TCP\n");
  fflush(stderr);
  exit(1);
}

int
main(argc, argv)
  int     argc;
  char  **argv;
{
  nasd_error_string_t err_str;
  double accum_time, d, avg;
  nasd_status_t nasd_status;
  nasd_rpc_status_t status;
  int done, niters, i, rc;
  nasd_drive_handle_t h;
  nasd_timespec_t diff;
  nasd_timer_t timer;
  char *server_name;
  int binding_type;
  int c;

  progname = argv[0];

  binding_type = NASD_BIND_DCE_DIRECT;

  while((c = getopt(argc, argv, "MT")) != EOF) {
    switch(c) {
      case 'M':
        binding_type = NASD_BIND_MSGQ;
        break;
      case 'T':
        binding_type = NASD_BIND_DCE_DIRECT_TCP;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }

  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;

  if (optind >= argc)
    usage();
  if (sscanf(argv[optind], "%d", &niters) != 1)
    usage();
  optind++;
  if (niters < 1)
    usage();

  rc = nasd_cl_p_init();
  if (rc) {
    fprintf(stderr,
      "ERROR: cannot initialize client library, error 0x%x (%s)\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_bind_to_drive(server_name, NASD_PDRIVE_PORT,
    binding_type, NULL, 0, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s error 0x%x (%s)\n",
      server_name, rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  done = 0;
  accum_time = (double)0.0;
  for(i=0;i<niters;i++) {
    NASD_TM_START(&timer);
    nasd_cl_p_null_dr(h, &nasd_status, &status);
    if (status || nasd_status) {
      fprintf(stderr,
        "ERROR from drive noop, status=0x%x (%s) nasd_status=0x%x (%s)\n",
        status, nasd_cl_error_string(h, status, err_str),
        nasd_status, nasd_error_string(nasd_status));
      exit(1);
    }
    NASD_TM_STOP(&timer);
    NASD_TM_ELAPSED_TS(&timer, &diff);
    d = (double)diff.ts_nsec;
    d /= (double)1000000000.0;
    d += (double)diff.ts_sec;
    accum_time += d;
    done++;
  }

  printf("%f seconds to do %d noops\n", accum_time, done);
  d = (double)done;
  avg = accum_time / d;
  printf("%f seconds average\n", avg);
  printf("%f ops / sec\n", (d/accum_time));

  nasd_unbind_drive(&h);
  nasd_cl_p_shutdown();

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
