/*
 * pcreatedelete.c
 *
 * Test drive create/delete by creating and deleting many objects,
 * and verifying contents of list-of-objects-object
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jason Grosman
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */


#include <nasd/nasd_options.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_security_capability.h>
#include <nasd/nasd_security_time.h>
#include <nasd/nasd_mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>

int do_delete(nasd_identifier_t nasdid);

#define OBJ_BUF_LEN 10

#if 0
#undef protection
#define protection (NASD_INTEGRITY_ARGS | NASD_INTEGRITY_DATA)
#endif

int  Verbose = 0;
int  Dirty = 0;
int  NumberObjs;
int  Filled = 0;

int randomize_deletions = 0;
int npasses             = 1;
int disable_pass_checks = 0;
int flush_before_delete = 0;
int sync_between_passes = 0;

nasd_error_string_t error_text; 
nasd_uint16 protection;         
nasd_drive_handle_t    h;       
nasd_identifier_t      nasdid;  
char                  *server_name;
int                    partnum;  
nasd_rpc_status_t      status;   
nasd_status_t          rc;       
nasd_key_t             part_key, master_key, drive_key, red_key, black_key;

int                           binding_type;
int                           binding_args_len;
void                         *binding_args;
char                         *binding_port = NASD_PDRIVE_PORT;
nasd_drive_param_dce_kpdev_t  kpdev_args;
int                           nondefault_binding = 0;


char *master_password; 
char * progname;



void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername partnum master_password\n", progname);
  fflush(stderr); fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -D Leave partition dirty. (Don't delete objects created)\n");
  fprintf(stderr, "  -d number of deletions(>=0). (default : random)\n");
  fprintf(stderr, "  -e randomize deletions\n");
  fprintf(stderr, "  -f Running on a non-empty file-system\n");
  fprintf(stderr, "  -F flush objects before deleting\n");
  fprintf(stderr, "  -k use kernel device\n");
  fprintf(stderr, "  -l use colocated drive\n");
  fprintf(stderr, "  -n number of objects to create (default : 20)\n");
  fprintf(stderr, "  -p number of passes\n");
  fprintf(stderr, "  -P disable per-pass checks\n");
  fprintf(stderr, "  -R run Regression test. (multiple passes of creates/deletes)\n");
  fprintf(stderr, "  -s Seed to use for the randomness.\n");
  fprintf(stderr, "  -S sync between passes\n");
  fprintf(stderr, "  -T use TCP\n");
  fprintf(stderr, "  -v Verbose mode \n");
  exit(1);
}

void endgame(nasd_identifier_t *array, int Num)
{
int i;
 if (!Dirty) {
 printf("Deleting objects created by this test.\n");
 for (i=0;i<Num;i++)
   {
     if (array[i] == NASDID_NULL) continue;
     do_delete(array[i]);
   }
 }
}

nasd_identifier_t
do_create()
{
  nasd_p_create_dr_args_t args;
  nasd_p_create_dr_res_t res;
  nasd_attribute_t *ap;

  bzero((char *)&args, sizeof(args));

  nasd_security_password_to_keys(master_password,master_key,drive_key,partnum,part_key,red_key,black_key); 
  bzero((char *)&args,sizeof(nasd_p_create_dr_args_t));
  args.in_security_param.type=NASD_BLACK_KEY;
  args.in_security_param.partnum = partnum;
  args.in_security_param.actual_protection = protection; /*set this above*/



  args.in_partnum = partnum;
 
  ap = &args.in_attribute;
  ap->object_len = 3000;
  /* Testing standard AV */ 

  args.in_attribute.av=7;
  args.in_fieldmask = (NASD_ATTR_OBJECT_LEN | NASD_ATTR_AV);
  nasd_cl_p_create_dr(h, black_key,&args, &res, &status); 

  if (status) {
    fprintf(stderr, "ERROR: status=0x%x (%s) nasd_status=0x%x (%s)\n",
      status, nasd_cl_error_string(h, status, error_text),
      res.nasd_status, nasd_error_string(res.nasd_status));
    fflush(stderr);
    exit(2);
  }

  if (res.nasd_status) {
    fprintf(stderr, "NASD ERROR: 0x%x (%s)\n", res.nasd_status,
      nasd_error_string(res.nasd_status));
    fflush(stderr);
    exit(2);
  }
 nasdid = res.out_identifier;




 if (Verbose) printf("New object identifier 0x%" NASD_ID_FMT "\n", nasdid);
  return nasdid; 
}


int
do_delete(nasd_identifier_t nasdid)
{
  nasd_p_remove_dr_args_t args;
  nasd_p_remove_dr_res_t res;
  nasd_p_flush_obj_dr_args_t fl_args;
  nasd_p_flush_obj_dr_res_t fl_res;

  bzero((char *)&args, sizeof(args)); 

  if (flush_before_delete) {
    bzero((char *)&fl_args, sizeof(fl_args)); 
    /* xxx cookie here */
    fl_args.in_identifier = nasdid;
    fl_args.in_partnum = partnum;
    nasd_cl_p_flush_obj_dr(h, black_key, &fl_args, &fl_res, &status);
    if (fl_res.nasd_status || status) {
      fprintf(stderr, "WARNING: got nasd_status 0x%x (%s) status 0x%x (%s) flushing 0x%" NASD_ID_FMT "\n",
        fl_res.nasd_status, nasd_error_string(fl_res.nasd_status),
        status, nasd_cl_error_string(h, status, error_text));
      fflush(stderr);
    }
  }

  args.in_partnum = partnum;
  args.in_identifier = nasdid;

  
 
  nasd_security_password_to_keys(master_password,master_key,drive_key,partnum,part_key,red_key,black_key); 

  /* Fill in capability and security param */ 
  args.in_security_param.type = NASD_BLACK_KEY;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;
  args.in_security_param.partnum=partnum;
  nasd_cl_p_remove_dr(h, black_key,&args, &res, &status); 

  if (res.nasd_status || status) {
    fprintf(stderr, "nasd_status=0x%x (%s) status=0x%x (%s)\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text));
    return 1;
  }

  
  if (Verbose)  printf("Deleted identifier 0x%" NASD_ID_FMT "\n", nasdid);
  return 0 ;
}


int act_vs_exp(nasd_identifier_t *array, int Num)
{
  nasd_identifier_t objs[OBJ_BUF_LEN];
  nasd_p_smpl_op_dr_args_t rd_args;
  nasd_p_fastread_dr_res_t rd_res;
  nasd_error_string_t err_str;
  nasd_cookie_t cookie;
  nasd_ctrl_part_info_t ptinfo;
  nasd_timespec_t tm;
  nasd_offset_t off;
  nasd_len_t len;
  char c;
  int i,j;


  /* No security on control objects */ 
  bzero((char *)&cookie, sizeof(cookie));

   rc = nasd_cl_p_ctrl_get_part_info(h,black_key, cookie.capability_args, partnum, &ptinfo);

   if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) from nasd_cl_p_get_drive_info\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  off = 0;
  do {
    rd_args.in_partnum = partnum;
    rd_args.in_identifier = (nasd_identifier_t)NASD_CTRL_PART_OBJS;
    rd_args.in_offset = off;
    rd_args.in_len = sizeof(objs);
    rd_args.in_capability_args = cookie.capability_args;
    rd_args.in_security_param.type = cookie.capability_args.type;
    rd_args.in_security_param.partnum = partnum;
    rd_args.in_security_param.actual_protection = cookie.capability_args.min_protection;
    nasd_cl_p_read2_simple_dr(h, cookie.key, &rd_args, objs, &rd_res, &status);
    len = rd_res.out_datalen; 
    if (status || rd_res.nasd_status) {
      fprintf(stderr,
        "Error issuing read, off=%lu status=0x%x (%s) nasd_status=0x%x (%s)\n",
        off, status, nasd_cl_error_string(h, status, err_str),
        rd_res.nasd_status, nasd_error_string(rd_res.nasd_status));
      fflush(stderr);
      exit(2);
    }
    for(i=0;i<len/sizeof(nasd_identifier_t);i++) {
      if (Verbose) printf("Searching for 0x%" NASD_ID_FMT "\n", objs[i]);
      if (objs[i] == ptinfo.first_obj) continue;
      for (j=0;j<Num;j++){
	if (array[j] == objs[i]) {
	  if (Verbose) printf("found  0x%" NASD_ID_FMT "\n", objs[i]);
	  break;
	}
      }
    
      if (j == Num) {
        fprintf(stderr, "Test Failed\n");
        fprintf(stderr, "There was a NASD object in the partition that was not the root object or created by this test.\n");
        fprintf(stderr, "Offending  object : 0x%" NASD_ID_FMT "\n", objs[i]);
        endgame(array, Num);
        exit(2);
      }
    }
    off += len;
  } while(len);

  printf("Success : All objects in partition accounted for.\n");
  return 0;
}

int exp_vs_act(nasd_identifier_t nasdid)
{
  nasd_identifier_t objs[OBJ_BUF_LEN];
  nasd_p_smpl_op_dr_args_t rd_args;
  nasd_p_fastread_dr_res_t rd_res;
  nasd_error_string_t err_str;
  nasd_cookie_t cookie;
  nasd_timespec_t tm;
  nasd_offset_t off;
  nasd_len_t len;
  char c;
  int i;


  /* No security on control objects */ 
  bzero((char *)&cookie, sizeof(cookie));

  off = 0;
  do {
    rd_args.in_partnum = partnum;
    rd_args.in_identifier = (nasd_identifier_t)NASD_CTRL_PART_OBJS;
    rd_args.in_offset = off;
    rd_args.in_len = sizeof(objs);
    rd_args.in_capability_args = cookie.capability_args;
    rd_args.in_security_param.type = cookie.capability_args.type;
    rd_args.in_security_param.partnum = partnum;
    rd_args.in_security_param.actual_protection = cookie.capability_args.min_protection;
    nasd_cl_p_read2_simple_dr(h, cookie.key, &rd_args, objs, &rd_res, &status);
    len = rd_res.out_datalen;
    if (status || rd_res.nasd_status) {
      fprintf(stderr,
        "Error issuing read, off=%lu status=0x%x (%s) nasd_status=0x%x (%s)\n",
        off, status, nasd_cl_error_string(h, status, err_str),
        rd_res.nasd_status, nasd_error_string(rd_res.nasd_status));
      fflush(stderr);
      exit(2);
    }
    for(i=0;i<len/sizeof(nasd_identifier_t);i++) {
      if (objs[i] == nasdid)
	{
	  if (Verbose) printf("Found 0x%" NASD_ID_FMT "\n", nasdid);
	  return 0;
	}
    }
    off += len;
  } while(len);
  
  return -1;
}



void do_checks(nasd_identifier_t * array, int Num)
{
  int i;
  
  if (!Filled) 
    printf("Checking...\n");
    act_vs_exp(array, Num);
  

  
  printf("Checking...\n");
  for (i=0;i<Num;i++)
    {  
      if (array[i] == NASDID_NULL) continue;
      if (Verbose) printf("Searching for: 0x%" NASD_ID_FMT "\n", array[i]);
       if (exp_vs_act(array[i]) == -1) {
         fprintf(stderr, "Test Failed : An object created by this test was not found in the partition\n");
         fprintf(stderr, "Offending Object :  0x%" NASD_ID_FMT "\n", array[i]);
         endgame(array, Num);
         exit(2);
       }
    }
    printf("Success : All objects created accounted for.\n");
    

}


void
do_regress ()
{
  int num,i,j,place;
  nasd_identifier_t * array;

  j= 0;

  /*init array*/ 
  NASD_Malloc(array, sizeof(nasd_identifier_t) * 200, (nasd_identifier_t *));
  for (i = 0;i<150;i++)
    array[i] = NASDID_NULL;
  
  
  /*pass one*/
  printf ("Pass One.\n");
  num = rand()%20 +  80;
  if (Verbose) printf("Creating %d objects.\n",num);

  for (i = 0;i<num;i++) {
    array[j] = do_create();
    j++;
  }
  
  do_checks(array,j);
  
  num = rand()%10 + 10;
  if (Verbose) printf("Deleting %d objects.\n",num);
  
  for(i=0;i<num;i++){
    do {place = rand()%j;}
	 while (!array[place]);
     do_delete(array[place]);
     array[place] = NASDID_NULL;
  }

  do_checks(array, j);

  /*pass two*/
  printf ("Pass Two.\n");
  num = rand()%10 +  20;
  if (Verbose) printf("Creating %d objects.\n",num);
  
  for (i = 0;i<num;i++) {
    array[j] = do_create();
    j++;
  }
  
  do_checks(array, j);
  
  num = rand()%10 + 30;
    if (Verbose) printf("Deleting %d objects.\n",num);
  
  for(i=0;i<num;i++){
    do {place = rand()%j;}
	 while (!array[place]);
     do_delete(array[place]);
     array[place] = NASDID_NULL;
  }

  do_checks(array, j);

  /*pass three*/
  printf ("Pass Three.\n");
  num = rand()%10 +  10;
  if (Verbose) printf("Creating %d objects.\n",num);

  for (i = 0;i<num;i++) {
    array[j] = do_create();
    j++;
  }
  
  do_checks(array, j);
  
  num = rand()%10 + 10;
  if (Verbose) printf("Deleting %d objects.\n",num);
  
  for(i=0;i<num;i++){
    do {place = rand()%j;}
	 while (!array[place]);
     do_delete(array[place]);
     array[place] = NASDID_NULL;
  }

  do_checks(array, j);
  
  /*pass four*/
  printf ("Pass Four.\n");

  num = rand()%10 +  10;
  if (Verbose) printf("Creating %d objects.\n",num);
  
  for (i = 0;i<num;i++) {
    array[j] = do_create();
    j++;
  }
  
  do_checks(array, j);
  
  num = rand()%5 + 5;
  if (Verbose) printf("Deleting %d objects.\n",num);
  
  for(i=0;i<num;i++){
    do {place = rand()%j;}
	 while (!array[place]);
     do_delete(array[place]);
     array[place] = NASDID_NULL;
  }

  do_checks(array, j);

  endgame(array, j);
  printf("Regression test successfully completed.\n");
  exit(0);
}


void
do_sync()
{
  nasd_rpc_status_t status;
  nasd_status_t rc;

  nasd_cl_p_sync_dr(h, &rc, &status);

  if (status || rc) {
    fprintf(stderr, "ERROR: nasd_cl_p_sync_dr status=0x%x (%s) nasd_status=0x%x (%s)\n",
      status, nasd_cl_error_string(h, status, error_text),
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(2);
  }
}




int
main (int argc, char **argv)
{
  
  char c;
  int i,numToDel,numOfDel = -1, pass;

  int Regress = 0, array2cnt, array1cnt, arraytmpcnt, j;
  int seed = getpid();
  nasd_identifier_t *array, *array2, *array_tmp;

  NumberObjs = 20; /*default*/


  progname = argv[0];

  binding_type = NASD_BIND_DCE_DIRECT;
  binding_args = NULL;
  binding_args_len = 0;


  while ((c = getopt(argc, argv, "Dd:efFkln:p:PRs:STv")) != EOF) {
    switch(c) {
    case 'e':
      randomize_deletions = 1;
      break;
    case 'S':
      sync_between_passes = 1;
      break;
    case 'F':
      flush_before_delete = 1;
      break;
    case 'P':
      disable_pass_checks = 1;
      break;
    case 'v':
      Verbose = 1;
      break;
    case 'f':
      Filled = 1;
      break;
    case 'n':
      if (sscanf(optarg, "%d", &NumberObjs) != 1) {
	usage();
      }
      break; 
    case 'p':
      if (sscanf(optarg, "%d", &npasses) != 1) {
	usage();
      }
      break; 
    case 'k':
      if (nondefault_binding)
	usage();
      nondefault_binding = 1;
      binding_type = NASD_BIND_DCE_KPDEV;
      binding_args = &kpdev_args;
      binding_args_len = sizeof(kpdev_args);
      strcpy(kpdev_args.devname, "/dev/nasdkp0");
      break;
    case 'l':
      if (nondefault_binding)
	usage();
      nondefault_binding = 1;
      binding_type = NASD_BIND_COLOCATE;
      binding_args = &kpdev_args;
      binding_args_len = sizeof(kpdev_args);
      strcpy(kpdev_args.devname, "/dev/nasdkp0");
      break;
    case 's':
      if (sscanf(optarg, "%d", &seed) != 1) {
	usage();
      }
      break;
    case 'd' :
      if (sscanf(optarg, "%d", &numOfDel) != 1) {
	usage();
      }
      if (numOfDel < 0) {
	usage();
      }
      break;
    case 'D':
      Dirty = 1;
      break;
    case 'R':
      Regress = 1;
      break;
    case 'T':
      if (nondefault_binding)
	usage();
      nondefault_binding = 1;
      binding_type = NASD_BIND_DCE_DIRECT_TCP;
      break;
    default:
      fprintf(stderr, "Unknown option '%c'\n", c);
      usage();
    }
  }
  
  if (optind >= argc)
    usage();
  

  
  server_name = argv[optind];

  optind++;
  if (optind >= argc)
    usage();

  if (sscanf(argv[optind], "%d", &partnum) != 1)
    usage();

  optind++;
  if (optind >= argc)
    usage();

  master_password=argv[optind];


  rc = nasd_cl_p_init(); 
  if (rc) {
    printf("ERROR (%s:%d): cannot init client library, rc=%ld\n",
	   __FILE__, __LINE__, (long)rc);
    return(rc);
  }

  NASD_Malloc(array, NumberObjs * sizeof(nasd_identifier_t), (nasd_identifier_t *));


  rc = nasd_bind_to_drive(server_name, binding_port,
			  binding_type,binding_args,binding_args_len, &h); 
  
  if (rc) { 
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    return 1;
  }


  printf("Random Seed is %d\n",seed);
  srand(seed);

  if (Regress) 
    do_regress(); /*Instead of doing the standard tests, switch to regrssion
		    test mode.*/

 array2 = NULL;
 for(pass=0;pass<npasses;pass++) {
  printf("Creating %d NASD objects in partition %d pass %d.\n",NumberObjs,partnum,pass);
  array1cnt = 0;
  for (i=0;i<NumberObjs;i++)
    {
      array[i] = do_create();
      if (array[i] != NASDID_NULL)
        array1cnt++;
    }


  if (numOfDel == -1) {
    numOfDel = (rand() % (NumberObjs-1)) + 1;
  }

  printf("Number of deletions : %d\n",numOfDel);
  if (numOfDel > 0) {
    for (i=0;i<numOfDel;i++)
      {
	if (randomize_deletions)
	  numToDel = rand() % (NumberObjs);
	else
	  numToDel = i;
	if (array[numToDel] == NASDID_NULL) continue;
	if (Verbose) printf("Deleting : array# : %d, ID :0x%" NASD_ID_FMT "\n",numToDel, array[numToDel]);	
	do_delete(array[numToDel]);
	array[numToDel] = NASDID_NULL;
        array1cnt--;
      }
  }

  if (array2) {
    arraytmpcnt = array1cnt + array2cnt;
    NASD_Malloc(array_tmp, arraytmpcnt * sizeof(nasd_identifier_t),
      (nasd_identifier_t *));
    for(i=0;i<array2cnt;i++) {
      array_tmp[i] = array2[i];
    }
    NASD_Free(array2, array2cnt * sizeof(nasd_identifier_t));
  }
  else {
    array2cnt = 0;
    arraytmpcnt = array1cnt;
    NASD_Malloc(array_tmp, arraytmpcnt * sizeof(nasd_identifier_t),
      (nasd_identifier_t *));
  }
  for(i=j=0;i<NumberObjs;i++) {
    if (array[i] != NASDID_NULL) {
      array_tmp[j+array2cnt] = array[i];
      j++;
    }
  }
  array2 = array_tmp;
  array2cnt = arraytmpcnt;

  if (disable_pass_checks == 0) {    
    do_checks(array2, array2cnt);
  }
  printf("Done pass %d\n", pass);

  if (sync_between_passes) {
    do_sync();
  }
 }

  if (disable_pass_checks) {
    do_checks(array2, array2cnt);
  }
  endgame(array2, array2cnt);
  printf("Successfully Completed.\n");

  exit(0);
}  
