/*
 * pcrmpdrspeed.c
 *
 * Create the objects necessary for an mpdrspeed run.
 * Built on the wreckage of pcrobj.c
 *
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Authors: Jim Zelenka, David Rochberg
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_security_time.h>
#include <nasd/nasd_security_capability.h>
#include <nasd/nasd_mem.h>
#include <dce/dce_error.h>

#include <nasd/mpdrspeed_common.h>


nasd_error_string_t error_text;
char *progname;
nasd_key_t part_key,master_key,drive_key,red_key,black_key;

nasd_status_t nasd_security_get_red_key(nasd_partnum_t partnum,nasd_key_t key){
  bcopy((char *)red_key,(char *)key,sizeof(nasd_key_t));
  return(NASD_SUCCESS);
}

nasd_status_t nasd_security_get_black_key(nasd_partnum_t partnum,nasd_key_t key){
  bcopy((char *)black_key,(char *)key,sizeof(nasd_key_t));
  return(NASD_SUCCESS);
}

extern char *optarg;
extern int   optind;


int	size = 1048576;
int	nobjects = 16;
char	*buf;

#if NASD_SECURITY_ON > 0

#define BUILD_CAPABILITY(_args_,_powers_,_partnum_,_cookie_)						\
{													\
  nasd_timespec_t __tm;											\
  (_args_)->in_security_param.type=NASD_BLACK_KEY;							\
  nasd_security_get_drive_time(0,&tm);									\
  __tm.ts_sec+=(60*60);											\
  nasd_security_build_capability(0,partnum,0,(_powers_),0,__tm.ts_sec,					\
                                 NASD_SECURE_OPS,NASD_BLACK_CAPABILITY,0,0,7,				\
                                 (_cookie_));								\
  bcopy((_cookie_).capability_args, &((_args_)->in_capability_args),sizeof(nasd_capability_args_t));	\
  (_args_)->in_security_param.type = (_args_)->in_capability_args.type; 				\
  (_args_)->in_security_param.partnum = (_partnum_);							\
  (_args_)->in_security_param.actual_protection = NASD_SECURE_OPS;					\
}
{									\
  (_args_)->in_security_param.partnum = _partnum_;			\
  (_args_)->in_security_param.actual_protection = NASD_SECURE_OPS;	\
}

#else /* NASD_SECURITY_ON > 0 */
#define BUILD_CAPABILITY(_args_,_powers_,_partnum_,_cookie_)		
#endif /* NASD_SECURITY_ON > 0 */



void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername master_password\n", progname);
  fprintf(stderr, "  -p  partition number\n");
  fflush(stderr);
  exit(1);
}



void
check_status(char *string,nasd_drive_handle_t h,nasd_status_t *nasd_status,nasd_rpc_status_t *rpc_status)
{
  nasd_error_string_t           err_str;
  if (rpc_status && *rpc_status) {
    fprintf(stderr, "%s: ERROR: status=0x%x (%s) nasd_status=0x%x (%s)\n",
	    string,
	    *rpc_status, nasd_cl_error_string(h, *rpc_status, err_str),
	    *nasd_status, nasd_error_string(*nasd_status));
    fflush(stderr);
    exit(2);
  }

  if (nasd_status && *nasd_status) {
    fprintf(stderr, "%s: NASD ERROR: %s\n", string, nasd_error_string(*nasd_status));
    fflush(stderr);
    exit(2);
  }
}



nasd_identifier_t
get_first_object(
  nasd_drive_handle_t   drive,
  int                   partnum,
  char                 *name)
{
  nasd_ctrl_part_info_t ptinfo;
  nasd_status_t         rc;
  nasd_rpc_status_t     status;
  nasd_cookie_t		cookie;
  

  /* If the drive is freshly started with no partitions, it seems to
     not want to accept I/Os */
  nasd_cl_p_null_dr(drive, &rc, &status);
  check_status("Initial null RPC in get_first_object",drive,&rc,&status);


  
  BUILD_CAPABILITY(&ptinfo,NASD_ACCESS_RIGHTS_GETINFO,partnum,&cookie);
#if NASD_SECURITY_ON > 0
  printf ("The get partition info cookie is sad.  Sorry --rochberg\n");
#endif /* NASD_SECURITY_ON > 0 */
  rc = nasd_cl_p_ctrl_get_part_info(drive, cookie.key, cookie.capability_args, partnum, &ptinfo);

  if ((rc == NASD_CTRL_ID_CHECK_FAILED) && (ptinfo.ctrl_id == 0)) {
    /* got back a page of zeroes - no such partition */
    fprintf(stderr, "ERROR: drive %s partition %d does not exist\n", name, partnum);
    fflush(stderr);
    exit(1);
  }
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) getting partition info\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }
  return ptinfo.first_obj;
  
}


main(argc, argv)
  int     argc;
  char  **argv;
{
  char			       *server_name, *master_password;
  nasd_cookie_t			create_cookie;
  nasd_cookie_t			write_cookie;
  nasd_p_create_dr_args_t	create_args;
  nasd_error_string_t		err_str;
  nasd_p_create_dr_res_t	res;
  nasd_rpc_status_t		status;
  nasd_drive_handle_t		h;
  nasd_timespec_t		tm;
  mpdrspeed_header_t		header;
  char				*pattern="Plus, we have no use for weapons grade nuclear materials, preferring instead to carry out our plans for world domination through total control of the world's data storage.";
  int				patlen;
  nasd_identifier_t		firstobj;
  nasd_len_t			out_len;
  int				i;
  nasd_status_t			rc;
  int				partnum;
  char				c;
  int				binding_type = NASD_BIND_DCE_DIRECT;
  nasd_p_smpl_op_dr_args_t wr_args;
  nasd_p_fastwrite_dr_res_t wr_res;

  progname = argv[0];

  bzero((char *)&create_args, sizeof(create_args));
  partnum = 1;

  while ((c = getopt(argc, argv, "p:b:s:n:T")) != EOF) {
    switch(c) {
      case 's':
      case 'b':
	if (sscanf(optarg, "%d",&size) != 1) 
	  usage();
	break;
      case 'n':
	if (sscanf(optarg, "%d",&nobjects) != 1) 
	  usage();
	if (nobjects > MPDRSPEED_MAX_OBJS) {
	  fprintf (stderr,"Only %d obejcts supported in mpdrspeedfs\n",MPDRSPEED_MAX_OBJS);
	  exit(2);
	}
	break;
      case 'p':
	if (sscanf(optarg, "%d", &partnum) != 1)
	  usage();
	break;
      case 'T':
	binding_type = NASD_BIND_DCE_DIRECT_TCP;
	break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }
  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;

  if (optind >= argc)
    usage();
  master_password = argv[optind];
  optind++;

  if (optind < argc)
    usage();

  if (partnum < 0)
    usage();


  NASD_Malloc(buf, size, (char *));
  if (buf == NULL) {
    fprintf(stderr,"Could not allocate %d bytes for buffer\n",size);
    exit(1);
  }
  patlen = strlen(pattern);
  
  for (i = 0 ; i < (size / patlen); i++) {
    strcpy (buf+(i*patlen),pattern,patlen);
  }
  strcpy(buf+(i*patlen),pattern,size%patlen);


  rc = nasd_cl_p_init();
  if (rc) {
    printf("ERROR (%s:%d): cannot init client library, rc=%ld\n",
      __FILE__, __LINE__, (long)rc);
    return(rc);
  }


  rc = nasd_bind_to_drive(server_name, NASD_PDRIVE_PORT,
    binding_type, NULL, 0, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    exit(1);
  }
  

  nasd_security_password_to_keys(master_password,master_key,drive_key,partnum,part_key,red_key,black_key);


  
  bzero(&header,sizeof(header)); 
  strcpy(header.magic,MPDRSPEED_MULTI_MAGIC,strlen(MPDRSPEED_MULTI_MAGIC));
 
  bzero(&create_args,sizeof(nasd_p_create_dr_args_t));
  BUILD_CAPABILITY(&create_args,NASD_ACCESS_RIGHTS_CREATE,partnum,&create_cookie);  
  for (i = 0; i< nobjects;i++) {
    create_args.in_fieldmask=NASD_ATTR_AV;
    create_args.in_attribute.av=7; /* Goodness knows where this 7 came from */
    create_args.in_partnum = partnum;
    create_args.in_security_param.partnum = partnum;
    create_args.in_security_param.type=NASD_BLACK_KEY;
    create_args.in_security_param.actual_protection = NASD_SECURE_OPS;
    nasd_cl_p_create_dr(h, black_key, &create_args, &res, &status);
    check_status("Creating object",h,&res.nasd_status,&status);
    printf("Creating object 0x%" NASD_ID_FMT "\n",res.out_identifier);
    header.objects[i] = res.out_identifier;
    
    nasd_security_get_drive_time(0,&tm);
    tm.ts_sec+=(60*60);

    nasd_security_build_capability(0,				/* drive_name */
				   partnum,			/* partnum */
				   header.objects[i],		/* ni */
				   NASD_ACCESS_RIGHTS_WRITE,	/* rights */
				   0,				/* Audit ID */
				   tm.ts_sec,			/* Expiry time */
				   NASD_SECURE_OPS,	        /* min protection? */
				   NASD_BLACK_CAPABILITY,      	/* type */
				   0,				/* start */
				   size,			/* end */
				   7, /* Or this 7 */		/* AV */
				   &write_cookie);		/* dest */
    


    wr_args.in_partnum = partnum;
    wr_args.in_identifier = header.objects[i];
    wr_args.in_offset = 0;
    wr_args.in_len = size;
    wr_args.in_capability_args = write_cookie.capability_args;
    wr_args.in_security_param.type = write_cookie.capability_args.type;
    wr_args.in_security_param.partnum = partnum;
    wr_args.in_security_param.actual_protection = write_cookie.capability_args.min_protection;
    nasd_cl_p_write_simple_dr(h, write_cookie.key, &wr_args, buf, &wr_res, &status);
    out_len = wr_res.out_datalen;
    rc = wr_res.nasd_status;
    check_status("Writing data",h,&rc,&status);
    if (out_len != size ) {
      fprintf(stderr,"Wrote only %ld bytes while writing data in iteration %d\n",out_len,i);
      exit(1);
    }
  }

  firstobj = get_first_object(h,partnum,server_name);
  /*  bzero(&write_cookie,sizeof(write_cookie));
  bzero(&wr_args, sizeof(wr_args));*/

  nasd_security_get_drive_time(0,&tm);
  tm.ts_sec+=(60*60);
  nasd_security_build_capability(0,				/* drive_name */
				 partnum,			/* partnum */
				 firstobj,			/* ni */
				 NASD_ACCESS_RIGHTS_WRITE,	/* rights */
				 0,				/* Audit ID */
				 tm.ts_sec,			/* Expiry time */
				 NASD_SECURE_OPS,	        /* min protection? */
				 NASD_BLACK_CAPABILITY,				/* type */
				 0,				/* start */
				 sizeof(header),		/* end */
				 0, /* Or this 7 */		/* AV */
				 &write_cookie);		/* dest */
  wr_args.in_partnum = partnum;
  wr_args.in_identifier = firstobj;
  wr_args.in_offset = 0;
  wr_args.in_len = sizeof(header); 
  wr_args.in_capability_args = write_cookie.capability_args;
  wr_args.in_security_param.type = write_cookie.capability_args.type;
  wr_args.in_security_param.partnum = partnum;
  wr_args.in_security_param.actual_protection = write_cookie.capability_args.min_protection;


  nasd_cl_p_write_simple_dr(h, write_cookie.key, &wr_args, &header, &wr_res, &status);
  out_len = wr_res.out_datalen;
  rc = wr_res.nasd_status;
  check_status("Writing header",h,&rc,&status);

  if (out_len != sizeof(header) ) {
    fprintf(stderr,"Wrote only %ld bytes while writing header, sizeof(header) = %d \n",out_len,sizeof(header));
    exit(1);
  }

  nasd_cl_p_sync_dr(h, &rc, &status);
  check_status("Forcing drive writeback", h, &rc, &status);
  exit(0);
}
