/*
 * pcrobj.c
 *
 * Create object on NASD partition
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <nasd/nasd_pdrive.h>

#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_security.h>

#include <nasd/nasd_security_time.h>

char *progname;
nasd_key_t part_key,master_key,drive_key,red_key,black_key;

int                           binding_type;
int                           binding_args_len;
void                         *binding_args;
char                         *binding_port = NASD_PDRIVE_PORT;
nasd_drive_param_dce_kpdev_t  kpdev_args;
int                           nondefault_binding = 0;
int                           nobjs;

extern char *optarg;
extern int   optind;

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername master_password\n",
    progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -k use kernel device\n");
  fprintf(stderr, "  -M use message queues\n");
  fprintf(stderr, "  -n number of objects to create [default 1]\n");
  fprintf(stderr, "  -p partition number\n");
  fprintf(stderr, "  -T use TCP\n");
  fflush(stderr);
  exit(1);
}

int
main(argc, argv)
  int     argc;
  char  **argv;
{
  char *server_name, *master_password;
  nasd_cookie_t chocolate_chip;
  nasd_p_create_dr_args_t args;
  nasd_error_string_t err_str;
  nasd_p_create_dr_res_t res;
  nasd_rpc_status_t status;
  nasd_drive_handle_t h;
  nasd_timespec_t tm;
  nasd_status_t rc;
  int partnum, i;
  char c;

  progname = argv[0];

  nobjs = 1;

  bzero((char *)&args, sizeof(args));
  partnum = 1;

  binding_type = NASD_BIND_DCE_DIRECT;
  binding_args = NULL;
  binding_args_len = 0;

  while ((c = getopt(argc, argv, "kMn:p:T")) != EOF) {
    switch(c) {
      case 'k':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_KPDEV;
        binding_args = &kpdev_args;
        binding_args_len = sizeof(kpdev_args);
        strcpy(kpdev_args.devname, "/dev/nasdkp0");
        break;
      case 'p':
        if (sscanf(optarg, "%d", &partnum) != 1)
          usage();
        break;
      case 'n':
        if (sscanf(optarg, "%d", &nobjs) != 1)
          usage();
        if (nobjs <= 0)
          usage();
        break;
      case 'M':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_MSGQ;
        break;
      case 'T':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_DIRECT_TCP;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }
  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;

  if (optind >= argc)
    usage();
  master_password = argv[optind];
  optind++;

  if (optind < argc)
    usage();

  if (partnum < 0)
    usage();

  rc = nasd_cl_p_init();
  if (rc) {
    printf("ERROR (%s:%d): cannot init client library, rc=%ld\n",
      __FILE__, __LINE__, (long)rc);
    return(rc);
  }

  rc = nasd_bind_to_drive(server_name, binding_port,
    binding_type, binding_args, binding_args_len, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    exit(1);
  }

  for(i=0;i<nobjs;i++) {
    /* Fill in capability and security param */
    nasd_security_password_to_keys(master_password,
      master_key, drive_key, partnum, part_key, red_key, black_key);
    bzero((char *)&args,sizeof(nasd_p_create_dr_args_t));
    args.in_security_param.type=NASD_BLACK_KEY;

    args.in_security_param.partnum = partnum;
    args.in_security_param.actual_protection = NASD_SECURE_OPS;

    /* Testing standard AV */ 
    args.in_fieldmask=NASD_ATTR_AV;
    args.in_attribute.av=7;

    args.in_partnum = partnum;
    nasd_cl_p_create_dr(h, black_key, &args, &res, &status);

    if (status) {
      fprintf(stderr, "ERROR: status=0x%x (%s) nasd_status=0x%x (%s)\n",
        status, nasd_cl_error_string(h, status, err_str),
        res.nasd_status, nasd_error_string(res.nasd_status));
      fflush(stderr);
      exit(2);
    }

    if (res.nasd_status) {
      fprintf(stderr, "ERROR: nasd_status=0x%x (%s) i=%d\n",
        res.nasd_status, nasd_error_string(res.nasd_status), i);
      fflush(stderr);
      exit(2);
    }

    if (nobjs > 1) {
      printf("Succeeded, id=0x%" NASD_ID_FMT " obj #%d\n",
        res.out_identifier, i+1);
    }
    else {
      printf("Succeeded, id=0x%" NASD_ID_FMT "\n", res.out_identifier);
    }
    fflush(stdout);
  }

  rc = nasd_unbind_drive(&h);
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) from nasd_unbind_drive()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(2);
  }

  nasd_cl_p_shutdown();

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
