/*
 * pcrpart.c
 *
 * Create partition on a NASD drive.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>

char *progname;

extern char *optarg;
extern int   optind;

int binding_type = NASD_BIND_DCE_DIRECT;

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername partnum nblocks protection_level master_password\n", progname);
  fprintf(stderr, "    partnum starts at 1\n");
  fprintf(stderr, "    nblocks in NASD starts at 1\n");

  fprintf(stderr, "  where protection levels are \n" );
  fprintf(stderr,"     0 none\n");
#if NASD_SECURE_RPCS_ENABLE > 0
  fprintf(stderr,"     1 INTEGRITY_ARGS\n");
  fprintf(stderr,"     2 INTEGRITY_DATA\n");
  fprintf(stderr,"     3 INTEGRITY_ARGS | INTEGRITY_DATA\n");
  fprintf(stderr,"     4 PRIVACY_DATA\n");
  fprintf(stderr,"     5 PRIVACY_DATA | PRIVACY_ARG\n");
  fprintf(stderr,"     6 {INTEGRITY,PRIVACY}_{ARGS,DATA}\n");
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -M use message queues\n");
  fprintf(stderr, "  -T use TCP\n");

  fflush(stderr);
  exit(1);
}

int
main(argc, argv)
  int     argc;
  char  **argv;
{
  nasd_key_t part_key,master_key,drive_key,red_key,black_key;
  nasd_p_part_creat_dr_args_t args;
  nasd_p_part_creat_dr_res_t res;
  nasd_error_string_t error_text;
  nasd_rpc_status_t status;
  nasd_drive_handle_t h;
  int partnum, nblocks;
  char *server_name, c;
  nasd_status_t rc;
  int sec_level=0;

  progname = argv[0];

  while((c = getopt(argc, argv, "MT")) != EOF) {
    switch(c) {
      case 'M':
        binding_type = NASD_BIND_MSGQ;
        break;
      case 'T':
        binding_type = NASD_BIND_DCE_DIRECT_TCP;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }

  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;

  if (optind >= argc)
    usage();
  if (sscanf(argv[optind], "%d", &partnum) != 1)
    usage();
  optind++;
  if (partnum < 1)
    usage();

  if (optind >= argc)
    usage();
  if (sscanf(argv[optind], "%d", &nblocks) != 1)
    usage();
  optind++;
  if (nblocks < 1)
    usage();

  if (optind >= argc)
    usage();
  if (sscanf(argv[optind],"%d",&sec_level) != 1 ) 
    usage();
  optind++;

  bzero((char *)&args, sizeof(nasd_p_part_creat_dr_args_t)); 

  if (sec_level== 0)
    args.in_min_protection=NASD_NO_PROTECTION;
#if NASD_SECURE_RPCS_ENABLE > 0
  else if (sec_level==1)
    args.in_min_protection=NASD_INTEGRITY_ARGS;
  else if (sec_level==2)
    args.in_min_protection=NASD_INTEGRITY_DATA;
  else if (sec_level==3)
    args.in_min_protection=NASD_INTEGRITY_ARGS | NASD_INTEGRITY_DATA;
  else if (sec_level==4)
    args.in_min_protection=NASD_PRIVACY_DATA;
  else if (sec_level==5)
    args.in_min_protection=NASD_PRIVACY_DATA | NASD_PRIVACY_ARGS;
  else if (sec_level==6)
    args.in_min_protection=NASD_PRIVACY_DATA | NASD_PRIVACY_ARGS | NASD_INTEGRITY_DATA | NASD_INTEGRITY_ARGS;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  else {
    fprintf(stderr, "Unknown security level %d\n", sec_level);
    usage();
  }
  printf("Security Level %d   protection %d\n",sec_level,args.in_min_protection);

  rc = nasd_cl_p_init();
  if (rc) {
    printf("ERROR (%s:%d): cannot init client library, rc=%ld\n",
      __FILE__, __LINE__, (long)rc);
    return(rc);
  }

  if (optind >= argc)
    usage();
  nasd_security_password_to_keys(argv[optind], master_key, drive_key,
    partnum, part_key, red_key, black_key);
  optind++;

  if (optind < argc)
    usage();

  rc = nasd_bind_to_drive(server_name, NASD_PDRIVE_PORT,
    binding_type, NULL, 0, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    exit(1);
  }

  /* Fill in capability and security param */

  args.in_security_param.type = NASD_DRIVE_KEY;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;
  args.in_partnum = partnum;
  args.in_blkcnt = nblocks;
  bcopy(part_key,args.in_partition_key,sizeof(nasd_key_t));
  bcopy(red_key,args.in_red_key,sizeof(nasd_key_t));
  bcopy(black_key,args.in_black_key,sizeof(nasd_key_t));

  nasd_cl_p_part_creat_dr(h, drive_key, &args, &res, &status);

  if (res.nasd_status || status) {
    fprintf(stderr, "ERROR: nasd_status 0x%x (%s) status 0x%x (%s)\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text));
    exit(2);
  }

  nasd_unbind_drive(&h);
  nasd_cl_p_shutdown();

  printf("Succeeded\n");
  fflush(stdout);

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
