/*
 * pgetattr.c
 *
 * send getattr to NASD drive
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_security_capability.h>
#include <nasd/nasd_security_time.h>

char *progname;

int                           binding_type;
int                           binding_args_len;
void                         *binding_args;
char                         *binding_port = NASD_PDRIVE_PORT;
nasd_drive_param_dce_kpdev_t  kpdev_args;
int                           nondefault_binding = 0;

extern char *optarg;
extern int   optind;
nasd_key_t part_key,master_key,drive_key,red_key,black_key;
nasd_error_string_t error_text;

nasd_status_t
nasd_security_get_red_key(
  nasd_partnum_t  partnum,
  nasd_key_t      key)
{
  bcopy(red_key,key,sizeof(nasd_key_t));
  return 0;
}

nasd_status_t
nasd_security_get_black_key(
  nasd_partnum_t  partnum,
  nasd_key_t      key)
{
  bcopy(black_key,key,sizeof(nasd_key_t));
  return 0;
}

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername partnum nasdid master_password\n", progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -k use kernel device\n");
  fprintf(stderr, "  -M use message queues\n");
  fprintf(stderr, "  -T use TCP\n");
  fflush(stderr);
  exit(1);
}

int
main(
  int     argc,
  char  **argv)
{
  nasd_p_getattr_dr_args_t args;
  nasd_p_getattr_dr_res_t res;
  nasd_rpc_status_t status;
  nasd_drive_handle_t h;
  char *server_name;
  nasd_status_t rc;
  int p;
  char *master_password;
  nasd_timespec_t tm;
  nasd_cookie_t chocolate_chip;
  char c;

  bzero((char *)&args,sizeof(nasd_p_getattr_dr_args_t));
  progname = argv[0];

  binding_type = NASD_BIND_DCE_DIRECT;
  binding_args = NULL;
  binding_args_len = 0;

  while ((c = getopt(argc, argv, "kMT")) != EOF) {
    switch(c) {
      case 'k':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_KPDEV;
        binding_args = &kpdev_args;
        binding_args_len = sizeof(kpdev_args);
        strcpy(kpdev_args.devname, "/dev/nasdkp0");
        break;
      case 'M':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_MSGQ;
        break;
      case 'T':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_DIRECT_TCP;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }

  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;

  bzero((char *)&args, sizeof(args));

  if (optind >= argc)
    usage();
  if (sscanf(argv[optind], "%d", &p) != 1)
    usage();
  optind++;
  args.in_partnum = p;

  if (optind >= argc)
    usage();
  rc = nasd_str_to_nasd_id(argv[optind], &args.in_identifier);
  optind++;
  if (rc)
    usage();

  if (optind >= argc)
    usage();
  master_password=argv[optind];
  optind++;

  if (optind < argc)
    usage();

  rc = nasd_cl_p_init();
  if (rc) {
    fprintf(stderr, "ERROR: cannot init client library, rc=0x%x (%s)\n",
      rc, nasd_error_string(rc));
    exit(1);
  }

  rc = nasd_bind_to_drive(server_name, binding_port,
    binding_type, binding_args, binding_args_len, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    exit(1);
  }

 /* Fill in capability and security param */
  nasd_security_password_to_keys(master_password,master_key,drive_key,args.in_partnum,part_key,red_key,black_key);
 
  args.in_security_param.type=NASD_BLACK_KEY;

  nasd_security_get_drive_time(0,&tm);
  tm.ts_sec+=(60*60);
  nasd_security_build_capability(0,args.in_partnum,args.in_identifier,NASD_ACCESS_RIGHTS_GETATTR,0,tm.ts_sec,
                                 NASD_SECURE_OPS,NASD_BLACK_CAPABILITY,0,0,7,
                                 &chocolate_chip);
  bcopy((char *)&chocolate_chip.capability_args,
    (char *)&args.in_capability_args, sizeof(nasd_capability_args_t));
  args.in_security_param.type = args.in_capability_args.type; 


  args.in_security_param.partnum = args.in_partnum;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;


  nasd_cl_p_getattr_dr(h,chocolate_chip.key,&args, &res, &status);

  if (res.nasd_status || status) {
    fprintf(stderr, "ERROR: nasd_status=0x%x (%s) status=0x%x (%s)\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text));
    exit(1);
  }

  nasd_printstat(&res.out_attribute);

  rc = nasd_unbind_drive(&h);
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) from nasd_unbind_drive()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(2);
  }

  nasd_cl_p_shutdown();

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
