/*
 * plspart.c
 *
 * List objects in a partition
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_security_capability.h>
#include <nasd/nasd_security_time.h>
#include <nasd/nasd_nfs_types.h>

char *progname;

nasd_drive_handle_t  h;
int                  partnum;

int                           binding_type;
int                           binding_args_len;
void                         *binding_args;
char                         *binding_port = NASD_PDRIVE_PORT;
nasd_drive_param_dce_kpdev_t  kpdev_args;
int                           nondefault_binding = 0;

extern char *optarg;
extern int   optind;
nasd_key_t part_key,master_key,drive_key,red_key,black_key;

#define OBJ_BUF_LEN 10

void (*show_entry)(nasd_identifier_t nid);

extern char *_nasd_timestr_r();
#define DOTIME(_ts_,_str_) nasd_timestr_r((_ts_),_str_), (_ts_).ts_sec, (_ts_).ts_nsec

nasd_status_t
nasd_security_get_red_key(
  nasd_partnum_t  partnum,
  nasd_key_t      key)
{
  bcopy(red_key,key,sizeof(nasd_key_t));
  return 0;
}

nasd_status_t
nasd_security_get_black_key(
  nasd_partnum_t  partnum,
  nasd_key_t      key)
{
  bcopy(black_key,key,sizeof(nasd_key_t));
  return 0;
}

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername partnum master_password\n",
    progname);
  fflush(stderr);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -k use kernel device\n");
  fprintf(stderr, "  -M use message queues\n");
  fprintf(stderr, "  -T use TCP\n");
  exit(1);
}

void
show_entry_simple(
  nasd_identifier_t  nid)
{
  printf("0x%" NASD_ID_FMT "\n", nid);
}

void
show_entry_jimz(
  nasd_identifier_t  nid)
{
  char t1[128], t2[128], t3[128], t4[128], t5[128];
  nasd_error_string_t error_text;
  nasd_p_getattr_dr_args_t args;
  nasd_p_getattr_dr_res_t res;
  nasd_rpc_status_t status;
  nasd_attribute_t *attr;
  nasd_cookie_t cookie;
  nasd_timespec_t tm;

  args.in_partnum = partnum;
  args.in_identifier = nid;

  args.in_security_param.type = NASD_BLACK_KEY;
  nasd_security_get_drive_time(0, &tm);
  tm.ts_sec+=(60*60);
  nasd_security_build_capability(0, args.in_partnum, args.in_identifier,
    NASD_ACCESS_RIGHTS_GETATTR, 0, tm.ts_sec,
    NASD_SECURE_OPS, NASD_BLACK_CAPABILITY, 0,
    0, 7, &cookie);
  args.in_security_param.type = args.in_capability_args.type; 
  args.in_security_param.partnum = args.in_partnum;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;

  nasd_cl_p_getattr_dr(h, cookie.key, &args, &res, &status);
  if (res.nasd_status || status) {
    fprintf(stderr, "ERROR: nasd_status=0x%x (%s) status=0x%x (%s) nid=0x%" NASD_ID_FMT "\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text), nid);
    exit(1);
  }

  attr = &res.out_attribute;

  printf("0x%" NASD_ID_FMT "  len %16" NASD_64u_FMT
    "  amtime=%s(%d:%09d) mdtime=%s(%d:%09d) crtime=%s(%d:%09d)"
    " fsamtime=%s(%d:%09d) fsmdtime=%s(%d:%09d)\n",
    nid, (nasd_uint64)attr->object_len,
    DOTIME(attr->attr_modify_time, t1),
    DOTIME(attr->object_modify_time, t2),
    DOTIME(attr->object_create_time, t3),
    DOTIME(attr->fs_attr_modify_time, t4),
    DOTIME(attr->fs_object_modify_time, t5));
}

void
show_entry_jimz_nfs(
  nasd_identifier_t  nid)
{
  nasd_nfs3_attributes_t *nfs3attr;
  char t1[128], t2[128], t3[128];
  nasd_error_string_t error_text;
  nasd_p_getattr_dr_args_t args;
  nasd_p_getattr_dr_res_t res;
  nasd_rpc_status_t status;
  nasd_attribute_t *attr;
  nasd_cookie_t cookie;
  nasd_timespec_t tm;

  args.in_partnum = partnum;
  args.in_identifier = nid;

  args.in_security_param.type = NASD_BLACK_KEY;
  nasd_security_get_drive_time(0, &tm);
  tm.ts_sec+=(60*60);
  nasd_security_build_capability(0, args.in_partnum, args.in_identifier,
    NASD_ACCESS_RIGHTS_GETATTR, 0, tm.ts_sec,
    NASD_SECURE_OPS, NASD_BLACK_CAPABILITY, 0,
    0, 7, &cookie);
  args.in_security_param.type = args.in_capability_args.type; 
  args.in_security_param.partnum = args.in_partnum;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;

  nasd_cl_p_getattr_dr(h, cookie.key, &args, &res, &status);
  if (res.nasd_status || status) {
    fprintf(stderr, "ERROR: nasd_status=0x%x (%s) status=0x%x (%s) nid=0x%" NASD_ID_FMT "\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text), nid);
    exit(1);
  }

  attr = &res.out_attribute;
  nfs3attr = (nasd_nfs3_attributes_t *)attr->fs_specific;
  printf("0x%" NASD_ID_FMT "  type 0x%02x mode 0x%x nlink %d uid %u gid %u",
    nid, nfs3attr->type&0xff, nfs3attr->mode, nfs3attr->nlink,
    nfs3attr->uid, nfs3attr->gid);
  if (nfs3attr->name_debug[0]) {
    printf(" \"%s\"", nfs3attr->name_debug);
  }
  printf("\n");
}

void
show_entry_jimz_blocks(
  nasd_identifier_t  nid)
{
  char t1[128], t2[128], t3[128];
  nasd_error_string_t error_text;
  nasd_p_getattr_dr_args_t args;
  nasd_p_getattr_dr_res_t res;
  nasd_rpc_status_t status;
  nasd_attribute_t *attr;
  nasd_cookie_t cookie;
  nasd_timespec_t tm;

  args.in_partnum = partnum;
  args.in_identifier = nid;

  args.in_security_param.type = NASD_BLACK_KEY;
  nasd_security_get_drive_time(0, &tm);
  tm.ts_sec+=(60*60);
  nasd_security_build_capability(0, args.in_partnum, args.in_identifier,
    NASD_ACCESS_RIGHTS_GETATTR, 0, tm.ts_sec,
    NASD_SECURE_OPS, NASD_BLACK_CAPABILITY, 0,
    0, 7, &cookie);
  args.in_security_param.type = args.in_capability_args.type; 
  args.in_security_param.partnum = args.in_partnum;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;

  nasd_cl_p_getattr_dr(h, cookie.key, &args, &res, &status);
  if (res.nasd_status || status) {
    fprintf(stderr, "ERROR: nasd_status=0x%x (%s) status=0x%x (%s) nid=0x%" NASD_ID_FMT "\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text), nid);
    exit(1);
  }

  attr = &res.out_attribute;

  printf("0x%" NASD_ID_FMT "  len %16" NASD_64u_FMT "  blocks %16" NASD_64u_FMT "  prealloc %16" NASD_64u_FMT "\n",
    nid, (unsigned long)attr->object_len,
    (unsigned long)attr->blocks_used,
    (unsigned long)attr->block_preallocation);
}

int
main(
  int     argc,
  char  **argv)
{
  nasd_identifier_t objs[OBJ_BUF_LEN];
  char *master_password, *server_name;
  nasd_p_smpl_op_dr_args_t rd_args;
  nasd_p_fastread_dr_res_t rd_res;
  nasd_error_string_t err_str;
  nasd_rpc_status_t status;
  nasd_cookie_t cookie;
  nasd_timespec_t tm;
  nasd_offset_t off;
  nasd_status_t rc;
  nasd_len_t len;
  char c;
  int i;

  progname = argv[0];
  show_entry = show_entry_simple;

  binding_type = NASD_BIND_DCE_DIRECT;
  binding_args = NULL;
  binding_args_len = 0;

  while((c = getopt(argc, argv, "jkMnTv")) != EOF) {
    switch(c) {
      case 'j':
        show_entry = show_entry_jimz;
        break;
      case 'n':
        show_entry = show_entry_jimz_nfs;
        break;
      case 'k':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_KPDEV;
        binding_args = &kpdev_args;
        binding_args_len = sizeof(kpdev_args);
        strcpy(kpdev_args.devname, "/dev/nasdkp0");
        break;
      case 'M':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_MSGQ;
        break;
      case 'T':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_DIRECT_TCP;
        break;
      case 'v':
        show_entry = show_entry_jimz_blocks;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }

  if (optind >= argc)
    usage();

  server_name = argv[optind];

  optind++;
  if (optind >= argc)
    usage();

  if (sscanf(argv[optind], "%d", &partnum) != 1)
    usage();

  optind++;
  if (optind >= argc)
    usage();

  master_password = argv[optind];

  optind++;
  if (optind < argc)
    usage();

  rc = nasd_cl_p_init();
  if (rc) {
    fprintf(stderr,"ERROR (%s:%d): cannot init client library, rc=0x%x (%s)\n",
      __FILE__, __LINE__, rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_bind_to_drive(server_name, binding_port,
    binding_type, binding_args, binding_args_len, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    exit(1);
  }

  /* Fill in capability and security param */
  nasd_security_password_to_keys(master_password, master_key,
    drive_key, partnum, part_key, red_key, black_key);

  /* No security on control objects */ 
  bzero((char *)&cookie, sizeof(cookie));

  off = 0;
  do {
    rd_args.in_partnum = partnum;
    rd_args.in_identifier = (nasd_identifier_t)NASD_CTRL_PART_OBJS;
    rd_args.in_offset = off;
    rd_args.in_len = sizeof(objs);
    rd_args.in_capability_args = cookie.capability_args;
    rd_args.in_security_param.type = cookie.capability_args.type;
    rd_args.in_security_param.partnum = partnum;
    rd_args.in_security_param.actual_protection = cookie.capability_args.min_protection;
    nasd_cl_p_read_simple_dr(h, cookie.key, &rd_args, objs, &rd_res, &status);
    len = rd_res.out_datalen;
    if (status || rd_res.nasd_status) {
      fprintf(stderr,
        "Error issuing read, off=%lu status=0x%x (%s) nasd_status=0x%x (%s)\n",
        off, status, nasd_cl_error_string(h, status, err_str),
        rd_res.nasd_status, nasd_error_string(rd_res.nasd_status));
      fflush(stderr);
      exit(1);
    }
    for(i=0;i<len/sizeof(nasd_identifier_t);i++) {
      show_entry(objs[i]);
    }
    off += len;
  } while(len);

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
