/*
 *  gframe.c -- generation of frames for assignment.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

/*
 *  We start by generating "morsels" of byte- or sub-byte-sized source and
 *  destination bitstrings.  The morsel lists are merged to create a list
 *  of "frames" combining a source and destination morsel.  Frames are
 *  also created for extra destination bits that must be zero filled or
 *  sign extended.  The frame list is then sorted in source byte order,
 *  and adjacent frames are combined, before being returned.
 */



#include "usc.h"

/* 
 *  Morsel:  a set of contiguous source or destination bits within one byte.
 *  Mlist:   a list of morsels.
 */
typedef struct {
    int m_offset;			/* offset to byte */
    char m_nbits;			/* number of bits */
    char m_shift;			/* offset within byte */
    char m_signed;			/* nonzero if part of a signed field */
} Morsel;

typedef struct {
    int m_length;			/* number of morsels */
    Morsel m_data[MAXINTBITS];		/* morsel data */
} Mlist;

static Mlist srclist, dstlist;		/* source and destination lists */
static Frame anchor;			/* pointer to head/tail of frame list */
static int nframes;			/* number of frames */



/*  local functions  */ 

static void gfobj(Type *lt, int loff, Type *rt, int roff); /*gen frame for obj*/
static void gfint(Type *lt, int loff, Type *rt, int roff); /*gen frame for int*/
static void gmorsels(Mlist *list, Type *t, int off);	/* gen morsels for int*/
static void addmorsel(Mlist *list, int nbits, int offset, int shift, int sflag);
static void marry(void);		/* combine morsel lists to gen frames */
static void addframe			/* add frame to list */
    (int dstoff, int srcoff, int nbits, int dshift, int sshift, int flags);
static int framecmp(const void *p1, const void *p2);	/* sorting comparison */
static void collapse(void);		/* combine adjacent frames */
static void pframe(Frame *f);		/* print frame on stderr */



/*  gframes(ltype, rtype) -- generate frames for assignment  */

Frame *gframes(Type *ltype, Type *rtype)
{
    Frame *f, *flist;

    anchor.f_prev = NULL;		/* init the list to empty */
    anchor.f_next = NULL;
    nframes = 0;

    gfobj(ltype, 0, rtype, 0);		/* generate frames (recursively) */
    flist = anchor.f_next;		/* save head of frame list */

    if (dbflags[DB_UFRAMES]) {
	fprintf(stderr, "unsorted frames:\n");
	for (f = flist; f; f = f->f_next)
	    pframe(f);
    }

    flist = sortframes(flist,		/* sort frames in src order */
	offsetof(Frame, f_srcoff), offsetof(Frame, f_srcshift));

    if (dbflags[DB_SFRAMES]) {
	fprintf(stderr, "sorted frames:\n");
	for (f = flist; f; f = f->f_next)
	    pframe(f);
    }

    collapse();				/* combine adjacent frames */

    if (dbflags[DB_FRAMES]) {
	fprintf(stderr, "final frame list:\n");
	for (f = flist; f; f = f->f_next)
	    pframe(f);
    }

    return flist;			/* return head of chain */
}



/*  gfobj(ltype, loffset, rtype, roffset) -- generate frames for object  */

static void gfobj(Type *ltype, int loffset, Type *rtype, int roffset)
{
    Symptr ls, rs;
    Type *lu, *ru;
    int i, n, lz, rz;

    if (ltype->t_tflags & T_INT)
	gfint(ltype, loffset, rtype, roffset);
    else if (ltype->t_tcode == T_STRUCT)
	for (ls = ltype->t_mbr, rs = rtype->t_mbr; ls && rs;
	    ls = ls->s_next, rs = rs->s_next) {
		if (dbflags[DB_MORSELS] || dbflags[DB_IFRAMES])
		    fprintf(stderr, "field: %s\n", ls->s_name);
		gfobj(ls->s_type, loffset + ls->s_type->t_annot.a_align,
		    rs->s_type, roffset + rs->s_type->t_annot.a_align);
	}
    else if (ltype->t_tcode == T_ARRAY) {
	n = ltype->t_bound;
	lu = ltype->t_utype;
	ru = rtype->t_utype;
	lz = lu->t_annot.a_msize;
	rz = ru->t_annot.a_msize;
	if (n == A_NOBOUNDS || n == A_VARBOUNDS)
	    tfatal(NULL, "need fixed bounds to copy array", NULL);
	else
	    for (i = 0; i < n; i++)
		gfobj(lu, loffset + i * lz, ru, roffset + i * rz);
    } else
	MALFUNC("bad type in gfobj");
}



/*  gfint(ltype, loffset, rtype, roffset) -- generate frames for integer  */

static void gfint(Type *ltype, int loffset, Type *rtype, int roffset)
{
    if (dbflags[DB_MORSELS])
	fprintf(stderr, "dst morsels:\n");
    gmorsels(&dstlist, ltype, loffset);		/* make list of dest fields */

    if (dbflags[DB_MORSELS])
	fprintf(stderr, "src morsels:\n");
    gmorsels(&srclist, rtype, roffset);		/* make list of source fields */

    if (dbflags[DB_IFRAMES])
	fprintf(stderr, "uncombined frames:\n");
    marry();					/* generate frames from lists */
}



/*  gmorsels(list, type, offset) -- make list of morsels <=8 bits, LSB first  */

static void gmorsels(Mlist *list, Type *t, int off)
{
    int i, av, nbits, fbit, sflag;
    Order *o, *b;
    Oseg *g;

    list->m_length = 0;				/* clear out list */
    o = t->t_annot.a_order;			/* byte order */
    b = t->t_annot.a_bitorder;			/* bit order */
    sflag = !(t->t_tflags & TF_UNSIGNED);	/* signed flag */

    if (!b) {				/* if not a bitfield */
	for (i = 0; i < t->t_annot.a_tsize; i++)
	    addmorsel(list, BYTE, off + boff(o, i), 0, sflag);
					/* full-byte morsels */
	return;
    }

    /* need to specify a bitfield */
    for (i = 0; i < b->o_nsegs; i++) {		/* for each order segment: */
	g = &b->o_oseg[i];
	nbits = g->o_length;			/* number of bits needed */
	fbit = g->o_start;			/* location of first bit */
	while (nbits > 0) {
	    av = BYTE - (g->o_start % BYTE);	/* bits available in byte */
	    if (av > nbits)
		av = nbits;			/* don't use more than needed */
	    if (g->o_dir < 0)
		av = 1;				/* if bits go backwards */

	    addmorsel(list, av, off + boff(o, fbit / BYTE), fbit % BYTE, sflag);

	    nbits -= av;			/* account for bits selected */
	    fbit += av * g->o_dir;		/* advance bit position */
	}
    }
}



/*  addmorsel(list, nbits, offset, shift, sflag) -- add one morsel to list  */

static void addmorsel(Mlist *list, int nbits, int offset, int shift, int sflag)
{
    Morsel *b;

    if (dbflags[DB_MORSELS])
	fprintf(stderr, "%4d%c@ %+2d >>%d\n", nbits, sflag ? 's' : 'u',
	    offset, shift);
    b = &list->m_data[list->m_length++];
    b->m_nbits = nbits;
    b->m_offset = offset;
    b->m_shift = shift;
    b->m_signed = sflag;
}



/*
 *  marry() -- generate frames of contiguous moves from morsel lists.
 *
 *  No attempt is made to combine adjacent bytes yet.
 */

static void marry(void)
{
    int ndst, nsrc, nbits;
    Morsel *pdst, *psrc;
    Morsel d, s;

    /* set up pointers and counters */
    ndst = dstlist.m_length;
    pdst = dstlist.m_data;
    nsrc = srclist.m_length;
    psrc = srclist.m_data;

    /* init current morsels empty */
    d.m_nbits = 0;
    s.m_nbits = 0;

    /* loop until all destination morsels have been processed */
    for (;;) {

	/* load destination morsel */
	if (d.m_nbits == 0) {
	    if (ndst-- > 0)
		d = *pdst++;
	    else
		break;
	}

	/* load source morsel */
	if (s.m_nbits == 0) {
	    if (nsrc-- > 0)
		s = *psrc++;
	    else {
		/* no more source bits available */
		if (s.m_signed) {
		    /* get sign bit from previous morsel */
		    s = psrc[-1];
		    addframe(d.m_offset, s.m_offset, d.m_nbits, d.m_shift,
			s.m_shift + s.m_nbits - 1, MV_SIGN);
		    s.m_nbits = 0;
		} else {
		    /* zero fill */
		    addframe(d.m_offset, 0, d.m_nbits, d.m_shift, 0, MV_ZERO);
		}
		d.m_nbits = 0;
		continue;
	    }
	}

	/* move the smaller of the destination and source sizes */
	nbits = d.m_nbits;
	if (nbits > s.m_nbits)
	    nbits = s.m_nbits;
	addframe(d.m_offset, s.m_offset, nbits, d.m_shift, s.m_shift, 0);

	/* adjust the working copies of the morsels to reflect consumed bits */
	d.m_shift += nbits;
	s.m_shift += nbits;
	d.m_nbits -= nbits;
	s.m_nbits -= nbits;
    }
}



/*
 *  addframe(dstoff, srcoff, nbits, dshift, sshift, flags) -- add to list.
 *
 *  Frames of 8 bits are turned into frames of one byte.
 *  New full-byte frames are combined with the previous frame when possible.
 */
static void addframe
    (int dstoff, int srcoff, int nbits, int dshift, int sshift, int flags)
{
    Frame *f, ftmp;
    int n, nbytes;

    if (nbits == BYTE) {
	nbytes = 1;
	nbits = 0;
    } else {
	nbytes = 0;
    }

    ftmp.f_dstoff = dstoff;
    ftmp.f_srcoff = srcoff;
    ftmp.f_nbytes = nbytes;
    ftmp.f_nbits = nbits;
    ftmp.f_dstshift = dshift;
    ftmp.f_srcshift = sshift;
    ftmp.f_flags = flags;

    if (dbflags[DB_IFRAMES])
	pframe(&ftmp);

    /*
     * See if frame can be appended or prepended to previous frame.
     * Destinations must be contiguous.  Sources must be contiguous
     * for a move or identical for a sign-extend.
     */
    f = anchor.f_prev;
    if (nbytes > 0 && f != NULL && flags == f->f_flags && (n=f->f_nbytes) > 0) {

	if ((f->f_dstoff + n == dstoff) && (
	    (flags == 0 && f->f_srcoff + n == srcoff) ||
	    (flags == MV_ZERO) ||
	    (flags == MV_SIGN && f->f_srcoff == srcoff && f->f_srcshift==sshift)
	)) {
	    f->f_nbytes += nbytes;
	    return;
	}

	if ((f->f_dstoff - nbytes == dstoff) && (
	    (flags == 0 && f->f_srcoff - nbytes == srcoff) ||
	    (flags == MV_ZERO) ||
	    (flags == MV_SIGN && f->f_srcoff == srcoff && f->f_srcshift==sshift)
	)) {
	    f->f_dstoff = dstoff;
	    f->f_srcoff = srcoff;
	    f->f_nbytes += nbytes;
	    return;

	}
    }

    /* allocate and initialize new Frame struct */
    f = zalloc(sizeof(Frame));
    *f = ftmp;

    /* link into chain */
    if (anchor.f_prev)
	anchor.f_prev->f_next = f;
    else
	anchor.f_next = f;
    f->f_prev = anchor.f_prev;
    f->f_next = NULL;
    anchor.f_prev = f;
    nframes++;
}



/*
 *  sortframes(flist, maj, min) -- sort frames based on offset and shift.
 *
 *  maj is offset to major key, an int-sized value holding an offset.
 *  min is offset to minor key, a char-sized value holding a shift count.
 */

static int majoff, minoff;	/* Frame struct offsets for current sort */

Frame *sortframes(Frame *flist, int maj, int min)
{
    Frame **ptrlist, **p, *f;
    int i;

    if (nframes < 2)
	return flist;				/* nothing to do */

    /* make a linear list of pointers */
    p = ptrlist = zalloc((nframes) * sizeof(Frame *));
    for (f = flist; f; f = f->f_next)
	*p++ = f;

    /* sort the list */
    majoff = maj;
    minoff = min;
    qsort(ptrlist, nframes, sizeof(Frame *), framecmp);

    /* rebuild the linked list */

    f = flist = ptrlist[0];			/* head element */
    f->f_prev = NULL;
    f->f_next = ptrlist[1];

    for (i = 1; i < nframes - 1; i++) {		/* interior elements */
	f = ptrlist[i];
	f->f_prev = ptrlist[i-1];
	f->f_next = ptrlist[i+1];
    }

    f = ptrlist[nframes - 1];			/* tail element */
    f->f_prev = ptrlist[nframes - 2];
    f->f_next = NULL;

    free(ptrlist);
    return flist;
}



/*  framecmp(*f1, *f2) -- compare frames for sorting using majoff, minoff  */

static int framecmp(const void *p1, const void *p2)
{
    Frame *f1, *f2;
    int n;

    f1 = * (Frame **) p1;
    f2 = * (Frame **) p2;
    n = IntAt(f1, majoff) - IntAt(f2, majoff);
    if (n != 0)
	return n;
    else
	return CharAt(f1, minoff) - CharAt(f2, minoff);
}



/*  collapse() -- combine adjacent byte-move frames  */

/*  This probably isn't needed with the "smart" code generator.  */

static void collapse(void)
{
    Frame *f, *g;

    for (f = anchor.f_next; f; f = f->f_next)
	while ((g = f->f_next) != NULL
	    && f->f_nbytes > 0
	    && g->f_nbytes > 0
	    && f->f_flags == g->f_flags
	    && f->f_dstoff + f->f_nbytes == g->f_dstoff
	    && (
		(f->f_flags == 0 && f->f_srcoff + f->f_nbytes == g->f_srcoff) ||
	        (f->f_flags == MV_ZERO) ||
	        (f->f_flags == MV_SIGN && f->f_srcoff == g->f_srcoff
		    && f->f_srcshift == g->f_srcshift)
	)) {
	    f->f_nbytes += g->f_nbytes;
	    f->f_next = g->f_next;
	    if (f->f_next)
		f->f_next->f_prev = f;
	    else
		anchor.f_prev = f;
	    free(g);
	    nframes--;
	}
}



/*  pframe(f) -- print a frame on stderr  */

static void pframe(Frame *f)
{
    fprintf(stderr, "   d%+d s%+d %d : %d d>>%d s>>%d", f->f_dstoff,
	f->f_srcoff, f->f_nbytes, f->f_nbits, f->f_dstshift, f->f_srcshift);
    if (f->f_flags & MV_SIGN)
	fprintf(stderr, " s");
    if (f->f_flags & MV_ZERO)
	fprintf(stderr, " z");
    fprintf(stderr, "\n");
}
