/*
 *  grammar.y -- yacc grammar for USC, Version 2.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */


%{

#include "usc.h"

#define YYDEBUG 1		/* enable yacc tracing */

static Scode curcode;		/* symbol code of current declaration */
static Type *curtype;		/* datatype of current declaration */
static Annot curannot;		/* current annotation */

static Symbol tempsym;		/* temporary symbol structure */

%}


%start file


/*  YYSTYPE union members  */

%type <y_long>	bound funcmods
%type <y_sym>   declarator declarators att member members mdecls mdecl mname
		funclabel parmlist parameters param
%type <y_type>  tlist sdecl typespec structure mlist optss opttype abstype
%type <y_annot>	annot annospec optannot
%type <y_order> ordering ordlist
%type <y_oseg> 	orditem
%type <y_mdef>  attlist mdspec
%type <y_node>	adecl idecl ddecl stmt retexpr expr cexpr uexpr pfexpr


/*  terminals  */

%token <y_string> TK_ID		/* identifier */
%token <y_type>	TK_TNAME	/* type name */
%token <y_long>	TK_ICONST	/* integer constant */

%token TK_LANGLE	/*  <   */
%token TK_RANGLE	/*  >   */
%token TK_LBRACE	/*  {   */
%token TK_RBRACE	/*  }   */
%token TK_LBRACKET	/*  [   */
%token TK_RBRACKET	/*  ]   */
%token TK_LPAREN	/*  (   */
%token TK_RPAREN	/*  )   */

%token TK_ASSIGN	/*  =   */
%token TK_ASGPAD	/*  =:  */
%token TK_SEMI		/*  ;	*/

%token TK_COMMA		/*  ,   */
%token TK_PERIOD	/*  .   */
%token TK_AMPER		/*  &   */

%token TK_PLUS		/*  +   */
%token TK_MINUS		/*  -   */
%token TK_ASTER		/*  *   */
%token TK_SLASH		/*  /   */

%token TK_ARROW		/*  ->  */

%token TK_COLON		/*  :   */
%token TK_THRU		/*  ..  */

%token TK_INLINE	/* keywords */
%token TK_MACRO
%token TK_MDEF
%token TK_RETURN
%token TK_SIZEOF
%token TK_STATIC
%token TK_STRUCT
%token TK_TYPEDEF
%token TK_UNION

%token TK_ERROR		/* unrecognized by lex */

/* the following are used only in the parse tree, not in the grammar */
%token TK_UPLUS		/*  +e   */
%token TK_UMINUS	/*  -e   */
%token TK_UASTER	/*  *e   */
%token TK_UAMPER	/*  &e   */
%token TK_SUBSCR	/*  e[e]  */
%token TK_CAST		/*  (type)e  */
%token TK_TMPVAL	/*  expr can be stored in a temp if needed  */


/*  precedence  */

%left	TK_PLUS TK_MINUS
%left	TK_ASTER TK_SLASH


%%


/*------------------------- overall grammar -------------------------*/

file:
	/* empty */		{ fatal("no valid USC input seen", NULL); }
    |	uscfile
    ;

uscfile:
	filepart
    |	uscfile filepart
    ;

filepart:
        mdef
    |	typeannot
    |	typedef
    |	stub
    |	error
    ;


/* typedef (tsize/msize,align,<order>) basictype; */
typeannot:
	TK_TYPEDEF annot TK_TNAME TK_SEMI
				{ if (!($3->t_tflags & TF_BASIC))
				      fatal("not a basic integral type", NULL); 
				  else if ($3->t_tflags & TF_ANNOT)
				      fatal("duplicate annotation", NULL);
				  else if (!okalign($2))
				      fatal("alignment must be power of two",
					 NULL);
				  else if ($3 == chartype && $2.a_msize != 1)
				      fatal("sizeof(char) must be 1", NULL);
#if __alpha
				  else if ($2.a_msize > sizeof(long))
#else /* __alpha */
				  else if ($2.a_msize > sizeof(long long))
#endif /* __alpha */
				      fatal(
			  "target size is too large to build on this platform",
					  NULL);
				  else {
				      $3->t_annot = $2;
				      $3->t_tflags |= TF_ANNOT;
				  }
				}
					      
    ;

/* C-style typedef */
typedef:
	TK_TYPEDEF typespec {curcode = S_TYPEDEF;} declarators TK_SEMI 
    ;


/*------------------------- mdefs -------------------------*/

/* mdef {read=1, write=0} foo; */
mdef:
	TK_MDEF TK_LBRACE attlist TK_RBRACE TK_ID TK_SEMI
				{ Symptr sym;
				  if (sym = defsym($5, S_MDEF)) {
				      sym->s_mdef = $3;
				      sym->s_mdef->m_name = sym->s_name;
				      }
				}
    ;

attlist:
	att			{ $$ = mattrib(NULL, $1); }
    |	attlist TK_COMMA att	{ $$ = mattrib($1, $3); }
    ;

att:
	TK_ID TK_ASSIGN TK_ICONST
				{ tempsym.s_name = $1;
				  tempsym.s_val = $3;
				  $$ = &tempsym;
				}
    ;


/*------------------------- types -------------------------*/

typespec:
	tlist			{ $$ = curtype = tfix($1); }
    |	sdecl			{ $$ = curtype = $1; }
    ;

/* unsigned short int, etc. */
tlist:
	TK_TNAME		{ $$ = tclone($1); }
    |	tlist TK_TNAME		{ $$ = tmerge($1, $2); }
    ;

/* struct|union [id] [{members}] */
sdecl:
    	TK_STRUCT structure	{ $$ = $2;
				  if ($2->t_tcode == T_UNION)
				     fatal("union tag reused as struct", NULL);
				  $2->t_tcode = T_STRUCT;
				}
    |	TK_UNION structure	{ $$ = $2;
				  if ($2->t_tcode == T_STRUCT)
				     fatal("struct tag reused as union", NULL);
				  warning("seeing union tag", NULL);
				  $2->t_tcode = T_UNION;
				}
    ;
	
structure:
	TK_ID optss		{ Symptr sym = lookup($1, S_TAG);
				  if (sym->s_type == NULL)
				      sym->s_type = $2;
				  else if (sym->s_type->t_mbr == NULL)
				      sym->s_type->t_mbr = $2->t_mbr;
				  else if ($2->t_mbr != NULL)
				      fatal("redefinition of tag", $1);
				  sym->s_type->t_tag = sym;
				  $$ = sym->s_type;
				}
    |	mlist			{ $1->t_tag = mktag($1); $$ = $1; } 
    ;

optss:
	/* empty */		{ $$ = newtype(T_UNK, NULL); }
    |	mlist			{ $$ = $1; }
    ;

/* { list of struct/union members } */
mlist:
	TK_LBRACE members TK_RBRACE
				{ $$ = newtype(T_UNK, NULL);
				  $$->t_mbr = $2;
				}
    ;

members:
	member			{ $$ = $1; }
    |	members member		{ $$ = mappend($1, $2); }
    ;

member:
	typespec		{ curcode = S_MEMBER;
				  curtype->t_annot = noannot;
				}
	    mdecls TK_SEMI
				{ $$ = $3; }
    |	error			{ $$ = NULL; }
    ;

mdecls:
	mname			{ $$ = $1; }
    |	mdecls TK_COMMA mname	{ $$ = mappend($1, $3); }
    ;

mname:
	mdecl			{ if (!okmember($1->s_type))
				      fatal("invalid member type", $1->s_name);
				}
    ;

mdecl:
	declarator
    |	declarator TK_COLON TK_ICONST annot	{ bitorder($1, $3, $4.a_order);}
    |	declarator TK_COLON TK_ICONST ordering	{ bitorder($1, $3, $4); }
    ;


/*------------------------- declarators -------------------------*/

/*
 * A declarator is an identifier with accompanying baggage appearing in a
 * typedef, struct, etc.  For example "int x, *y, z(4,2,<0..3>);" contains
 * three declarators "x", "*y", and "z(4,2,<0..3>)".
 */

declarators:
	declarator
    |	declarators TK_COMMA declarator
    ;

declarator:
	adecl			{ $$ = setdecl($1, curtype, curannot);
				  if (curcode == S_TYPEDEF) {
				      $$->s_type->t_name = $$->s_name;
				      if (!okalign(curannot))
				         fatal("alignment must be power of two",
					     NULL);
				  }
				}
    ;

adecl:
	ddecl			{ $$ = $1; }
    |	TK_ASTER adecl		{ $$ = newnode(TK_UASTER, NULL, $2); }
    ;


ddecl:
	idecl			{ $$ = $1; }
	
    |	ddecl TK_LBRACKET bound TK_RBRACKET {
				  $$ = newnode(TK_SUBSCR, $1, NULL);
				  $$->n_long = $3;
				}
    |	TK_LPAREN adecl TK_RPAREN
				{ $$ = $2; }
    ;


idecl:
	TK_ID optannot		{ Symptr sym = defsym($1, curcode);
				  if (sym == NULL)
				      sym = &tempsym;	/* error recovery */
				  curannot = $2;
				  $$ = newnode(TK_ID, NULL, NULL);
				  $$->n_sym = sym;
				}
    ;


bound:
	/* empty */		{ $$ = A_NOBOUNDS; }
    |	TK_ICONST		{ $$ = $1; }
    |	TK_ID			{ fatal("nonconstant array bounds", NULL);
				  $$ = 1;	/* error recovery */
				}
    ;

/*------------------------- annotations -------------------------*/

optannot:
	/* empty */			{ $$ = noannot; }
    |	annot				{ $$ = $1; }
    ;

annot:
	TK_LPAREN annospec TK_RPAREN	{ $$ = $2; }
    ;

annospec:
	TK_ICONST                    TK_COMMA TK_ICONST TK_COMMA ordering
					{ $$ = mkannot($1, $1, $3, $5); }
    |	TK_ICONST TK_SLASH TK_ICONST TK_COMMA TK_ICONST TK_COMMA ordering
					{ $$ = mkannot($1, $3, $5, $7); }
    |	TK_ICONST TK_COMMA TK_ICONST TK_COMMA TK_ICONST TK_COMMA ordering
					{ $$ = mkannot($1, $3, $5, $7); }
    ;

ordering:
	TK_LANGLE ordlist TK_RANGLE	{ $$ = oclone($2); }
    |	TK_ICONST			{ if ($1 != 0)
					      fatal("illegal order value",NULL);
					  $$ = NULL;
					}
    ;

ordlist:
	orditem				{ $$ = ostart($1); }
    |	ordlist TK_COMMA orditem	{ $$ = oappend($3); }
    ;

orditem:
	TK_ICONST			{ $$ = ointerval($1, $1); }
    |	TK_ICONST TK_THRU TK_ICONST	{ $$ = ointerval($1, $3); }
    ;


/*------------------------- stub header -------------------------*/

stub:
	funclabel TK_LPAREN parmlist TK_RPAREN
					{ curfunc = $1;
					  $1->s_next = $3;
					  genheader($1);
					}
	    body  
					{ gentail($1);
					  rmparms();
					}
    ;

funclabel:
	funcmods opttype TK_ID		{
					  $2 = newtype(T_FUNC, $2);
					  $2->t_tflags |= $1;
					  functype($2);
					  if ($$ = defsym($3, S_FUNCTION))
					      $$->s_type = $2;
					}
    ;

funcmods:
	/* empty */			{ $$ = 0; }
    | 	funcmods TK_STATIC		{ $$ = $1 | TF_STATIC; }
    | 	funcmods TK_INLINE		{ $$ = $1 | TF_INLINE; }
    | 	funcmods TK_MACRO		{ $$ = $1 | TF_MACRO; }
    ;

opttype:
	/* empty */			{ $$ = inttype; }
    |	typespec			{ $$ = $1; }
    |	opttype TK_ASTER		{ $$ = newtype(T_PTR, $1); }
    ;

parmlist:
	/* empty */	{ warning("assuming parameter list of (void)", NULL);
			  $$ = NULL; 
			}
    |	TK_TNAME	{ if ($1->t_tcode != T_VOID)
			      fatal("illegal parameter list", NULL);
			  $$ = NULL; 
			}
    |	parameters	{ $$ = $1; }
    ;

parameters:
	param				{ $$ = $1; }
    |	parameters TK_COMMA param	{ $$ = mappend($1, $3); }
    ;

param:
	typespec {curcode = S_PARAM;} declarator mdspec
					{ if (($$ = $3) != NULL) {
					      $$->s_mdef = $4;
					      /* change array parm to pointer */
					      if ($$->s_type->t_tcode==T_ARRAY)
					      	  $$->s_type->t_tcode = T_PTR;
					      if (!okparm($$->s_type))
						 fatal("illegal parameter type",
						     $$->s_name);
					  }
					}
    ;

mdspec:
	/* empty */		{ $$ = NULL; }
    |	TK_COLON TK_ID		{ Symptr sym = lookup($2, 0);
				  if (sym != NULL && sym->s_scode == S_MDEF)
				      $$ = sym->s_mdef;
				  else {
				      fatal("not an mdef", $2);
				      $$ = NULL;
				  }
				      
				}
    ;


/*------------------------- stub body -------------------------*/

body:
	TK_LBRACE statements TK_RBRACE
    ;

statements:
	/* empty */
    |	statements stmt			{ if ($2 != NULL) statement($2); }
    ;

stmt:
	expr TK_ASSIGN expr TK_SEMI	{ $$ = newnode(TK_ASSIGN, $1, $3); }
    |   expr TK_ASGPAD expr TK_SEMI	{ $$ = newnode(TK_ASGPAD, $1, $3); }
    |	TK_RETURN retexpr TK_SEMI	{ $$ = newnode(TK_RETURN, NULL, $2);
					  $$->n_sym = curfunc;
					}
    |	error				{ $$ = NULL; }
    ;

retexpr:
	/* empty */			{ $$ = NULL; } 
    |	expr				{ $$ = $1; }
    ;

/*------------------------- expressions -------------------------*/

expr:
	cexpr				{ $$ = $1; }
    |	expr TK_PLUS cexpr		{ $$ = newnode(TK_PLUS, $1, $3); }
    |	expr TK_MINUS cexpr		{ $$ = newnode(TK_MINUS, $1, $3); }
    |	expr TK_ASTER cexpr		{ $$ = newnode(TK_ASTER, $1, $3); }
    |	expr TK_SLASH cexpr		{ $$ = newnode(TK_SLASH, $1, $3); }
    ;

cexpr:
	uexpr				{ $$ = $1; }
    |	TK_LPAREN abstype TK_RPAREN uexpr
					{ $$ = newnode(TK_CAST, NULL, $4);
					  $$->n_l=newnode(TK_TYPEDEF,NULL,NULL);
					  $$->n_l->n_type = $2;
					}
    ;

uexpr:
	pfexpr				{ $$ = $1; }
    |	TK_PLUS uexpr			{ $$ = newnode(TK_UPLUS, NULL, $2); }
    |	TK_MINUS uexpr			{ $$ = newnode(TK_UMINUS, NULL, $2); }
    |	TK_ASTER uexpr			{ $$ = newnode(TK_UASTER, NULL, $2); }
    |	TK_AMPER uexpr			{ $$ = newnode(TK_UAMPER, NULL, $2); }
    |	TK_SIZEOF uexpr			{ $$ = newnode(TK_SIZEOF, NULL, $2); }
    |	TK_SIZEOF TK_LPAREN abstype TK_RPAREN
					{ $$ = newnode(TK_SIZEOF, NULL, NULL);
					  $$->n_r=newnode(TK_TYPEDEF,NULL,NULL);
					  $$->n_r->n_type = $3;
					}
    ;

pfexpr:
	TK_ID				{ $$ = newnode(TK_ID, NULL, NULL);
					  $$->n_name = $1;
					} 
    |	TK_ICONST			{ $$ = newnode(TK_ICONST, NULL, NULL);
					  $$->n_long = $1;
					}
    |	TK_LPAREN expr TK_RPAREN	{ $$ = $2; }
    |	pfexpr TK_LBRACKET expr TK_RBRACKET  { $$ = newnode(TK_SUBSCR, $1, $3);}
    |	pfexpr TK_PERIOD TK_ID		{ $$ = newnode(TK_PERIOD, $1, NULL);
					  $$->n_name = $3;
					}
    |	pfexpr TK_ARROW TK_ID		{ $$ = newnode(TK_UASTER, NULL, $1);
					  $$ = newnode(TK_PERIOD, $$, NULL); 
					  $$->n_name = $3;
					}
    ;

abstype:		/* abstract type for sizeof() or cast */
	typespec			{ $$ = $1; }
    |	abstype TK_ASTER		{ $$ = newtype(T_PTR, $1); }
    |	abstype TK_LBRACKET TK_ICONST TK_RBRACKET
					{ $$ = newtype(T_ARRAY, $1);
					  $$->t_bound = $3;
					}
    ;



%%


/*  yyerror(s) -- report syntax error diagnosed by Yacc parser  */

/*ARGSUSED*/
int yyerror (char *s)
{
   fatal("syntax error at token", strlen(yystring()) ? yystring() : "<EOF>");
   return 0;
}
