/*
 *  structs.h -- USC internal data structures.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */



/*
 *  Frame -- a set of contiguous bits or full bytes to be moved.
 *
 *  NOTE:  If member types change, also modify sortframes() and makechunks().
 */
typedef struct frame {
    int f_dstoff;		/* destination offset (bytes) */
    int f_srcoff;		/* source offset (bytes) */
    int f_nbytes;		/* number of full bytes */
    char f_nbits;		/* number of full bits */
    char f_dstshift;		/* destination shift */
    char f_srcshift;		/* source shift */
    char f_flags;		/* flags (see below) */
    struct frame *f_prev;	/* previous frame */
    struct frame *f_next;	/* next frame */
} Frame;

#define MV_ZERO 1		/* move zero bits in, ignore source */
#define MV_SIGN 2		/* source is one bit for sign extn */



/*
 *  Mdef -- a machine definition.
 *
 *  This is not really complex enough to require a unique structure,
 *  but it is anticipated that mdefs may later become more sophisticated.
 */
typedef struct mdef {
    char *m_name;		/* mdef name */
    int m_read;			/* read value; -1 if not set */
    int m_write;		/* write value; -1 if not set */
} Mdef;



/*
 *  Order -- a byte- or bit-order specification.
 *
 *  An order struct contains one or more "order segments" preceded
 *  by a header giving the number of segments.
 */
typedef struct oseg {
    unsigned char o_length;	/* length, in bytes or bits */
    unsigned char o_start;	/* starting point */
    signed char o_dir;		/* direction (+1, -1, or 0 if length 1) */
} Oseg;
typedef struct order {
    int o_nsegs;		/* number of segments */
    Oseg o_oseg[1];		/* variable array of segments */
} Order;



/*
 *  Annot -- a USC type annotation.
 */
typedef struct annot {
    int a_tsize;		/* tsize */
    int a_msize;		/* msize (if zero, no annotation given) */
    int a_align;		/* alignment or offset */
    Order *a_order;		/* byte order */
    Order *a_bitorder;		/* bit order */
} Annot;



/*
 *  Type -- a USC datatype.
 *
 *  A USC datatype is a DAG of structures.  The t_tcode field gives the
 *  top-level classification of a type (int, pointer, union, etc.), and
 *  other fields lead to underlying information (e.g. fields of a struct).
 */
typedef enum e_tcode {		/* type code */
    T_BOGUS,			/* if seen, this is an error */
    T_UNK,			/* not yet specified */
    T_VOID,			/* void */
    T_CHAR,			/* char */
    T_SHORT,			/* short */
    T_INT,			/* int */
    T_LONG,			/* long */
#ifndef __alpha
    T_HYPER,			/* usitll */
#endif /* __alpha */
    T_PTR,			/* pointer to t_utype */
    T_ARRAY,			/* array[t_bound] of t_utype */
    T_STRUCT,			/* struct t_tag of {t_mbr...} */
    T_UNION,			/* union t_tag of {t_mbr...} */
    T_FUNC,			/* stub function returning this type */
    TCOUNT			/* number of defined Tcode values */
} Tcode;

#define TF_INT 1		/* flag set for integral type */
#define TF_SIGNED 2		/* flag set for signed datatype */
#define TF_UNSIGNED 4		/* flag set for unsigned datatype */
#define TF_ANNOT 8		/* flag set if explicitly annotated */

#define TF_STATIC 0x10		/* flag set if static stub */
#define TF_INLINE 0x20		/* flag set if inline stub */
#define TF_MACRO  0x40		/* flag set if macro stub */

#define TF_BASIC 0x100		/* flag set if original char/short/int/long */

#define A_NOBOUNDS -1		/* array with no bounds, e.g.  a[]  */
#define A_VARBOUNDS -2		/* array with variable bounds, e.g.  a[k]  */

typedef struct type {
    Tcode t_tcode;		/* type code (T_xxxx) */
    short t_tflags;		/* flags (TF_xxxx) */
    Annot t_annot;		/* annotation */
    long t_bound;		/* array bound (or A_NOBOUNDS or A_VARBOUNDS) */
    struct type *t_utype;	/* underlying type (array, ptr, stub return) */
    struct sym *t_tag;		/* struct/union tag */
    struct sym *t_mbr;		/* link to first member */
    char *t_name;		/* name, if typedef */
} Type;



/*
 *  Symbol -- a symbol table entry.
 *
 *  The symbol table holds identifiers, keywords, and constants.
 *  Identifiers play various roles as distinguished by the Scode.
 */
typedef enum e_scode {		/* symbol code (S_xxxx) */
    S_BOGUS,			/* if seen, this is an error */
    S_UNKNOWN,			/* unknown identifier */
    S_KEYWORD,			/* USC language keyword */
    S_TYPEDEF,			/* predefined or user-defined type name */
    S_MDEF,			/* machine attribute definition */
    S_FUNCTION,			/* stub function name */
    S_PARAM,			/* stub parameter name */
    S_TAG,			/* struct or union tag */
    S_MEMBER,			/* struct or union member */
    SCOUNT			/* number of defined Scode values */
} Scode;

typedef struct sym {
    Scode s_scode;		/* what kind of symbol? */
    char *s_name;		/* symbol name */
    long s_val;			/* enum value; keyword token code */
    struct sym *s_next;		/* next symbol in member or parameter list */
    struct sym *s_chain;	/* next symbol in hash chain */
    Type *s_type;		/* associated type */
    Mdef *s_mdef;		/* associated mdef */
} Symbol, *Symptr;



/*
 *  Node -- a parse tree node.
 *
 *  Not all fields of the node are used in all cases; it depends on context.
 */
typedef struct node {
    short n_lno;		/* source line, for diagnostics & debugging */
    short n_tok;		/* token code (TK_PLUS, TK_ASSIGN, etc.) */
    Type *n_type;		/* datatype of result */
    struct node *n_l;		/* left child */
    struct node *n_r;		/* right child */
    struct node *n_chain;	/* allocation chain */
    char *n_name;		/* struct/union member name */
    Symbol *n_sym;		/* symbol table entry */
    long n_long;		/* integer constant */
} Node, *Nodeptr;



/*
 *  YYSTYPE -- YACC stack entry type.
 */
typedef union {
				/* values stored directly */
    long y_long;		/*     arbitrary integer value */
    Annot y_annot;		/*     annotation specification */
    Oseg y_oseg;		/*     order segment */
				/* pointer values */
    char *y_string;		/*     string */
    Symbol *y_sym;		/*     symbol table entry */
    Type *y_type;		/*     datatype */
    Order *y_order;		/*     order specification */
    Mdef *y_mdef;		/*     mdef specification */
    Node *y_node;		/*     parse tree node */
} YYSTYPE;
