/*
 *  symtab.c -- symbol table manipulation.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

static int symcmp(const void *p1, const void *p2);	/* sorting comparison */
static int nspace(Symptr s);		/* classify namespace for sorting */
static void pslist(Symptr s);		/* print member or parameter list */

#define SBINS 251		/* number of hash bins -- should be odd prime */
static Symbol *symtab[SBINS];	/* symbol table */
static int nsyms;		/* number of symbols in table */



/*
 *  savename(name) -- save name in allocated memory.
 *
 *  This is used to save a copy of a token string before it gets overwritten.
 *  A hash table maps all duplicates of a particular string map to the same
 *  address, allowing comparison of string pointers instead of string contents.
 *  These names are never freed.
 */

typedef struct nbin {		/* structure for remembering a name */
    struct nbin *r_next;	/* next entry in hash chain */
    char r_name[1];		/* name itself (allocated at size needed) */
} Nbin;

#define NBINS 241		/* number of savename() bins */
static Nbin *rtable[NBINS];	/* table of hash bins */

char *savename(char *name)
{
    unsigned int c, h;
    char *p;
    Nbin *r;

    h = 0;
    for (p = name; c = *p++; )
	h = 37 * h + c;			/* calc hash value */
    h %= NBINS;				/* calc bin number */

    for (r = rtable[h]; r != NULL && strcmp(name,r->r_name) != 0; r = r->r_next)
	;

    if (r == NULL) {
	r = zalloc(sizeof(struct nbin *) + strlen(name) + 1);
	strcpy(r->r_name, name);
	r->r_next = rtable[h];
	rtable[h] = r;
    }

    return r->r_name;
}



/*
 *  lookup(name, kind) -- look up symbol table entry.
 *
 *  name must be a unique pointer from an earlier call to savename().
 *  If kind = 0, return pointer to existing entry or NULL.
 *  If kind > 0, install with kind specified if not found.
 */

Symbol *lookup(char *name, int kind)
{
    unsigned int h;
    Symbol *s;
    int istag;

    istag = (kind == S_TAG);			

    h = ((unsigned long) name) % SBINS;		/* address yields hash value */
    for (s = symtab[h]; s != NULL; s = s->s_chain) {
	if (name != s->s_name)
	    continue;				/* if name mismatch */
	if (istag ? (s->s_scode == S_TAG) : (s->s_scode != S_TAG))
	    break;				/* if in same name space */
    }

    if (s == NULL && kind > 0) {
	/* install new definition at head of chain */
	s = zalloc(sizeof(Symbol));
	s->s_name = name;
	s->s_scode = kind;
	s->s_chain = symtab[h];
	symtab[h] = s;
	nsyms++;
	}
    return s;
}


/*
 *  defsym(name, sc) -- define a symbol as type sc.
 *
 *  name must be a unique pointer from an earlier call to savename().
 *
 *  If sc != S_MEMBER, the name is entered in the symbol table.  defname
 *  returns NULL, and issues an error message, if redefinition is attempted.
 *
 *  If sc == S_MEMBER, a new symbol structure is always created, but it is
 *  not entered in the symbol table.  This implements a separate name space
 *  for the members of each struct or union.
 */
Symbol *defsym(char *name, Scode sc)
{
    Symbol *sym;

    if (sc == S_MEMBER) {
	sym = zalloc(sizeof(Symbol));
	sym->s_name = name;
	sym->s_scode = sc;
    } else {
	sym = lookup(name, S_UNKNOWN);
	if (sym->s_scode != S_UNKNOWN) {
	    fatal("redefinition of symbol", sym->s_name);
	    return NULL;
	}
    }
    sym->s_scode = sc;
    return sym;
}



/*
 *  mappend(old, new) -- append new members to struct/union member list.
 *
 *  (A member list is just a linked list of Symbol structures.)
 */
Symbol *mappend(Symbol *old, Symbol *new)
{
    Symbol *s;

    if (old == NULL)
	return new;
    if (new == NULL)
	return old;
    s = old; 
    for (;;) {
	if (s->s_name == new->s_name)
	    fatal("duplicate member name", new->s_name);
	if (s->s_next == NULL)
	    break;
	s = s->s_next;
    }
    s->s_next = new;
    return old;
}



/*
 *  rmparms() -- remove parameter declarations from symbol table
 *
 *  rmparms() removes parameter entries from the symbol table proper
 *  when they go out of scope at the end of a stub.  The Symtab structs
 *  remain linked to the function header, however.
 */

void rmparms(void)
{
    int i;

    for (i = 0; i < SBINS; i++)
	while (symtab[i] != NULL && symtab[i]->s_scode == S_PARAM) {
	    symtab[i] = symtab[i]->s_chain;
	    nsyms--;
	}
}



/*  dumpsyms() -- dump symbol table to stderr  */

void dumpsyms(void)
{
    int i, nsprev;
    Symbol *s, **list, **p;
    Type *t;
    char *sc, sbuf[10];

    fprintf(stderr, "\nSymbol table:\n");

    /* make and sort a list of pointers to all symbols */
    p = list = zalloc((nsyms + 1) * sizeof(Symptr));
    for (i = 0; i < SBINS; i++)
	for (s = symtab[i]; s != NULL; s = s->s_chain)
	    *p++ = s;
    qsort(list, nsyms, sizeof(Symptr), symcmp);

    /* traverse the list and print each symbol */
    *p++ = NULL;			/* append NULL to list */
    nsprev = 1;				/* previous namespace */
    for (p = list; s = *p++; ) {
	if (nspace(s) != nsprev) {
	    fprintf(stderr, "   ---------------------\n");
	    nsprev = nspace(s);		/* indicate change in namespace */
	}
	switch (s->s_scode) {
	    case S_UNKNOWN:	sc = "....";	break;
	    case S_KEYWORD:	sc = "KYWD";	break;
	    case S_TYPEDEF:	sc = "TYPE";	break;
	    case S_MDEF:	sc = "MDEF";	break;
	    case S_FUNCTION:	sc = "FUNC";	break;
	    case S_PARAM:	sc = "PARM";	break;
	    case S_TAG:		sc = "TAG ";	break;
	    case S_MEMBER:	sc = "MEMB";	break;
	    default:		sprintf(sc = sbuf, "=%d=", s->s_scode); 
	}
	fprintf(stderr, "   %-4s %-10s %5ld  ", sc,
	    s->s_name ? s->s_name : "-", s->s_val);

	/* print mdef values */
	if (s->s_scode == S_MDEF && s->s_mdef) {
	    fprintf(stderr, "read=%d, write=%d\n",
		s->s_mdef->m_read, s->s_mdef->m_write);
	    continue;
	}

	/* print type and annotation */
	if ((t = s->s_type) != NULL) {
	    ptype(t, s->s_scode == S_TYPEDEF);	/* print type */
	    pannot(t);				/* print annotation */
	}
	fprintf(stderr, "\n");

	/* print struct/union members */
	if (s->s_scode == S_TAG && t != NULL)
	    pslist(s->s_type->t_mbr);
	else if (s->s_scode == S_FUNCTION)
	    pslist(s->s_next);
    }

    free(list);
}

/*
 *  symcmp(*p1, *p2) -- compare symbols for sorting list of Symptrs
 *
 *  Symbols are sorted first by namespace and then by name.
 */
static int symcmp(const void *p1, const void *p2)
{
    Symptr s1, s2;
    int n;

    s1 = * (Symptr *) p1;
    s2 = * (Symptr *) p2;
    n = nspace(s1) - nspace(s2);
    if (n != 0)
	return n;
    else
	return strcmp(s1->s_name, s2->s_name);
}

/*  nspace(s) -- classify namespace of symbol  */

static int nspace(Symptr s)
{
    switch (s->s_scode) {
	case S_KEYWORD:	return 1;
	case S_TYPEDEF:	return 1;
	default:	return 2;
	case S_TAG:	return 3;
    }
}


/*  pslist(s) -- print member or parameter list  */

static void pslist(Symptr s)
{
    while (s != NULL) {
	fprintf(stderr, "      ");
	ptype(s->s_type, 0);
	fprintf(stderr, " %s", s->s_name ? s->s_name : "<?>");
	pannot(s->s_type);
	if (s->s_mdef)
	    fprintf(stderr, " :%s", s->s_mdef->m_name);
	fprintf(stderr, "\n");
	s = s->s_next;
    }
}
