/*
 *  type.c -- the type system.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

static Type *initt(char *name, Tcode tc, int flags);
static int gcd(int i, int j);
static Type *asgtype(Node *n, Type *ltype, Type *rtype);
static int copytype(Type *ltype, Type *rtype);	/* are types copy compatible? */
static int sameorder(Order *o1, Order *o2);	/* are Orders identical? */



/*  inittypes() -- initialize predefined datatypes  */

void inittypes(void)
{
    initt("void",  T_VOID,  0);
    chartype = initt("char",  T_CHAR,  TF_INT | TF_BASIC);
    shorttype = initt("short", T_SHORT, TF_INT | TF_BASIC);
    inttype = initt("int", T_INT, TF_INT | TF_BASIC);
    longtype = initt("long", T_LONG, TF_INT | TF_BASIC);
#ifndef __alpha
    hypertype = initt("usitll", T_HYPER, TF_INT | TF_BASIC);
    uhypertype = initt("unusitll", T_HYPER, TF_INT | TF_BASIC);
#endif /* !__alpha */
    initt("signed",   T_UNK, TF_INT | TF_SIGNED);
    initt("unsigned", T_UNK, TF_INT | TF_UNSIGNED);
}

static Type *initt(char *name, Tcode tc, int flags)
{
    Symbol *s;
    Type *t;

    name = savename(name);

    t = newtype(tc, NULL);
    t->t_tflags = flags;
    t->t_name = name;

    s = lookup(name, S_TYPEDEF);
    s->s_type = t;
    return t;
}



/*  mktag(t) -- invent a tag for an untagged struct/union  */

Symptr mktag(Type *t)
{
    static int n;
    char buf[10];
    Symptr s;

    sprintf(buf, "%03d", ++n);
    s = lookup(savename(buf), S_TAG);
    s->s_type = t;
    return s;
}



/*  newtype(tcode, utype) -- alloc and initialize a new type structure  */

Type *newtype(Tcode tc, Type *utype)
{
    Type *t;

    t = zalloc(sizeof(Type));
    t->t_tcode = tc;
    t->t_utype = utype;
    return t;
}



/*  tclone(t) -- copy a type (to be able to make changes to it)  */

Type *tclone(Type *t)
{
    Type *t2;

    t2 = zalloc(sizeof(Type));
    *t2 = *t;			/* copy type structure contents */
    t2->t_tflags &= ~TF_BASIC;	/* no longer the original basic type struct */
    return t2;
}


/*
 *  tfix(t) -- clean up loose ends of type declaration.
 *
 *  This fixes a type structure for cases where a variable is simply declared
 *  "unsigned" (vs. "unsigned int"), and also diagnoses illegal combinations.
 */
Type *tfix(Type *t)
{
    int tc, s, u;

    tc = t->t_tcode;
    s = t->t_tflags & TF_SIGNED;
    u = t->t_tflags & TF_UNSIGNED;

    if (tc == T_UNK) {
	tc = t->t_tcode = T_INT;
	t->t_annot = inttype->t_annot;
	}

    if ((s && u) || ((s | u) && (tc == T_VOID))) {
	fatal("illegal type specification", NULL);
	s = u = t->t_tflags = 0;
    }
    return t;
}


/*
 *  tmerge(old, mod) -- merge type specifications.
 *
 *  The type "old" is modified by the attributes of "mod".
 *  Some conflicts are caught here; some are caught later in tfix().
 */
Type *tmerge(Type *old, Type *mod)
{
    int modt;

    if (old->t_tflags & mod->t_tflags & (TF_SIGNED | TF_UNSIGNED)) {
	fatal("illegal type specification", NULL);
	return old;
    }
    old->t_tflags |= mod->t_tflags & (TF_SIGNED | TF_UNSIGNED);
    old->t_name = NULL;		/* modification invalidates old typedef name */

    modt = mod->t_tcode;
    switch (old->t_tcode) {
	case T_UNK:
	    old->t_tcode = modt;
	    old->t_annot = mod->t_annot;
	    break;
#ifndef __alpha
        case T_HYPER:
#endif /* !__alpha */
	case T_LONG:
	case T_SHORT:
	    if (modt != T_INT && modt != T_UNK)
		fatal("illegal type specification", NULL);
	    break;
	case T_INT:
	    if (
#if __alpha
		modt == T_SHORT || modt == T_LONG
#else /* __alpha */
		modt == T_SHORT || modt == T_LONG || modt == T_HYPER
#endif /* __alpha */
		) {
		old->t_tcode = modt;
		old->t_annot = mod->t_annot;
	    } else if (modt != T_UNK)
		fatal("illegal type specification", NULL);
	    break;
	default:
	    if (modt != T_UNK)
		fatal("illegal type specification", NULL);
	    break;
    }
    return old;
}



/*  setdecl(tree, type, annot) -- set declarator type from parse tree  */

Symptr setdecl(Node *n, Type *t, Annot a)
{
    int aok;		/* annotation ok? */
    Symptr sym;

    t = tclone(t);
    aok = setannot(t, a);
    while (n->n_tok != TK_ID) {
	if (n->n_tok == TK_UASTER) {
	    t = newtype(T_PTR, t);
	    n = n->n_r;
	} else if (n->n_tok == TK_SUBSCR) {
	    t = newtype(T_ARRAY, t);
	    t->t_bound = n->n_long;
	    t->t_annot = t->t_utype->t_annot;
	    if (n->n_long > 0)
		t->t_annot.a_msize *= n->n_long;
	    n = n->n_l;
	} else
	    MALFUNC("malformed declaration tree");
    }
    sym = n->n_sym;
    sym->s_type = t;
    if (!aok)
	fatal("missing annotation", sym->s_name);
    return sym;
}



/*  functype(t) -- validate function type and issue message  */

void functype(Type *t)
{
    if (t->t_utype->t_tcode == T_VOID)
	return;				/* void is always okay */
    if (!okparm(t->t_utype))
	fatal("illegal return type", NULL);
    else if (t->t_tflags & TF_MACRO)
	fatal("macro stub must return void", NULL);
    else if (t->t_utype->t_tcode != T_PTR && t->t_utype->t_annot.a_msize == 0)
	fatal("return type has not been previously annotated", NULL);
}



/*  okparm(t) -- is type t a valid parameter type?  */

int okparm(Type *t)
{
    if (t->t_tflags & TF_INT)
	return 1;			/* any integral type */
    else if (t->t_tcode == T_PTR)
	return 1;			/* pointer */
    else
	return 0;
}



/*  okmember(t) -- is type t a valid type for a struct/union member?  */

int okmember(Type *t)
{
    switch (t->t_tcode) {
	case T_CHAR:
	case T_SHORT:
	case T_INT:
	case T_LONG:
#ifndef __alpha
        case T_HYPER:
#endif /* !__alpha */
	case T_ARRAY:
	case T_STRUCT:
	case T_UNION:
	    return 1;
	default:
	    return 0;
    }
}



/*
 *  typecheck(tree) -- check and set types in parse tree
 *
 *  Check each node in bottom-up fashion and set its type.
 *  Leave type NULL if an error precludes figuring out a sensible value.
 */

void typecheck(Nodeptr tree)
{
    Node *l = tree->n_l;
    Node *r = tree->n_r;
    Symbol *s;
    Type *t;
    int a;

    if (l != NULL)
	typecheck(l);
    if (r != NULL)
	typecheck(r);

    if ((l != NULL && l->n_type == NULL) || (r != NULL && r->n_type == NULL))
	return;					/* earlier error */

    switch (tree->n_tok) {

	case TK_ASSIGN:
	case TK_ASGPAD:
	    tree->n_type = asgtype(tree, l->n_type, r->n_type);
	    if (tree->n_type != NULL && tree->n_type->t_tcode == T_PTR) {
		tfatal(tree, "cannot assign a pointer", NULL);
		tree->n_type = NULL;
	    }
	    break;

	case TK_RETURN:
	    if (tree->n_sym != NULL && tree->n_sym->s_type != NULL)
		tree->n_type = 
		    asgtype(tree, tree->n_sym->s_type->t_utype, r->n_type);
	    break;

	case TK_PLUS:
	case TK_MINUS:
	case TK_ASTER:
	case TK_SLASH:
	    /* for our purposes, all arithmetic results are type "long" */
	    if (l->n_type->t_tflags & r->n_type->t_tflags & TF_INT)
		tree->n_type = longtype;
	    else
		tfatal(tree, "arithmetic operands are not integral", NULL);
	    break;

	case TK_UPLUS:
	case TK_UMINUS:
	    if ((r->n_type->t_tflags & TF_INT))
		tree->n_type = longtype;
	    else
		tfatal(tree, "operand of unary +/- is not integral", NULL);
	    break;

	case TK_CAST:
	    /* casts are allowed for any pointers, and only for pointers */
	    if (l->n_type->t_tcode == T_PTR && r->n_type->t_tcode == T_PTR)
		tree->n_type = l->n_type;
	    else
		tfatal(tree, "casts are allowed only for pointers", NULL);
	    break;

	case TK_UAMPER:
	    tree->n_type = newtype(T_PTR, r->n_type);
	    break;

	case TK_UASTER:
	    if (r->n_type->t_tcode != T_PTR)
		tfatal(tree, "not a pointer", NULL);
	    else
		tree->n_type = r->n_type->t_utype;
	    break;

	case TK_SUBSCR:
	    if (l->n_type->t_tcode != T_ARRAY && l->n_type->t_tcode != T_PTR)
		tfatal(tree, "subscripted value is not an array", NULL);
	    else if (!(r->n_type->t_tflags & TF_INT))
		tfatal(tree, "subscript is not an integral value", NULL);
	    else {
		tree->n_type = tclone(l->n_type->t_utype);
		a = tree->n_type->t_annot.a_align;
		a = gcd(a, l->n_type->t_annot.a_align);
		a = gcd(a, tree->n_type->t_annot.a_msize);
		tree->n_type->t_annot.a_align = a;
	    }
	    break;

	case TK_PERIOD:
	    t = l->n_type;
	    if (t->t_tcode != T_STRUCT && t->t_tcode != T_UNION)
		{ tfatal(l, "not a struct/union", NULL); break; }
	    for (s = t->t_mbr; s != NULL; s = s->s_next)
		if (s->s_name == tree->n_name)
		    break;
	    if (s != NULL) {
		tree->n_type = tclone(s->s_type);
		a = tree->n_type->t_annot.a_align;
		a = gcd(a, l->n_type->t_annot.a_align);
		tree->n_type->t_annot.a_align = a;
	    } else
		tfatal(tree, "unrecognized struct/union member", tree->n_name);
	    tree->n_sym = s;		/* always set sym, even if NULL */
	    break;

	case TK_ID:
	    s = lookup(tree->n_name, 0);
	    if (s != NULL && s->s_scode == S_PARAM)
		tree->n_type = s->s_type;
	    else
		tfatal(tree, "not a parameter name", tree->n_name);
	    tree->n_sym = s;		/* always set sym, even if NULL */
	    break;

	case TK_ICONST:
	    tree->n_type = inttype;
	    break;

	case TK_SIZEOF:
	    tree->n_type = longtype;
	    break;

	case TK_TYPEDEF:
	case TK_ERROR:
	    /* nothing to do */
	    break;

	default:
	    fatal("unimplemented in typecheck", tokname(tree->n_tok));
	    break;
    }
}



/*  gcd(i, j) -- return greatest common divisor  */

static int gcd(int i, int j)
{
    int r;

    if (i <= 0 || j <= 0)
	return 1;
    while ((r = i % j) != 0) {
	i = j;
	j = r;
	}
    return j;
}



/*  asgtype(node, ltype, rtype) -- check type for assignment or return  */

static Type *asgtype(Node *n, Type *ltype, Type *rtype)
{
    if (ltype == NULL || rtype == NULL)
	return NULL;			/* earlier error */

    if (ltype->t_tcode == T_VOID)
	{ tfatal(n, "void stub cannot return a value", NULL); return NULL; }

    if (ltype->t_tcode == T_UNION)
	{ tfatal(n, "cannot assign or return a union", NULL); return NULL; }
	
    if (copytype(ltype, rtype))
	return ltype;
    else
	{ tfatal(n, "incompatible assignment or return", NULL); return NULL; }
}



/*  copytype(ltype, rtype) -- check if two types are copy-compatible  */

static int copytype(Type *ltype, Type *rtype)
{
    Symbol *lm, *rm;

    if (ltype->t_tflags & rtype->t_tflags & TF_INT)
	return 1;			/* both integral */

    if (ltype->t_tcode != rtype->t_tcode)
	return 0;			/* basically incompatible */

    if (ltype->t_tcode == T_ARRAY)
	return copytype(ltype->t_utype, rtype->t_utype);

    if (ltype->t_tcode == T_PTR)
	return
	    ltype->t_utype->t_tcode == T_VOID ||
	    rtype->t_utype->t_tcode == T_VOID ||
	    sametype(ltype->t_utype, rtype->t_utype);
	
    if (ltype->t_tcode == T_STRUCT) {	/* structure */
	lm = ltype->t_mbr;
	rm = rtype->t_mbr;
	for (; lm != NULL && rm != NULL; lm = lm->s_next, rm = rm->s_next)
	    if (!copytype(lm->s_type, rm->s_type))
		break;
	return lm == NULL && rm == NULL;
    }

    return 0;				/* unrecognized type */
}



/*  sametype(ltype, rtype) -- check if two types are *identical*  */

int sametype(Type *ltype, Type *rtype)
{
    Symbol *lm, *rm;
    Annot la, ra;
    int tc;

    if (ltype == rtype)
	return 1;

    if (ltype == NULL || rtype == NULL)
	return 1;			/* early error; don't flag another */

    if (ltype->t_tcode != rtype->t_tcode)
	return 0;

    tc = ltype->t_tcode;

    if (tc == T_ARRAY && ltype->t_bound != rtype->t_bound)
	return 0;

    if (tc == T_PTR || tc  == T_ARRAY)
	return sametype(ltype->t_utype, rtype->t_utype);

    if (tc == T_STRUCT || tc == T_UNION) {
	lm = ltype->t_mbr;
	rm = rtype->t_mbr;
	for (; lm != NULL && rm != NULL; lm = lm->s_next, rm = rm->s_next)
	    if (!sametype(lm->s_type, rm->s_type))
		break;
	return lm == NULL && rm == NULL;
    }

    if (ltype->t_tflags & rtype->t_tflags & TF_INT) {
	la = ltype->t_annot;
	ra = rtype->t_annot;
	if (la.a_tsize != ra.a_tsize || la.a_msize != ra.a_msize)
	    return 0;
	if (!sameorder(la.a_order, ra.a_order))
	    return 0;
	if (!sameorder(la.a_bitorder, ra.a_bitorder))
	    return 0;
	return 1;
    }

    MALFUNC("confusion in sametype()");
    /*NOTREACHED*/
    return(-1); /* shut up gcc */
}



/*  sameorder(o1, o2) -- are Orders identical?  */

static int sameorder(Order *o1, Order *o2)
{
    int i;
    Oseg *s1, *s2;

    if (o1 == NULL && o2 == NULL)
	return 1;
    if (o1 == NULL || o2 == NULL)
	return 0;
    if (o1->o_nsegs != o2->o_nsegs)
	return 0;
    for (i = 0; i < o1->o_nsegs; i++) {
	s1 = &o1->o_oseg[i];
	s2 = &o2->o_oseg[i];
	if (s1->o_length != s2->o_length
	||  s1->o_start != s2->o_start
	||  s1->o_dir != s2->o_dir)
	    return 0;
    }
    return 1;
}



/*
 *  ptype(t, xtdef) -- print synopsys of type t on stderr for debugging.
 *
 *  If xtdef is nonzero, expand typedef names to underlying definitions.
 *  Arrays and pointers use postfix notation, not C notation.
 */
void ptype(Type *t, int xtdef)
{
    if (!t)				/* if untyped, print nothing */
	return;

    if (t->t_name != NULL && xtdef == 0) {
	fprintf(stderr, "%s", t->t_name);	/* just print typedef name */
	return;
    }

    if (t->t_tflags & TF_STATIC) fprintf(stderr, "static ");
    if (t->t_tflags & TF_INLINE) fprintf(stderr, "inline ");
    if (t->t_tflags & TF_MACRO)  fprintf(stderr, "macro ");
    if (t->t_tflags & TF_SIGNED) fprintf(stderr, "signed ");
    if (t->t_tflags & TF_UNSIGNED) fprintf(stderr, "unsigned ");

    switch (t->t_tcode) {
	case T_UNK:	fprintf(stderr, "???");		break;
	case T_VOID:	fprintf(stderr, "void");	break;
	case T_CHAR:	fprintf(stderr, "char");	break;
	case T_SHORT:	fprintf(stderr, "short");	break;
	case T_INT:	fprintf(stderr, "int");		break;
	case T_LONG:	fprintf(stderr, "long");	break;
#ifndef __alpha
        case T_HYPER:	fprintf(stderr, "long long");	break;
#endif /* !__alpha */
	case T_PTR:
	    ptype(t->t_utype, xtdef);
	    fprintf(stderr, "*");
	    break;
	case T_ARRAY:
	    ptype(t->t_utype, xtdef);
	    if (t->t_bound == A_NOBOUNDS)
	        fprintf(stderr, "[]");
	    else if (t->t_bound == A_VARBOUNDS)
	        fprintf(stderr, "[%s]", t->t_tag->s_name);
	    else
	        fprintf(stderr, "[%ld]", t->t_bound);
	    break;
	case T_STRUCT:
	    fprintf(stderr, "struct %s", t->t_tag->s_name);
	    break;
	case T_UNION:
	    fprintf(stderr, "union %s", t->t_tag->s_name);
	    break;
	case T_FUNC:
	    ptype(t->t_utype, xtdef);
	    fprintf(stderr, "()");
	    break;
	default:
	    fprintf(stderr, "=t%d=", t->t_tcode);
	    break;
    }
}
