/*
 *  expr.c -- expression handling:
 *  printing, optimization, preprocessing before intermediate code generation.
 *
 *  Copyright 1994, 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"
#include "symbol.h"
#include "utils.h"
#include "expr.h"
#include "y.tab.h"

void bit_expr_to_byte_expr (Expr *theexpr)
{
  if (theexpr) {
    switch (EXPR_TYPE (theexpr)) {
    case CONST_EXPR:
      if (EXPR_CONST_UNIT (theexpr) == NO_UNIT) {
	crash ("internal error in bit_expr_to_byte_expr: no unit");
      }
      if (EXPR_CONST_UNIT (theexpr) == BIT_UNIT) {
	EXPR_CONST_UNIT (theexpr) = BYTE_UNIT;
	if (EXPR_CONST_TYPE (theexpr) == INT_CONST) 
	  EXPR_CONST_INT (theexpr) /= 8;
	else
	  crash ("internal error in bit_expr_to_byte_expr: not an int!");
      }
      break;

    case OP_EXPR:
      bit_expr_to_byte_expr (EXPR_OP_LEFT (theexpr));
      bit_expr_to_byte_expr (EXPR_OP_RIGHT (theexpr));
      break;

    case SYMINFO_EXPR:
      bit_expr_to_byte_expr (EXPR_OFFSET (theexpr));
      bit_expr_to_byte_expr (EXPR_STRUCTOFFSET (theexpr));
      break;

    case IDENT_EXPR:
      break;

    default:
      crash ("internal error in bit_expr_to_byte_expr: bad expr");
      break;
    }
  }
}

void free_expr (Expr *theexpr)
{

  /* don't use this. */

  switch (EXPR_TYPE (theexpr)) {
  case CONST_EXPR:
    break;

  case OP_EXPR:
    break;

  case SYMINFO_EXPR:
    break;

  case IDENT_EXPR:
    break;
  }
}

ConstantType type_cast (Constant *left, Constant *right)
{
  ConstantType thetype;
  if (CONST_TYPE(left) == INT_CONST) {
    if (CONST_TYPE(right) == DOUBLE_CONST) {
      CONST_TYPE(left) = DOUBLE_CONST;
      CONST_DOUBLE(left) = (double)CONST_INT(left);
      thetype = DOUBLE_CONST;
    }
    else if (CONST_TYPE(right) == INT_CONST) {
      thetype = INT_CONST;
    }
    else if (CONST_TYPE(right) == STRING_CONST) {
      crash ("illegal operation on a string"); 
    }
  }
  else {
    if (CONST_TYPE(right) == INT_CONST) {
      CONST_TYPE(right) = DOUBLE_CONST;
      CONST_DOUBLE(right) = (double)CONST_INT(right);
      thetype = DOUBLE_CONST;
    }
    else if (CONST_TYPE(right) == DOUBLE_CONST) {
      thetype = DOUBLE_CONST;
    }
    else if (CONST_TYPE (right) == STRING_CONST) {
      crash ("illegal operation on a string");
    }
  }
  return thetype;
}

Constant *int_to_const (int x)
{
  Constant *new = (Constant *)TempMalloc (sizeof (Constant));
  CONST_TYPE(new) = INT_CONST;
  CONST_INT(new) = x;
  return new;
}

Constant *double_to_const (double x)
{
  Constant *new = (Constant *)TempMalloc (sizeof (Constant));
  CONST_TYPE(new) = DOUBLE_CONST;
  CONST_DOUBLE(new) = x;
  return new;
}

Expr *const_to_expr (Constant *theconst)
{
  Expr *new = (Expr *)TempMalloc (sizeof (Expr));
  EXPR_TYPE(new) = CONST_EXPR;
  EXPR_CONST(new) = (Constant *)TempMalloc (sizeof (Constant));
  *EXPR_CONST(new) = *theconst;
  return new;
}

Expr *make_bit_unit (Expr *p)
{
  EXPR_CONST_UNIT (p) = BIT_UNIT;
  return p;
}

Expr *int_to_expr (int x)
{
  Expr *new = (Expr *)TempMalloc (sizeof (Expr));
  EXPR_TYPE(new) = CONST_EXPR;
  EXPR_CONST(new) = int_to_const (x);
  EXPR_CONST_UNIT(new) = NO_UNIT;
  return new;
}

Expr *bit_to_expr (int x)
{
  Expr *new = make_bit_unit (int_to_expr (x));
  return new;
}

Constant *add_const (Constant *left, Constant *right)
{
  Constant *new;
  ConstantType thetype;
  thetype = type_cast (left, right);
  if (thetype == INT_CONST) 
    new = int_to_const (CONST_INT(left) + CONST_INT(right));
  else
    new = double_to_const (CONST_DOUBLE(left) + CONST_DOUBLE(right));
  return new;
}

Constant *sub_const (Constant *left, Constant *right)
{
  Constant *new;
  ConstantType thetype;
  thetype = type_cast (left, right);
  if (thetype == INT_CONST)
    new = int_to_const (CONST_INT(left) - CONST_INT(right));
  else
    new = double_to_const (CONST_DOUBLE(left) - CONST_DOUBLE(right));
  return new;
}

Constant *mult_const (Constant *left, Constant *right)
{
  Constant *new;
  ConstantType thetype;
  thetype = type_cast (left, right);
  if (thetype == INT_CONST)
    new = int_to_const (CONST_INT(left) * CONST_INT(right));
  else
    new = double_to_const (CONST_DOUBLE(left) * CONST_DOUBLE(right));
  return new;
}

Constant *div_const (Constant *left, Constant *right)
{
  Constant *new;
  ConstantType thetype;
  thetype = type_cast (left, right);
  if (thetype == INT_CONST) {
    if (CONST_INT(right) == 0)
      crash ("divide by zero");
    new = int_to_const (CONST_INT(left) / CONST_INT(right));
  }
  else {
    if (CONST_DOUBLE(right) == 0.0)
      crash ("divide by zero");
    new = double_to_const (CONST_DOUBLE(left) / CONST_DOUBLE(right));
  }
  return new;
}

Expr *copy_expr (Expr *theexpr)
{
  Expr *new;
  new = (Expr *)TempMalloc (sizeof (Expr));
  *new = *theexpr;
  switch (EXPR_TYPE (new)) {
  case CONST_EXPR:
    EXPR_CONST (new) = (Constant *)TempMalloc (sizeof (Constant));
    *EXPR_CONST(new) = *EXPR_CONST (theexpr);
    break;

  case OP_EXPR:
    EXPR_OP (new) = (BinOp *)TempMalloc (sizeof (BinOp));
    *EXPR_OP(new) = *EXPR_OP (theexpr);
    break;
    
  case SYMINFO_EXPR:
    break;

  case IDENT_EXPR:
    break;

  default:
    crash ("internal error in copy_expr");
    break;
  }
  return new;
}

Expr *make_op (Expr *left, int op, Expr *right)
{
  Expr *new = NULL;
  int flag = 1;


  switch (op) {
  case PLUS:
    if (EXPR_IS_CONST (right)) {
      new = left;
      left = right;
      right = new;
    }
    else {
      new = right;
    }
    if (EXPR_IS_PLUS(right) && EXPR_TYPE(EXPR_OP_LEFT(right)) == CONST_EXPR) {
      new = copy_expr (right);
      EXPR_OP_LEFT(new) = 
        const_to_expr (add_const (EXPR_CONST(left), 
                                  EXPR_CONST (EXPR_OP_LEFT(right))));
      flag = 0;

    }
    break;

  case STAR:
    break;

  default:
    break;
  }
  
  if (flag) {
    new = (Expr *)TempMalloc (sizeof (Expr));
    EXPR_TYPE(new) = OP_EXPR;
    EXPR_OP(new) = (BinOp *)TempMalloc(sizeof (BinOp));
    EXPR_OP_TYPE(new) = op;
    EXPR_OP_LEFT(new) = left ? copy_expr(left) : NULL;
    EXPR_OP_RIGHT(new) = right ? copy_expr(right) : NULL;
    EXPR_NEXT(new) = NULL;
  }
  return new;
}

Expr *do_expr_op (Expr *left, int op, Expr *right)
{
  Expr *new = NULL;

  if (EXPR_TYPE(left) != CONST_EXPR || EXPR_TYPE(right) != CONST_EXPR) {
    new = make_op (left, op, right);
  }
  else {
    switch (op) {
    case PLUS:
      new = const_to_expr (add_const (EXPR_CONST(left), EXPR_CONST(right)));
      if ( (EXPR_CONST_UNIT (left) == BIT_UNIT)
	 ||(EXPR_CONST_UNIT (right)== BIT_UNIT) )
	EXPR_CONST_UNIT(new) = BIT_UNIT;
      else
      if ( EXPR_CONST_UNIT (left) == NO_UNIT )
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (right);
      else
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (left);
      break;
      
    case MINUS:
      new = const_to_expr (sub_const (EXPR_CONST(left), EXPR_CONST(right)));
      if ( (EXPR_CONST_UNIT (left) == BIT_UNIT)
	 ||(EXPR_CONST_UNIT (right)== BIT_UNIT) )
	EXPR_CONST_UNIT(new) = BIT_UNIT;
      else
      if ( EXPR_CONST_UNIT (left) == NO_UNIT )
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (right);
      else
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (left);
      break;
      
    case STAR:
      new = const_to_expr (mult_const (EXPR_CONST(left), EXPR_CONST(right)));
      if ( (EXPR_CONST_UNIT (left) == BIT_UNIT)
	 ||(EXPR_CONST_UNIT (right)== BIT_UNIT) )
	EXPR_CONST_UNIT(new) = BIT_UNIT;
      else
      if ( EXPR_CONST_UNIT (left) == NO_UNIT )
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (right);
      else
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (left);
      break;

      
    case DIV:
      new = const_to_expr (div_const (EXPR_CONST(left), EXPR_CONST(right)));
      if ( (EXPR_CONST_UNIT (left) == BIT_UNIT)
	 ||(EXPR_CONST_UNIT (right)== BIT_UNIT) )
	EXPR_CONST_UNIT(new) = BIT_UNIT;
      else
      if ( EXPR_CONST_UNIT (left) == NO_UNIT )
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (right);
      else
	EXPR_CONST_UNIT(new) = EXPR_CONST_UNIT (left);
      break;
      
    default:
      new = make_op (left, op, right);
      break;
    }
  }

  DEBUG(900) {
    dprintf (0, "---------------------------");
    dprintf (0, "do_expr_op:");
    dprintf (0, "left:");
    print_expr (0, 0, left);
    dprintf (0, "right");
    print_expr (0, 0, right);
    dprintf (0, "new");
    print_expr(0, 0, new);
    dprintf (0, "---------------------------");
  }

  return new;
}

void negate_const (Constant *theconst)
{
  if (CONST_TYPE(theconst) == INT_CONST)
    CONST_INT(theconst) = -CONST_INT(theconst);
  else if (CONST_TYPE(theconst) == DOUBLE_CONST)
    CONST_DOUBLE(theconst) = -CONST_DOUBLE(theconst);
  else
    crash ("can't negate a string");
}

void tilde_const (Constant *theconst) 
{
  if ( CONST_TYPE(theconst) == INT_CONST ) 
    CONST_INT(theconst) = ~CONST_INT(theconst);
  else
    crash ("can't take a one's compliment of a non integer");
}

void bang_const (Constant *theconst)
{
  if ( CONST_TYPE(theconst) == INT_CONST ) 
    CONST_INT(theconst) = !CONST_INT(theconst);
  else
    crash ("can't take a one's compliment of a non integer");
}

Expr *postfix_op (int op, Expr *theexpr) 
{
  Expr *new = NULL;

  if (EXPR_TYPE(theexpr) == CONST_EXPR) {
    crash ("illegal use of rvalue");
  }
  else {
    new = make_op (theexpr, op, NULL);
  }
  return new;
}

Expr *prefix_op (int op, Expr *theexpr)
{
  Expr *new;
  Constant *temp;
  if (EXPR_TYPE(theexpr) != CONST_EXPR) {
    new = make_op (NULL, op, theexpr);
  }
  else {
    switch (op) {
    case SIZEOF:
      crash("Unable to get size of a constant");
      break;
    case POINTER:
      crash("Unable to de-reference a constant");
      break;
    case ADDROF:
      crash("Unable to get address of a constant");
      break;

    case PLUS:
      break;
      
    case MINUS:
      temp = EXPR_CONST(theexpr);
      negate_const (temp);
      break;
      
    case TILDE:
      tilde_const (EXPR_CONST(theexpr));
      break;
      
    case BANG:
      bang_const (EXPR_CONST(theexpr));
      break;

    case PLUSPLUS:
      crash ("Can't increment an rvalue");
      break;

    case MINMIN:
      crash ("Can't decrement an rvalue");
      break;
    
    case RETURN:
      theexpr = make_op (NULL, op, theexpr);
      break;
      
    default:
      crash ("internal error in prefix_op");
      break;
    }
    new = theexpr;
  }
  return new;
}

Expr *ident_to_expr (Constant *theconst)
{
  Expr *new = (Expr *)TempMalloc (sizeof (Expr));
  if (CONST_TYPE(theconst) != STRING_CONST)
    crash ("internal error in indent_to_expr");
  EXPR_TYPE(new) = IDENT_EXPR;
  EXPR_IDENT(new) = CONST_STRING(theconst); /* already malloc'd */
  return new;
}

void ident_to_syminfo (Expr **theident)
{
  Symbol *sym;
  
  dprintf (500, "Entering ident_to_syminfo");

  if (EXPR_TYPE (*theident) != IDENT_EXPR) {
    crash ("Internal error in ident_to_syminfo");
  }
  else {
    sym = find_sym (EXPR_IDENT (*theident));
    if (!sym) {
      crash ("Symbol %s not declared", EXPR_IDENT (*theident));
    }
    EXPR_SYMINFO (*theident) = (SymInfo *)TempMalloc (sizeof (SymInfo));
    *EXPR_SYM (*theident) = *sym;
    *EXPR_BASESYM (*theident) = *sym;
    EXPR_LASTSTRUCT (*theident)->name = NULL;
    EXPR_OFFSET (*theident) = bit_to_expr (0);
    EXPR_STRUCTOFFSET (*theident) = bit_to_expr (0);
    EXPR_TYPEDATA (*theident) = SYM_TYPEDATA (sym);
    EXPR_TYPE (*theident) = SYMINFO_EXPR;
  }
  print_expr (500, 0, *theident);
  dprintf (500, "Exiting ident_to_syminfo");
}

Symbol *find_field (Symbol *thesym, char *field)
{
  Symbol *fields;
  int found = 0;
  dprintf (500, "Entering find_field: %s", field);
  print_sym (700, 0, "field", thesym);
  if (!TYPE_IS_STRUCT_OR_UNION (SYM_TYPE (thesym)))
    crash ("Trying to find a field of something which is not"
           "a struct or union");
  for (fields = TYPE_STRUCTDEF_FIELDS (SYM_TYPE (thesym)); !found && fields; 
       fields =  fields->next) {
    if (!strcmp (field, SYM_NAME (fields))) {
      found = 1;
      break;
    }
  }
  if (found) {
    dprintf (500, "field found");
    return fields;
  }
  else {
    crash ("%s is not a field of this struct or union", field);
  }
  dprintf (500, "Exiting find_field");
  return NULL;
}

Expr *add_exprs (Expr *left, Expr *right)
{
  Expr *new = do_expr_op (left, PLUS, right);
  return new;
}

Expr *mul_exprs (Expr *left, Expr *right)
{
  return do_expr_op (left, STAR, right);
}

Expr *clean_op (Expr *theexpr)
{
  Expr *new = theexpr, *newoffset;
  Symbol *thesym;
  char *field;
  SymInfo *theinfo;
  int tempoffset;
  Type *thetype;

  switch (EXPR_OP_TYPE (theexpr)) {
  case DOT:
    if (EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) != IDENT_EXPR) {
      crash ("illegal field");
    }
    else if (!EXPR_OP_LEFT (theexpr)) {
      crash ("illegal expression: no left side to DOT operator");
    }
    else if (EXPR_TYPE (EXPR_OP_LEFT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_LEFT (theexpr));
    }
    if ( !EXPR_IS_SYMINFO (EXPR_OP_LEFT (theexpr)) || !EXPR_SYM (theexpr)) {
      crash ("illegal expression: left side of DOT is not symbol");
    }
    else if (!(field = EXPR_IDENT (EXPR_OP_RIGHT (theexpr)))) {
      crash ("field is empty string. not good");
    }
    else {
      theinfo = EXPR_SYMINFO (EXPR_OP_LEFT (theexpr));
      *SYMINFO_LASTSTRUCT (theinfo) = *SYMINFO_SYM (theinfo);
      SYMINFO_STRUCTOFFSET (theinfo) = SYMINFO_OFFSET (theinfo);
      thetype = SYM_TYPE (SYMINFO_SYM (theinfo));
      SYM_TYPE (SYMINFO_LASTSTRUCT (theinfo)) = 
	copy_type (thetype);
      *SYMINFO_SYM (theinfo) = *find_field (SYMINFO_SYM (theinfo), field);
      SYMINFO_TYPEDATA (theinfo) = SYM_TYPEDATA (SYMINFO_SYM (theinfo));
      free_expr (EXPR_OP_RIGHT (theexpr));
      EXPR_TYPE (theexpr) = SYMINFO_EXPR;
      EXPR_SYMINFO (theexpr) = theinfo;
      EXPR_OFFSET (theexpr) = 
	add_exprs (EXPR_OFFSET (theexpr),
		   bit_to_expr (8 * SYM_OFFSET (EXPR_SYM (theexpr)))); 
      /* translate to bits */
    }
    break;

  case ARRAY:
    if (!EXPR_OP_LEFT (theexpr) || !EXPR_OP_RIGHT (theexpr))
      crash ("Illegal use of binary array operator");
    if (EXPR_TYPE (EXPR_OP_LEFT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_LEFT (theexpr));
    }
    if (EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_RIGHT (theexpr));
    }
    if ( !EXPR_IS_SYMINFO (EXPR_OP_LEFT (theexpr)) ||
         !(thesym = EXPR_SYM (EXPR_OP_LEFT (theexpr)))) {
      crash ("illegal expression: left side of [] is not a symbol");
    }
    else {
      thetype = SYM_TYPE (thesym);
      if (!TYPE_IS_ARRAY (thetype))
        crash ("this is not an array: %s", SYM_NAME (thesym));
      theinfo = EXPR_SYMINFO (EXPR_OP_LEFT (theexpr));
      tempoffset = TYPE_MEMSIZE (thetype) * 8;
      dprintf (500, "tempoffset: %d", tempoffset);
      print_expr (500, 0, EXPR_OP_RIGHT (theexpr));
      
      newoffset = mul_exprs (EXPR_OP_RIGHT (theexpr), 
			     bit_to_expr(tempoffset));
      print_expr (500, 0, newoffset);

      newoffset = add_exprs (EXPR_OFFSET (EXPR_OP_LEFT (theexpr)), newoffset);
      print_expr (500, 0, newoffset);
      SYM_TYPE (thesym) = TYPE_NEXT (SYM_TYPE (thesym));
      EXPR_TYPE (theexpr) = SYMINFO_EXPR;
      EXPR_SYMINFO (theexpr) = theinfo;
      EXPR_OFFSET (theexpr) = newoffset;
    }
    break;

  case SIZEOF:
    if (EXPR_OP_LEFT (theexpr) || !EXPR_OP_RIGHT (theexpr))
      crash ("Illegal use of sizeof operator");
    if (EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_RIGHT (theexpr));
    }
    if (!EXPR_IS_SYMINFO (EXPR_OP_RIGHT (theexpr)) ||
        !(thesym = EXPR_SYM (EXPR_OP_RIGHT (theexpr)))) {
      crash ("illegal expression: argument of sizeof is not a symbol");
    }
    else {
      new = clean_expr( sizeof_type(SYM_TYPE (thesym) ) );
    }
    break;

  case POINTER:
    if (EXPR_OP_LEFT (theexpr) || !EXPR_OP_RIGHT (theexpr))
      crash ("Illegal use of unary dereference operator");
    if (EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_RIGHT (theexpr));
    }
    if (!EXPR_IS_SYMINFO (EXPR_OP_RIGHT (theexpr)) ||
        !(thesym = EXPR_SYM (EXPR_OP_RIGHT (theexpr)))) {
      crash ("illegal expression: argument of dereference is not a symbol");
    }
    else {
      thetype = SYM_TYPE (thesym);
      if (!TYPE_IS_POINTER (thetype))
	crash ("this is not a pointer: %s", SYM_NAME (thesym));
      theinfo = EXPR_SYMINFO (EXPR_OP_RIGHT (theexpr));
      SYM_TYPE (thesym) = TYPE_NEXT (SYM_TYPE (thesym));
      SYM_TYPE (EXPR_BASESYM (EXPR_OP_RIGHT (theexpr))) = SYM_TYPE (thesym);
      EXPR_TYPE (theexpr) = SYMINFO_EXPR;
      EXPR_SYMINFO (theexpr) = theinfo;
    }
    break;

  case ADDROF:
    if (EXPR_OP_LEFT (theexpr) || !EXPR_OP_RIGHT (theexpr))
      crash ("Illegal use of unary address of operator");
    if (EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_RIGHT (theexpr));
    }
    if (!EXPR_IS_SYMINFO (EXPR_OP_RIGHT (theexpr)) ||
        !(thesym = EXPR_SYM (EXPR_OP_RIGHT (theexpr)))) {
      crash ("illegal expression: argument of address of is not a symbol");
    }
    else if ( !do_i_deref( SYM_NAME( thesym ) ) ) {
	crash("can only take address of non-local variables");
    }
    break;

  case ARROW:
    if (EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) != IDENT_EXPR) {
      crash ("illegal field");
    }
    else if (!EXPR_OP_LEFT (theexpr)) {
      crash ("illegal expression: no left side to ARROW operator");
    }
    else if (EXPR_TYPE (EXPR_OP_LEFT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_LEFT (theexpr));
    }
    if ( !EXPR_IS_SYMINFO (EXPR_OP_LEFT (theexpr)) || 
	 !(thesym = EXPR_SYM (EXPR_OP_LEFT (theexpr)))) {
      crash ("illegal expression: left side of ARROW is not symbol");
    }
    else if (!(field = EXPR_IDENT (EXPR_OP_RIGHT (theexpr)))) {
      crash ("field is empty string. not good");
    }
    else {
      thetype = SYM_TYPE (thesym);
      if (!TYPE_IS_POINTER (thetype))
	crash ("this is not a pointer: %s", SYM_NAME (thesym));

      /* first de-reference left */
      theinfo = EXPR_SYMINFO (EXPR_OP_LEFT (theexpr));
      SYM_TYPE (thesym) = TYPE_NEXT (SYM_TYPE (thesym));
      SYM_TYPE (EXPR_BASESYM (EXPR_OP_LEFT (theexpr))) = SYM_TYPE (thesym);

      EXPR_TYPE (EXPR_OP_LEFT (theexpr)) = SYMINFO_EXPR;
      EXPR_SYMINFO (EXPR_OP_LEFT (theexpr)) = theinfo;

      /* now act as if it were a DOT operator */
      SYMINFO_STRUCTOFFSET (theinfo) = SYMINFO_OFFSET (theinfo);
      *SYMINFO_LASTSTRUCT (theinfo) = *SYMINFO_SYM (theinfo);
      SYM_TYPE (SYMINFO_LASTSTRUCT (theinfo)) = 
	copy_type (SYM_TYPE (SYMINFO_SYM (theinfo)) );
      *SYMINFO_SYM (theinfo) = *find_field (SYMINFO_SYM (theinfo), field);
      SYMINFO_TYPEDATA (theinfo) = SYM_TYPEDATA (SYMINFO_SYM (theinfo));
      free_expr (EXPR_OP_RIGHT (theexpr));
      EXPR_TYPE (theexpr) = SYMINFO_EXPR;
      EXPR_SYMINFO (theexpr) = theinfo;
      EXPR_OFFSET (theexpr) = 
	add_exprs (EXPR_OFFSET (theexpr),
		   bit_to_expr(8 * SYM_OFFSET (EXPR_SYM (theexpr)))); 
    }
    break;

  case RETURN:
  case PLUS:
  case MINUS:
  case STAR:
  case DIV:
  case MOD:
  case AND:
  case PLUSPLUS:
  case MINMIN:
  case LSHIFT:
  case RSHIFT:
  case GE:
  case LE:
  case NEQ:
  case EQEQ:
  case TILDE:
  case BANG:
  case LT:
  case GT:
  case EQUAL:
    if (EXPR_OP_LEFT (theexpr) && 
        EXPR_TYPE (EXPR_OP_LEFT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_LEFT (theexpr));
    }
    if (EXPR_OP_RIGHT (theexpr) &&
        EXPR_TYPE (EXPR_OP_RIGHT (theexpr)) == IDENT_EXPR) {
      ident_to_syminfo (&EXPR_OP_RIGHT (theexpr));
    }
    break;

  default:
    break;
  }
  
  return new;
}

Expr *clean_expr (Expr *theexpr)
{
  Expr *new = theexpr;

  if (theexpr) {
    switch (EXPR_TYPE(theexpr)) {
    case CONST_EXPR:
      EXPR_CONST_UNIT(theexpr) = BYTE_UNIT;
      break;

    case OP_EXPR:
      EXPR_OP_LEFT (theexpr) = clean_expr (EXPR_OP_LEFT (theexpr));
      EXPR_OP_RIGHT (theexpr) = clean_expr (EXPR_OP_RIGHT (theexpr));
      new = clean_op (theexpr);
      break;

    case SYMINFO_EXPR:
      break;

    case IDENT_EXPR:
      break;
      
    default:
      crash ("internal error in clean_expr: %d\n", EXPR_TYPE(theexpr));
      break;
    }
  }
  return new;
}
