%{
/*
 *  grammar.y
 *
 *  Copyright 1994, 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"
#include "symbol.h"
#include "debug.h"
#include "buffer.h"
#include "utils.h"
#include "expr.h"
#include <stdlib.h>

extern int yylex();
extern int yyerror(char *s);

%}

%union {
   Constant theconst[1];
   Constant *theconstptr;
   Symbol *thesym;
   Type *thetype;
   Expr *theexpr;
   int theop;
}

%token <theconst> IDENT INT LONG LONGLONG ULONGLONG CHAR SHORT FLOAT DOUBLE
%token <theconst> SIGNED UNSIGNED
%token <theconst> STRUCT UNION TYPEDEF
%token <theconst> ICON FCON STRING
%token <theconst> INLINE MACRO

%token <thetype>  TYPEDEFNAME

%token <theop>    ARROW DOT PLUSPLUS MINMIN LSHIFT RSHIFT
%token <theop>    ARRAY POINTER SIZEOF ADDROF
%token <theop>    LE GE GT LT EQEQ NEQ ANDAND OROR
%token <theop>    '+' '-' '*' '/' '%'
%token <theop>    PLUS MINUS STAR DIV MOD
%token <theop>    AND OR XOR
%token <theop>    '~' '!' ','
%token <theop>    TILDE BANG COMMA
%token <theop>    EQUAL RETURN

%token VOID STATIC ENUM
%token DOTDOT DOTDOTDOT
%token IF SWITCH WHILE
%token MULEQ DIVEQ MODEQ ADDEQ SUBEQ LSHIFTEQ RSHIFTEQ ANDEQ XOREQ OREQ

%type <thesym> translation_unit
%type <thesym> declaration
%type <thesym> declarator declarator2
%type <thesym> init_declarator init_declarator_list
%type <thesym> parameter_list parameter_type_list parameter_declaration
%type <thesym> struct_declarator struct_declarator_list
%type <thesym> struct_declaration_list struct_declaration

%type <thesym> struct_field_declarator struct_field_declarator2

%type <thetype> type_specifier type_specifier_list
%type <thetype> type_name
%type <thetype> declaration_specifiers
%type <thetype> struct_or_union struct_or_union_specifier
%type <thetype> bitfield_description

%type <theop>   structop incop unary_operator shiftop relop equop
%type <theop>   pointer assignment_operator

%type <theexpr> expr


%type <theexpr> primary_expr postfix_expr argument_expr_list unary_expr
%type <theexpr> cast_expr multiplicative_expr additive_expr shift_expr
%type <theexpr> relational_expr equality_expr and_expr exclusive_or_expr
%type <theexpr> inclusive_or_expr logical_and_expr logical_or_expr
%type <theexpr> conditional_expr assignment_expr constant_expr

%type <theconst> simple_type storage_class_specifier

%start translation_unit
%%

primary_expr
    : IDENT        { $$ = ident_to_expr ($1); print_expr (1000, 0, $$); }
    | ICON         { $$ = const_to_expr ($1); print_expr (1000, 0, $$); }
    | FCON         { $$ = const_to_expr ($1); print_expr (1000, 0, $$); }
    | STRING       { $$ = const_to_expr ($1); print_expr (1000, 0, $$); }
    | '(' expr ')' { $$ = $2; }
    ;

structop
    : ARROW | DOT;

incop
    : PLUSPLUS | MINMIN ;

postfix_expr
    : primary_expr
    | postfix_expr '[' expr ']'   { $$ = do_expr_op($1, ARRAY, $3); }
    | postfix_expr structop IDENT { $$ = do_expr_op($1, $2, ident_to_expr($3));}
    | postfix_expr incop          { $$ = postfix_op ($2, $1); }
    | postfix_expr '(' ')' {
      $$ = NULL;
      error("procedure call not allowed");
    }
    | postfix_expr '(' argument_expr_list ')' {
      $$ = NULL;
      error("procedure call not allowed");
    }
    ;

argument_expr_list
    : assignment_expr
    | argument_expr_list ',' assignment_expr
    ;

unary_expr
    : postfix_expr
    | incop unary_expr         { $$ = prefix_op ($1, $2); }
    | unary_operator cast_expr { $$ = prefix_op ($1, $2); }
    | SIZEOF '(' type_name ')' {
      $$ = sizeof_type( $3 );
    }
    | SIZEOF unary_expr        {
      $$ = prefix_op( $1, $2 );
    }
    ;

unary_operator
    : AND { $$ = ADDROF; }
    | '*' { $$ = POINTER; }
    | '+'
    | '-'
    | '~'
    | '!'
    ;

cast_expr
    : unary_expr
    | '(' type_name ')' cast_expr {
      $$ = $4;
      error("typecasting ignored");
    }
    ;

multiplicative_expr
    : cast_expr
    | multiplicative_expr '*' cast_expr { $$ = do_expr_op($1, $2, $3); }
    | multiplicative_expr '/' cast_expr { $$ = do_expr_op($1, $2, $3); }
    | multiplicative_expr '%' cast_expr { $$ = do_expr_op($1, $2, $3); }
    ;

additive_expr
    : multiplicative_expr
    | additive_expr '+' multiplicative_expr { $$ = do_expr_op($1, $2, $3); }
    | additive_expr '-' multiplicative_expr { $$ = do_expr_op($1, $2, $3); }
    ;

shiftop
    : LSHIFT | RSHIFT ;

shift_expr
    : additive_expr
    | shift_expr shiftop additive_expr { $$ = do_expr_op($1, $2, $3); }
    ;

relop   : LE | GE | LT | GT

relational_expr
    : shift_expr
    | relational_expr relop shift_expr { $$ = do_expr_op($1, $2, $3); }
    ;

equop   : EQEQ | NEQ ;

equality_expr
    : relational_expr
    | equality_expr equop relational_expr { $$ = do_expr_op($1, $2, $3); }
    ;

and_expr
    : equality_expr
    | and_expr AND equality_expr { $$ = do_expr_op($1, $2, $3); }
    ;

exclusive_or_expr
    : and_expr
    | exclusive_or_expr XOR and_expr { $$ = do_expr_op($1, $2, $3); }
    ;

inclusive_or_expr
    : exclusive_or_expr
    | inclusive_or_expr OR exclusive_or_expr { $$ = do_expr_op($1, $2, $3); }
    ;

logical_and_expr
    : inclusive_or_expr
    | logical_and_expr ANDAND inclusive_or_expr { $$ = do_expr_op($1, $2, $3); }
    ;

logical_or_expr
    : logical_and_expr
    | logical_or_expr OROR logical_and_expr { $$ = do_expr_op($1, $2, $3); }
    ;

conditional_expr
    : logical_or_expr
    | logical_or_expr '?' logical_or_expr ':' conditional_expr {
      $$ = NULL;
      error("'?' operator not implemented yet");
    }
    ;

assignment_expr
    : conditional_expr
    | unary_expr assignment_operator assignment_expr {
      $$ = do_expr_op ($1, $2, $3);
    }
    ;

assignment_operator
    : EQUAL
    | ADDEQ     { error("'+=' not allowed"); }
    | SUBEQ     { error("'-=' not allowed"); }
    | MULEQ     { error("'*=' not allowed"); }
    | DIVEQ     { error("'/=' not allowed"); }
    | MODEQ     { error("'%=' not allowed"); }
    | ANDEQ     { error("'&=' not allowed"); }
    | XOREQ     { error("'|=' not allowed"); }
    | OREQ      { error("'^=' not allowed"); }
    | LSHIFTEQ  { error("'<<=' not allowed"); }
    | RSHIFTEQ  { error("'>>=' not allowed"); }
    ;

expr
    : assignment_expr { $$ = $1; print_expr (700, 0, $$); }
    | expr ',' assignment_expr {
      $$ = do_expr_op ($1, $2, $3);
      print_expr (700, 0, $$);
    }
    ;

constant_expr
    : conditional_expr { print_expr (1000, 0, $1); }
    ;

declaration
    : declaration_specifiers ';'  { $$ = NULL; }
    | declaration_specifiers init_declarator_list ';' {
      Symbol *sym;
      sym = distribute_specifier ($1, $2);
      $$ = add_to_sym_or_type_table (sym);
    }
    ;

declaration_specifiers
    : storage_class_specifier {
      $$ = NULL;
      error("'typedef' must be followed by type definition");
    }
    | storage_class_specifier declaration_specifiers {
      /* must confirm storage_class_specifier is typedef? */
      $$ = make_typedef ($2);
    }
    | type_specifier
    | type_specifier declaration_specifiers {
      print_type_table(900);
      $$ = or_specifier ($1, $2);
      print_type_table(900);
    }
    ;

init_declarator_list
    : init_declarator
    | init_declarator_list ',' init_declarator { $$ = cat_sym( $1, $3 ); }
    ;

init_declarator
    : declarator
    | declarator EQUAL initializer { 
      $$ = $1;
      error("no variable initialization allowed");
    }
    ;

storage_class_specifier
    : TYPEDEF
    ;

simple_type
    : CHAR
    | SHORT
    | INT
    | LONG
    | LONGLONG
    | ULONGLONG
    | SIGNED
    | UNSIGNED
    ;


type_specifier
    : simple_type {
      $$ = make_specifier ($1);
      print_type (800, 0, "type_specifier", $$);
    }
    | FLOAT       { $$=NULL; error("'float' type not allowed"); }
    | DOUBLE      { $$=NULL; error("'double' type not allowed"); }
    | VOID        { $$=NULL; }
    | struct_or_union_specifier
    | enum_specifier { $$=NULL; }
    | TYPEDEFNAME
    ;

struct_or_union_specifier
    : struct_or_union IDENT '{' struct_declaration_list'}' {
      Type *temp;
      print_type (800, 0, "struct_or_union", $1);
      temp = find_struct_def (CONST_STRING($2));
      if (temp)
	crash ("Struct <%s> already defined", CONST_STRING($2));
      else {
	temp = finish_struct_def ($1, CONST_STRING($2), $4);
	add_to_struct_defs (temp, NestLevel);
	$$ = temp;
      }
    }
    | struct_or_union '{' struct_declaration_list '}' {
      Type *temp;
      temp = finish_struct_def ($1, NULL, $3);
      $$ = temp;
    }
    | struct_or_union IDENT {
      $$ = find_struct_def (CONST_STRING($2));
      if ($$ == NULL) crash ("Undefined struct: %s", CONST_STRING($2));
    }
    ;

struct_or_union
    : STRUCT { $$ = make_specifier ($1); }
    | UNION  { $$ = make_specifier ($1); }
    ;

struct_declaration_list
    : struct_declaration
    | struct_declaration_list struct_declaration {
      $$ = cat_sym ($1, $2);
      print_sym_list (800, 0,"struct_decltn_list", $$);
    }
    ;

struct_declaration
    : type_specifier_list struct_declarator_list ';' {
      $$ = distribute_specifier ($1, $2);
      print_sym_list(800, 0,"struct_declaration", $$);
    }
    ;

struct_declarator_list
    : struct_declarator
    | struct_declarator_list ',' struct_declarator {
      $$ = cat_sym ($1, $3);
      print_sym_list (800, 0, "struct_declr_list", $$);
    }
    ;

struct_declarator
    : struct_field_declarator { print_sym(800,0,"struct_declr", $1); }
    | ':' bitfield_description {
      $$ = NULL;
      error("not named bitfields not allowed");
    }
    | struct_field_declarator ':' bitfield_description {
      $$=make_bitfield($1,$3);
    }
    ;

/* ---------- */
struct_field_declarator
    : struct_field_declarator2       
    | pointer struct_field_declarator2 {
      error("no pointers allowed as struct fields");
      $$ = NULL;
      /* must insert pointer info */
    }
    ;

struct_field_declarator2
    : IDENT                                           {
      $$ = ident_to_sym(CONST_STRING($1));
    }
    | '(' struct_field_declarator ')'                 { $$ = $2; }
    | struct_field_declarator2 '[' ']'                {
      error( "array definition misses length" );
    }
    | struct_field_declarator2 '[' constant_expr ']'  {
      $$ = make_array ($1, $3);
    }
    | struct_field_declarator2 '(' ')' {
      $$ = NULL;
      error("functions are not allowed as fields");
    }
    | struct_field_declarator2 '(' parameter_type_list ')' {
      $$ = NULL;
      error("functions are not allowed as fields");
      /* must combine parameter type info */
    }
    | struct_field_declarator2 '(' parameter_identifier_list ')' {
      $$ = NULL;
      error("functions are not allowed as fields");
    }
    ;


/* --------- */

bitfield_description
    : ICON /* constant_expr */ {
      $$ = make_native_bitfield (const_to_expr($1));
    }
    ;

enum_specifier
    : ENUM '{' enumerator_list '}'
    | ENUM IDENT '{' enumerator_list '}'
    | ENUM IDENT
    ;

enumerator_list
    : enumerator
    | enumerator_list ',' enumerator
    ;

enumerator
    : IDENT
    | IDENT EQUAL constant_expr
    ;

declarator
    : declarator2 {
      $$ = $1;
    }
    | pointer declarator2 {
      if ( $1 == STAR ) {
	$$ = make_pointer( $2 );
      } else {
	$$ = $2;
      }
    }
    ;

declarator2
    : IDENT                              { $$ = ident_to_sym(CONST_STRING($1));}
    | '(' declarator ')'                 { $$ = $2; }
    | declarator2 '[' ']'                {
      error( "array definition misses length" );
    }
    | declarator2 '[' constant_expr ']'  { $$ = make_array ($1, $3); }
    | declarator2 '(' ')' {
      $$ = make_function( $1, NULL );
    }
    | declarator2 '(' parameter_type_list ')' {
      $$ = make_function( $1, $3 );
    }
    | declarator2 '(' parameter_identifier_list ')' {
      $$ = NULL;
      error("parameters in parameter list must be typed");
    }
    ;

pointer
    : '*'
    | '*' type_specifier_list           {
      $$ = ' ';
      error("only a simple '*' is allowed");
    }
    | '*' pointer {
      $$ = ' ';
      error("only a simple '*' is allowed");
    }
    | '*' type_specifier_list pointer {
      $$ = ' ';
      error("only a simple '*' is allowed");
    }
    ;

type_specifier_list
    : type_specifier
    | type_specifier type_specifier_list { $$ = or_specifier($1, $2); }
    ;

parameter_identifier_list
    : identifier_list
    | identifier_list ',' DOTDOTDOT {
      error("stubs must have fixed parameter lists");
    }
    ;

identifier_list
    : IDENT
    | identifier_list ',' IDENT
    ;

parameter_type_list
    : parameter_list
    | parameter_list ',' DOTDOTDOT {
      $$ = $1;
      error("stubs must have fixed parameter lists");
    }
    ;

parameter_list
    : parameter_declaration {
      $$ = $1;
    }
    | parameter_list ',' parameter_declaration {
      if ( $1 != NULL )
	$$ = cat_sym( $1, $3 );
      else
	$$ = $3;
    }
    ;

parameter_declaration
    : type_specifier_list declarator {
      Symbol *temp;
      temp = distribute_specifier ($1, $2);
      NestLevel++;
      $$ = add_to_sym_or_type_table (temp);

      if ( SYM_IS_STRUCT_OR_UNION(temp) ) {
	error("structs and unions cannot be passed by value to stubs");
      }
      else if ( !SYM_IS_ARRAY(temp) && !SYM_IS_POINTER(temp) ) {
	dont_deref( SYM_LOCAL_NAME(temp) );
      }

      NestLevel--;
    }
    | type_name {
      $$ = NULL;
      error("parameters must have names associated with them");
    }
    ;

type_name
    : type_specifier_list
    | type_specifier_list abstract_declarator {
      error("only symple USC type allowed as 'type name'");
      $$ = NULL;
    }
    ;

abstract_declarator
    : pointer
    | abstract_declarator2
    | pointer abstract_declarator2
    ;

abstract_declarator2
    : '(' abstract_declarator ')'
    | '[' ']'
    | '[' constant_expr ']'
    | abstract_declarator2 '[' ']'
    | abstract_declarator2 '[' constant_expr ']'
    | '(' ')'
    | '(' parameter_type_list ')'
    | abstract_declarator2 '(' ')'
    | abstract_declarator2 '(' parameter_type_list ')'
    ;

initializer
    : assignment_expr
    | '{' initializer_list '}'
    | '{' initializer_list ',' '}'
    ;

initializer_list
    : initializer
    | initializer_list ',' initializer
    ;

translation_unit
    : /* empty */			{ $$ = 0; }
    | translation_unit declaration	{ dump_source(); }
    ;

%%
