/*
 *  symbol.c -- symbol table functions.
 *
 *  Copyright 1994, 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"
#include "symbol.h"
#include "expr.h"
#include "utils.h"

int NestLevel;
Hash_Tab *Type_Tab, *Symbol_Tab, *Struct_Defs, *Union_Defs;

int structcleanfunc (Type **t);
int symbolcleanfunc (Symbol **sym);

char *spec_type_to_string (Type* type)
{
  static char s[128];
  SpecifierTypeField spec_type;

  s[0] = '\0';

  if ( type == NULL )
    return "void";

  spec_type = TYPE_SPEC_TYPE(type);

  if (spec_type & SIGNED_TYPE)
    strcat (s, "signed ");
  if (spec_type & UNSIGNED_TYPE)
    strcat (s, "unsigned ");
  if (spec_type & CHAR_TYPE)
    strcat (s, "char ");
  if (spec_type & SHORT_TYPE)
    strcat (s, "short ");
  if (spec_type & INT_TYPE)
    strcat (s, "int ");
  if (spec_type & LONG_TYPE)
    strcat (s, "long ");
#ifndef __alpha
  if (spec_type & ULONGLONG_TYPE)
    strcat (s, "usitll ");
  else if (spec_type & LONGLONG_TYPE)
    strcat (s, "usitll ");
#endif /* !__alpha */

  if ( preserve_type_names ) {
    if (spec_type & STRUCT_TYPE)
      strcat (s, "struct ");
    if (spec_type & ENUM_TYPE)
      strcat (s, "enum ");
    if (spec_type & UNION_TYPE)
      strcat (s, "union ");
  }
  else if ( s[0] == '\0' )
    strcpy(s,"void");

  return s;
}

void pushlevel (void)
{
  dprintf (1000, ">>>>>>>>>>>>pushing level");
  NestLevel++;
}

void poplevel (void)
{

  dprintf (1000, ">>>>>>>>>>>>popping level");
  cleantab (Struct_Defs, structcleanfunc);
  cleantab (Union_Defs, structcleanfunc);
  cleantab (Symbol_Tab, symbolcleanfunc);
  NestLevel--;
}

char *local_name (char *name)
{
  char *temp = (char *)Malloc ((strmem (name) + strlen (LOCAL_PREFIX)) *
                               sizeof (char));
  sprintf (temp, "%s%s", LOCAL_PREFIX, name);
  return temp;
}

char *return_name (char *name)
{
  char *temp = (char *)Malloc ((strmem (name) + strlen (RETURN_PREFIX)) *
                               sizeof (char));
  sprintf (temp, "%s%s", RETURN_PREFIX, name);
  return temp;
}

int compute_memsize( int size, int align )
{
  if ( !align )
    crash ("Can't have an alignment of 0");
  return size + (align - size % align) % align;
}

Type *make_basic_declarator (void)
{
  Type *decl;
  decl = (Type *)Malloc (sizeof (Type));
  
  TYPE_CLASS(decl) = DECLARATOR;
  TYPE_MODIFIERS(decl) = NULL;
  TYPE_TDEF(decl) = 0;
  TYPE_NAME(decl) = NULL;
  TYPE_NEXT(decl) = NULL;
  TYPE_ETYPE(decl) = decl;
  TYPE_NATIVE(decl) = 1;
  TYPE_ADDNATIVE(decl) = 0;
  TYPE_PREVBRO(decl) = TYPE_NEXTBRO(decl) = decl;
  TYPE_USIT_SEEN(decl,0) = USIT_NOT_SEEN;
  TYPE_USIT_SEEN(decl,1) = USIT_NOT_SEEN;
  return decl;
}

int symbolcleanfunc (Symbol **sym)
{
  if (*sym) {
    return (SYM_LEVEL(*sym) >= NestLevel);
  }
  else {
    crash ("internal error in symbolcleanfunc");
  }
  return 0;
}

int structcleanfunc (Type **t)
{
  if (TYPE_IS_STRUCT(*t) || TYPE_IS_UNION(*t)) {
    return (TYPE_STRUCTDEF_LEVEL(*t) >= NestLevel);
  }
  else {
    crash ("internal error in structcleanfunc");
  }
  return 0;
}

int sym_cmp (Symbol **sym1, Symbol **sym2)
{
  int flag;
  dprintf (1000, "Entering sym_cmp: <%s> <%s>", (*sym1)->name, (*sym2)->name);
  flag = strcmp ((*sym1)->name, (*sym2)->name);
  dprintf (1000, "Exiting sym_cmp: %d\n", flag);
  return flag;
}

int struct_cmp (Type **std1, Type **std2)
{
  int flag;
  dprintf (900, "Entering struct_cmp");
  flag = strcmp (TYPE_STRUCTDEF_NAME(*std1),
                 TYPE_STRUCTDEF_NAME(*std2));
  dprintf (900, "Exiting struct_cmp");
  return flag;
}

unsigned hash_function(Symbol **sym)
{
  unsigned hash_val = 0;
  unsigned i;
  char *name = (*sym)->name;

  for ( ; *name; ++name) {
    hash_val = (hash_val << 2) + *name;
    if ( (i = (hash_val & 0x3fff)) )
      hash_val = (hash_val ^ (i >> 12)) & ~0x3fff;
  }

  return hash_val;
}

void init_symbols (void)
{
  Symbol_Tab = maketab(0, hash_function, sym_cmp);
  Type_Tab = maketab(0, hash_function, sym_cmp);
  /* a NULL hash function means a hash table of one entry which = a stack */
  Struct_Defs = maketab(1, NULL, struct_cmp);  
  Union_Defs = maketab(1, NULL, struct_cmp);
  
  NestLevel = 0;
}

TypeData *make_default_typedata (void)
{
  TypeData *new = (TypeData *)Malloc (sizeof (TypeData));
  TYPEDATA_MEMSIZE(new) = TYPEDATA_SIZE(new) = DEFAULT_SIZE;
  TYPEDATA_ALIGN(new) = DEFAULT_ALIGN;
  TYPEDATA_BORDER(new) = DEFAULT_BORDER;
  TYPEDATA_OFFSET(new) = DEFAULT_OFFSET;
  TYPEDATA_COMPLETE(new) = 1;
  return new;
}

void set_to_string (Constant *theconst, char *string)
{
  int i;
  dprintf (1000, "Entering set_to_string: %s", string);
  /* remove leading and trailing quotes if necessary */
  i = strlen (string); 
  if (string[0] == '\"') {
    string++;
    if (string[i-2] == '\"')
      string[i-2] = '\0';
  }
  CONST_TYPE(theconst) = STRING_CONST;
  CONST_STRING(theconst) = allocate_new_string (string);
  AddToGarbage (CONST_STRING (theconst), strmem (CONST_STRING (theconst))); 

  dprintf (1000, "Exiting set_to_string: %s %s", 
           CONST_STRING(theconst), string);
}

void set_to_int (Constant *theconst, char *string)
{
  dprintf (1000, "Entering set_to_int");
  CONST_TYPE(theconst) = INT_CONST;
  CONST_INT(theconst) = atoi (string);
  dprintf (1000, "Exiting set_to_int");
}

void set_to_float (Constant *theconst, char *string)
{
  dprintf (1000, "Entering set_to_float");
  CONST_TYPE(theconst) = DOUBLE_CONST;
  CONST_DOUBLE(theconst) = atof (string);
  dprintf (1000, "Exiting set_to_float");
}

Symbol *ident_to_sym (char *theident)
{
  Symbol *new = NULL;

  if (theident) {
    new = (Symbol *)Malloc (sizeof (Symbol));
    SYM_NAME(new) = allocate_new_string (theident);
    SYM_LOCAL_NAME(new) = local_name (theident);
    SYM_NEXT(new) = NULL;
    SYM_END(new) = new;
    SYM_TYPE(new) = NULL;
    SYM_ETYPE(new) = NULL;
    SYM_TYPEDATA(new) = NULL;
    SYM_LEVEL(new) = 0;
  }

  return new;
}

Symbol *cat_sym (Symbol *one, Symbol *two) 
{
  if (!one || !two)
    crash ("internal error in cat_sym");
  SYM_END(one)->next = two;
  SYM_END(one) = SYM_END(two);
  return one;
}

Type *make_specifier (Constant *theconst)
{
  Type *new;
  dprintf (900, "Entering make_specifier");
  print_constant (900, theconst);
  if (CONST_TYPE(theconst) != STRING_CONST)
    crash ("internal error in make_specifier");
  new = (Type *)Malloc (sizeof (Type));
  TYPE_MODIFIERS(new) = NULL;
  TYPE_CLASS(new) = SPECIFIER;
  TYPE_NEXT(new) = NULL;
  TYPE_NAME(new) = NULL;
  TYPE_ETYPE(new) = new;
  TYPE_TDEF(new) = 0;
  TYPE_SPEC_STRUCTDEF(new) = NULL;
  TYPE_NATIVE(new) = 1;
  TYPE_ADDNATIVE(new) = 0;
  TYPE_PREVBRO(new) = TYPE_NEXTBRO(new) = new;
  TYPE_USIT_SEEN(new,0) = USIT_NOT_SEEN;
  TYPE_USIT_SEEN(new,1) = USIT_NOT_SEEN;

  /* XXXmju this decision tree seems faulty to me; is there a reason not
     to do full string comparisons here? */
  /* XXXjimz I could not agree more. */
  switch (CONST_STRING(theconst)[0]) {
  case 'c':
    TYPE_SPEC_TYPE(new) = CHAR_TYPE;
    break;

  case 's':
    if (CONST_STRING(theconst)[1] == 'h')
      TYPE_SPEC_TYPE(new) = SHORT_TYPE;
    else if (CONST_STRING(theconst)[1] == 'i')
      TYPE_SPEC_TYPE(new) = SIGNED_TYPE;
    else
      TYPE_SPEC_TYPE(new) = STRUCT_TYPE;
    break;
    
  case 'l':
    TYPE_SPEC_TYPE(new) = LONG_TYPE;
    break;

  case 'u':
    if (CONST_STRING(theconst)[2] == 's')
      TYPE_SPEC_TYPE(new) = UNSIGNED_TYPE;
#ifndef __alpha
    else if (CONST_STRING(theconst)[1] == 's') /* usitll */
      TYPE_SPEC_TYPE(new) = LONGLONG_TYPE;
    else if (CONST_STRING(theconst)[2] == 'u') /* unusitll */
      TYPE_SPEC_TYPE(new) = ULONGLONG_TYPE;
#endif /* !__alpha */
    else
      TYPE_SPEC_TYPE(new) = UNION_TYPE;
    break;

  case 'i':
    TYPE_SPEC_TYPE(new) = INT_TYPE;
    break;

  default:
    crash ("internal error in make_specifier: %s", CONST_STRING(theconst));
    break;
  }
  return new;
}

Type *or_specifier (Type *one, Type *two)
{
  dprintf (500, "Entering or_specifier");
  print_type_list (500, 0, "one", one);
  print_type_list (500, 0, "two", two);
  if (!one || !two ||
      TYPE_CLASS(one) != SPECIFIER || TYPE_CLASS(two) != SPECIFIER)
    {
      crash ("internal error in or_specifier");
    }
  if (TYPE_TDEF(one) || TYPE_TDEF(two)) {
    crash ("Can't modify a typedef");
  }
  if (TYPE_MODIFIERS(one) || TYPE_MODIFIERS(two)) {
    crash ("Can't modify this type with specifiers");
  }
  if (TYPE_SPEC_STRUCTDEF(one) || TYPE_SPEC_STRUCTDEF(two)) {
    crash ("Can't modify a \'struct\' or \'union\' with specifiers");
  }
  TYPE_SPEC_TYPE(one) |= TYPE_SPEC_TYPE(two);
  print_type_list (500, 0, "one2", one);
  dprintf (500, "Exiting or_specifier");
  return one;
}

Type *copy_type (Type *thetype)
{
  Type *new = NULL;
  if (thetype) {
    new = (Type *)Malloc (sizeof (Type));
    *new = *thetype;
    if (TYPE_MODIFIERS (new)) {
      TYPE_MODIFIERS (new) = (TypeData *)Malloc (sizeof (TypeData));
      *TYPE_MODIFIERS (new) = *TYPE_MODIFIERS (thetype);
    }
    TYPE_NEXT(new) = copy_type(TYPE_NEXT(thetype));
    if (TYPE_NEXT(new))
      TYPE_ETYPE(new) = TYPE_ETYPE(TYPE_NEXT(new));
    else
      TYPE_ETYPE(new) = new;
    TYPE_NEXTBRO(new) = TYPE_NEXTBRO(thetype);
    TYPE_NEXTBRO(thetype) = new;
    TYPE_PREVBRO(new) = thetype;
    TYPE_PREVBRO (TYPE_NEXTBRO (new)) = new;
  }
  return new;
}

TypeData *merge_modifiers (TypeData *old, TypeData *new)
{ 
  int tempi;
  ByteOrder temps;

  if (new) {
    if ((tempi = TYPEDATA_SIZE(new)) != DEFAULT_SIZE) {
      if (TYPEDATA_SIZE (old) != DEFAULT_SIZE && TYPEDATA_SIZE (old) != tempi)
        warning ("Redefining size modifier");
      TYPEDATA_SIZE(old) = tempi;
    }
    if ((tempi = TYPEDATA_MEMSIZE(new)) != DEFAULT_SIZE) {
      if (TYPEDATA_MEMSIZE(old) != DEFAULT_SIZE&&TYPEDATA_MEMSIZE(old) != tempi)
        warning ("Redefining memsize modifier");
      TYPEDATA_MEMSIZE(old) = tempi;
    }
    if ((tempi = TYPEDATA_ALIGN(new)) != DEFAULT_ALIGN) {
      if ( TYPEDATA_ALIGN (old) != DEFAULT_ALIGN && 
          tempi % TYPEDATA_ALIGN(old))
        warning ("Redefining alignment modifier");
      TYPEDATA_ALIGN(old) = TYPEDATA_ALIGN(new);
    }
    if ((temps = TYPEDATA_BORDER(new)) != DEFAULT_BORDER) {
      if (TYPEDATA_BORDER (old) != DEFAULT_BORDER)
	if ( !equal_byteorder(TYPEDATA_BORDER (old), TYPEDATA_BORDER(new) ) )
	  warning ("Redefining byte order modifier");
      TYPEDATA_BORDER(old) = temps;
    }
    if ((tempi = TYPEDATA_OFFSET(new)) != DEFAULT_OFFSET) {
      if (TYPEDATA_OFFSET (old) != DEFAULT_OFFSET)
        warning ("Redefining offset modifier");
      TYPEDATA_OFFSET(old) = tempi;
    }
  }
  return old;
}

Symbol *distribute_specifier (Type *thetype, Symbol *symlist)
{
  Symbol *thelist = symlist;
  Type *temp;
  int tdef, native;
  dprintf (500, "Entering distribute_specifier");
  print_type_table(600);
  print_type_list (600, 0, "type", thetype);

  if (!thetype)
    crash ("Internal error in distribute_specifier");
  native = TYPE_ADDNATIVE(thetype);
  tdef = TYPE_TDEF(thetype);
  dprintf (500, "tdef: %d\n", tdef);
  
  while (thelist) {
    print_type_list (900, 1, "dist_spec1", SYM_TYPE(thelist));
    dprintf (900, "\n");
    temp = copy_type(thetype);

    if (SYM_TYPE(thelist)) {
      TYPE_NEXT(SYM_ETYPE(thelist)) = temp;
      SYM_ETYPE(thelist) = TYPE_ETYPE(temp);
      if (TYPE_IS_MODIFIER(SYM_TYPE(thelist))) {
        if (TYPE_MODIFIERS(temp)) {
          TYPE_MODIFIERS (temp) = 
            merge_modifiers (TYPE_MODIFIERS (temp),
                             TYPE_DECL_TYPEDATA(SYM_TYPE(thelist)));
        }
        else {
          TYPE_MODIFIERS(temp) = TYPE_DECL_TYPEDATA(SYM_TYPE(thelist));
        }
        SYM_TYPE(thelist) = TYPE_NEXT(SYM_TYPE(thelist));
	TYPE_NATIVE (SYM_TYPE (thelist)) = 0;
	if (TYPE_ADDNATIVE (thetype) && !TYPE_NATIVE (SYM_TYPE (thelist)))
          crash ("Can't declare this as native");
        TYPE_ETYPE(SYM_TYPE(thelist)) = SYM_ETYPE(thelist);
      }
    }
    else {
      SYM_TYPE(thelist) = temp;
      SYM_ETYPE(thelist) = TYPE_ETYPE(temp);
    }
    TYPE_TDEF(SYM_TYPE(thelist)) |= tdef;
    TYPE_ADDNATIVE (SYM_TYPE (thelist)) = native;
    if ( tdef && preserve_type_names )
      TYPE_NAME(SYM_TYPE(thelist)) = allocate_new_string(SYM_NAME(thelist));


    print_type_list (900, 1, "dist_spec2", SYM_TYPE(thelist));
    dprintf (900, "\n\n");
    thelist = thelist->next;
  }
  print_type_table (600);
  dprintf (500, "Exiting distribute_specifier");
  return symlist;
}

Type *name_to_type_hack (char *name)
{
  static StructDef stdef;
  static Type thetype;
  Type *hack = &thetype;
  TYPE_CLASS(hack) = SPECIFIER;
  TYPE_SPEC_TYPE(hack) = STRUCT_TYPE;
  TYPE_SPEC_STRUCTDEF(hack) = &stdef;
  TYPE_STRUCTDEF_NAME(hack) = allocate_new_string(name);
  return hack;
}

Type *find_struct_def (char *name)
{
  Type **h, *hack;

  dprintf (900, "Entering find_struct_def: %s\n", name);

  hack = name_to_type_hack (name);
  print_type (900, 0, "", hack);
  if ((h = findsym (Struct_Defs, &hack)) || 
      (h = findsym (Union_Defs, &hack))) {
    dprintf (900, "Entering find_struct_def: %s\n", name);
    Free (TYPE_STRUCTDEF_NAME(hack), strmem (TYPE_STRUCTDEF_NAME(hack)));
    return *h;
  }
  else {
    dprintf (900, "Entering find_struct_def: %s\n", name);
    Free (TYPE_STRUCTDEF_NAME(hack), strmem (TYPE_STRUCTDEF_NAME(hack)));
    return NULL;
  }
}

char all_fields_are_native (Symbol *syms)
{
  Symbol *p;

  for (p = syms; p; p = SYM_NEXT (p))
    if (!SYM_NATIVE (p))
      return 0;
  return 1;
}

Type *finish_struct_def (Type *spec, char *name, Symbol *syms)
{
  StructDef *stdef = (StructDef *)Malloc (sizeof (StructDef));
  dprintf (500, "Entering finish_struct_def");
  print_sym_list (800, 0, "", syms);
  stdef->fields = syms;
  stdef->name = allocate_new_string (name);
  stdef->level = -1;
  spec->native = all_fields_are_native (syms);

  print_type (800, 0, "1", spec);
  TYPE_SPEC_STRUCTDEF(spec) = stdef;

  print_type (500, 0, "2", spec);
  dprintf (500, "Exiting finish_struct_def");
  return spec;
}

Type *add_to_struct_defs (Type *stdef, int level)
{
  Type **hashstdef;
  
  dprintf (500, "Entering add_to_struct_defs");
  dprintf (500, "NestLevel: %d\n", level);
  print_type (500, 0, "", stdef);
  
  if (TYPE_IS_STRUCT(stdef) && TYPE_IS_UNION(stdef))
    crash ("A type can't be both a struct and union");
  
  hashstdef = (Type **)newsym (sizeof (Type *));
  *hashstdef = stdef;
  TYPE_STRUCTDEF_LEVEL(*hashstdef) = level;
  
  if (TYPE_IS_STRUCT(stdef)) {
    addsym (Struct_Defs, hashstdef);
    print_struct_table (900);
  }
  else {
    addsym (Union_Defs, hashstdef);
    print_union_table (900);
  }
      
  dprintf (500, "Exiting add_to_struct_defs");
  return stdef;
}

Type *make_typedef (Type *thetype)
{
  if (TYPE_TDEF(thetype)) {
    crash ("Already typedef'd!");
  }
  TYPE_TDEF(thetype) = 1;
  return thetype;
}

Symbol *add_to_sym_or_type_table (Symbol *syms)
{
  Symbol *s = syms, **new;
  Hash_Tab *table;
  dprintf (500, "Entering add_to_sym_or_type_table");
  print_sym_list (1000, 0, "", syms);

  if (!syms) {
    crash ("internal error in add_to_sym_or_type_table");
  }

  if (TYPE_TDEF(syms->type)) {
    table = Type_Tab;
    dprintf (500, "TypeTab:");
  }
  else {
    table = Symbol_Tab;
    dprintf (500, "SymTab:");
  }

  while (s) {
    dprintf (900, "looking for: %s\n", s->name);
    if (table == Symbol_Tab) {
      if ( ( NestLevel == 0 ) &&(!SYM_IS_FUNCTION( s ) ) ) {
	error("No global variables allowed. Discarding '%s'", SYM_NAME(s) );
	s = SYM_NEXT(s);
	continue;
      }
      else if ( SYM_LEVEL( s ) == 0 ) {
	SYM_LEVEL( s ) = NestLevel;
      }
    }
    if (findsym (table, &s)) {
      crash ("symbol <%s> already declared", SYM_NAME(s) );
    }
    else {
      dprintf (900, "adding the symbol: %s", s->name);
      new = newsym (sizeof (Symbol *));
      *new = s;
      addsym (table, new);
    }
    s = SYM_NEXT(s);
  }
  if (table == Symbol_Tab){
    print_sym_table (700);
  }
  else {
    print_type_table (700);
  }

  dprintf (500, "exiting add_to_sym_or_type_table");
  return syms;
}

Type *find_type (char *name)
{
  static Symbol sym;
  Symbol *s = &sym;
  Symbol **ans;
  Type *rvar;

  dprintf (500, "entering find_type: %s", name);

  SYM_NAME(&sym) = allocate_new_string(name);
  ans = findsym (Type_Tab, &s);
  Free (SYM_NAME (&sym), strmem (SYM_NAME(&sym)));
  if (!ans) {
    dprintf (500, "type not found: %s", name);
    rvar = NULL;
  }
  else {
    TYPE_TDEF(SYM_TYPE(*ans)) = 0;
    print_sym (600, 0, "found", *ans);
    dprintf (500, "type found: %s", name);
    rvar =  SYM_TYPE(*ans);
  }
  print_type_table(900);
  dprintf (500, "exiting find_type");
  return rvar;
}

Symbol *find_sym (char *name)
{
  static Symbol sym;
  Symbol *s = &sym;
  Symbol **ans;

  dprintf (500, "entering find_sym: %s", name);

  SYM_NAME(&sym) = allocate_new_string(name);
  ans = findsym (Symbol_Tab, &s);
  Free (SYM_NAME (&sym), strmem (SYM_NAME (&sym)));
  if (!ans) {
    dprintf (500, "symbol not found: %s", name);
    dprintf (500, "exiting find_sym");
    return NULL;
  }
  else {
    TYPE_TDEF(SYM_TYPE(*ans)) = 0;
    print_sym (500, 0, "found", *ans);
    dprintf (600, "symbol found: %s", name);
  }
  print_sym_table(900);
  dprintf (500, "exiting find_sym");
  return *ans;
}

Symbol *make_array (Symbol *sym, Expr *theexpr)
{
  int size = 0;
  Type *decl;
  Expr* var_array = NULL;

  print_sym_list (800, 0, "make_array", sym);

  if (theexpr) {
    if ( EXPR_IS_CONST(theexpr) ) {
      size = EXPR_CONST_INT(theexpr);
      if ( size < 0 )
	crash("negative array size");
    }
    else if ( EXPR_IS_IDENT(theexpr) ) {
      size = -1;
      var_array = (Expr*)malloc(sizeof(Expr));
      *var_array = *theexpr;
      EXPR_IDENT(var_array) = malloc(strlen(EXPR_IDENT( theexpr)) +1 );
      strcpy(EXPR_IDENT(var_array),EXPR_IDENT( theexpr));
    }
  }
  else {
    crash("Internal error in make_array");
    size = 0;
  }

  decl = make_basic_declarator ();
  
  TYPE_DECL_TYPE(decl) = ARRAY_TYPE;
  if (size < 0)
    size = -1;
  TYPE_DECL_NUMEL(decl) = size;
  TYPE_DECL_VAR_ARRAY(decl) = var_array;

  if (SYM_ETYPE(sym)) {
    TYPE_NEXT(SYM_ETYPE(sym)) = decl;
    SYM_ETYPE(sym) = decl;
  }
  else {
    SYM_TYPE(sym) = SYM_ETYPE(sym) = decl;
  }

  return sym;
}

Symbol *make_pointer (Symbol *sym)
{
  Type *decl;
  print_sym_list (800, 0, "make_pointer", sym);

  decl = make_basic_declarator ();
  
  TYPE_DECL_TYPE(decl) = POINTER_TYPE;

  if (SYM_ETYPE(sym)) {
    TYPE_NEXT(SYM_ETYPE(sym)) = decl;
    SYM_ETYPE(sym) = decl;
  }
  else {
    SYM_TYPE(sym) = SYM_ETYPE(sym) = decl;
  }

  return sym;
}

Symbol *make_function (Symbol *sym, Symbol *params)
{
  int pcount = 0;
  Symbol *p = params;
  Type *decl;
  char* r_name = SYM_LOCAL_NAME(sym);

  print_sym(80, 0, "make_function - function", sym);
  print_sym_list (80, 0, "make_function - params", params);

  Free(r_name, strmem(r_name) );
  SYM_LOCAL_NAME(sym) = return_name(SYM_NAME(sym) );
  dont_deref(SYM_LOCAL_NAME(sym));

  decl = make_basic_declarator ();
  
  TYPE_DECL_TYPE(decl) = FUNCTION_TYPE;
  TYPE_DECL_PARAMS(decl) = params;

  if (SYM_ETYPE(sym)) {
    TYPE_NEXT(SYM_ETYPE(sym)) = decl;
    SYM_ETYPE(sym) = decl;
  }
  else {
    SYM_TYPE(sym) = SYM_ETYPE(sym) = decl;
  }

  while (p) {
    pcount++;
    p = SYM_NEXT(p);
  }
  TYPE_DECL_NUMEL(decl) = pcount;

  return sym;
}

void check_spec (Type *spec, TypeData *data)
{
  SpecifierTypeField field;

  TYPEDATA_MEMSIZE(data) = TYPEDATA_SIZE(data) = DEFAULT_SIZE;
  TYPEDATA_ALIGN(data)  = DEFAULT_ALIGN;
  TYPEDATA_BORDER(data) = DEFAULT_BORDER;
  TYPEDATA_OFFSET(data) = DEFAULT_OFFSET;

  if (TYPE_IS_SPECIFIER(spec)) {
    
    field = TYPE_SPEC_TYPE(spec);
    
    if ((field & STRUCT_TYPE) && (field & (~STRUCT_TYPE))) {
      crash ("Can't modify a struct with specifiers: %s", 
             spec_type_to_string (spec));
    }
    if ((field & UNION_TYPE) && (field & (~UNION_TYPE))) {
      crash ("Can't modify a union with specifiers");
    }
    if ((field & ENUM_TYPE) && (field & (~ENUM_TYPE))) {
      crash ("Can't modify an enum with specifiers");
    }
    if ((field & SIGNED_TYPE) && (field & UNSIGNED_TYPE)) {
      error ("type is declared both signed and unsigned");
      field &= ~UNSIGNED_TYPE;
    }
    if (field & LONG_TYPE) {
      field &= LONG_TYPE | SIGNED_TYPE | UNSIGNED_TYPE;
    }
#ifndef __alpha
    if (field & LONGLONG_TYPE) {
      field &= ULONGLONG_TYPE | LONGLONG_TYPE | SIGNED_TYPE | UNSIGNED_TYPE;
    }
    if (field & ULONGLONG_TYPE) {
      field &= ULONGLONG_TYPE | LONGLONG_TYPE | SIGNED_TYPE | UNSIGNED_TYPE;
    }
#endif /* !__alpha */
    if ((field & INT_TYPE) && (field & CHAR_TYPE)) {
      error ("a char int?");
      field &= INT_TYPE | SIGNED_TYPE | UNSIGNED_TYPE | SHORT_TYPE;
    }
    if ((field & SHORT_TYPE) && (field & CHAR_TYPE)) {
      field &= CHAR_TYPE | SIGNED_TYPE | UNSIGNED_TYPE;
    }
    if (field & SHORT_TYPE) {
      field &= SHORT_TYPE | SIGNED_TYPE | UNSIGNED_TYPE;
    }
    if (!(field & (SIGNED_TYPE | UNSIGNED_TYPE | STRUCT_TYPE | UNION_TYPE |
#ifndef __alpha
                   ULONGLONG_TYPE |
#endif /* !__alpha */
                   ENUM_TYPE))) {
      field |= SIGNED_TYPE;
    }
#if 0
    if (field & ULONGLONG_TYPE) {
      field |= UNSIGNED_TYPE;
    }
#endif
/*
    if (field & (STRUCT_TYPE | UNION_TYPE | ULONGLONG_TYPE)) {
      field |= UNSIGNED_TYPE;
    }
*/

    TYPE_SPEC_TYPE(spec) = field;
    
#ifndef __alpha
    if (field & LONGLONG_TYPE) {
      TYPEDATA_SIZE(data) = machine_info.longlong_info.size;
      TYPEDATA_MEMSIZE(data) = machine_info.longlong_info.memsize;
      TYPEDATA_ALIGN(data) = machine_info.longlong_info.alignment;
      TYPEDATA_BORDER(data) = machine_info.longlong_info.byteorder;
    } else
    if (field & ULONGLONG_TYPE) {
      TYPEDATA_SIZE(data) = machine_info.ulonglong_info.size;
      TYPEDATA_MEMSIZE(data) = machine_info.ulonglong_info.memsize;
      TYPEDATA_ALIGN(data) = machine_info.ulonglong_info.alignment;
      TYPEDATA_BORDER(data) = machine_info.ulonglong_info.byteorder;
    } else
#endif /* !__alpha */
    if (field & LONG_TYPE) {
      TYPEDATA_SIZE(data) = machine_info.long_info.size;
      TYPEDATA_MEMSIZE(data) = machine_info.long_info.memsize;
      TYPEDATA_ALIGN(data) = machine_info.long_info.alignment;
      TYPEDATA_BORDER(data) = machine_info.long_info.byteorder;
    }
    else if (field & INT_TYPE) {
      TYPEDATA_SIZE(data) = machine_info.int_info.size;
      TYPEDATA_MEMSIZE(data) = machine_info.int_info.memsize;
      TYPEDATA_ALIGN(data) = machine_info.int_info.alignment;
      TYPEDATA_BORDER(data) = machine_info.int_info.byteorder;
    }
    else if (field & SHORT_TYPE) {
      TYPEDATA_SIZE(data) = machine_info.short_info.size;
      TYPEDATA_MEMSIZE(data) = machine_info.short_info.memsize;
      TYPEDATA_ALIGN(data) = machine_info.short_info.alignment;
      TYPEDATA_BORDER(data) = machine_info.short_info.byteorder;
    }
    else if (field & CHAR_TYPE) {
      TYPEDATA_SIZE(data) = machine_info.char_info.size;
      TYPEDATA_MEMSIZE(data) = machine_info.char_info.memsize;
      TYPEDATA_ALIGN(data) = machine_info.char_info.alignment;
      TYPEDATA_BORDER(data) = machine_info.char_info.byteorder;
    }
  }
  else if (TYPE_IS_ARRAY(spec)) {
    TYPEDATA_SIZE(data) = 0;
    TYPEDATA_MEMSIZE(data) = 0;
    TYPEDATA_ALIGN(data) = 0;
    TYPEDATA_BORDER(data) = NULL;
  }
  else if (TYPE_IS_BITFIELD(spec)) {
    TYPEDATA_MEMSIZE(data) = 0;
    TYPEDATA_SIZE(data) = 0;
    TYPEDATA_ALIGN(data) = 0;
    TYPEDATA_BORDER(data) = NULL;
  }
  else if (TYPE_IS_POINTER(spec)) {
    TYPEDATA_MEMSIZE(data) = 0;
    TYPEDATA_SIZE(data) = 0;
    TYPEDATA_ALIGN(data) = 0;
    TYPEDATA_BORDER(data) = NULL;
  }
  else {
    crash ("internal error in check_spec");
  }
}

TypeData *merge_type_data (TypeData *data, Type *spec)
{
  TypeData *new;
  int tempi;
  ByteOrder temps;

  new = make_default_typedata();

  check_spec (spec, new);
  
  if (data) {
    if ((tempi = TYPEDATA_SIZE(data)) != DEFAULT_SIZE) {
      TYPEDATA_SIZE(new) = tempi;
      dprintf (20, "merge_type_data: setting size: %d", tempi);
    }
    else {
      dprintf (20, "Setting size to default: %d\n", DEFAULT_SIZE);
    }
    if ((tempi = TYPEDATA_ALIGN(data)) != DEFAULT_ALIGN)
      TYPEDATA_ALIGN(new) = tempi;
    if ((temps = TYPEDATA_BORDER(data)) != DEFAULT_BORDER)
      TYPEDATA_BORDER(new) = temps;
    if ((tempi = TYPEDATA_OFFSET(data)) != DEFAULT_OFFSET)
      TYPEDATA_OFFSET(new) = tempi;
    if ((tempi = TYPEDATA_MEMSIZE(data))!= DEFAULT_SIZE)
      TYPEDATA_MEMSIZE(new) = tempi;

    TYPEDATA_COMPLETE(new) = TYPEDATA_COMPLETE(data);
  }
  if (TYPE_IS_SPECIFIER(spec) &&( TYPEDATA_MEMSIZE(new) == DEFAULT_SIZE ) ) {
    TYPEDATA_MEMSIZE(new) = compute_memsize(TYPEDATA_SIZE(new),
                                            TYPEDATA_ALIGN(new));
  }
  if (TYPEDATA_BORDER (new) && 
      byteorder_len (TYPEDATA_BORDER(new)) != TYPEDATA_SIZE(new))
    crash ("Illegal byte order specification: %d %s", 
	   TYPEDATA_SIZE (new), byteorder_to_string(TYPEDATA_BORDER(new)));
  return new;
}

TypeData calculate_typedata (Type *type)
{
  TypeData ans = { 
    -1, 
    -1, 
    -1,
    NULL, 
    DEFAULT_OFFSET, 
    -1,
  };
  TypeData temp;
  int bitsize = 0;

  if (type) {
    if ( TYPE_NEXT(type) && TYPE_IS_POINTER(TYPE_NEXT(type)) )
      crash("pointers cannot occur inside other types");
    temp = calculate_typedata (TYPE_NEXT(type));
    if (TYPE_IS_ARRAY(type)) {
      if (!type->next || TYPE_IS_BITFIELD (type->next)) {
        crash ("Array has invalid type");
      }
      if (TYPE_MODIFIERS(type)) {
        ans.offset = TYPEDATA_OFFSET(TYPE_MODIFIERS(type));
      }
      if (!temp.complete_type) {
        crash ("Incomplete array");
      }
      if (temp.size < 0) {
        crash ("internal error in calculate_typedata");
      }
      else {
        ans.size = temp.size;
        ans.alignment = temp.alignment;
        ans.offset = temp.offset;
        ans.byte_order = allocate_new_string(temp.byte_order);
        if (TYPE_DECL_NUMEL (type) < 0) {
          ans.complete_type = 0;
          ans.memsize = temp.memsize;
          if (!temp.complete_type)
            crash ("Illegal array declaration");
        }
        else {
          ans.memsize = TYPE_DECL_NUMEL(type) * temp.memsize;
          ans.complete_type = temp.complete_type;
        }
        
        if (TYPE_MODIFIERS (type)) {
          TYPEDATA_MEMSIZE (TYPE_MODIFIERS(type)) = temp.memsize;
          if ( (TYPEDATA_ALIGN (TYPE_MODIFIERS(type)) != DEFAULT_ALIGN )
	     &&(TYPEDATA_ALIGN (TYPE_MODIFIERS(type)) != temp.alignment ) )
            warning ("you are overriding the natural alignment of an array"
		     ":%d %d", TYPEDATA_ALIGN (TYPE_MODIFIERS(type)),
			      temp.alignment
		    );
          else
            TYPEDATA_ALIGN (TYPE_MODIFIERS(type)) = temp.alignment;
          TYPEDATA_COMPLETE (TYPE_MODIFIERS(type)) = ans.complete_type;
        }
        else {
          crash ("internal error in calculate_typedata");
        }
      }
    }
    else if (TYPE_IS_BITFIELD (type)) {
      if (!type->next || !TYPE_IS_SPECIFIER (type->next) ||
          TYPE_IS_STRUCT_OR_UNION (type->next)) {
        crash ("Bitfield has invalid type");
      }
      if (TYPE_DECL_BITINFO (type))
        bitsize = TYPE_DECL_MEMSIZE(type);
      else
	bitsize = TYPE_DECL_NUMEL (type);
      dprintf (100, "bitfield size: %d", bitsize);
      if (bitsize > temp.size * 8)
        crash ("Size of bitfield exceeds size of its type: %d %d", 
               bitsize, temp.size);
      ans = temp;
      ans.byte_order = allocate_new_string (temp.byte_order);
      ans.size = ((bitsize + 7) & ~7) / 8;
      ans.memsize = ans.size;

      if (TYPE_MODIFIERS (type)) {
        *TYPE_MODIFIERS (type) = ans;
      }
      else {
        TYPE_MODIFIERS (type) = (TypeData *)Malloc (sizeof (TypeData));
        *TYPE_MODIFIERS (type) = ans;
      }
    }
    else if (TYPE_IS_SPECIFIER(type)) {
      ans = *TYPE_MODIFIERS(type);
    }
    else if (TYPE_IS_POINTER(type)) {
      if (!type->next || TYPE_IS_BITFIELD (type->next)) {
        crash ("Pointer to invalid type");
      }
      if (temp.size < 0) {
        crash ("internal error in calculate_typedata");
      }
      else {
        ans = temp;
       
	if (TYPE_MODIFIERS (type)) {
	  *TYPE_MODIFIERS (type) = ans;
	}
	else {
	  TYPE_MODIFIERS (type) = (TypeData *)Malloc (sizeof (TypeData));
	  *TYPE_MODIFIERS (type) = ans;
	 ans = temp;
	}
      }
    }
    else {
      crash ("internal error in calculate_typedata");
    }
  }
  
  return ans;
}

TypeData *compute_symbol_typedata (Symbol *sym)
{
  TypeData *new = (TypeData *)Malloc (sizeof (TypeData));
  *new = calculate_typedata (SYM_TYPE(sym));
  return new;
}

void check_and_set_type_data (Symbol *sym)
{
  int tdef;
  Type *t;

  dprintf (500, "Entering check_and_set_type_data");
  dprintf (500, "checking type of: %s", SYM_NAME(sym));
  print_sym (500, 0, "1", sym);

  tdef = TYPE_TDEF(SYM_TYPE(sym));
  for (t = SYM_TYPE(sym); t; t = t->next) {
    if (TYPE_IS_MODIFIER(t)) {
      crash ("internal error in check_and_set_type_info");
    }
    else {
      TYPE_MODIFIERS(t) = merge_type_data (TYPE_MODIFIERS(t), t);
    }
    TYPE_TDEF(SYM_TYPE(sym)) = tdef;
  }
  SYM_TYPEDATA(sym) = compute_symbol_typedata(sym);
  print_sym (500, 0, "2", sym);
  
  dprintf (500, "Exiting check_and_set_type_data");
}

void set_specifier (Type *spec)
{
  Symbol *field;
  dprintf (100, "Entering set_specifier");
  if (TYPE_IS_STRUCT_OR_UNION (spec)) {
    for (field = TYPE_STRUCTDEF_FIELDS (spec); field; 
         field = SYM_NEXT (field)){
      set_specifier (SYM_ETYPE (field));
      check_and_set_type_data (field);
    }
  }
  dprintf (100, "Exiting set_specifier");
}

void set_all_sym_type_info (Symbol *sym)
{
  Symbol *s;
  dprintf (100, "Entering set_all_sym_type_info");
  print_sym_list (500, 0, "", sym);
  for (s = sym; s; s = SYM_NEXT(s)) {
    set_specifier (SYM_ETYPE (s));
    check_and_set_type_data (s);
  }
  dprintf (100, "Exiting set_all_sym_type_info");
}

Type *make_native_bitfield (Expr *theexpr)
{
  Type *new = make_basic_declarator();
  TYPE_DECL_TYPE (new) = BITFIELD_TYPE;
  TYPE_DECL_BITINFO (new) = NULL;

  if (EXPR_IS_INT (theexpr)) {
    TYPE_DECL_NUMEL (new) = EXPR_INT (theexpr);
  }
  else {
    crash ("internal error in make_native_bitfield");
  }
  return new;
}

Symbol *make_bitfield (Symbol *sym, Type *bitfield)
{
  dprintf (100, "Entering make_bitfield");
  if (!bitfield)
    crash ("internal error in make_bitfield");
  
  if (SYM_ETYPE(sym)) {
    TYPE_NEXT(SYM_ETYPE(sym)) = bitfield;
    SYM_ETYPE(sym) = bitfield;
  }
  else {
    SYM_TYPE(sym) = SYM_ETYPE(sym) = bitfield;
  }
  TYPE_NEXT (bitfield) = NULL;
  print_sym (500, 0, "", sym);
  dprintf (100, "Exiting make_bitfield");
  return sym;
}

void variable_length_array( Type* var_len_type, VarLen* var_len_data )
{
  Symbol*   fields;

  if ( !var_len_type || !TYPE_MODIFIERS(var_len_type) ) {
    crash("internal error in variable_length_array");
  }
  if ( TYPE_COMPLETE( var_len_type ) )
    crash("incomplete type expected in variable_length_array");

  if ( TYPE_IS_ARRAY( var_len_type ) ) {
    VAR_ARRAY_ELEM_SIZE( var_len_data ) = TYPE_MEMSIZE( var_len_type );
    VAR_ARRAY_LENGTH_ID( var_len_data ) =
			EXPR_IDENT(TYPE_DECL_VAR_ARRAY( var_len_type ) );
    return;
  }
  else if ( TYPE_IS_STRUCT( var_len_type ) ) {
    fields = TYPE_STRUCTDEF_FIELDS( var_len_type );
    if ( !fields ) 
      crash("no fields in struct in variable_length_array");
    while ( SYM_NEXT( fields ) )
      fields = SYM_NEXT( fields );
    variable_length_array( SYM_TYPE(fields), var_len_data );
    return;
  }
  else if ( TYPE_IS_POINTER( var_len_type ) ) {
    var_len_type = TYPE_NEXT( var_len_type );
    variable_length_array( var_len_type, var_len_data );
    return;
  }
  else {
    crash("unexpected type in variable_length_array");
  }
}

Expr* sizeof_type( Type* type )
{
  Expr* expr = NULL;
  int size = 0;

  if ( !type ) {
    crash ("internal error in sizeof_type");
  }
  if ( !TYPE_MODIFIERS(type) ) {
    SpecifierTypeField spec_type = TYPE_SPEC_TYPE(type);

#ifndef __alpha
    if (spec_type & ULONGLONG_TYPE) {
      expr = int_to_expr( machine_info.longlong_info.memsize );
    }
    else if (spec_type & LONGLONG_TYPE) {
      expr = int_to_expr( machine_info.ulonglong_info.memsize );
    }
#endif /* !__alpha */
    if (spec_type & LONG_TYPE) {
      expr = int_to_expr( machine_info.long_info.memsize );
    }
    else if (spec_type & INT_TYPE) {
      expr = int_to_expr( machine_info.int_info.memsize );
    }
    else if (spec_type & SHORT_TYPE) {
      expr = int_to_expr( machine_info.short_info.memsize );
    }
    else if (spec_type & CHAR_TYPE ) {
      expr = int_to_expr( machine_info.char_info.memsize );
    }
    else{
      crash("unexpected specifier in sizeof_type");
    }
    EXPR_CONST_UNIT(expr) = BYTE_UNIT;
  }
  else if ( TYPE_IS_POINTER(type) ) {
    expr = int_to_expr( machine_info.int_info.memsize );
    EXPR_CONST_UNIT(expr) = BYTE_UNIT;
  }
  else if ( TYPE_COMPLETE( type ) ) {
    size = TYPE_MEMSIZE( type );
    if ( TYPE_IS_ARRAY( type ) )
      size *= TYPE_DECL_NUMEL( type );
    expr = int_to_expr( size );
    EXPR_CONST_UNIT(expr) = BYTE_UNIT;
  }
  else {
    VarLen   var_data;
    Constant length_identifier;
    int      base_struct_size;

    variable_length_array( type, &var_data );

    set_to_string( &length_identifier, VAR_ARRAY_LENGTH_ID( &var_data ) );

    base_struct_size = TYPE_SIZE(type);
    if ( base_struct_size > 0 )
      base_struct_size -= VAR_ARRAY_ELEM_SIZE( &var_data );

    if ( base_struct_size < 0 ) {
      crash("internal error in sizeof_type (%d,%d)",
	     TYPE_SIZE(type),
	     VAR_ARRAY_ELEM_SIZE(&var_data));
    }
    expr = int_to_expr( VAR_ARRAY_ELEM_SIZE(&var_data) );
    EXPR_CONST_UNIT(expr) = BYTE_UNIT;

    expr = add_exprs( int_to_expr( base_struct_size ),
		      mul_exprs( expr, ident_to_expr( &length_identifier )) );
    EXPR_CONST_UNIT(EXPR_OP_LEFT(expr)) = BYTE_UNIT;
  }
  return expr;
}
